"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});

/**
 * ブートストラップ用のユーティリティ
 *
 * @class
 */
let arBootstrapUtils = Object.freeze({
  /**
   * resource URI に登録する
   *
   * @param   {string} resourceName
   *          リソースのディレクトリ名
   * @param   {Object} data
   *          Bootstrap データ
   */
  registerResource: function(resourceName, data) {
    let alias = Services.io.newFileURI(data.installPath);
    if (data.installPath.isDirectory()) {
      alias.spec += 'res/';
    } else {
      alias = Services.io.newURI("jar:" + alias.spec + "!/res/", null, null);
    }

    Services.io.getProtocolHandler("resource")
      .QueryInterface(Ci.nsIResProtocolHandler)
      .setSubstitution(resourceName, alias);
  },

  /**
   * resource URI の登録を解除する
   *
   * @param   {string} resourceName
   *          リソースのディレクトリ名
   * @param   {Object} data
   *          Bootstrap データ
   */
  unregisterResource: function(resourceName, data) {
    Services.io.getProtocolHandler("resource")
      .QueryInterface(Ci.nsIResProtocolHandler)
      .setSubstitution(resourceName, null);
  }
});

/* exported startup */

/**
 * 有効な場合の起動時、無効から有効への変更時 (更新含む)
 *
 * @param   {Object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            1: APP_STARTUP
 *            3: ADDON_ENABLE
 *            5: ADDON_INSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function startup(data, reason) {
  try {
    arBootstrapUtils.registerResource("unmht", data);

    Cu.import("resource://unmht/modules/startup.jsm", {})
      .startup(data, reason, "8.3.2");
  } catch (e) {
    Cu.reportError(e + "\n"
                   + e.fileName + ":" + e.lineNumber + "\n"
                   + e.stack);
  }
}

/* exported shutdown */

/**
 * 有効な場合の終了時、有効から無効への変更時 (更新含む)
 *
 * @param   {Object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            2: APP_SHUTDOWN
 *            4: ADDON_DISABLE
 *            6: ADDON_UNINSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function shutdown(data, reason) {
  try {
    Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {})
      .arShutdownEvent.broadcast(data, reason);
  } catch (e) {
    Cu.reportError(e + "\n"
                   + e.fileName + ":" + e.lineNumber + "\n"
                   + e.stack);
  }

  let APP_SHUTDOWN = 2;

  if (reason == APP_SHUTDOWN) {
    return;
  }

  let modules = [
    "UnMHTAdditionalDialog",
    "UnMHTCache",
    "UnMHTConfirmDialog",
    "UnMHTContentModifier",
    "UnMHTContentPolicy",
    "UnMHTContentSniffer",
    "UnMHTCustomFilename",
    "UnMHTDLFRegisterer",
    "UnMHTDOMConverter",
    "UnMHTDocumentCategory",
    "UnMHTDocumentLoaderFactory",
    "UnMHTExtractor",
    "UnMHTFennecMenu",
    "UnMHTFennecPrefPage",
    "UnMHTFileConverter",
    "UnMHTFileLoadHook",
    "UnMHTFileLoader",
    "UnMHTFileLoader_content",
    "UnMHTFileSaver",
    "UnMHTFilename",
    "UnMHTKeyboardShortcut",
    "UnMHTLoadModifier",
    "UnMHTLocationBar",
    "UnMHTMAFIntegration",
    "UnMHTMail",
    "UnMHTMailURLUtils",
    "UnMHTMenu",
    "UnMHTPageInfo",
    "UnMHTPref",
    "UnMHTPrefWindow",
    "UnMHTProtocolHandler",
    "UnMHTSave",
    "UnMHTSaveDialog",
    "UnMHTSaveMulti",
    "UnMHTSaveMultiDialog",
    "UnMHTSaveOptionsDialog",
    "UnMHTSaveProgress",
    "UnMHTSaveState",
    "UnMHTScheme",
    "UnMHTScrapBookIntegration",
    "UnMHTStringBundle",
    "UnMHTSystemDirectory",
    "UnMHTToolbar",
    "base/arAlertUtils",
    "base/arAppInfo",
    "base/arArrayBufferUtils",
    "base/arArrayUtils",
    "base/arComponentLoader",
    "base/arCustomizableUI",
    "base/arDLFRegisterer",
    "base/arDOMUtils",
    "base/arDialog",
    "base/arEvent",
    "base/arFennecMenu",
    "base/arFilePicker",
    "base/arFileUtils",
    "base/arFormatter",
    "base/arListView",
    "base/arLocaleUtils",
    "base/arMIMEDecoder",
    "base/arMIMEEncoder",
    "base/arMIMEParams",
    "base/arMIMEParser",
    "base/arMIMEPart",
    "base/arMIMEUtils",
    "base/arNetFileLoader",
    "base/arPathUtils",
    "base/arPrefClass",
    "base/arPrototype",
    "base/arPseudoID",
    "base/arScoped",
    "base/arShutdownEvent",
    "base/arStringBundle",
    "base/arStylesheetLoader",
    "base/arSystemDirectory",
    "base/arTaskUtils",
    "base/arTreeWalker",
    "base/arUconv",
    "base/arWinRegUtils",
    "base/arWindowOverlay",
    "base/arWindowUtils",
    "base/arWindowWatcher",
    "startup"
  ];

  for (let module of modules) {
    Cu.unload("resource://unmht/modules/" + module + ".jsm");
  }

  try {
    arBootstrapUtils.unregisterResource("unmht", data);
  } catch (e) {
    Cu.reportError(e + "\n"
                   + e.fileName + ":" + e.lineNumber + "\n"
                   + e.stack);
  }
}

/* exported install */

/**
 * インストール時、更新時
 * ただしアプリケーションに互換性が無い場合は呼ばれない
 *
 * @param   {Object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            5: ADDON_INSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function install(data, reason) {
}

/* exported uninstall */

/**
 * アンインストール時、更新時
 * ただしアプリケーションに互換性が無い場合は呼ばれない
 *
 * @param   {Object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            6: ADDON_UNINSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function uninstall(data, reason) {
}
