"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTCache"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arFormatter } = Cu.import("resource://unmht/modules/base/arFormatter.jsm", {});
let { arPathUtils } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { UnMHTSaveState, UnMHTSaveStateUtils } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});
let { toUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arMIMEDecoder",
                                  "resource://unmht/modules/base/arMIMEDecoder.jsm");

/**
 * ファイルの展開情報のキャッシュ
 *
 * @class
 * @implements arIShutdownEventListener
 */
let UnMHTCache = Object.freeze({
  /**
   * 初期化処理
   */
  init: function() {
    arShutdownEvent.addListener(this);
  },

  /**
   * 一意性を確保するためにファイル名を正規化する
   *
   * @param   {string} original
   *          ファイル名
   * @returns {string}
   *          正規化したファイル名
   */
  normalizeName: function(original) {
    if (!original) {
      return "";
    }

    original = arPathUtils.unique(original);

    for (let i = 0; i < 10; i += 1) {
      let last = original;
      try {
        original = decodeURIComponent(original);
      } catch (e) {
        original = lazy.arMIMEDecoder.decodeExtOctet(original);
      }
      original = original
        .replace(/%u([A-Fa-f0-9]{4})/g,
                 (matched, octet) => {
                   return String.fromCharCode(Number.parseInt(octet, 16));
                 });
      original = arDOMUtils.unescapeEntity(original);
      original = original.replace(/&#([0-9]+);/g, (m, n) => {
        return String.fromCharCode(Number.parseInt(n, 10));
      }).replace(/&#x([0-9]+);/g, (m, n) => {
        return String.fromCharCode(Number.parseInt(n, 16));
      });
      if (last == original) {
        break;
      }
    }

    return original;
  },

  /**
   * キャッシュに展開情報を格納する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  add: function(eFileInfo) {
    let normalizedBaseURI = this.normalizeName(eFileInfo.baseURI);

    this._eFileInfoMap.set(normalizedBaseURI, eFileInfo);
  },

  /**
   * キャッシュから展開情報を返す
   *
   * @param   {string} baseURI
   *          展開情報の baseURI
   * @returns {?UnMHTExtractFileInfo}
   *          展開情報
   *          見付からなければ null
   */
  get: function(baseURI) {
    let normalizedBaseURI = this.normalizeName(baseURI);

    if (!this._eFileInfoMap.has(normalizedBaseURI)) {
      return null;
    }

    let eFileInfo = this._eFileInfoMap.get(normalizedBaseURI);
    eFileInfo.lastAccessDate = new Date();

    return eFileInfo;
  },

  /**
   * キャッシュから古い展開情報を削除する
   *
   * @param   {number=} expire
   *          有効期限 [ms]
   */
  removeOld: function(expire=300000) {
    let opened = new Set();
    for (let window of arWindowUtils.getBrowserWindows()) {
      for (let browser of arWindowUtils.getBrowsers(window)) {
        let location;
        try {
          location = browser.currentURI.spec;
        } catch (e) {
          /* 対象が生きていない */
          continue;
        }

        let param = UnMHTScheme.getUnMHTURIParam(location);
        if (param.baseURI) {
          opened.add(param.baseURI);
        }
      }
    }

    let now = Date.now();
    for (let [uri, eFileInfo] of this._eFileInfoMap) {
      if (opened.has(eFileInfo.baseURI)) {
        continue;
      }

      if (now < eFileInfo.lastAccessDate.getTime() + expire) {
        continue;
      }

      eFileInfo.destruct();

      this._eFileInfoMap.delete(uri);
    }
  },

  /**
   * キャッシュから展開情報の配列を返す
   *
   * @returns {Array.<UnMHTExtractFileInfo>}
   *          展開情報
   */
  getList: function() {
    return [...this._eFileInfoMap.values()];
  },

  /**
   * unmht 形式の URI から展開情報を取得する
   *
   * @param   {string} uri
   *          unmht: 形式のファイルの URI
   * @returns {Tuple.<?UnMHTExtractFileInfo,?arMIMEPart>}
   *          展開情報と対応するパート
   *          見付からなかった場合はそれぞれ null
   */
  getPart: function(uri) {
    if (!UnMHTScheme.isUnMHTURI(uri)) {
      return [null, null];
    }

    let param = UnMHTScheme.getUnMHTURIParam(uri);
    let baseURI = param.baseURI;
    let refName = param.refName;

    let eFileInfo = this.get(baseURI);
    if (!eFileInfo) {
      return [null, null];
    }

    let part = null;
    if (refName) {
      if (eFileInfo.refParts.has(refName)) {
        part = eFileInfo.refParts.get(refName);
      }
    } else {
      part = eFileInfo.startPart;
    }

    return [eFileInfo, part];
  },

  /**
   * mht に含まれるファイルの URI を取得する
   *
   * @param   {string} contentLocation
   *          対象の URI
   * @param   {string} requestOrigin
   *          呼び出し元の URI
   *          unmht 形式
   * @returns {?string}
   *          ファイルの URI
   *          存在しなかった場合は null
   */
  getMHTFileURI: function(contentLocation, requestOrigin) {
    let [eFileInfo, part] = this.getPart(requestOrigin);
    if (!(eFileInfo && part)) {
      return null;
    }

    let [targetPart, fragment]
      = this.findPart(eFileInfo, part, contentLocation, false);
    if (!targetPart) {
      return null;
    }

    return eFileInfo.baseURI + targetPart.eParam.refName + fragment;
  },

  /**
   * Content-Location に対応するパスからパートを探す
   * 階層や multipart を考慮する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} originPart
   *          呼び出し元のパート
   * @param   {string} path
   *          対象のパス
   * @param   {boolean} referred
   *          referredBaseDir をチェックするか
   * @returns {Tuple.<?arMIMEPart,string>}
   *          対象のパートとフラグメント(#つき)
   *          見付からなければパートが null
   */
  findPart: function(eFileInfo, originPart, path, referred) {
    let [targetPart, fragment]
      = this.findPartSimple(eFileInfo, originPart, path, referred);
    if (targetPart) {
      let parentPath = targetPart.eParam.parentPath;
      if (originPart.isMixed) {
        if (!originPart.eParam.mixedPath.some(path => path.startsWith(parentPath))) {
          targetPart = null;
        }
      } else if (!originPart.eParam.path.startsWith(parentPath)) {
        targetPart = null;
      }
    }

    if (targetPart) {
      if (targetPart.isMultipart) {
        targetPart = targetPart.eParam.startPart;
      }
    }

    return [targetPart, fragment];
  },

  /**
   * Content-Location に対応するパスからパートを探す
   * 階層や multipart は考慮しない
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} originPart
   *          呼び出し元のパート
   * @param   {string} path
   *          対象のパス
   * @param   {boolean} referred
   *          referredBaseDir をチェックするか
   * @returns {Tuple.<?arMIMEPart,string>}
   *          対象のパートとフラグメント(#つき)
   *          見付からなければパートが null
   */
  findPartSimple: function(eFileInfo, originPart, path, referred) {
    let sharpFragment = "";
    let fragment = "";
    [path, fragment] = arPathUtils.splitFragment(path);
    if (fragment) {
      sharpFragment = "#" + fragment;
    }
    if (fragment && path === "") {
      return [originPart, sharpFragment];
    }

    let normalFragment = this.normalizeName(fragment);

    let normalPath = this.normalizeName(path);

    let part;

    /* Content-ID でチェック */
    let m = normalPath.match(/^cid:(.+)$/);
    if (m) {
      let cid = m[1];

      part
        = eFileInfo.parts
        .find(part => cid == part.contentID);
      if (part) {
        return [part, sharpFragment];
      }
    }

    /* フルパスでの参照をチェック */

    /* フラグメント付き */
    if (normalFragment) {
      part
        = eFileInfo.parts
        .find(part => (normalPath == part.eParam.normalLocation &&
                       normalFragment == part.eParam.normalFragment));
      if (part) {
        return [part, ""];
      }
    }
    /* フラグメント無し */
    part
      = eFileInfo.parts
      .find(part => normalPath == part.eParam.normalLocation);
    if (part) {
      return [part, sharpFragment];
    }

    if (originPart.eParam.baseDir) {
      /* フルパス以外での参照をチェック */
      let resolvedPath = arPathUtils.resolve(originPart.eParam.baseDir,
                                             normalPath);
      let normalResolved = this.normalizeName(resolvedPath);

      /* フラグメント付き */
      if (normalFragment) {
        part
          = eFileInfo.parts
          .find(part => (normalResolved == part.eParam.normalLocation &&
                         normalFragment == part.eParam.normalFragment));
        if (part) {
          return [part, ""];
        }
      }
      /* フラグメント無し */
      part
        = eFileInfo.parts
        .find(part => normalResolved == part.eParam.normalLocation);
      if (part) {
        return [part, sharpFragment];
      }
    }

    /* unmht: を元に戻してチェック */
    if (UnMHTScheme.isUnMHTURI(normalPath)) {
      let param = UnMHTScheme.getUnMHTURIParam(normalPath);
      let refName = param.refName;

      if (param.original == eFileInfo.original) {
        if (!refName) {
          return [eFileInfo.startPart, sharpFragment];
        }

        if (eFileInfo.refParts.has(refName)) {
          return [eFileInfo.refParts.get(refName), sharpFragment];
        }

        return [null, ""];
      }
    }

    /* IE のバグを考慮してチェック */
    if (referred &&
        originPart.eParam.referredBaseDir !== "") {
      let referredPath = arPathUtils.resolve(originPart.eParam.referredBaseDir,
                                             normalPath);
      let normalReferred = this.normalizeName(referredPath);

      part
        = eFileInfo.parts
        .find(part => normalReferred == part.eParam.normalLocation);
      if (part) {
        return [part, sharpFragment];
      }
    }

    /* 含まれない */
    return [null, ""];
  },

  /**
   * キャッシュページを作成する
   *
   * @returns {string}
   *          キャッシュページのソース
   */
  createDocument: function() {
    let tableRow = (id, header, data) => {
      return (`  <tr id="${id}_row">
    <th id="${id}_header">${header}</th>
    <td id="${id}_data">${data}</td>
  </tr>`);
    };
    let ESCAPE = text => arDOMUtils.escapeEntity(text);
    let ETEXT = name => ESCAPE(TEXT(name));

    let size = 0;

    let eFileInfoList = this.getList()
      .sort((x, y) => y.lastAccessDate.getTime() - x.lastAccessDate.getTime());

    let bodies
      = eFileInfoList.map((eFileInfo, i) => {
        let body = [];

        let _row = (id, header, data) => {
          body.push(tableRow("entry_" + i + "_" + id, header, data));
        };

        size += eFileInfo.size;

        let param = UnMHTScheme.getUnMHTURIParam(eFileInfo.baseURI);
        let original = param.original;
        if (param.scheme == "file") {
          original = arFileUtils.URLSpecToPath(original);
        }

        body.push(`<table class="file">`);

        _row("uri", "URI:",
             `<a href="${ESCAPE(eFileInfo.baseURI)}">${ESCAPE(eFileInfo.baseURI)}</a>`);
        _row("original_uri", ETEXT("info.original_address"),
             ESCAPE(original));
        _row("opened_time", ETEXT("cache.opened_time"),
             ESCAPE(arFormatter.date(eFileInfo.extractDate)));
        _row("access_time", ETEXT("cache.last_access_time"),
             ESCAPE(arFormatter.date(eFileInfo.lastAccessDate)));
        _row("part_count", ETEXT("cache.file_count"),
             eFileInfo.parts.length);

        if (eFileInfo.date) {
          let savedTime = arFormatter.date(eFileInfo.date);

          _row("saved_date", ETEXT("info.saved_date"),
               ESCAPE(savedTime));
        }

        if (eFileInfo.topPart.hasField("X-UnMHT-Save-State")) {
          let state = UnMHTSaveState.UNKNOWN;

          switch (eFileInfo.topPart.getField("X-UnMHT-Save-State")) {
            case "Current-State": {
              state = UnMHTSaveState.CURRENT;
              break;
            }
            case "Current-State-Without-Script": {
              state = UnMHTSaveState.CURRENT_NOJS;
              break;
            }
            case "Original-File": {
              state = UnMHTSaveState.ORIGINAL;
              break;
            }
          }

          let saveState = UnMHTSaveStateUtils.saveStateToReadableString(state);

          if (saveState) {
            _row("save_state", ETEXT("info.save_state"),
               ESCAPE(saveState));
          }
        }

        if (eFileInfo.subject) {
          _row("subject", "Subject:",
               ESCAPE(eFileInfo.subject));
        }

        if (eFileInfo.topPart.hasField("From")) {
          let from = eFileInfo.topPart.getField("From");
          _row("from", "From:",
               ESCAPE(from));
        }

        if (eFileInfo.topPart.hasField("X-MAF-Information")) {
          _row("maf_information", "X-MAF-Information:",
               ESCAPE(eFileInfo.topPart.getField("X-MAF-Information")));
        }
        if (eFileInfo.topPart.hasField("X-MimeOLE")) {
          _row("ole_information", "X-MimeOLE:",
               ESCAPE(eFileInfo.topPart.getField("X-MimeOLE")));
        }


        _row("file_size", ETEXT("cache.file_size"),
             arFormatter.size(eFileInfo.source.length)
             + " " + ETEXT("cache.bytes"));

        body.push("</table>");

        eFileInfo.parts.forEach((part, j) => {
          let _row2 = (id, header, data) => {
            _row("part_" + j + "_" + id, header, data);
          };

          body.push(`<table class="part">`);
          if (part.eParam.refName) {
            let href = ESCAPE(eFileInfo.baseURI) + ESCAPE(part.eParam.refName);
            let start = "";
            if (eFileInfo.startPart == part) {
              start = " (" + ETEXT("cache.start_part") + ")";
            }
            _row2("refname", ETEXT("cache.ref_name"),
                  `<a href="${href}">${ESCAPE(part.eParam.refName)}</a>${start}`);
          }
          if (part.contentLocation) {
            _row2("content_location", "Content-Location:",
                  ESCAPE(part.contentLocation));
          }
          if (part.contentID) {
            _row2("content_id", "Content-ID:",
                  ESCAPE(part.contentID));
          }
          if (part.contentDispositionType &&
              part.contentDispositionFilename) {
            _row2("content_disposition", "Content-Disposition:",
                  ESCAPE(part.contentDispositionType)
                  + "; filename=" + ESCAPE(part.contentDispositionFilename));
          }
          _row2("content-type", "Content-Type:",
                ESCAPE(part.eParam.mimetype)
                + (part.eParam.charset
                   ? ("; charset=" + ESCAPE(part.eParam.charset))
                   : ""));
          if (part.contentTransferEncoding) {
            _row2("content-encoding", "Content-Encoding:",
                  ESCAPE(part.contentTransferEncoding));
          }
          if (part.hasField("X-UnMHT-Last-Modified")) {
            try {
              let lastModified = new Date(part.getField("X-UnMHT-Last-Modified").replace(/^\"/, "").replace(/\"$/, ""));
              if (!Number.isNaN(lastModified.getFullYear())) {
                _row("ole_information", ETEXT("cache.last_modified"),
                     ESCAPE(arFormatter.date(lastModified)));
              }
            } catch (e) {
            }
          }

          _row2("file_size", ETEXT("cache.file_size"),
                arFormatter.size(part.eParam.content.length)
                + " " + ETEXT("cache.bytes"));
          body.push("</table>");
        });

        return body.join("\n");
      });

    let css = "resource://unmht/css/cache.css";

    let css2
      = arAppInfo.isFennec
      ? "resource://unmht/css/cache_fennec.css"
      : "resource://unmht/css/cache_firefox.css";

    let data
      = `<!DOCTYPE html>
<html>
<head>
  <meta charset="UTF-8">
  <title>${ETEXT("cache.title")}</title>
  <link rel="stylesheet" href="${css}">
  <link rel="stylesheet" href="${css2}">
</head>
<body>
<h1>${ETEXT("cache.title")}</h1>
<table id="entries">
${tableRow("cache_count", ETEXT("cache.entry"), eFileInfoList.length)}
${tableRow("cache_size", ETEXT("cache.size"), arFormatter.size(size) + " " + ETEXT("cache.bytes"))}
</table>
<hr>
${bodies.join("<hr>")}
</body>
</html>
`;

    return toUTF8(data);
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    for (let eFileInfo of this._eFileInfoMap.values()) {
      eFileInfo.destruct();
    }
    this._eFileInfoMap.clear();
  },

  /* ==== private ==== */

  /**
   * 正規化した baseURI と展開情報のマップ
   * @type {Map.<string,UnMHTExtractFileInfo>}
   */
  _eFileInfoMap: new Map()
});
UnMHTCache.init();

module.UnMHTCache = UnMHTCache;

})(this);
