"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTConfirmDialog"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { arDialog, arDialogClass } = Cu.import("resource://unmht/modules/base/arDialog.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arPrototype } = Cu.import("resource://unmht/modules/base/arPrototype.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});

/**
 * 汎用確認ダイアログクラス
 *
 * ウィンドウの引数 {Object}
 *   {
 *     focus:    {string=}
 *               デフォルトでフォーカスする対象 (オプショナル)
 *                 (yes/no/abort/skip/rename),
 *
 *     title:    {string}
 *               ダイアログのタイトル,
 *     message:  {string=}
 *               ダイアログのメッセージ,
 *
 *     log:      {string=}
 *               ログ等の複数行テキスト,
 *
 *     accept:   {string}
 *               OK ボタン [yes] のラベル,
 *     cancel:   {string=}
 *               キャンセルボタン [no] のラベル,
 *     extra1:   {string=}
 *               スキップボタン [skip] のラベル,
 *     extra2:   {string=}
 *               中断ボタン [abort] のラベル,
 *     toall:    {string=}
 *               全てに適用チェックのラベル,
 *
 *     rename:   {string=}
 *               リネームボタン [rename] のラベル,
 *     filePath: {string=}
 *               既存のファイルのパス (rename により必須),
 *     exist:    {(function(string):string)=}
 *               ファイルが存在するメッセージのフォーマッタ (rename により必須)
 *                 function(leafName)
 *                   *param   {string} leafName
 *                            ファイル名
 *                   *returns {string}
 *                            ファイルが存在するメッセージ
 *   }
 *
 * 結果 {Promise.<Object>}
 * @resolves {
 *     action:     {string}
 *                 押したボタン (yes/no/abort/skip/rename),
 *     applyToAll: {boolean}
 *                 全てに適用するか,
 *     renamePath: {?string}
 *                 リネームするファイルのパス
 *   }
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialogClass
 * @implements arIDialog
 * @implements arIDialogExtra
 */
function UnMHTConfirmDialogClass(window) {
  /**
   * 既存のファイルのパス
   * @type {?string}
   */
  this._filePath = null;

  /**
   * ファイルが存在するメッセージのフォーマッタ
   * @type {?function(string):string}
   */
  this._existFormatter = null;

  arDialogClass.call(this, window);
}
UnMHTConfirmDialogClass.prototype = Object.freeze({
  /**
   * 各種ボタンを押したイベント
   *
   * @param   {string} type
   *          ボタンの種類 (yes/no/abort/skip/rename)
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onClick: function(type) {
    this.$("warning").value = "";
    this.$("warning").hidden = true;

    let result = {
      action: type,
      applyToAll: this.$("apply_to_all").checked,
      renamePath: null
    };

    if (type == "rename") {
      Task.spawn(function*() {
        let parent = OS.Path.dirname(this._filePath);
        let base = UnMHTFilename.removeExtension(this.$("filename").value);
        let suffix = "." + UnMHTFilename.defaultExtension;
        let options = { allowExists: true, noCreate: true };
        let pResult = yield arFileUtils.prepareFile(parent, base, suffix,
                                                   options);
        if (pResult.exists) {
          this.$("warning").value
            = this._existFormatter(OS.Path.basename(pResult.path));
          this.$("warning").hidden = false;
          this.$("filename").focus();
          return;
        }

        result.renamePath = pResult.path;

        this.resolve(result);

        this._window.close();
      }.bind(this)).catch(e => {
        Cu.reportError(e + "\n" + e.stack);
      });

      return false;
    }

    this.resolve(result);

    if (type == "abort" ||
        type == "skip") {
      /* ダイアログのボタンではないので手動で閉じる */
      this._window.close();
    }

    return true;
  },

  /* ==== arIDialog ==== */

  /**
   * 破棄
   */
  destruct: function() {
    this._filePath = null;
    try {
      arDialogClass.prototype.destruct.call(this);
    } catch (e) {
      /* モジュールのアンロード後 */
    }
  },

  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this.arg;

    let dialog = this.$("unmht_confirm");

    dialog.setAttribute("title", arg.title);

    if ("message" in arg && arg.message) {
      this.$("message").appendChild(this._window.document.createTextNode(arg.message));
      this.$("message").hidden = false;
    }

    if ("log" in arg && arg.log) {
      this.$("log").value = arg.log;
      this.$("log").hidden = false;
    }

    dialog.getButton("accept").setAttribute("label", arg.accept);

    if ("cancel" in arg && arg.cancel) {
      dialog.getButton("cancel").setAttribute("label", arg.cancel);
    } else {
      dialog.getButton("cancel").hidden = true;
    }

    if ("extra1" in arg && arg.extra1) {
      dialog.getButton("extra1").setAttribute("label", arg.extra1);
    } else {
      dialog.getButton("extra1").hidden = true;
    }

    if ("extra2" in arg && arg.extra2) {
      dialog.getButton("extra2").setAttribute("label", arg.extra2);
    } else {
      dialog.getButton("extra2").hidden = true;
    }

    if ("toall" in arg && arg.toall) {
      this.$("apply_to_all").checked = false;
      this.$("apply_to_all").label = arg.toall;
      this.$("apply_to_all").hidden = false;
    }

    if ("rename" in arg && arg.rename) {
      this.$("rename").label = arg.rename;

      if ("filePath" in arg && arg.filePath) {
        this.$("filename").value = OS.Path.basename(arg.filePath);
        this._filePath = arg.filePath;
      }

      this.$("renamebox").hidden = false;
    }

    if ("exist" in arg && arg.exist) {
      this._existFormatter = arg.exist;
    }

    if ("focus" in arg && arg.focus) {
      switch (arg.focus) {
        case "yes": {
          dialog.getButton("accept").focus();
          break;
        }
        case "no": {
          dialog.getButton("cancel").focus();
          break;
        }
        case "abort": {
          dialog.getButton("extra2").focus();
          break;
        }
        case "skip": {
          dialog.getButton("extra1").focus();
          break;
        }
        case "rename": {
          this.$("rename").focus();
          break;
        }
      }
    }

  },

  /**
   * ダイアログを OK したイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    try {
      return this.onClick("yes");
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }
    return false;
  },

  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    try {
      return this.onClick("no");
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }
    return false;
  },

  /* ==== arIDialogExtra ==== */

  /**
   * ダイアログの extra1 を押したイベント
   */
  onDialogExtra1: function() {
    try {
      this.onClick("skip");
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }
  },

  /**
   * ダイアログの extra2 を押したイベント
   */
  onDialogExtra2: function() {
    try {
      this.onClick("abort");
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }
  }
});
arPrototype.extend(UnMHTConfirmDialogClass, arDialogClass);

/**
 * 汎用確認ダイアログ
 * @type {arDialog}
 */
let UnMHTConfirmDialog = new arDialog(UnMHTConfirmDialogClass, {
  url: "chrome://unmht/content/confirm.xul",
  features: "chrome,titlebar,centerscreen",
  single: false,
  usePromise: true
});

module.UnMHTConfirmDialog = UnMHTConfirmDialog;

})(this);
