"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTContentModifier"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { arPathUtils } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { arUconv } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { UnMHTCache } = Cu.import("resource://unmht/modules/UnMHTCache.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

/**
 * コンテンツを修正する
 *
 * @class
 */
let UnMHTContentModifier = Object.freeze({
  /**
   * コンテンツを修正する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} part
   *          対象のパート
   */
  modifyContents: function(eFileInfo, part) {
    this._modifyAttribute(eFileInfo, part);

    this._modifyCSS(eFileInfo, part);
  },

  /**
   * コンテンツの修正を解除する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {string} content
   *          対象のドキュメント
   * @param   {string} charset
   *          対象のドキュメントの charset
   * @returns {string}
   *          変換したドキュメント
   */
  unmodifyContents: function(eFileInfo, content, charset) {
    content = this._unmodifyAttribute(eFileInfo, content, charset);

    content = this._unmodifyCSS(eFileInfo, content, charset);

    return content;
  },

  /* ==== private ==== */

  /*
   * URL-rule = URL-rule-open ( quoted-string | URL-paren ) URL-rule-close
   * URL-rule-open  = <any char except ALPHA, DIGIT, "_"> "url" "("
   * URL-rule-close = ")"
   *
   * import-no-URL-rule = "@import" quoted-string
   *
   * quoted-string = dquoted-string | squoted-string
   *
   * duoted-string = ( ( DQUOTE   dquoted-string-body DQUOTE )
   *                   ( "&quot;" dquoted-string-body "&quot;" )
   *                   ( "&quot;" dquoted-string-body "&quot;" )
   *                   ( "&#x22;" dquoted-string-body "&#x22;" )
   *                   ( "&#34;"  dquoted-string-body "&#34;" ) )
   * dquoted-string-body = *( ( "\" DQUOTE ) | dquoted--char )
   * dquoted--char = <any char except DQUOTE>
   *
   * suoted-string = ( ( QUOTE squoted-string-body QUOTE )
   *                   ( "&apos;" squoted-string-body "&apos;" )
   *                   ( "&#x27;" squoted-string-body "&#x27;" )
   *                   ( "&#39;" squoted-string-body "&#39;" ) )
   * squoted-string-body = *( ( "\" QUOTE ) | squoted--char )
   * squoted--char = <any char except QUOTE>
   *
   * URL-paren = +URL-paren-char
   * URL-paren-char = <any char except DQUOTE, QUOTE, "(", ")">
   */

  /*
   * 1: url(
   * 2: dquote
   * 3: dquoted-string-body
   * 4: squote
   * 5: squoted-string-body
   * 6: URL
   * 7: )
   */
  _cssURLRe: /([^a-zA-Z0-9_]url\((?:\s+|\/\*(?:.|\r|\n)*?\*\/)*)(?:(\"|&quot;|&#x22;|&#34;)((?:\\\"|[^\"])*?)\2|(\'|&apos;|&#x27;|&#39;)((?:\\\'|[^\'])*?)\4|([^\"\'\(\)]+))((?:\s+|\/\*(?:.|\r|\n)*?\*\/)*\))/ig,

  /*
   * 1: @import
   * 2: dquote
   * 3: dquoted-string-body
   * 4: squote
   * 5: squoted-string-body
   */
  _cssImportRe: /(@import(?:\s+|\/\*(?:.|\r|\n)*?\*\/)*)(?:(\"|&quot;|&#x22;|&#34;)((?:\\\"|[^\"])*?)\2|(\'|&apos;|&#x27;|&#39;)((?:\\\'|[^\'])*?)\4)/ig,

  _elemRe: /(<([\?A-Za-z0-9_:\-]+))(\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)(>)/g,
  _attrRe: /(\s(class|src|href|background|action|data|integrity)\s*=\s*)(?:(\")((?:\\\"|[^\"])*)\"|(\')((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))/ig,
  _baseRe: /<base\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?href\s*=\s*(?:\"((?:\\\"|[^\"])*)\"|\'((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*>/ig,
  _pptRe: /<\/?(o|p|oa|v):(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?>/g,

  /**
   * CSS 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} part
   *          対象のパート
   * @param   {boolean} isURL
   *          url() か
   * @param   {string} matched
   *          マッチした全体
   * @param   {string} prev
   * @param   {?string} dquote
   * @param   {?string} dqpath
   * @param   {?string} quote
   * @param   {?string} qpath
   * @param   {?string} path
   * @param   {?string} next
   *          以下参照
   *
   *         dquote
   *         -
   * @import "12345.jpg";
   * -------- ---------
   * prev     dqpath
   *
   *      quote      next
   *      -          -
   * url ('12345.jpg');
   * ----- ---------
   * prev  qpath
   *
   * @returns {string}
   *          変換した結果
   */
  _cssFunc: function(eFileInfo, part, isURL,
                     matched, prev, dquote, dqpath, quote, qpath,
                     path, next) {
    if (!isURL) {
      path = "";
      next = "";
    }
    if (dqpath) {
      path = dqpath.replace(/\\\"/g, "\"").replace(/\\\\/g, "\\");
    } else if (qpath) {
      path = qpath.replace(/\\\'/g, "\'").replace(/\\\\/g, "\\");
    } else {
      path = path || "";
      path = path.replace(/ +$/, "");
    }
    quote = quote || dquote || "";

    path = arUconv.toUnicode(path, part.eParam.charset);

    if (path.startsWith("mailto:") || path.startsWith("data:")) {
      return matched;
    }

    let [targetPart, fragment]
      = UnMHTCache.findPart(eFileInfo, part, path, part.eParam.isCSS);
    if (targetPart) {
      targetPart.eParam.referredBaseDir = part.eParam.baseDir;
      if (targetPart == part && fragment) {
        return prev + quote
          + arUconv.fromUnicode(fragment, part.eParam.charset)
          + quote + next;
      }

      let location = eFileInfo.baseURI + targetPart.eParam.refName + fragment;
      return prev + quote
        + arUconv.fromUnicode(location, part.eParam.charset)
        + quote + next;
    }

    if (path.startsWith("file:")) {
      return prev + quote + quote + next;
    }

    return matched;
  },

  /**
   * CSS 中の unmht: 形式の URI を解除する
   * replace に渡す関数
   *
   * 引数と返り値は part が charset になる以外は _cssFunc と同じ
   */
  _unCssFunc: function(eFileInfo, charset, isURL,
                       matched, prev, dquote, dqpath, quote, qpath,
                       path, next) {
    if (!isURL) {
      path = "";
      next = "";
    }
    if (dqpath) {
      path = dqpath.replace(/\\\"/g, "\"").replace(/\\\\/g, "\\");
    } else if (qpath) {
      path = qpath.replace(/\\\'/g, "\'").replace(/\\\\/g, "\\");
    } else {
      path = path || "";
      path = path.replace(/ +$/, "");
    }
    quote = quote || dquote || "";

    path = arUconv.toUnicode(path, charset);

    if (!path.startsWith("unmht:")) {
      return matched;
    }

    let param = UnMHTScheme.getUnMHTURIParam(path);
    let refName = param.refName;
    let targetPart = null;
    if (refName) {
      if (eFileInfo.refParts.has(refName)) {
        targetPart = eFileInfo.refParts.get(refName);
      }
    } else {
      targetPart = eFileInfo.startPart;
    }
    if (targetPart) {
      let location = targetPart.contentLocation + (param.fragment
                                                   ? ("#" + param.fragment)
                                                   : "");
      if (!location) {
        location = "cid:" + targetPart.eParam.cid;
      }

      return prev + quote
        + arUconv.fromUnicode(location, charset)
        + quote + next;
    }

    return matched;
  },

  /**
   * CSS 中の URI を unmht: 形式に変換する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} part
   *          対象のパート
   */
  _modifyCSS: function(eFileInfo, part) {
    if (part.eParam.isHTML ||
        part.eParam.isCSS) {
      part.eParam.content
        = part.eParam.content
        .replace(this._cssURLRe, this._cssFunc.bind(this, eFileInfo,
                                                    part, true))
        .replace(this._cssImportRe, this._cssFunc.bind(this, eFileInfo,
                                                       part, false));
    }
  },

  /**
   * HTML 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} part
   *          対象のパート
   * @param   {string} matched
   *          マッチした全体
   * @param   {string} prev
   * @param   {string} attrs
   * @param   {string} next
   * @returns {string}
   *          変換した結果
   */
  _elemFunc: function(eFileInfo, part,
                      matched, prev, name, attrs, next) {
    if (/^base$/i.test(name)) {
      return matched;
    }
    let isObject = false;
    if (/^object$/i.test(name) || /^embed$/i.test(name)) {
      isObject = true;
    }
    let isAnchor = false;
    if (/^a$/i.test(name)) {
      isAnchor = true;
    }
    let isLink = false;
    if (/^link$/i.test(name)) {
      isLink = true;
    }
    let isOriginalLink = false;

    attrs
      = attrs
      .replace(this._attrRe, (matched, prev, attr,
                              dquote, dqpath, quote, qpath, path) => {
        if (dqpath) {
          path = dqpath.replace(/\\\"/g, "\"");
        } else if (qpath) {
          path = qpath.replace(/\\\'/g, "\'");
        } else {
          path = path || "";
        }
        quote = quote || dquote || "\"";

        path = arUconv.toUnicode(path, part.eParam.charset);

        if (attr == "integrity") {
          /* unmht:// スキーム内では SRI は無効化 */
          return "";
        }

        if (attr == "class") {
          if (isAnchor && path == "unmht_link_to_original") {
            isOriginalLink = true;
          }

          return matched;
        }

        if (isOriginalLink && attr == "href") {
          return matched;
        }

        if (path.startsWith("mailto:") || path.startsWith("data:")) {
          return matched;
        }

        if (!isObject && /^data$/i.test(attr)) {
          return matched;
        }

        let isRelative = false;
        if (!/^[A-Za-z0-9\-]+:/.test(path)) {
          isRelative = true;
        }
        if (isRelative && !isAnchor && !isLink) {
          /* 相対パスは newURI の方で変える */
          return matched;
        }

        let [targetPart, fragment]
          = UnMHTCache.findPart(eFileInfo, part, path, false);
        if (targetPart) {
          if (targetPart == part && fragment) {
            return prev + quote
              + arUconv.fromUnicode(fragment, part.eParam.charset)
              + quote;
          }

          if (isLink) {
            targetPart.eParam.referredBaseDir = part.eParam.baseDir;
          }

          let location = eFileInfo.baseURI + targetPart.eParam.refName + fragment;
          return prev + quote
            + arUconv.fromUnicode(location, part.eParam.charset)
            + quote;
        }

        if (path.startsWith("file:") ||
            path.startsWith("cid:")) {
          return prev + quote + quote;
        }

        if (isRelative && isAnchor) {
          /* リンク先はあらかじめ展開しておく */
          try {
            let resolved = Services.io.newURI(part.eParam.baseDir, null, null).resolve(path);
            return prev + quote
              + arUconv.fromUnicode(resolved, part.eParam.charset)
              + quote;
          } catch (e) {
            /* 展開失敗した場合は諦める */
          }
        }

        return matched;
      });

    return prev + attrs + next;
  },

  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   * replace に渡す関数
   *
   * 引数と返り値は part が charset になる以外は _elementFunc と同じ
   */
  _unElemFunc: function(eFileInfo, charset,
                        matched, prev, name, attrs, next) {
    let isBase = false;
    if (/^base$/i.test(name)) {
      isBase = true;
    }

    attrs
      = attrs
      .replace(this._attrRe, (matched, prev, attr,
                              dquote, dqpath, quote, qpath, path) => {
        if (dqpath) {
          path = dqpath.replace(/\\\"/g, "\"");
        } else if (qpath) {
          path = qpath.replace(/\\\'/g, "\'");
        } else {
          path = path || "";
        }
        quote = quote || dquote || "\"";

        path = arUconv.toUnicode(path, charset);

        if (!path.startsWith("unmht:")) {
          return matched;
        }

        let param = UnMHTScheme.getUnMHTURIParam(path);
        let refName = param.refName;
        let targetPart = null;
        if (refName) {
          if (eFileInfo.refParts.has(refName)) {
            targetPart = eFileInfo.refParts.get(refName);
          }
        } else {
          targetPart = eFileInfo.startPart;
        }
        if (targetPart) {
          let location = targetPart.contentLocation;
          if (isBase) {
            location = targetPart.eParam.baseDir;
          }
          if (param.fragment) {
            location += "#" + param.fragment;
          }
          if (!location && !isBase) {
            location = "cid:" + targetPart.eParam.cid;
          }

          return prev + quote
            + arUconv.fromUnicode(location, charset)
            + quote;
        }

        return matched;
      });

    return prev + attrs + next;
  },

  /**
   * HTML 中の BASE の値を収集して削除する
   * replace に渡す関数
   *
   * 引数と返り値は _elemFunc と同じ
   */
  _baseFunc: function(eFileInfo, part,
                      matched, dqpath, qpath, path) {
    if (dqpath) {
      path = dqpath.replace(/\\\"/g, "\"");
    } else if (qpath) {
      path = qpath.replace(/\\\'/g, "\'");
    } else {
      path = path || "";
    }

    path = arUconv.toUnicode(path, part.eParam.charset);
    /* BASE の属性は必ずディレクトリであると仮定する */
    if (!path.endsWith("/") && !path.endsWith("\\")) {
      path += "/";
    }
    let href = arPathUtils.getBaseDir(path);
    part.eParam.baseDir = arPathUtils.resolve(part.eParam.baseDir, href);

    return "";
  },

  /**
   * HTML 中の URI を unmht: 形式に変換する
   * また、BASE 要素と Office 用の要素を削除する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} part
   *          対象のパート
   */
  _modifyAttribute: function(eFileInfo, part) {
    if (part.eParam.isHTML) {
      part.eParam.content
        = part.eParam.content
        .replace(this._baseRe, this._baseFunc.bind(this, eFileInfo, part))
        .replace(this._elemRe, this._elemFunc.bind(this, eFileInfo, part));

      if (part.eParam.content.indexOf("xmlns:v=\"urn:schemas-microsoft-com:vml\"") != -1) {
        part.eParam.content = part.eParam.content.replace(this._pptRe, "");
      }
    }
  },

  /**
   * CSS 中の unmht: 形式の URI を解除する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {string} content
   *          対象のドキュメント
   * @param   {string} charset
   *          対象のドキュメントの charset
   * @returns {string}
   *          変換したドキュメント
   */
  _unmodifyCSS: function(eFileInfo, content, charset) {
    return content
      .replace(this._cssURLRe, this._unCssFunc.bind(this, eFileInfo,
                                                    charset, true))
      .replace(this._cssImportRe, this._unCssFunc.bind(this, eFileInfo,
                                                       charset, false));
  },

  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {string} content
   *          対象のドキュメント
   * @param   {string} charset
   *          対象のドキュメントの charset
   * @returns {string}
   *          変換したドキュメント
   */
  _unmodifyAttribute: function(eFileInfo, content, charset) {
    return content.replace(this._elemRe,
                           this._unElemFunc.bind(this, eFileInfo, charset));
  }
});

module.UnMHTContentModifier = UnMHTContentModifier;

})(this);
