"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTContentPolicy"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTLoadModifier",
                                  "resource://unmht/modules/UnMHTLoadModifier.jsm");

/**
 * mht から参照されているファイルのチェック、変換を行う
 *
 * @constructor
 * @implements nsIContentPolicy
 */
function UnMHTContentPolicy() {
  Object.seal(this);
}
UnMHTContentPolicy.prototype = Object.freeze({
  classID:           Components.ID("{5c21c8ee-2d1e-46c1-9d6f-bc4876c13890}"),
  contractID:        "@unmht.org/unmht-content-policy;1",
  classDescription:  "UnMHT Content Policy JS Component",
  _xpcom_categories: [{ category: "content-policy" }],
  QueryInterface:    XPCOMUtils.generateQI([Ci.nsIContentPolicy]),

  /**
   * nsIContentPolicy の定数
   * モジュールのアンロードからコンポーネントのアンロードまでの間
   * 外の環境が見えない状態で動くので保持しておく
   * @type {number}
   */
  ACCEPT:       Ci.nsIContentPolicy.ACCEPT,
  REJECT_OTHER: Ci.nsIContentPolicy.REJECT_OTHER,

  /* ==== nsIContentPolicy ==== */

  /**
   * ロードするかどうかのチェック
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {?nsIURI} requestOrigin
   *          参照元の URI (オプショナル)
   * @param   {?nsIDOMElement} context
   *          ロード先 (オプショナル)
   * @param   {?string} mimeTypeGuess
   *          予想される MIME-Type (オプショナル)
   * @param   {?nsISupports} extra
   *          付加データ (オプショナル)
   */
  shouldLoad: function(contentType, contentLocation,
                       requestOrigin, context,
                       mimeTypeGuess, extra) {
    if (this._shouldLoadPreCheck(contentType, contentLocation,
                                 requestOrigin, context)) {
      return this.ACCEPT;
    }

    return this.REJECT_OTHER;
  },

  /**
   * 処理するかどうかのチェック
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {?nsIURI} contentLocation
   *          対象の URI (オプショナル)
   * @param   {?nsIURI} requestOrigin
   *          参照元の URI (オプショナル)
   * @param   {?nsIDOMElement} context
   *          ロード先 (オプショナル)
   * @param   {string} mimeType
   *          MIME-Type
   * @param   {?nsISupports} extra
   *          付加データ (オプショナル)
   */
  shouldProcess: function(contentType, contentLocation,
                          requestOrigin, context,
                          mimeTypeGuess, extra) {
    return this.ACCEPT;
  },

  /* ==== private ==== */

  /**
   * shouldLoad のうち UnMHTContentPolicy で行うもの
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {nsIURI} requestOrigin
   *          参照元の URI
   * @param   {nsIDOMElement} context
   *          ロード先
   * @returns {boolean}
   *          許可するかどうか
   */
  _shouldLoadPreCheck: function(contentType, contentLocation,
                                requestOrigin, context) {
    if (!requestOrigin) {
      /* 呼び出し元が無い場合チェック不可能なので許可する */
      return true;
    }

    let contentScheme = contentLocation.scheme;

    if (contentScheme == "chrome" ||
        contentScheme == "resource" ||
        contentScheme == "data") {
      /* chrome, resource, data への参照は対象外 */
      return true;
    }

    let originScheme = requestOrigin.scheme;

    if (contentScheme == "unmht" &&
        UnMHTScheme.isUnMHTURI(contentLocation.spec)) {
      let cParam = UnMHTScheme.getUnMHTURIParam(contentLocation.spec);

      if (originScheme == "unmht") {
        /* unmht から unmht への読み込みは
         * 同じ MHT ファイル内、もしくはシステムスキームからのみ */
        if (!UnMHTScheme.isUnMHTURI(requestOrigin.spec)) {
          return false;
        }
        let oParam = UnMHTScheme.getUnMHTURIParam(requestOrigin.spec);
        if (cParam.original != oParam.original) {
          if (!oParam.scheme.startsWith("unmht-")) {
            return false;
          }
        }
      } else {
        /* 他のプロトコルから unmht への読み込みは
         * chrome (browser.xul など)
         * moz-nullprincipal ([背景画像だけを表示] など)
         * およびプロトコルが同じもののみ */
        if (!(cParam.scheme == originScheme ||
              originScheme == "chrome" ||
              originScheme == "moz-nullprincipal")) {
          return false;
        }
      }
    }

    if (originScheme != "unmht") {
      /* unmht 以外からの参照は対象外 */
      return true;
    }

    /* ---- ここから unmht からの参照 ---- */

    if (!context) {
      /* ロード先が無い場合 URI の置換えができないので許可する */
      return true;
    }

    if (context instanceof Ci.nsIDOMXULElement) {
      /* タブを開く場合対象外 */
      return true;
    }

    /* ---- ここからドキュメント内へのロード ---- */

    return lazy.UnMHTLoadModifier.shouldLoad(contentType, contentLocation,
                                             requestOrigin, context);
  }
});

module.UnMHTContentPolicy = UnMHTContentPolicy;

})(this);
