"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTContentSniffer"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});

/**
 * mht ファイルを検出する
 *
 * @constructor
 * @implements nsIContentSniffer
 */
function UnMHTContentSniffer() {
  this._mimeExtensions = null;

  this._loadMIMEExtentions();

  Object.seal(this);
}
UnMHTContentSniffer.prototype = Object.freeze({
  classID:           Components.ID("{4fc5587e-f418-4270-b384-0d4ca638fb1d}"),
  contractID:        "@unmht.org/unmht-content-sniffer;1",
  classDescription:  "UnMHT Content Sniffer JS Component",
  _xpcom_categories: [{ category: "net-content-sniffers" }],
  QueryInterface:    XPCOMUtils.generateQI([Ci.nsIContentSniffer]),

  /* ==== nsIContentSniffer ==== */

  /**
   * コンテントの MIME-Type を識別する
   *
   * @param   {nsIRequest} request
   *          データが送られてきたリクエスト
   * @param   {Object} data
   *          データ
   * @param   {number} length
   *          データの長さ
   * @returns {string}
   *          識別された MIME-Type
   */
  getMIMETypeFromContent: function(request, data, length) {
    if (!request || !request.name) {
      return "";
    }

    if (!request.name.startsWith("file://")) {
      /* 検出方法が適当なので file 以外では危険 */
      return "";
    }

    let mimetype = this._getMimeTypeFromLocation(request.name);
    if (!mimetype) {
      return "";
    }

    let head = String.fromCharCode.apply(String, Array.slice(data, 0, 1024));
    if (/MIME-Version: *1\.0/i.test(head)) {
      return mimetype;
    }

    return "";
  },

  /* ==== private ==== */

  /**
   * 拡張子から MIME-Type の関連付けのリストを読み込む
   */
  _loadMIMEExtentions: function() {
    let tmp = UnMHTPref.getPref("open.mime.extensions");
    let list = [];
    try {
      list = JSON.parse(tmp);
      if (!list.every(n => typeof(n[0]) == "string" && typeof(n[1]) == "string")) {
        list = [];
      }
    } catch (e) {
      list = [];
    }

    if (arAppInfo.isSeaMonkey) {
      list = list.filter(([ext, mimetype]) => ext != "eml");
    }

    this._mimeExtensions = list;
  },

  /**
   * 拡張子に対応する MIME-Type を返す
   *
   * @param   {string} location
   *          対象のドキュメントのアドレス
   * @returns {string}
   *          MIME-Type
   *          未知なら null
   */
  _getMimeTypeFromLocation: function(location) {
    return (this._mimeExtensions
            .find(([ext, mimetype]) => location.endsWith("." + ext))
            || [null, null])[1];
  }
});

module.UnMHTContentSniffer = UnMHTContentSniffer;

})(this);
