"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTCustomFilename"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { arFormatter } = Cu.import("resource://unmht/modules/base/arFormatter.jsm", {});
let { arPathUtils } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

/**
 * StringBundle 中の曜日の名前
 * @type {Array.<string>}
 */
const weekdays = Object.freeze([
  "sun",
  "mon",
  "tue",
  "wed",
  "thu",
  "fri",
  "sat"
]);

/**
 * カスタムファイル名フォーマッタ
 *
 * @class
 */
let UnMHTCustomFilename = Object.freeze({
  /**
   * カスタムファイル名をフォーマットする
   * 設定は自動取得する
   *
   * @param   {string} TITTLE
   *          ページのタイトル
   * @param   {string} FILENAME
   *          ページのファイル名
   * @param   {string} URL
   *          ページの URL
   * @param   {Date} date
   *          日時
   * @returns {string}
   *          フォーマットされた文字列
   */
  format: function(TITLE, FILENAME, URL, date) {
    return this.formatWith(TITLE, FILENAME, URL, date,
                           UnMHTPref.getPref("save.filename.custom"),
                           UnMHTPref.getPref("save.filename.custom.am"),
                           UnMHTPref.getPref("save.filename.custom.pm"));
  },

  /**
   * カスタムファイル名をフォーマットする
   *
   * @param   {string} TITTLE
   *          ページのタイトル
   * @param   {string} FILENAME
   *          ページのファイル名
   * @param   {string} URL
   *          ページの URL
   * @param   {Date} date
   *          日時
   * @param   {string} pattern
   *          カスタムファイル名フォーマット文字列
   * @param   {string} am
   *          a.m. の文字列
   * @param   {string} pm
   *          p.m. の文字列
   * @returns {string}
   *          フォーマットされた文字列
   */
  formatWith: function(TITLE, FILENAME, URL, date, pattern, am, pm) {
    let uri = Services.io.newURI(URL, null, null);

    let host = "";
    try {
      host = uri.host;
    } catch (e) {
      /* about:home 等は host をサポートしない */
    }

    let port = -1;
    try {
      port = uri.port;
    } catch (e) {
      /* about:home 等は port をサポートしない */
    }
    if (port == -1) {
      if (uri.scheme == "http") {
        port = 80;
      } else if (uri.scheme == "https") {
        port = 443;
      } else if (uri.scheme == "ftp") {
        port = 21;
      }
    }

    let path, fragment, query;
    [path, fragment] = arPathUtils.splitFragment(uri.path);
    let m = path.match(/^([^\?]*)\?([^\?]*)$/);
    if (m) {
      [, path, query] = m;
    }
    query = query || "";
    fragment = fragment || "";

    let decodedURL = URL;
    try {
      decodedURL = decodeURIComponent(URL);
    } catch (e) {
    }
    try {
      query = decodeURIComponent(query);
    } catch (e) {
    }
    try {
      fragment = decodeURIComponent(fragment);
    } catch (e) {
    }

    if (UnMHTPref.getPref("save.filename.custom.limit")) {
      let limitCount = UnMHTPref.getPref("save.filename.custom.limit.count");
      TITLE = TITLE.slice(0, limitCount);
      FILENAME = FILENAME.slice(0, limitCount);
    }

    let vars = {
      YYYY: date.getFullYear(),
      Y:    date.getFullYear() % 100,
      YY:   arFormatter.pad(2, date.getFullYear() % 100),
      M:    date.getMonth() + 1,
      MM:   arFormatter.pad(2, date.getMonth() + 1),
      MMM:  this._weekdaysShort[date.getDay()],
      MMMM: this._weekdaysLong[date.getDay()],
      D:    date.getDate(),
      DD:   arFormatter.pad(2, date.getDate()),
      H:    date.getHours(),
      HH:   arFormatter.pad(2, date.getHours()),
      h:    date.getHours() % 12,
      hh:   arFormatter.pad(2, date.getHours() % 12),
      m:    date.getMinutes(),
      mm:   arFormatter.pad(2, date.getMinutes()),
      s:    date.getSeconds(),
      ss:   arFormatter.pad(2, date.getSeconds()),
      ampm: date.getHours() >= 12 ? pm : am,

      TITLE:    TITLE,
      FILENAME: FILENAME,
      URL:      decodedURL,
      URL2:     decodedURL.replace(/^[A-Za-z0-9]+:\/\//, ""),
      SCHEME:   uri.scheme,
      HOST:     host,
      PORT:     port,
      FULLPATH: uri.path,
      PATH:     path,
      QUERY:    query,
      FRAGMENT: fragment
    };

    let rexp = new RegExp("%(" + Object.keys(vars).join("|") + ")%", "g");

    return pattern.replace(rexp, (matched, name) => vars[name]);
  },

  /* ==== private ==== */

  /**
   * 曜日 (省略)
   * @type {Array.<string>}
   */
  get _weekdaysShort() {
    return weekdays.map(n => TEXT("weekday." + n + ".short"));
  },

  /**
   * 曜日
   * @type {Array.<string>}
   */
  get _weekdaysLong() {
    return weekdays.map(n => TEXT("weekday." + n + ".long"));
  }
});

module.UnMHTCustomFilename = UnMHTCustomFilename;

})(this);
