"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTDLFRegisterer"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arDLFRegisterer } = Cu.import("resource://unmht/modules/base/arDLFRegisterer.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});
let { UnMHTDocumentLoaderFactory } = Cu.import("resource://unmht/modules/UnMHTDocumentLoaderFactory.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyServiceGetter(lazy, "MIMEService",
                                   "@mozilla.org/mime;1",
                                   "nsIMIMEService");

/**
 * 起動時に Document Loader Factory を登録する
 *
 * @constructor
 * @implements nsIObserver
 */
function UnMHTDLFRegistererClass() {
  /**
   * 登録完了を知らせる Promise の resolve 関数
   * @type {function(*)}
   */
  this._resolve = null;

  Object.seal(this);
}
UnMHTDLFRegistererClass.prototype = Object.freeze({
  QueryInterface:    XPCOMUtils.generateQI([Ci.nsIObserver]),

  /**
   * Document Loader Factory の登録を開始し、完了したら解決する Promise を返す
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   * @returns {Promise.<void>}
   * @resolves 登録完了
   */
  register: function(data, reason) {
    return new Promise((resolve, reject) => {
      this._resolve = resolve;

      let APP_STARTUP = 1;

      if (reason == APP_STARTUP) {
        /* profile-after-change イベント前に
         * nsIMIMEService.getTypeFromExtension を呼ぶと
         * いくつかのプラグインがクラッシュするので待つ */
        Services.obs.addObserver(this, "profile-after-change", false);
        arShutdownEvent.addFunc((data, reason) => {
          Services.obs.removeObserver(this, "profile-after-change");
        });
      } else {
        this._registerDLF();
      }
    });
  },

  /* ==== nsIObserver ==== */

  /**
   * 通知を受けたイベント
   *
   * @param   {nsISupports} subject
   *          変更やアクションのあったオブジェクト
   * @param   {string} topic
   *          イベントの種類
   * @param   {string} data
   *          イベントのデータ
   */
  observe: function(subject, topic, data) {
    if (topic == "profile-after-change") {
      /* MAF は SeaMonkey の組み込みのメール用の
       * Gecko-Content-Viewers: message/rfc822 を削除しようとするが
       * MAF より前に UnMHT のコードが走ると
       * UnMHT の Gecko-Content-Viewers: message/rfc822 が削除されるので
       * MAF より後に走らせるためにイベントを抜けるまで待つ */
      setTimeout(() => {
        this._registerDLF();
      }, 0);
    }
  },

  /* ==== private ==== */

  /**
   * Document Loader Factory を登録する
   */
  _registerDLF: function() {
    let contractID = UnMHTDocumentLoaderFactory.prototype.contractID;
    let registerer = new arDLFRegisterer(contractID);

    for (let type of this._getMimeTypeList()) {
      registerer.addMimeType(type, true);
    }

    for (let [ext, type] of this._getMimeExtensionList()) {
      registerer.addFileExtension(ext, type, true);
    }

    try {
      registerer.addMimeType(lazy.MIMEService.getTypeFromExtension("mht"),
                             true);
    } catch (e) {
    }

    registerer.register();
    arShutdownEvent.addFunc((data, reason) => {
      registerer.unregister();
    });

    this._resolve(true);
    this._resolve = null;
  },

  /**
   * ハンドルする MIME-Type のリストを返す
   *
   * @returns {Array.<string>}
   *          ハンドルする MIME-Type のリスト
   */
  _getMimeTypeList: function() {
    let tmp = UnMHTPref.getPref("open.mime.types");
    let list = [];
    try {
      list = JSON.parse(tmp);
      if (!list.every(n => typeof(n) == "string")) {
        list = [];
      }
    } catch (e) {
      list = [];
    }

    return list;
  },

  /**
   * 拡張子から MIME-Type の関連付けのリストを返す
   *
   * @returns {Array.<Tuple.<string,string>>}
   *          拡張子から MIME-Type の関連付けのリスト
   */
  _getMimeExtensionList: function() {
    let tmp = UnMHTPref.getPref("open.mime.extensions");
    let list = [];
    try {
      list = JSON.parse(tmp);
      if (!list.every(n => typeof(n[0]) == "string" && typeof(n[1]) == "string")) {
        list = [];
      }
    } catch (e) {
      list = [];
    }

    return list;
  }
});

/**
 * 起動時に Document Loader Factory を登録する
 */
let UnMHTDLFRegisterer = new UnMHTDLFRegistererClass();

module.UnMHTDLFRegisterer = UnMHTDLFRegisterer;

})(this);
