"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTDOMConverter"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { UnMHTMailURLUtils } = Cu.import("resource://unmht/modules/UnMHTMailURLUtils.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arPseudoID",
                                  "resource://unmht/modules/base/arPseudoID.jsm");

/**
 * DOM ツリーを文字列に変換する
 *
 * @class
 */
let UnMHTDOMConverter = Object.freeze({
  /**
   * DOM ツリーを文字列に変換する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} location
   *          ドキュメントの URI
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {boolean} useSelection
   *          選択範囲を使用するか
   * @param   {Object} params
   *            {
   *              noScript:   {boolean=}
   *                          スクリプトを削除するか
   *              isMail:     {boolean=}
   *                          メールか
   *              saveCanvas: {boolean=}
   *                          キャンバスを保存するか
   *              frameIDs:   {Map.<nsIDOMElement,string>=}
   *                          フレームの Content-ID
   *              markerID:   {string=}
   *                          マーカーの ID
   *                          使用しない場合は "" か undefined
   *              isXML:      {boolean=}
   *                          XML か
   *            }
   * @returns {string}
   *          要素全体の HTML
   */
  toPlainText: function(window, location, targetDocument,
                        useSelection, params) {
    if (!targetDocument) {
      return "";
    }

    let targetWindow = targetDocument.defaultView;

    let targetSelection = null;
    if (useSelection) {
      targetSelection = targetWindow.getSelection();
      if (!targetSelection ||
          targetSelection.isCollapsed ||
          targetSelection.rangeCount === 0) {
        targetSelection = null;
      }
    }

    let modName = n => {
      if (n.startsWith("xmlns")) {
        return "A" + n;
      } else if (n.startsWith("http-equiv")) {
        return "B" + n;
      } else {
        return "C" + n;
      }
    };

    let ignoreNodes = ("ignoreNodes" in params) ? params.ignoreNodes : [];
    let isMail = ("isMail" in params) ? params.isMail : false;
    let noScript = ("noScript" in params) ? params.noScript : false;
    let saveCanvas = ("saveCanvas" in params) ? params.saveCanvas : false;
    let frameIDs = ("frameIDs" in params) ? params.frameIDs : null;
    let markerID = ("markerID" in params) ? params.markerID : null;
    let isXML = ("isXML" in params) ? params.isXML : false;

    let conv = (node, targetSelection, texts) => {
      if (targetSelection) {
        if (node instanceof targetWindow.HTMLHeadElement ||
            node instanceof targetWindow.HTMLStyleElement ||
            node instanceof targetWindow.HTMLTitleElement) {
          targetSelection = null;
        }
      }

      if (targetSelection) {
        if (!arDOMUtils.isNodeInSelection(node, targetSelection)) {
          return;
        }
      }

      if (ignoreNodes.indexOf(node) != -1) {
        return;
      }

      if (node instanceof targetWindow.Text) {
        let text = node.nodeValue;

        if (targetSelection) {
          let textList = [];
          for (let i = 0, len = targetSelection.rangeCount; i < len; i += 1) {
            let t = text;
            let isPartial = false;

            let range = targetSelection.getRangeAt(i);

            if (range.endContainer == node) {
              t = t.slice(0, range.endOffset);
              isPartial = true;
            }
            if (range.startContainer == node) {
              t = t.slice(range.startOffset);
              isPartial = true;
            }

            if (isPartial) {
              textList.push(t);
            }
          }

          if (textList.length > 0) {
            text = textList.join("");
          }
        }
        text = arDOMUtils.escapeEntity(text);

        texts.push(text);
      } else if (node instanceof targetWindow.Comment) {
        /* コメントをコメントとして展開する
         * コメント中の実体参照はそのまま nodeValue に入っているので
         * escapeEntity を呼ぶ必要は無い */
        let value = node.nodeValue;
        if (/^\[if[ \t\r\n]+[A-Za-z0-9\. \t\r\n!\&\|\(\)]*\]$/.test(value) ||
            /^\[endif\]$/.test(value)) {
          /* 不正な条件付きコメント */
          texts.push("<!" + value + ">");
        } else {
          texts.push("<!--" + value + "-->");
        }
      } else if (node instanceof targetWindow.HTMLBaseElement &&
                 isMail) {
      } else if (node instanceof targetWindow.HTMLScriptElement &&
                 noScript) {
      } else {
        let attributes = [];

        let addStylePre = "";

        if (saveCanvas &&
            node instanceof targetWindow.HTMLCanvasElement) {
          addStylePre
            += "background-image: url(\'" + node.toDataURL("image/png") + "\');";
        }

        if (node.attributes) {
          for (let attr of Array.from(node.attributes)) {
            let name = attr.name.toLowerCase();
            let value = attr.value;

            if (node instanceof targetWindow.HTMLOptionElement) {
              if (name == "selected") {
                continue;
              }
            }

            if (noScript &&
                name.startsWith("on")) {
              continue;
            }

            if (node instanceof targetWindow.HTMLInputElement) {
              if (node.type == "checkbox" ||
                  node.type == "radio") {
                if (name == "checked") {
                  continue;
                }
              } else {
                if (name == "value") {
                  continue;
                }
              }
            }

            if ((node instanceof targetWindow.HTMLFrameElement ||
                 node instanceof targetWindow.HTMLIFrameElement) &&
                name == "src" ||
                name == "srcdoc") {
              continue;
            }

            if (/^(src|href|background|action|formaction|data)$/.test(name) &&
                isMail) {
              value = UnMHTMailURLUtils.hideMailURL(value, location);
            }

            if (/^(src|href|background|action|formaction|data)$/.test(name) &&
                noScript &&
                /^\s*javascript:/i.test(value)) {
              value = "";
            }

            if (name == "style") {
              if (addStylePre) {
                value = addStylePre + value;
                addStylePre = "";
              }
            }

            if (/^(src|href|background|action|formaction|data)$/.test(name)) {
              let nop = value.replace(/%/g, "");

              let malformed = false;
              try {
                encodeURI(nop);
              } catch (e) {
                malformed = true;
              }

              if (!malformed && encodeURI(nop) != nop) {
                /* % を除いてもエンコード対象を含む場合のみエンコードする
                 * ただし先頭と末尾の空白文字はエンコードしない */
                let leading = "", trailing = "";
                value = value.replace(/^(\s+)/, matched => {
                  leading = matched;
                  return "";
                });
                value = value.replace(/(\s+)$/, matched => {
                  trailing = matched;
                  return "";
                });
                value = leading + encodeURI(value) + trailing;
              }
            }
            value = arDOMUtils.escapeEntity(value);

            attributes.push(name +  "=" + "\"" + value + "\"");
          }
        }

        if ((node instanceof targetWindow.HTMLFrameElement ||
             node instanceof targetWindow.HTMLIFrameElement)) {
          let name = "src";
          let value = "";
          if (frameIDs) {
            let id = "iframe." + frameIDs.size + "." + lazy.arPseudoID.gen()
              + "@unmht.org";
            frameIDs.set(node, id);
            value = "cid:" + id;
          } else {
            value = node.contentDocument.documentURI;
          }

          attributes.push(name +  "=" + "\"" + value + "\"");
        }

        if (addStylePre) {
          let name = "style";
          let value = addStylePre;
          addStylePre = "";

          value = arDOMUtils.escapeEntity(value);

          attributes.push(name +  "=" + "\"" + value + "\"");
        }

        if (node instanceof targetWindow.HTMLOptionElement) {
          if (node.selected) {
            let name = "selected";
            let value = "selected";
            attributes.push(name +  "=" + "\"" + value + "\"");
          }
        }

        if (node instanceof targetWindow.HTMLInputElement) {
          if (node.type == "checkbox" ||
              node.type == "radio") {
            if (node.checked) {
              let name = "checked";
              let value = "checked";
              attributes.push(name +  "=" + "\"" + value + "\"");
            }
          } else if (node.type == "file") {
          } else {
            let name = "value";
            let value = arDOMUtils.escapeEntity(node.value);
            attributes.push(name +  "=" + "\"" + value + "\"");
          }
        }

        let nodeName = node.nodeName.toLowerCase();

        let attrSorter = (a, b) => modName(a) > modName(b) ? 1 : -1;

        let childs = [];

        if (markerID &&
            ["html", "head", "body"].indexOf(nodeName.toLowerCase()) != -1) {
          childs.push(this.makeMarker(markerID,
                                      nodeName.toLowerCase(), "first"));
        }

        if ((nodeName == "noscript" ||
             nodeName == "noembed" ||
             nodeName == "noframes") &&
            node.firstChild &&
            node.firstChild instanceof targetWindow.Text &&
            !node.firstChild.nextSibling) {
          /* NOSCRIPT, NOEMBED, NOLAYER, NOFRAMES 要素が展開されていない場合 */
          childs.push(node.firstChild.nodeValue);
        } else if (node instanceof targetWindow.HTMLStyleElement ||
                   node instanceof targetWindow.HTMLScriptElement) {
          /* noScript の場合には SCRIPT 要素はここまで来ない */
          let childNode = node.firstChild;
          let tmp = [];
          while (childNode) {
            conv(childNode, targetSelection, tmp);
            childNode = childNode.nextSibling;
          }
          for (let t of tmp) {
            childs.push(arDOMUtils.unescapeEntity(t));
          }
        } else if (node instanceof targetWindow.HTMLTextAreaElement) {
          childs.push(arDOMUtils.escapeEntity(node.value));
        } else {
          let childNode = node.firstChild;
          while (childNode) {
            conv(childNode, targetSelection, childs);
            childNode = childNode.nextSibling;
          }
        }

        if (markerID &&
            ["html", "head", "body"].indexOf(nodeName.toLowerCase()) != -1) {
          childs.push(this.makeMarker(markerID,
                                      nodeName.toLowerCase(), "last"));
        }

        let attrs
          = attributes.length
          ? (" " + attributes.sort(attrSorter).join(" "))
          : "";
        if (childs.length || !this.isVoid(node)) {
          texts.push("<" + nodeName + attrs + ">");
          for (let t of childs) {
            texts.push(t);
          }
          texts.push("</" + nodeName + ">");
        } else {
          if (isXML) {
            texts.push("<" + nodeName + attrs + " />");
          } else {
            texts.push("<" + nodeName + attrs + ">");
          }
        }
      }
    };

    let texts = [];
    conv(targetDocument.documentElement, targetSelection, texts);

    return texts.join("");
  },

  /**
   * マーカーを作成する
   *
   * @param   {string} markerID
   *          マーカーの ID
   * @param   {string} name
   *          要素の名前
   * @param   {string} pos
   *          要素中の位置
   * @returns {string}
   *          マーカー
   */
  makeMarker: function(markerID, name, pos) {
    return "{" + name + "." + pos + "." + markerID + "}";
  },

  /**
   * マーカーのある位置にテキストを追加する
   *
   * @param   {string} text
   *          対象のテキスト
   * @param   {string} markerID
   *          マーカーの ID
   * @param   {string} name
   *          要素の名前
   *           "html" / "head" / "body"
   * @param   {string} pos
   *          要素中の位置
   *           "first" / "last"
   * @param   {string} subtext
   *          追加する対象のテキスト
   * @returns {string}
   *          追加したテキスト
   */
  appendToMarker: function(text, markerID, name, pos, subtext) {
    let marker = this.makeMarker(markerID, name, pos);
    if (text.indexOf(marker) == -1) {
      return {
        ok: false,
        text: ""
      };
    }

    return {
      ok: true,
      text: text.replace(marker, marker + subtext)
    };
  },

  /**
   * マーカーを削除する
   *
   * @param   {string} text
   *          対象のテキスト
   * @param   {string} markerID
   *          マーカーの ID
   * @returns {string}
   *          マーカーを削除したテキスト
   */
  removeMarker: function(text, markerID) {
    return text
      .replace(this.makeMarker(markerID, "html", "first"), "")
      .replace(this.makeMarker(markerID, "html", "last"), "")
      .replace(this.makeMarker(markerID, "head", "first"), "")
      .replace(this.makeMarker(markerID, "head", "last"), "")
      .replace(this.makeMarker(markerID, "body", "first"), "")
      .replace(this.makeMarker(markerID, "body", "last"), "");
  },

  /**
   * 空の要素がどうかを返す
   *
   * @param   {nsIDOMNode} node
   *          対象の要素
   * @returns {boolean}
   *          空の要素か
   */
  isVoid: function(node) {
    let nodeName = node.nodeName.toLowerCase();

    let voidNames = [
      /* HTML5 */
      "area",
      "base",
      "br",
      "col",
      /* embed is contained in HTML5 void element,
       * but HTML4 needs closing tag */
      "hr",
      "img",
      "input",
      "keygen",
      "link",
      "meta",
      "param",
      "source",
      "track",
      "wbr",

      /* HTML 4.01 */
      "basefont",
      "frame",
      "isindex"
    ];
    return voidNames.indexOf(nodeName) != -1;
  }
});

module.UnMHTDOMConverter = UnMHTDOMConverter;

})(this);
