"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTDocumentCategory"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

/**
 * ドキュメントの情報を取得する
 *
 * @class
 */
let UnMHTDocumentCategory = Object.freeze({
  /**
   * 保存可能かどうかを調べる
   *
   * @param   {?nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {boolean} allowOtherResource
   *          HTML 以外のファイルを許可するか
   * @returns {boolean}
   *          保存可能かどうか
   */
  isSavable: function(window, browser, allowOtherResource) {
    if (!browser) {
      return false;
    }

    if (window &&
        arWindowUtils.isMailer(window) &&
        browser.currentURI.spec == "about:blank") {
      /* 複数のメッセージを選択しているとき等 */
      return false;
    }
    if (window &&
        arWindowUtils.isMessagePaneCollapsed(window)) {
      return false;
    }

    let targetDocument = browser.contentDocument;
    if (!targetDocument) {
      return false;
    }

    if (targetDocument instanceof Ci.nsIDOMXULDocument) {
      return false;
    }

    if ("contentType" in targetDocument &&
        targetDocument.contentType &&
        (targetDocument.contentType.startsWith("text/") ||
         /\/x(ht)?ml(\+xml)?$/.test(targetDocument.contentType))) {
      return true;
    }

    return allowOtherResource;
  },

  /**
   * 保存可能かどうかを調べる
   *
   * @param   {?nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {boolean} allowOtherResource
   *          HTML 以外のファイルを許可するか
   * @returns {boolean}
   *          保存可能かどうか
   */
  isDocumentSavable: function(window, targetDocument, allowOtherResource) {
    if (!targetDocument) {
      return false;
    }

    if (window &&
        arWindowUtils.isMailer(window) &&
        targetDocument.documentURI == "about:blank") {
      /* 複数のメッセージを選択しているとき等 */
      return false;
    }
    if (window &&
        arWindowUtils.isMessagePaneCollapsed(window)) {
      return false;
    }

    if (targetDocument instanceof Ci.nsIDOMXULDocument) {
      return false;
    }

    if ("contentType" in targetDocument &&
        targetDocument.contentType &&
        (targetDocument.contentType.startsWith("text/") ||
         /\/x(ht)?ml(\+xml)?$/.test(targetDocument.contentType))) {
      return true;
    }

    return allowOtherResource;
  },

  /**
   * ローカルの MHT ファイルかどうかを返す
   *
   * @param   {?nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @returns {boolean}
   *          ローカルの MHT ファイルかどうか
   */
  isLocalMHT: function(window, browser) {
    if (!browser) {
      return false;
    }

    let location;

    try {
        location = browser.currentURI.spec;
    } catch (e) {
      /* 対象が生きていない */
      return false;
    }

    if (UnMHTScheme.isUnMHTURI(location)) {
      let param = UnMHTScheme.getUnMHTURIParam(location);
      return ("scheme" in param && param.scheme == "file");
    }

    return false;
  },

  /**
   * ローカルの MHT ファイルかどうかを返す
   *
   * @param   {?nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {boolean}
   *          ローカルの MHT ファイルかどうか
   */
  isDocumentLocalMHT: function(window, targetDocument) {
    if (!targetDocument) {
      return false;
    }

    let location;
    try {
      location = targetDocument.documentURI;
    } catch (e) {
      /* 対象が生きていない */
      return false;
    }
    if (UnMHTScheme.isUnMHTURI(location)) {
      let param = UnMHTScheme.getUnMHTURIParam(location);
      return ("scheme" in param && param.scheme == "file");
    }
    return false;
  },

  /**
   * ローカルファイルかどうかを返す
   * ローカルファイルであればローカルファイルを格納する
   *
   * @param   {string} location
   *          対象の URI
   * @param   {Object} filePath
   *            {
   *              value: {string}
   *                     ローカルファイルのパス
   *            }
   * @returns {boolean}
   *          ローカルファイルかどうか
   */
  isLocalFile: function(location, filePath) {
    if (UnMHTScheme.isUnMHTURI(location)) {
      location = UnMHTScheme.deUnMHTURI(location);
    }

    if (!location.startsWith("file://")) {
      return false;
    }

    filePath.value = arFileUtils.URLSpecToPath(location);

    return true;
  },

  /**
   * ブラウザが生きているかどうか調べる
   *
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @returns {boolean}
   *          ブラウザが生きているか
   */
  isAlive: function(browser) {
    if (!browser) {
      return false;
    }

    try {
      /* とりあえず適当にブラウザのプロパティにアクセスする
       * ブラウザが生きていなければ例外が飛ぶ */
      return browser.currentURI.spec !== "";
    } catch (e) {
      return false;
    }
  },

  /**
   * ドキュメントが生きているかどうか調べる
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {boolean}
   *          ドキュメントが生きているか
   */
  isDocumentAlive: function(targetDocument) {
    if (!targetDocument) {
      return false;
    }

    try {
      /* とりあえず適当にドキュメントのプロパティにアクセスする
       * ドキュメントが生きていなければ例外が飛ぶ */
      return targetDocument.location.href !== "";
    } catch (e) {
      return false;
    }
  },

  /**
   * HTML っぽいかどうかを返す
   *
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @returns {boolean}
   *          HTML っぽいか
   */
  maybeHTML: function(browser) {
    return this.maybeHTMLDocument(browser.contentDocument);
  },

  /**
   * HTML っぽいかどうかを返す
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {boolean}
   *          HTML っぽいか
   */
  maybeHTMLDocument: function(targetDocument) {
    if ("contentType" in targetDocument) {
      if (targetDocument.contentType == "text/html") {
        return true;
      }
    }

    return false;
  }
});

module.UnMHTDocumentCategory = UnMHTDocumentCategory;

})(this);
