"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTDocumentLoaderFactory",
                         "UnMHTDLFHandledLocations"];

(function(module) {

/**
 * @file
 * Defines a Document Loader Factory to handle the MHT document type.
 *
 * Currently, this implementation just fires the standard MHT loading process,
 *  and refuses to create a content viewer and a stream listener.
 *
 * For more information on the document loading process, see
 *  <http://www.mozilla.org/newlayout/doc/webwidget.html> and
 *  <https://developer.mozilla.org/En/The_life_of_an_HTML_HTTP_request>
 *  (retrieved 2008-10-07).
 */

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTExtractor",
                                  "resource://unmht/modules/UnMHTExtractor.jsm");
XPCOMUtils.defineLazyServiceGetter(lazy, "DocumentLoaderFactory",
                                   "@mozilla.org/content/document-loader-factory;1",
                                   "nsIDocumentLoaderFactory");

/**
 * DLF でハンドルした URI
 * @type {Set.<string>}
 */
let UnMHTDLFHandledLocations = new Set();

/**
 * ファイルを展開するリスナ
 *
 * @constructor
 * @param   {string} originalURISpec
 *          元のチャネルの URI
 * @param   {nsIWebNavigation} nav
 *          ドキュメントを開いているウィンドウの WebNavigation
 * @implements nsIRequestObserver
 * @implements nsIStreamListener
 */
function UnMHTDLFExtractListener(originalURISpec, nav) {
  /**
   * 元のチャネルの URI
   * @type {string}
   */
  this._originalURISpec = originalURISpec;

  /**
   * ドキュメントを開いているウィンドウの WebNavigation
   * @type {nsIWebNavigation}
   */
  this._nav = nav;

  /**
   * 取得したデータ
   * @type {string}
   */
  this._data = "";

  /**
   * バイナリ用入力ストリーム
   * @type {nsIBinaryInputStream}
   */
  this._bstream
    = Cc["@mozilla.org/binaryinputstream;1"]
    .createInstance(Ci.nsIBinaryInputStream);

  Object.seal(this);
}
UnMHTDLFExtractListener.prototype = Object.freeze({
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIRequestObserver,
                                         Ci.nsIStreamListener]),

  /**
   * 破棄
   * 自分で呼ぶ
   */
  destruct: function() {
    this._originalURISpec = "";
    this._nav = null;
    this._data = "";
    this._bstream = null;
  },

  /* ==== nsIRequestObserver ==== */

  /**
   * リクエスト開始のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {?nsISupports} context
   *          ユーザ定義
   */
  onStartRequest: function(request, context) {
  },

  /**
   * リクエスト終了のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {?nsISupports} context
   *          ユーザ定義
   * @param   {number} statusCode
   *          終了コード
   */
  onStopRequest: Task.async(function*(request, context, statusCode) {
    let eFileInfo
        = yield lazy.UnMHTExtractor.extractMHT(this._originalURISpec, this._data);

    if (eFileInfo) {
      try {
        this._nav.loadURI(eFileInfo.baseURI,
                          Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY,
                          null, null, null);
      } catch (e) {
        Cu.reportError("Failed to load URI\n"
                       + eFileInfo.baseURI + "\n"
                       + e + "\n" + e.stack);
      }
    }

    this.destruct();
  }),

  /* ==== nsIStreamListener ==== */

  /**
   * データ到着のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {?nsISupports} context
   *          ユーザ定義
   * @param   {nsIInputStream} inputStream
   *          データを取得するストリーム
   * @param   {number} offset
   *          データの位置
   * @param   {number} count
   *          データの長さ
   */
  onDataAvailable: function(request, context, inputStream, offset, count) {
    this._bstream.setInputStream(inputStream);
    this._data += this._bstream.readBytes(count);
  }
});

/**
 * Construct the UnMHTDocumentLoaderFactory object.
 *
 * Document loader factories are XPCOM services, so they should be constructed
 *  as singletons, for example as explained in
 *  <http://weblogs.mozillazine.org/weirdal/archives/019620.html> (retrieved
 *  2008-10-07). However, since this object doesn't store any state information,
 *  we might as well create it as a normal XPCOM object.
 *
 * Document loader factories must be registered with the
 *  "Gecko-Content-Viewers" category. For this extension, this is done
 *  dynamically on startup, because the list of MIME types that this document
 *  loader factory will handle is not known in advance.
 *
 * @constructor
 * @implements nsIDocumentLoaderFactory
 */
function UnMHTDocumentLoaderFactory() {
  Object.seal(this);
}
UnMHTDocumentLoaderFactory.prototype = Object.freeze({
  classID:          Components.ID("{fa22337c-baf4-4c05-ab92-5e1fb4a97aa9}"),
  contractID:       "@unmht.org/unmht-document-loader-factory;1",
  classDescription: "UnMHT Document Loader Factory JS Component",
  QueryInterface:   XPCOMUtils.generateQI([Ci.nsIDocumentLoaderFactory]),

  /* ==== nsIDocumentLoaderFactory ==== */

  /**
   * コンテントビューアを作成する
   *
   * @param   {string} aCommand
   *          実行するコマンド
   * @param   {nsIChannel} aChannel
   *          入力チャネル
   * @param   {nsILoadGroup} aLoadGroup
   *          ページを読み込むロードグループ
   * @param   {string} aContentType
   *          表示するコンテンツの MIME-Type
   * @param   {nsISupport} aContainer
   *          nsIDocShell
   *          (docshell/base/nsDocShell.cpp 参照)
   * @param   {nsISupport} aExtraInfo
   *          たぶん使われていない
   *          (docshell/base/nsDocShell.cpp 参照)
   * @param   {Object} aDocListenerResult
   *          入力チャネルのストリームのリスナ (出力)
   *            {
   *              value: {nsIStreamListener}
   *            }
   * @returns {nsIContentViewer}
   *          コンテントビューア
   * @throws  {number} Cr.NS_ERROR_NOT_IMPLEMENTED
   */
  createInstance: function(aCommand, aChannel, aLoadGroup, aContentType,
                           aContainer, aExtraInfo, aDocListenerResult) {
    if (aCommand != "view") {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }

    let originalURISpec = aChannel.URI.spec;

    if (/^(imap|mailbox|news|nntp|pop|snews):\/\//.test(originalURISpec)) {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }

    let loadingChannel
      = Services.io.newChannelFromURI2(Services.io.newURI(UnMHTScheme.createLoadingURI(), "utf-8", null),
                                       null, // aLoadingNode
                                       Services.scriptSecurityManager.getSystemPrincipal(),
                                       null, // aTriggeringPrincipal
                                       Ci.nsILoadInfo.SEC_NORMAL,
                                       Ci.nsIContentPolicy.TYPE_OTHER);
    let listener = { value: null };

    let viewer
      = lazy.DocumentLoaderFactory
      .createInstance("view", loadingChannel, aLoadGroup, "text/html",
                      aContainer, aExtraInfo, listener);

    loadingChannel.asyncOpen(listener.value.QueryInterface(Ci.nsIStreamListener),
                           null);

    let nav = null;
    try {
      nav = aContainer.QueryInterface(Ci.nsIInterfaceRequestor)
        .getInterface(Ci.nsIWebNavigation);
    } catch (e) {
      Cu.reportError("Failed get nsIWebNavigation\n"
                     + e + "\n" + e.stack);
    }

    UnMHTDLFHandledLocations.add(originalURISpec);

    aDocListenerResult.value
      = new UnMHTDLFExtractListener(originalURISpec, nav);

    return viewer;
  },

  /**
   * ドキュメントに対応するコンテントビューアの作成
   *
   * @param   {nsISupport} aContainer
   *          nsIDocShell
   *          (docshell/base/nsDocShell.cpp 参照)
   * @param   {nsIDocument} aDocment
   *          対象のドキュメント
   * @param   {string} aCommand
   *          実行するコマンド
   */
  createInstanceForDocument: function(aContainer, aDocument, aCommand) {
    /* This function should never have been called. As of 2008-10-07, it may
     *  be called only from nsDocShell::CreateAboutBlankContentViewer, that
     *  doesn't handle exceptions, so we return null instead of throwing
     *  NS_ERROR_NOT_IMPLEMENTED. */
    return null;
  },

  /**
   * 空のドキュメントの作成
   *
   * @param   {nsILoadGroup} aLoadGroup
   *          ページを読み込むロードグループ
   * @param   {nsIPrincipal} aPrincipal
   *          読み込むドキュメントの Principal
   */
  createBlankDocument: function(aLoadGroup, aPrincipal) {
    /* See the comment inside createInstanceForDocument. */
    return null;
  }
});

module.UnMHTDocumentLoaderFactory = UnMHTDocumentLoaderFactory;
module.UnMHTDLFHandledLocations = UnMHTDLFHandledLocations;

})(this);
