"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTExtractor",
                         "UnMHTContentModifier"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { arMIMEDecoder } = Cu.import("resource://unmht/modules/base/arMIMEDecoder.jsm", {});
let { arPathUtils, arPathTypeMask } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { arPseudoID } = Cu.import("resource://unmht/modules/base/arPseudoID.jsm", {});
let { arUconv, fromUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { UnMHTCache } = Cu.import("resource://unmht/modules/UnMHTCache.jsm", {});
let { UnMHTContentModifier } = Cu.import("resource://unmht/modules/UnMHTContentModifier.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

/**
 * パートの展開情報
 *
 * @constructor
 */
function UnMHTExtractParam() {
  /* ---- 階層 ---- */

  /**
   * 親パートのパス
   * @type {string}
   */
  this.parentPath = "";

  /**
   * パス
   * @type {string}
   */
  this.path = "";

  /* ---- 元のパートのヘッダの補完 ---- */

  /**
   * Content-Type のキャラクタセット
   * multipart/mixed で元のパートと異なる
   * @type {string}
   */
  this.charset = "";

  /**
   * Content-Type の MIME-Type
   * ボディが HTML である application/octet-stream や
   * multipart/mixed で元のパートと異なる
   * @type {string}
   */
  this.mimetype = "";

  /**
   * Content-Location
   * 元のパートが Content-Location を持たない場合や
   * multipart/mixed で元のパートと異なる
   * @type {string}
   */
  this.location = "";

  /**
   * Content-ID
   * 元のパートが Content-ID を持つ場合は同じ
   * 持たない場合は自動生成
   * @type {string}
   */
  this.cid = "part." + arPseudoID.gen() + "@unmht.org";

  /* ---- 種類 ---- */

  /**
   * HTML かどうか
   * @type {boolean}
   */
  this.isHTML = false;

  /**
   * CSS かどうか
   * @type {boolean}
   */
  this.isCSS = false;

  /**
   * multipart/mixed かどうか
   * @type {boolean}
   */
  this.isMixed = false;

  /**
   * mixed なドキュメントに使用したパートのパス
   * @type {Array.<string>}
   */
  this.mixedPath = [];

  /* ---- 元のパートのヘッダから生成する情報 ---- */

  /**
   * Content-Location の親ディレクトリ
   * @type {string}
   */
  this.baseDir = "";

  /**
   * Content-Location のファイル名
   * @type {string}
   */
  this.leafName = "";

  /**
   * Content-Location のフラグメント
   * @type {string}
   */
  this.fragment = "";

  /* ---- 元のパートのボディから生成する情報 ---- */

  /**
   * 修正済みのファイルの中身
   * @type {string}
   */
  this.content = "";

  /* ---- 参照用 ---- */

  /**
   * 正規化した Content-Location
   * フラグメントは含まない
   * @type {string}
   */
  this.normalLocation = "";

  /**
   * 正規化した Content-Location のファイル名
   * @type {string}
   */
  this.normalLeafName = "";

  /**
   * 正規化した Content-Location のフラグメント
   * @type {string}
   */
  this.normalFragment = "";

  /**
   * 参照用のファイル名
   * @type {string} */
  this.refName = "";

  /**
   * 参照元のファイルの親ディレクトリ
   * referredBaseDir は
   * IE で保存された MHT ファイルにおける以下の問題を回避するためのもの
   *
   * [状況]
   *   ファイル
   *     http://unmht.org/x.html
   *     http://unmht.org/style/a.css
   *     http://unmht.org/style/b.css
   *     http://unmht.org/style/x.png
   *   x.html 中の記述
   *     <link href="./style/a.css" ... />
   *   a.css 中の記述
   *     @import url("./b.css");
   *   b.css 中の記述
   *     background: url("./x.png");
   *
   * [期待される結果]
   *   相対パス "./b.css" を a.css の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/style/b.css
   *
   *   これにより a.css から b.css が参照できる
   *   また b.css から x.png が参照できる
   *
   * [実際の結果]
   *   相対パス "./b.css" を x.html の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/b.css
   *
   *   これにより a.css から b.css が参照できない
   *   また b.css から x.png が参照できない
   *
   * [対策]
   *   1. すべての X から参照されているすべての Y について
   *      Y.referredBaseDir を X.baseDir とする
   *   2. Z.referredBaseDir が空でない Z について
   *      Z 内の相対パス R による参照について
   *      Z.baseDir から R を解決した URI を A とする
   *      Z.referredBaseDir から R を解決した URI を B とする
   *   3. F.contentLocation が A に一致するような F が存在しない、かつ
   *      G.contentLocation が B に一致するような G が存在する場合
   *      R による参照先を G とする
   *      この時、G.referredBaseDir は F.baseDir となり 2 から再度適用する
   *
   * [例]
   *   0. {x.html}.baseDir = "http://unmht.org/"
   *      {a.css}.baseDir = "http://unmht.org/style/"
   *      {b.css}.contentLocation = "http://unmht.org/b.css"
   *      {x.png}.contentLocation = "http://unmht.org/style/x.png"
   *   1. {a.css}.referredBaseDir = {x.html}.baseDir
   *                                "http://unmht.org/"
   *   2. Z = {a.css}
   *      R = "./b.css"
   *      A = {a.css}.baseDir + R
   *        = "http://unmht.org/style/b.css"
   *      B = {a.css}.referredBaseDir + R
   *        = "http://unmht.org/b.css"
   *   3. A != {b.css}.contentLocation
   *      B == {b.css}.contentLocation
   *      # a.css から b.css が参照できる
   *      {b.css}.referredBaseDir = {a.css}.baseDir
   *                                "http://unmht.org/style/"
   *   2. Z = {b.css}
   *      A = {b.css}.baseDir + R
   *        = "http://unmht.org/" + R
   *      B = {b.css}.referredBaseDir + R
   *        = "http://unmht.org/style/" + R
   *      # b.css から "./x.png" で参照されたファイル x.png が
   *      #   {x.png}.contentLocation = "http://unmht.org/x.png"
   *      # となっていても
   *      #   {x.png}.contentLocation = "http://unmht.org/style/x.png"
   *      # となっていても参照できる
   *
   * [問題]
   *   誤って生成された Content-Location に一致するファイルが
   *   同 MHT 内に存在した場合これらを区別する事はできない
   * @type {string}
   */
  this.referredBaseDir = "";

  /**
   * findStartPart のキャッシュ
   * @type {arMIMEPart}
   */
  this.startPart = null;

  Object.seal(this);
}
UnMHTExtractParam.prototype = Object.freeze({
});

/**
 * ファイル全体の展開情報
 *
 * @constructor
 */
function UnMHTExtractFileInfo() {
  /**
   * 展開したファイル名の URI
   * @type {string}
   */
  this.original = "";

  /**
   * 元のファイルのソース
   * @type {string}
   */
  this.source = "";

  /**
   * 開始ページの URI
   * @type {string}
   */
  this.baseURI = "";

  /**
   * 開始ページのパート
   * @type {arMIMEPart}
   */
  this.startPart = null;

  /**
   * トップレベルのパート
   * @type {arMIMEPart}
   */
  this.topPart = null;

  /**
   * Subject の値
   * @type {string}
   */
  this.subject = "";

  /**
   * Date の値
   * @type {Date}
   */
  this.date = null;

  /**
   * mht ファイルの中にあるパート
   * @type {Array.<arMIMEPart>}
   */
  this.parts = [];

  /**
   * 参照できるパート
   * @type {Map.<string,arMIMEPart>}
   */
  this.refParts = new Map();

  /**
   * Content-ID とパートのマップ
   * @type {Map.<string,arMIMEPart>}
   */
  this.cids = new Map();

  /**
   * 展開時刻
   * @type {Date}
   */
  this.extractDate = null;

  /**
   * 最終アクセス時刻
   * @type {Date}
   */
  this.lastAccessDate = null;

  /**
   * 全キャッシュのサイズ
   * @type {number}
   */
  this.size = 0;

  Object.seal(this);
}
UnMHTExtractFileInfo.prototype = Object.freeze({
  /**
   * データを開放する
   */
  destruct: function() {
    for (let part of this.parts) {
      part.eParam.content = null;
      part.eParam = null;
    }
    this.parts = null;

    if (this.topPart) {
      this.topPart.destruct();
      this.topPart = null;
    }

    this.startPart = null;
    this.refParts = null;
    this.source = null;
  }
});

/**
 * ファイルの展開
 *
 * @class
 */
let UnMHTExtractor = Object.freeze({
  /**
   * mht ファイルを展開する
   *
   * @param   {string} originalURISpec
   *          展開したファイル名の URI 表記
   * @param   {string} text
   *          mht ファイルの内容
   * @returns {Primise.<UnMHTExtractFileInfo>}
   *          展開情報
   */
  extractMHT: function(originalURISpec, text) {
    let eFileInfo = new UnMHTExtractFileInfo();

    /* とりあえず特殊な文字はエスケープしておく */
    eFileInfo.original
      = originalURISpec
      .replace(/[\(\)\'\"\[\] ]/g, matched => {
        let hexString
          = "0" + matched.charCodeAt(0).toString(16).toUpperCase();
        return "%" + hexString.slice(-2);
      });

    eFileInfo.baseURI = UnMHTScheme.enUnMHTURI(eFileInfo.original);

    eFileInfo.source = text;

    eFileInfo.extractDate = new Date();

    eFileInfo.topPart = arMIMEDecoder.decodeMessage(text);
    if (!eFileInfo.topPart) {
      /* 改行が LF のみ、CR のみを想定してもう一度変換 */
      let crlfText = text.replace(/\r|\n/g, "\r\n");
      eFileInfo.topPart = arMIMEDecoder.decodeMessage(crlfText);
    }

    if (!eFileInfo.topPart || !eFileInfo.topPart.findStartPart()) {
      /* 展開に失敗した場合 */

      let createDummyMessage = (mimetype, body) => {
        let data
          = "From: <Created by UnMHT>" + "\r\n"
          + "Content-Type: " + mimetype + "\r\n"
          + "\r\n"
          + body;

        return data;
      };

      let maybeHTML= false;
      if (!/\.(eml|mbs)$/i.test(originalURISpec) && /^\s*</.test(text)) {
        maybeHTML = true;
      }

      if (maybeHTML) {
        eFileInfo.topPart
          = arMIMEDecoder.decodeMessage(createDummyMessage("text/html; charset=\"UTF-8\"", text));
      } else {
        eFileInfo.topPart
          = arMIMEDecoder.decodeMessage(createDummyMessage("text/plain; charset=\"UTF-8\"", text));
      }
    }

    eFileInfo.subject = eFileInfo.topPart.subject;

    eFileInfo.date = eFileInfo.topPart.date;

    let baseDir = "thismessage:///";
    if (originalURISpec) {
      baseDir = arPathUtils.getBaseDir(originalURISpec);
    }
    this._createExtractParam(eFileInfo, eFileInfo.topPart,
                             baseDir, "", "1");

    eFileInfo.startPart = eFileInfo.topPart.eParam.startPart;

    this._setRefName(eFileInfo);

    for (let part of eFileInfo.parts) {
      UnMHTContentModifier.modifyContents(eFileInfo, part);
    }

    if (UnMHTPref.getPref("open.ppt.skip_warning")) {
      this._skipPPTWarning(eFileInfo);
    }

    this._createMixedDocument(eFileInfo);

    this._gatherPartsInfo(eFileInfo);

    eFileInfo.lastAccessDate = new Date();

    UnMHTCache.add(eFileInfo);

    setTimeout(() => {
      try {
        UnMHTCache.removeOld();
      } catch (e) {
        /* モジュールがアンロードされた等 */
      }
    }, 1000);

    return Promise.resolve(eFileInfo);
  },

  /* ==== private ==== */

  /**
   * パートの情報を作成する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} part
   *          対象のパート
   * @param   {string} parentLocation
   *          親パートの解決済みの URI
   * @param   {string} parentPath
   *          親パートのパス
   * @param   {string} path
   *          パス
   */
  _createExtractParam: function(eFileInfo, part, parentLocation,
                                parentPath, path) {
    let eParam = new UnMHTExtractParam();

    eParam.parentPath = parentPath;
    eParam.path = path;

    if (part.contentID) {
      eParam.cid = part.contentID;
    }

    if (part.isMixed) {
      eParam.isMixed = true;
      eParam.content = "";

      eParam.mimetype = "text/html";
      eParam.charset = "utf-8";

      eParam.isHTML = true;
    } else {
      eParam.content = part.body;

      eParam.mimetype = part.mimetype;
      eParam.charset = part.charset;
      if (!eParam.mimetype) {
        eParam.mimetype = "text/plain";
        if (!eParam.charset) {
          eParam.charset = "us-ascii";
        }
      }
      if (eParam.mimetype == "text/css") {
        eParam.isCSS = true;
      } else if (eParam.mimetype == "text/html" ||
                 eParam.mimetype == "application/xhtml+xml") {
        eParam.isHTML = true;
      }

      if (eParam.mimetype == "application/octet-stream" &&
          part.body.slice(0, 32).indexOf("<") != -1) {
        /* CGI 生成のページ等が application/octet-stream として
         * 保存されている場合がある */
        if (part.body.search(/<html/i) != -1) {
          eParam.mimetype = "text/html";
          eParam.isHTML = true;
        }
      }
    }

    eParam.location = arPathUtils.resolve(arPathUtils.getBaseDir(parentLocation),
                                          part.contentLocation);
    part.eParam = eParam;

    this._getPathInfo(part);

    eFileInfo.parts.push(part);

    if (part.isMultipart) {
      let i = 1;
      for (let p of part.parts) {
        this._createExtractParam(eFileInfo, p, eParam.location,
                                 path + ".", path + "." + i);
        i += 1;
      }
    }

    part.eParam.startPart = part.findStartPart();
  },

  /**
   * パスの情報を取得する
   *
   * @param   {arMIMEPart} part
   *          対象のパート
   */
  _getPathInfo: function(part) {
    let eParam = part.eParam;

    let location = eParam.location;

    let info = arPathUtils.getPathInfo(location);

    if (info.type & arPathTypeMask.DATA) {
      eParam.baseDir = location;
      eParam.leafName = "";
      eParam.fragment = "";
    } else {
      // DATA 以外は pathComponents が必ず 1 個以上存在する
      eParam.leafName = info.pathComponents[info.pathComponents.length - 1];
      eParam.fragment = info.fragment || "";

      let baseInfo = info.clone();
      if (baseInfo.pathComponents.length) {
        baseInfo.pathComponents.pop();
      }
      baseInfo.query = undefined;
      baseInfo.fragment = undefined;
      eParam.baseDir = arPathUtils.createPath(baseInfo);
    }

    if (part.contentDispositionFilename) {
      eParam.leafName = part.contentDispositionFilename;
    }

    eParam.referredBaseDir = "";

    let noFragmentInfo = info.clone();
    noFragmentInfo.fragment = undefined;
    let noFragment = arPathUtils.createPath(noFragmentInfo);

    eParam.normalLocation = UnMHTCache.normalizeName(noFragment);
    eParam.normalLeafName = UnMHTCache.normalizeName(eParam.leafName);
    eParam.normalFragment = UnMHTCache.normalizeName(eParam.fragment);
  },

  /**
   * 参照名を重複しないように設定する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _setRefName: function(eFileInfo) {
    let indices = new Map();

    let nonameIndex = { value: 0 };
    for (let part of eFileInfo.parts) {
      if (part.isMultipart && !part.isMixed) {
        continue;
      }

      let [refBase, extension] = this._getRefBaseAndExt(part, nonameIndex);
      refBase = encodeURIComponent(refBase).replace(/\//g, "_");
      if (extension) {
        extension = "." + extension;
      }

      let refName = refBase + extension;

      if (indices.has(refName)) {
        part.eParam.refName
          = refBase + "." + indices.get(refName) + extension;
        indices.set(refName, indices.get(refName) + 1);
      } else {
        part.eParam.refName = refName;
        indices.set(refName, 1);
      }
    }
  },

  /**
   * 参照名のベース部分と拡張子を取得する
   *
   * @param   {arMIMEPart} part
   *          対象のパート
   * @param   {Object} nonameIndex
   *          適切な情報が無い場合に使用するファイルの番号
   * @returns {Tuple.<string,string>}
   *          ベース部分と拡張子
   */
  _getRefBaseAndExt: function(part, nonameIndex) {
    let m = part.eParam.leafName.match(/^(.+)\.([A-Za-z0-9]+)/);
    if (m) {
      return [m[1], m[2]];
    }

    if (part.eParam.leafName) {
      return [part.eParam.leafName, ""];
    }

    if (part.contentID) {
      return [part.contentID, ""];
    }

    if (part.eParam.mimetype == "text/html") {
      return ["index", "html"];
    }

    let i = nonameIndex.value;
    nonameIndex.value += 1;
    return [i.toString(10), ""];
  },

  /**
   * PowerPoint の警告ページをスキップする
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _skipPPTWarning: function(eFileInfo) {
    let isPPT = false;
    let framePart = null;

    for (let part of eFileInfo.parts) {
      if (part.eParam.refName == "frame.htm") {
        framePart = part;
      }

      if (part.eParam.isHTML &&
          part.eParam.content.indexOf("xmlns:v=\"urn:schemas-microsoft-com:vml\"") != -1) {
        isPPT = true;
      }
    }

    if (isPPT && framePart) {
      eFileInfo.startPart = framePart;
    }
  },

  /**
   * multipart/mixed なドキュメントを作成する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _createMixedDocument: function(eFileInfo) {
    let createBody = (part,
                      charsets,
                      doctype, htmlAttrs, headContents, bodyAttrs) => {
      let isAttachment = false;
      if (part.contentDispositionType &&
          part.contentDispositionType != "inline") {
        isAttachment = true;
      }

      let getUnicodeContent = part => {
        let content = part.eParam.content;

        if (part.eParam.charset &&
            part.eParam.charset.toLowerCase() == "utf-8") {
          return fromUTF8(content);
        }

        if (part.eParam.charset) {
          return arUconv.toUnicode(content, part.eParam.charset);
        }

        return content;
      };

      let content;
      if ((!isAttachment && part.eParam.mimetype == "text/html") ||
          part.eParam.isMixed) {
        if (part.eParam.charset) {
          charsets.add(part.eParam.charset);
        }

        content
          = getUnicodeContent(part)
          .replace(this._htmlbodyRe, (matched, name, attrs) => {
            if (attrs) {
              attrs
                .replace(this._attrRe, (matched, attr,
                                        dqvalue,
                                        qvalue,
                                        value) => {
                  if (dqvalue) {
                    value = dqvalue;
                  } else if (qvalue) {
                    value = qvalue;
                  }

                  if (name.toLowerCase() == "html") {
                    htmlAttrs.set(attr.toLowerCase(), value);
                  } else if (name.toLowerCase() == "body") {
                    bodyAttrs.set(attr.toLowerCase(), value);
                  }
                });
            }

            return "";
          })
          .replace(this._headRe, (matched, content) => {
            headContents.push(content);

            return "";
          })
          .replace(this._doctypeRe, (matched, name) => {
            if (name.toLowerCase() == "doctype") {
              doctype.value = matched;
            }
            return "";
          });
      } else if (!isAttachment && part.eParam.mimetype == "text/plain") {
        if (part.eParam.charset) {
          charsets.add(part.eParam.charset);
        }

        content
          = "<pre>"
          + arDOMUtils.escapeEntity(getUnicodeContent(part))
          .replace(/[ \t]/g, "&#xa0;")
          + "</pre>";
      } else {
        let href
          = arDOMUtils.escapeEntity(eFileInfo.baseURI + part.eParam.refName);
        let filename
          = arDOMUtils.escapeEntity(part.eParam.leafName || "???");

        if (!isAttachment && part.contentType == "image") {
          content
            = "<div style=\"text-align: center;\">"
            + "<img src=\"" + href + "\" alt=\"" + filename + "\" />"
            + "</div>";
        } else {
          content = "<a href=\"" + href + "\">" + filename + "</a>";
        }

        content = content;
      }

      if (part.contentDispositionFilename) {
        content
          = "<fieldset style=\"border: 1px solid graytext;\">"
          + "<legend style=\"color: graytext; font-size: 80%;\">"
          + part.contentDispositionFilename
          + "</legend>"
          + content
          + "</fieldset>";
      }

      return content;
    };

    /* ネストした multipart/mixed のために後ろから解決する */
    for (let part of eFileInfo.parts.slice().reverse()) {
      if (!part.eParam.isMixed) {
        continue;
      }

      let charsets = new Set();
      let doctype = { value: "" };
      let htmlAttrs = new Map();
      let headContents = [];
      let bodyAttrs = new Map();
      let body = part.parts
        .map(p => createBody(p.eParam.startPart,
                             charsets,
                             doctype, htmlAttrs, headContents, bodyAttrs))
        .join("");

      let a = [];
      for (let p of part.parts) {
        a.push(p.eParam.path);
        for (let path of p.eParam.mixedPath) {
          a.push(path);
        }
        a.push(p.eParam.startPart.eParam.path);
      }
      part.eParam.mixedPath = a;

      let htmlAttrString
        = [...htmlAttrs].map(([k, v]) => (" " + k + "=" + v + ""));
      let bodyAttrString
        = [...bodyAttrs].map(([k, v]) => (" " + k + "=" + v + ""));

      part.eParam.content
        = (doctype.value ? (doctype.value + "\r\n") : "")
        + "<html" + htmlAttrString + ">\r\n"
        + "<head>\r\n"
        + headContents.join("\r\n")
        + "</head>\r\n"
        + "<body" + bodyAttrString + ">\r\n"
        + body
        + "</body>\r\n"
        + "</html>\r\n";

      if (charsets.size == 1) {
        part.eParam.charset = [...charsets.keys()][0];

        part.eParam.content
          = arUconv.fromUnicodeWithEntity(part.eParam.content,
                                          part.eParam.charset);
      }
    }
  },

  /**
   * パートの情報を収集する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _gatherPartsInfo: function(eFileInfo) {
    eFileInfo.size = eFileInfo.source.length;

    for (let part of eFileInfo.parts) {
      eFileInfo.refParts.set(part.eParam.refName, part);
      if (part.eParam.cid) {
        eFileInfo.cids.set(part.eParam.cid, part);
      }

      eFileInfo.size += part.eParam.content.length;
    }
  },

  _attrRe: /\s*([\?A-Za-z0-9_:\-]+)\s*=\s*(?:(\"(?:\\\"|[^\"])*\")|(\'(?:\\\'|[^\'])*\')|([^\"\'\\ >]+))/g,
  _htmlbodyRe: /<\/?(html|body)(\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig,
  _headRe: /<head>([^\0]*)<\/head>/i,
  _doctypeRe: /<!([\?A-Za-z0-9_:\-]+)(?:\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig
});

module.UnMHTExtractor = UnMHTExtractor;
module.UnMHTContentModifier = UnMHTContentModifier;

})(this);
