"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTFileConverter",
                         "UnMHTFileConverterHTMLRefType"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arMIMEDecoder } = Cu.import("resource://unmht/modules/base/arMIMEDecoder.jsm", {});
let { arMIMEEncoder } = Cu.import("resource://unmht/modules/base/arMIMEEncoder.jsm", {});
let { arMIMEParams } = Cu.import("resource://unmht/modules/base/arMIMEParams.jsm", {});
let { arMIMEPart } = Cu.import("resource://unmht/modules/base/arMIMEPart.jsm", {});
let { arMIMEUtils } = Cu.import("resource://unmht/modules/base/arMIMEUtils.jsm", {});
let { arPathUtils } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});
let { arUconv } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTSave } = Cu.import("resource://unmht/modules/UnMHTSave.jsm", {});

/**
 * 属性値のエンコード方法
 * @enum {number}
 */
let UnMHTFileConverterHTMLRefType = Object.freeze({
  /** 未知 */
  UNKNOWN:       0x0000,
  /** そのまま (Chrome) */
  RAW:           0x0001,
  /** 全て URL エンコード (Firefox) */
  ENCODED:       0x0002,
  /** スペースだけ (IE) */
  SPACE_ENCODED: 0x0003
});

/**
 * 変換情報
 *
 * @constructor
 */
function UnMHTFileConvertInfo() {
  /**
   * Subject
   * @type {string}
   */
  this.subject = "";

  /**
   * MHT ファイルのパス
   * @type {string}
   */
  this.mhtPath = "";

  /**
   * HTML ファイルのパス
   * @type {string}
   */
  this.htmlPath = "";

  /**
   * HTML ファイルのファイル名
   * @type {string}
   */
  this.htmlBasename = "";

  /**
   * HTML ファイルのファイル名拡張子抜き
   * @type {string}
   */
  this.htmlBasenameNoExt = "";

  /**
   * 親ディレクトリのパス
   * @type {string}
   */
  this.parentDirPath = "";

  /**
   * リソースディレクトリのパス
   * @type {string}
   */
  this.resDirPath = "";

  /**
   * リソースファイルのパス
   * @type {Array.<string>}
   */
  this.resPaths = [];

  /**
   * HTML の charset
   * @type {string}
   */
  this.charset = "";

  /**
   * リソースファイルへの参照時のエンコード方法
   * @type {UnMHTFileConverterHTMLRefType}
   */
  this.refType = UnMHTFileConverterHTMLRefType.UNKNOWN;

  /**
   * リソースファイルへの参照時のディレクトリセパレータ
   * @type {string}
   */
  this.refSep = "";

  /**
   * リソースファイルへの参照時の実際に使用されたディレクトリ名
   * Content-Location に使用する
   * @type {string}
   */
  this.actualResDirPath = "";

  /**
   * リソースファイルへの参照が正確にマッチした
   * @type {boolean}
   */
  this.exactMatch = false;

  /**
   * リソースファイルへの参照がたぶんマッチした
   * @type {boolean}
   */
  this.maybeMatch = false;

  Object.seal(this);
}

/* global UnMHTFileConverter */

/**
 * 変換コンテキスト
 *
 * @constructor
 * @param   {Array.<string>} paths
 *          対象の HTML ファイル
 * @param   {string} suffix
 *          ディレクトリ名の末尾
 * @param   {IUnMHTFileConverterListener} listener
 *          イベントリスナ
 */
function UnMHTFileConvertContext(paths, suffix, listener) {
  /**
   * 対象の HTML ファイル
   * @type {Array.<string>}
   */
  this._paths = paths;

  /**
   * ディレクトリ名の末尾
   * @type {string}
   */
  this._suffix = suffix;

  /**
   * イベントリスナ
   * @type {IUnMHTFileConverterListener}
   */
  this._listener = listener;

  /**
   * 中断したか
   * @type {boolean}
   */
  this._aborted = false;

  Object.seal(this);
}
UnMHTFileConvertContext.prototype = Object.freeze({
  /**
   * 破棄
   * start 後は自分で呼ぶ
   */
  destruct: function() {
    this._paths = null;
    this._listener = null;
  },

  /**
   * 変換開始
   */
  start: function() {
    this._convert().catch(e => {
      Cu.reportError(e + "\n" + e.stack);
    });
  },

  /**
   * 変換中断
   */
  abort: function() {
    this._aborted = true;
  },

  /* ==== private ==== */

  /**
   * ファイルを変換する
   */
  _convert: Task.async(function*() {
    let converted = 0;

    for (let current = 0, total = this._paths.length;
         current < total; current += 1) {
      if (this._aborted) {
        this._listener.onAbort(this);
        this.destruct();
        return;
      }

      let info = new UnMHTFileConvertInfo();

      info.htmlPath = this._paths[current];

      if (/\.mht(ml)?$/i.test(info.htmlPath)) {
        this._listener.onIgnored(this, info, current, total);
        continue;
      }

      info.parentDirPath = OS.Path.dirname(info.htmlPath);
      info.htmlBasename = OS.Path.basename(info.htmlPath);
      info.htmlBasenameNoExt = info.htmlBasename.replace(/\.[^\.]+$/, "");
      info.subject = info.htmlBasename;

      info.resDirPath = yield* this._findResDirTask(info);
      if (!info.resDirPath) {
        this._listener.onResDirNotFound(this, info, current, total);
        continue;
      }

      info.mhtPath = yield UnMHTSave.getUniqueFilePath(info.parentDirPath,
                                                       info.htmlBasenameNoExt);

      this._listener.onStart(this, info, current, total);

      let resIter = new OS.File.DirectoryIterator(info.resDirPath);
      let resEntries = yield resIter.nextBatch();
      info.resPaths = resEntries
        .filter(entry => !entry.isDir)
        .sort((a, b) => a.name < b.name ? -1 : 1)
        .map(entry => entry.path);

      try {
        yield UnMHTFileConverter.convertFile(info);
        this._listener.onFinish(this, info, current, total);
        converted += 1;
      } catch (e) {
        Cu.reportError(e.message);
        this._listener.onFail(this, info, current, total);
      }

      yield arTaskUtils.wait(10);
    }

    this._listener.onComplete(this, converted);
    this.destruct();
  }),

  /**
   * リソースディレクトリを探す
   *
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @yields  {Promise.<*>}
   * @returns {?string}
   *          リソースディレクトリのパス
   *          見付からなければ null
   */
  _findResDirTask: function*(info) {
    let suffixes = new Set();
    suffixes.add(this._suffix);
    try {
      let bundle = Services.strings.createBundle("chrome://global/locale/contentAreaCommands.properties");
      let systemSuffix = bundle.formatStringFromName("filesFolder", [""], 1);
      suffixes.add(systemSuffix);
    } catch (e) {
    }
    suffixes.add("_files");
    suffixes.add(".files");
    for (let suffix of suffixes) {
      let dir = OS.Path.join(info.parentDirPath,
                             info.htmlBasenameNoExt + suffix);
      let exists = yield OS.File.exists(dir);
      if (!exists) {
        continue;
      }
      let stat = yield OS.File.stat(dir);
      if (!stat.isDir) {
        continue;
      }

      return dir;
    }
    return null;
  }
});

/* global -UnMHTFileConverter */

/**
 * HTML ファイルを MHT ファイルに変換
 *
 * @class
 */
let UnMHTFileConverter = Object.freeze({
  /**
   * 変換コンテキストを作成する
   *
   * @param   {Array.<string>} paths
   *          対象の HTML ファイル
   * @param   {string} suffix
   *          ディレクトリ名の末尾
   * @param   {IUnMHTFileConverterListener} listener
   *          イベントリスナ
   * @returns {UnMHTFileConvertContext}
   *          変換コンテキスト
   */
  createContext: function(paths, suffix, listener) {
    return new UnMHTFileConvertContext(paths, suffix, listener);
  },

  /**
   * HTML ファイルを MHT に変換
   *
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  convertFile: Task.async(function*(info) {
    let topPart = new arMIMEPart();
    topPart.isMultipart = true;
    topPart.boundary = arMIMEUtils.createUniqueBoundary();
    topPart.startMimetype = "text/html";

    topPart.addField("From",
                     "<Saved by UnMHT Ver." + UnMHTPref.version + ">");
    topPart.addField("Subject", info.subject);
    topPart.addField("Date",
                     arMIMEEncoder.encodeDate(new Date()));
    topPart.addField("MIME-Version", "1.0");
    let params = new arMIMEParams();
    params.type = "multipart";
    params.subtype = "related";
    params.addParam("boundary", topPart.boundary);
    params.addParam("type", topPart.startMimetype);
    topPart.setContentType(params);
    topPart.parts = [];

    let htmlBody = yield arFileUtils.readBinary(info.htmlPath);

    yield* this._getHTMLInfoTask(htmlBody, info);

    let location = this.encode(info.refType, info.htmlBasename);

    let part = new arMIMEPart();
    part.contentTransferEncoding = "quoted-printable";
    part.setContentType(new arMIMEParams("text/html"
                                         + (info.charset
                                            ? ("; charset=\""
                                               + info.charset + "\"")
                                            : "")));
    part.addField("Content-Transfer-Encoding",
                  part.contentTransferEncoding);
    part.addField("Content-Location", location);
    part.body = htmlBody;
    topPart.parts.push(part);

    for (let resPath of info.resPaths) {
      let resBody = yield arFileUtils.readBinary(resPath);
      let resBasename = OS.Path.basename(resPath);
      let ext = arPathUtils.getExtension(resBasename);
      let mimetype = arMIMEUtils.getTypeFromExtension(ext);
      if (ext == "php" || ext == "cgi" ||
          mimetype == "application/octet-stream") {
        /* 基本的に application/octet-stream なファイルは保存されていないハズ */
        mimetype = arMIMEUtils.guessTypeFromBody(resBody, "text/plain");
      }

      let part = new arMIMEPart();
      part.contentTransferEncoding = "base64";
      part.setContentType(new arMIMEParams(mimetype));
      part.addField("Content-Transfer-Encoding",
                    part.contentTransferEncoding);
      part.addField("Content-Location",
                    info.actualResDirPath + info.refSep
                    + this.encode(info.refType, resBasename));

      part.body = resBody;

      topPart.parts.push(part);
    }

    let data = arMIMEEncoder.encodeMessage(topPart);
    topPart.destruct();

    yield arFileUtils.createUTF8(info.mhtPath, data);

    let htmlStat = yield OS.File.stat(info.htmlPath);
    yield OS.File.setDates(info.mhtPath, null, htmlStat.lastModificationDate);
  }),

  /**
   * ファイル名をエンコード
   *
   * @param   {UnMHTFileConverterHTMLRefType} refType
   *          リソースファイルへの参照時のエンコード方法
   * @param   {string} basename
   *          ファイル名
   * @returns {string}
   *          エンコードしたファイル名
   */
  encode: function(refType, basename) {
    switch (refType) {
      case UnMHTFileConverterHTMLRefType.RAW: {
        break;
      }
      case UnMHTFileConverterHTMLRefType.ENCODED: {
        basename = encodeURIComponent(basename);
        break;
      }
      case UnMHTFileConverterHTMLRefType.SPACE_ENCODED: {
        basename = basename.replace(/ /g, "%20");
        break;
      }
    }
    return basename;
  },

  /**
   * HTML の情報を解析して取得する
   *
   * @param   {string} htmlBody
   *          HTML ファイルの内容
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @yields  {Promise.<*>}
   */
  _getHTMLInfoTask: function*(htmlBody, info) {
    info.charset = this.getCharset(htmlBody);

    function getBasenameVariants(path) {
      let basename = OS.Path.basename(path);
      let raw = basename;
      if (info.charset) {
        raw = arUconv.fromUnicode(basename, info.charset);
      }
      return {
        raw: raw,
        encoded: encodeURIComponent(basename),
        spaceEncoded: raw.replace(/ /g, "%20")
      };
    }

    let resDirBasenameVar = getBasenameVariants(info.resDirPath);
    let resBasenameVars = info.resPaths.map(getBasenameVariants);

    let refs = [];
    htmlBody
      .replace(this._elemRe, (matched, name, attrs) => {
        attrs
          .replace(this._attrRe, (matched, attr,
                                  dqpath, qpath, path) => {
            if (dqpath) {
              path = dqpath.replace(/\\\"/g, "\"");
            } else if (qpath) {
              path = qpath.replace(/\\\'/g, "\'");
            }

            refs.push(path);
          });
      });

    let maybeRefs = [];
    function pushMaybe(ref, relPath, refType, sep) {
      let dirBasename = ref.slice(0, -relPath.length);

      let prefix = "";
      if (dirBasename.startsWith("." + sep)) {
        prefix = dirBasename.slice(0, 1 + sep.length);
        dirBasename = dirBasename.slice(2);
      }
      if (dirBasename.indexOf(sep) != -1) {
        return;
      }

      if (info.charset) {
        dirBasename = arUconv.toUnicode(dirBasename, info.charset);
      }

      switch (refType) {
      case UnMHTFileConverterHTMLRefType.RAW: {
        break;
      }
      case UnMHTFileConverterHTMLRefType.ENCODED: {
        try {
          dirBasename = decodeURIComponent(dirBasename);
        } catch (e) {
          return;
        }
        break;
      }
      case UnMHTFileConverterHTMLRefType.SPACE_ENCODED: {
        dirBasename = dirBasename.replace(/%20/g, " ");
        break;
      }
      }

      maybeRefs.push([dirBasename, prefix, refType, sep]);
    }

    function checkRef(ref) {
      for (let resBasenameVar of resBasenameVars) {
        for (let sep of ["/", "\\"]) {
          let relPath = resDirBasenameVar.encoded + sep + resBasenameVar.encoded;
          if (ref.endsWith(relPath)) {
            info.refType = UnMHTFileConverterHTMLRefType.ENCODED;
            info.refSep = sep;
            let path = ref.slice(0, -(sep + resBasenameVar.encoded).length);
            if (info.charset) {
              path = arUconv.toUnicode(path, info.charset);
            }
            info.actualResDirPath = path;
            return true;
          }

          if (ref.indexOf("%20") != -1) {
            relPath = resDirBasenameVar.spaceEncoded + sep + resBasenameVar.spaceEncoded;
            if (ref.endsWith(relPath)) {
              info.refType = UnMHTFileConverterHTMLRefType.SPACE_ENCODED;
              info.refSep = sep;
              let path = ref.slice(0, -(sep + resBasenameVar.spaceEncoded).length);
              if (info.charset) {
                path = arUconv.toUnicode(path, info.charset);
              }
              info.actualResDirPath = path;
              return true;
            }
          }

          relPath = resDirBasenameVar.raw + sep + resBasenameVar.raw;
          if (ref.endsWith(relPath)) {
            info.refType = UnMHTFileConverterHTMLRefType.RAW;
            info.refSep = sep;
            let path = ref.slice(0, -(sep + resBasenameVar.raw).length);
            if (info.charset) {
              path = arUconv.toUnicode(path, info.charset);
            }
            info.actualResDirPath = path;
            return true;
          }

          /* ファイル名が NFD になっていて一致しない時など
           * 実際にファイルの存在を確かめる
           *
           * FIXME: String.prototype.normalize を先に試す */

          relPath = sep + resBasenameVar.encoded;
          if (ref.endsWith(relPath)) {
            pushMaybe(ref, relPath,
                      UnMHTFileConverterHTMLRefType.ENCODED, sep);
          }

          if (ref.indexOf("%20") != -1) {
            relPath = sep + resBasenameVar.spaceEncoded;
            if (ref.endsWith(relPath)) {
              pushMaybe(ref, relPath,
                        UnMHTFileConverterHTMLRefType.SPACE_ENCODED, sep);
            }
          }

          relPath = sep + resBasenameVar.raw;
          if (ref.endsWith(relPath)) {
            pushMaybe(ref, relPath,
                      UnMHTFileConverterHTMLRefType.RAW, sep);
          }
        }
      }

      return false;
    }

    let exactMatch = false;
    for (let ref of refs) {
      if (checkRef(ref)) {
        exactMatch = true;
        break;
      }
    }

    let maybeMatch = false;
    if (!exactMatch) {
      for (let [dirBasename, prefix, refType, refSep] of maybeRefs) {
        let dirPath = OS.Path.join(info.parentDirPath, dirBasename);
        let exists = yield OS.File.exists(dirPath);
        if (!exists) {
          continue;
        }

        maybeMatch = true;
        info.actualResDirPath = prefix + dirBasename;
        info.refType = refType;
        info.refSep = refSep;
        break;
      }

      if (!maybeMatch) {
        info.refType = UnMHTFileConverterHTMLRefType.RAW;
        let path = resDirBasenameVar.raw;
        if (info.charset) {
          path = arUconv.toUnicode(path, info.charset);
        }
        info.actualResDirPath = path;
        info.refSep = "/";
      }
    }

    info.exactMatch = exactMatch;
    info.maybeMatch = maybeMatch;
  },

  /**
   * HTML の charset を取得する
   *
   * @param   {string} htmlBody
   *          HTML ファイルの内容
   * @returns {string}
   *          charset
   */
  getCharset: function(htmlBody) {
    let charset = "";

    htmlBody
      .replace(this._metaRe, (matched, name, attrs) => {
        let metaHTTPEquiv = "";
        let metaContent = "";

        attrs
          .replace(this._metaAttrRe, (matched, attr,
                                      dqpath, qpath, path) => {
            if (dqpath) {
              path = dqpath.replace(/\\\"/g, "\"");
            } else if (qpath) {
              path = qpath.replace(/\\\'/g, "\'");
            }

            if (/^charset$/i.test(attr)) {
              if (!charset) {
                charset = path;
              }
            } else if (/^http-equiv$/i.test(attr)) {
              metaHTTPEquiv = path;
            } else if (/^content$/i.test(attr)) {
              metaContent = path;
            }
          });

        if (!charset) {
          if (metaHTTPEquiv.toLowerCase() == "content-type") {
            let params = arMIMEDecoder.decodeParam(metaContent);
            if (params.hasParam("charset")) {
              charset = params.getParam("charset");
            }
          }
        }
      });

    return charset;
  },

  _metaRe: /<(meta)(\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)>/ig,
  _metaAttrRe: /\s(charset|http-equiv|content)\s*=\s*(?:\"((?:\\\"|[^\"])*)\"|\'((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))/ig,

  _elemRe: /<([A-Za-z0-9_:\-]+)(\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)>/g,
  _attrRe: /\s(href|src|background|data)\s*=\s*(?:\"((?:\\\"|[^\"])*)\"|\'((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))/ig
});

module.UnMHTFileConverter = UnMHTFileConverter;
module.UnMHTFileConverterHTMLRefType = UnMHTFileConverterHTMLRefType;

})(this);
