"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTFileLoadHook", "UnMHTOpenExtensions"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arPathUtils } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});
let { UnMHTDLFHandledLocations } = Cu.import("resource://unmht/modules/UnMHTDocumentLoaderFactory.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileLoader",
                                  "resource://unmht/modules/UnMHTFileLoader.jsm");

/**
 * 開く拡張子
 *
 * @constructor
 */
function UnMHTOpenExtensionsClass() {
  /**
   * 開く拡張子
   * @type {Array.<string>}
   */
  this._extensions = null;

  this.load();

  Object.seal(this);
}
UnMHTOpenExtensionsClass.prototype = Object.freeze({
  /**
   * ハンドルする拡張子のリストを返す
   *
   * @returns {Array.<string>}
   *          ハンドルする拡張子のリスト
   */
  load: function() {
    let tmp = UnMHTPref.getPref("open.extensions");
    let list = [];
    try {
      list = JSON.parse(tmp);
      if (!list.every(n => typeof(n) == "string")) {
        list = [];
      }
    } catch (e) {
      list = [];
    }

    this._extensions = list;
  },

  /**
   * ハンドルする拡張子かどうかを返す
   *
   * @param   {string} location
   *          対象のドキュメントのアドレス
   * @returns {boolean}
   *          ハンドルする拡張子か
   */
  isExtensionForOpen: function(location) {
    if (!this._extensions.length) {
      return false;
    }

    let ext = arPathUtils.getExtension(location);
    if (ext && this._extensions.indexOf(ext.toLowerCase()) != -1) {
      return true;
    }

    return false;
  }
});
let UnMHTOpenExtensions = new UnMHTOpenExtensionsClass();

/**
 * ファイルロードのフック
 *
 * @class
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
let UnMHTFileLoadHook = Object.freeze({
  /**
   * ページがロードされたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のイベント
   * @param   {string} location
   *          アドレス
   */
  onPageLoad: function(window, browser, location) {
    if (UnMHTDLFHandledLocations.has(location)) {
      /* ハンドル済みの mht ファイル */
      UnMHTDLFHandledLocations.delete(location);

      return;
    }

    if (UnMHTOpenExtensions.isExtensionForOpen(location)) {
      /* 未ハンドルの mht ファイル */

      if (location.startsWith("chrome://ietab/")) {
        /* IETab がハンドル済み */
        return;
      }

      if (UnMHTDocumentCategory.maybeHTML(browser)) {
        /* HTML っぽい */
        return;
      }

      let loadParam = {
        window:         window,
        uri:            Services.io.newURI(location, null, null),
        browser:        browser,
        hasDocument:    true,
        replaceHistory: true,
        refName:        "",
        useCache:       true
      };
      lazy.UnMHTFileLoader.loadWithParam(loadParam);

      return;
    }

    if (!location.startsWith("unmht:")) {
      return;
    }

    if (UnMHTScheme.isUnMHTURI(location)) {
      lazy.UnMHTFileLoader.onUnMHTDocumentLoaded(window, browser);
    }
  },

  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    arWindowWatcher.addListener(this);

    let APP_STARTUP = 1;

    if (reason != APP_STARTUP) {
      setTimeout(() => {
        for (let window of arWindowUtils.getBrowserWindows()) {
          this._reloadTabs(window);
        }
      }, 100);
    }
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!arWindowUtils.isBrowser(window)) {
      return;
    }

    if (arAppInfo.isFennec) {
      arWindowUtils.addPageLoadListener(window, event => {
        let targetWindow = event.target.defaultView;
        let browser = window.BrowserApp.getBrowserForWindow(targetWindow);
        let location = browser.currentURI.spec;

        this.onPageLoad(window, browser, location);
      });

      return;
    }

    let self = this;
    arWindowUtils.addProgressListener(window, {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsISupportsWeakReference,
                                             Ci.nsIWebProgressListener]),
      onLocationChange: function(webProgress, request, location) {},
      onProgressChange: function(webProgress, request,
                                 curSelfProgress, maxSelfProgress,
                                 curTotalProgress, maxTotalProgress) {},
      onSecurityChange: function(webProgress, request, state) {},
      onStateChange: function(webProgress, request, flags, status) {
        if (flags & Ci.nsIWebProgressListener.STATE_STOP) {
          let location;
          try {
            location = request.QueryInterface(Ci.nsIChannel).URI.spec;
          } catch (e) {
            return;
          }
          let browser = arWindowUtils.getBrowserForWebProgress(window,
                                                               webProgress);
          if (!browser) {
            return;
          }
          try {
            self.onPageLoad(window, browser, location);
          } catch (e) {
            Cu.reportError(e + "\n" + e.stack);
          }
        }
      },
      onStatusChange: function(webProgress, request, status, message) {}
    });

    /* セッションリストアは addProgressListener では検出できない */
    let tabbrowser = arWindowUtils.getTabBrowser(window);
    if (tabbrowser &&
        "tabContainer" in tabbrowser &&
        tabbrowser.tabContainer) {
      arEvent.add(window, tabbrowser.tabContainer, "SSTabRestored", event => {
        let tab = event.target;
        if (!tab) {
          return;
        }
        let browser = tabbrowser.getBrowserForTab(tab);
        if (!browser) {
          return;
        }
        let location = browser.currentURI.spec;

        this.onPageLoad(window, browser, location);
      }, false);
    }

    this._reloadTabs(window);
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
  },

  /* ==== private ==== */

  /**
   * タブにある MHT ファイルを展開する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _reloadTabs: function(window) {
    for (let browser of arWindowUtils.getBrowsers(window)) {
      let location = browser.currentURI.spec;
      if (!UnMHTScheme.isUnMHTURI(location)) {
        continue;
      }

      let param = UnMHTScheme.getUnMHTURIParam(location);

      if (param.scheme == "unmht-view-source") {
        location = UnMHTScheme.deViewSourceURI(location);
        param = UnMHTScheme.getUnMHTURIParam(location);
      }

      if (!param.scheme.startsWith("unmht-")) {
        let uri;
        try {
          uri = Services.io.newURI(param.original, null, null);
        } catch (e) {
          continue;
        }

        let loadParam = {
          window:         window,
          uri:            uri,
          browser:        browser,
          hasDocument:    false,
          replaceHistory: true,
          refName:        param.refName,
          useCache:       false
        };
        lazy.UnMHTFileLoader.loadWithParam(loadParam);
      } else {
        let flags = Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY;
        browser.loadURIWithFlags(location, flags, null, null, null);
      }
    }
  }
});

module.UnMHTFileLoadHook = UnMHTFileLoadHook;
module.UnMHTOpenExtensions = UnMHTOpenExtensions;

})(this);
