"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTFileLoader"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { fromUint8Array } = Cu.import("resource://unmht/modules/base/arArrayBufferUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { toUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { UnMHTCache } = Cu.import("resource://unmht/modules/UnMHTCache.jsm", {});
let { UnMHTFileLoader_content } = Cu.import("resource://unmht/modules/UnMHTFileLoader_content.jsm", {});
let { UnMHTExtractor } = Cu.import("resource://unmht/modules/UnMHTExtractor.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arNetFileLoader",
                                  "resource://unmht/modules/base/arNetFileLoader.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arNetFileLoaderStatus",
                                  "resource://unmht/modules/base/arNetFileLoader.jsm");


/**
 * MHT ファイルのロード
 *
 * @class
 */
let UnMHTFileLoader = Object.freeze({
  /**
   * 展開して表示
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   */
  openWithExtract: function(window, browser) {
    let location = browser.currentURI.spec;
    let param = UnMHTScheme.getUnMHTURIParam(location);

    if (param.scheme == "unmht-view-source") {
      browser.loadURI(UnMHTScheme.deViewSourceURI(location), null, null);
    } else {
      this._loadFromBrowser(window, browser);
    }
  },

  /**
   * ソースを表示
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   */
  openWithViewSource: function(window, browser) {
    let location = browser.currentURI.spec;

    browser.loadURI(UnMHTScheme.enViewSourceURI(location), null, null);
  },

  /* ---- UI イベント ----- */

  /**
   * unmht: scheme のドキュメントがロードされたイベント
   * mht ファイルの展開、イベントのフックを行う
   * UnMHTFileLoadHook から呼ばれる
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   */
  onUnMHTDocumentLoaded: function(window, browser) {
    let location = browser.currentURI.spec;
    let param = UnMHTScheme.getUnMHTURIParam(location);
    let original = param.original;
    let refName = param.refName;
    let scheme = param.scheme;

    if (scheme == "unmht-error") {
      UnMHTFileLoader_content.hookRetryButton(browser);
      return;
    } else if (scheme.startsWith("unmht-")) {
      return;
    }

    let [eFileInfo, part] = UnMHTCache.getPart(location);
    if (!eFileInfo) {
      /* 未展開のファイルだった場合 */
      let uri;
      try {
        uri = Services.io.newURI(original, null, null);
      } catch (e) {
        return;
      }

      let loadParam = {
        window:         window,
        uri:            uri,
        browser:        browser,
        hasDocument:    false,
        replaceHistory: false,
        refName:        refName,
        useCache:       false
      };
      this.loadWithParam(loadParam);

      return;
    }

    if (part && part == eFileInfo.startPart) {
      let onReload = location => {
        let param = UnMHTScheme.getUnMHTURIParam(location);

        let loadParam = {
          window:         window,
          uri:            Services.io.newURI(param.original, null, null),
          browser:        browser,
          hasDocument:    false,
          replaceHistory: true,
          refName:        param.refName,
          useCache:       false,
          doReload:       true
        };
        UnMHTFileLoader.loadWithParam(loadParam);
      };

      UnMHTFileLoader_content.hookReload(browser, onReload);

      if (eFileInfo.topPart.hasField("X-UnMHT-Tabs")) {
        /* 複数のファイルを含む MHT の
         * インデックスを開いた */

        let action = UnMHTPref.getPref("open.multiple.action");
        if (action == "tabs" || action == "first") {
          this._replaceTabs(window, browser, action, eFileInfo);
        }
      }
    }
  },

  /**
   * mht ファイルを展開して開く
   *
   * @param   {Object} loadParam
   *          ロードのパラメータ
   *            {
   *              window:         {nsIDOMWindow}
   *                              実行中のウィンドウ
   *              uri:            {nsIURI}
   *                              開く URI
   *              browser:        {nsIXULElement}
   *                              対象のブラウザ
   *              hasDocument:    {boolean}
   *                              ドキュメントに内容を持っているか
   *              replaceHistory: {boolean}
   *                              履歴を置き換えるかどうか
   *              refName:        {string}
   *                              対象のファイル
   *              useCache:       {boolean}
   *                              キャッシュを使用するか
   *            }
   * @returns {Promise.<boolean>}
   *          開いたか
   */
  loadWithParam: Task.async(function*(loadParam) {
    let { uri } = loadParam;
    let result;

    if (uri.scheme == "file") {
      let filePath = arFileUtils.URLSpecToPath(uri.spec);
      result = yield this.loadWithParamFromFile(loadParam, filePath);
    } else {
      result = yield this.loadWithParamFromWeb(loadParam);
    }

    return result;
  }),

  /**
   * ファイルから mht ファイルを展開して開く
   *
   * @param   {Object} loadParam
   *          ロードのパラメータ
   * @param   {string} targetFilePath
   *          開くファイルのパス
   * @returns {Promise.<boolean>}
   *          開いたか
   */
  loadWithParamFromFile: Task.async(function*(loadParam, targetFilePath) {
    let { window, browser, uri, hasDocument } = loadParam;
    let data;
    try {
      data = yield arFileUtils.readBinary(targetFilePath);
    } catch (e) {
      data = "";
    }

    let result;

    if (data) {
      result = yield this.loadWithParamFromData(loadParam, data);
    } else {
      this.loadErrorPage(window, browser, uri.spec);
      result = false;
    }

    return result;
  }),

  /**
   * データから mht ファイルを展開して開く
   *
   * @param   {Object} loadParam
   *          ロードのパラメータ
   * @param   {string} data
   *          対象のデータ
   * @returns {Promise.<boolean>}
   *          開いたか
   */
  loadWithParamFromData: Task.async(function*(loadParam, data) {
    let { uri, browser, replaceHistory, refName } = loadParam;

    let eFileInfo = yield UnMHTExtractor.extractMHT(uri.spec, data);
    if (!eFileInfo) {
      return false;
    }

    if ("doReload" in loadParam && loadParam.doReload) {
      browser.reload();
      return true;
    }

    let urlspec = eFileInfo.baseURI;
    if (refName &&
        eFileInfo.refParts.has(refName)) {
      urlspec += refName;
    }

    let flags = replaceHistory
      ? Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY
      : 0;
    browser.loadURIWithFlags(urlspec, flags, null, null, null);

    return true;
  }),

  /**
   * ウェブのファイルから mht ファイルを展開して開く
   *
   * @param   {Object} loadParam
   *          ロードのパラメータ
   * @returns {Promise.<boolean>}
   *          開いたか
   */
  loadWithParamFromWeb: Task.async(function*(loadParam) {
    let { window, uri, browser, hasDocument, useCache } = loadParam;

    let referrer = null;
    if (hasDocument) {
      try {
        referrer
          = Services.io.newURI(browser.currentURI.spec, null, null);
      } catch (e) {
      }
    }

    let result = yield lazy.arNetFileLoader.loadData(uri, referrer, useCache,
                                                     window);
    if (result.status == lazy.arNetFileLoaderStatus.NG) {
      this.loadErrorPage(window, browser, uri.spec);
    } else {
      yield this.loadWithParamFromData(loadParam, result.data);
    }
  }),

  /**
   * エラーメッセージを表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のウィンドウ
   * @param   {string} location
   *          ファイルの URI
   */
  loadErrorPage: function(window, browser, location) {
    browser.loadURIWithFlags(UnMHTScheme.enErrorURI(location),
                             Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY,
                             null, null, null);
  },

  /* ==== private ==== */

  /**
   * インデックスページをタブに置き換える
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {string} action
   *          複数のタブを含む MHT を開く時
   *            tabs: 全てのページをタブとして開く
   *            index: インデックスページを表示する
   *            first: 最初のタブを開く
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          開始パートの展開情報
   */
  _replaceTabs: function(window, browser, action, eFileInfo) {
    let tabbrowser = arWindowUtils.getTabBrowser(window);

    let urls = null;
    if (eFileInfo.topPart.hasField("X-UnMHT-Tabs-List")) {
      urls = eFileInfo.topPart.getField("X-UnMHT-Tabs-List")
        .split(",").map(s => s.replace(/^<|>$/g, ""))
        .filter(cid => eFileInfo.cids.has(cid))
        .map(cid => eFileInfo.cids.get(cid))
        .filter(part => (part != eFileInfo.topPart &&
                         part != eFileInfo.startPart))
        .map(part => eFileInfo.baseURI + part.eParam.refName);
    } else {
      urls = UnMHTFileLoader_content.getTabsFromDocument(browser);
    }

    if (action == "first") {
      urls = urls.slice(0, 1);
    }

    if (urls.length === 0) {
      return;
    }

    browser.loadURIWithFlags(urls[0], 0, null, null, null);
    for (let url of urls.slice(1)) {
      tabbrowser.addTab(url, {relatedToCurrent: true, selected: false});
    }
  },

  /**
   * 対象のドキュメントから mht ファイルを開く
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   */
  _loadFromBrowser: function(window, browser) {
    let uri = Services.io.newURI(browser.currentURI.spec, null, null);

    if (UnMHTScheme.isUnMHTURI(uri.spec)) {
      uri.spec = UnMHTScheme.deUnMHTURI(uri.spec);
    }

    let loadParam = {
      window:         window,
      uri:            uri,
      browser:        browser,
      hasDocument:    true,
      replaceHistory: false,
      refName:        "",
      useCache:       true
    };
    this.loadWithParam(loadParam);
  }
});

module.UnMHTFileLoader = UnMHTFileLoader;

})(this);
