"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTFileLoader"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arScopeManager } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { UnMHTCache } = Cu.import("resource://unmht/modules/UnMHTCache.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

/**
 * MHT ファイルのロードのコンテント側
 *
 * @class
 */
let UnMHTFileLoader_content = Object.freeze({
  /**
   * MHT ファイルのタブに対応するリンクの URI を返す
   *
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @returns {Array.<string>}
   *          MHT ファイルのタブに対応するリンクの URI
   */
  getTabsFromDocument: function(browser) {
      let location = browser.currentURI.spec;
    let [eFileInfo, part] = UnMHTCache.getPart(location);

    let targetDocument = browser.contentDocument;
    let links = Array.from(targetDocument.getElementsByTagName("a"))
        .filter(node => ("className" in node &&
                         node.className == "unmht-tabs"))
        .map(node => node.href);

    let urls = links.map(link => {
      let [eFileInfo2, part] = UnMHTCache.getPart(link);
      if (!eFileInfo || !part) {
        return null;
      }
      if (eFileInfo2 != eFileInfo) {
        return null;
      }
      return part;
    })
        .filter(part => (part &&
                         part != eFileInfo.topPart &&
                         part != eFileInfo.startPart))
        .map(part => eFileInfo.baseURI + part.eParam.refName);

    return urls;
  },

  /**
   * リロードをフックする
   *
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {function(string)} onReload
   *          リロード時のコールバック
   */
  hookReload: function(browser, onReload) {
    let nav = browser.webNavigation;
    let targetWindow = browser.contentWindow;

    let listener = {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsISupportsWeakReference,
                                             Ci.nsISHistoryListener]),

      OnHistoryReload: null,
      OnHistoryGoBack: function() { return true; },
      OnHistoryGoForward: function() { return true; },
      OnHistoryGotoIndex: function() { return true; },
      OnHistoryNewEntry: function() { return true; },
      OnHistoryPurge: function() { return true; }
    };

    let onUnload = () => {
      try {
        nav.sessionHistory.removeSHistoryListener(listener);
      } catch (e) {
      }

      arScopeManager.leave(targetWindow);
    };

    listener.OnHistoryReload = (reloadURI, reloadFlags) => {
      if (!UnMHTScheme.isUnMHTURI(reloadURI.spec)) {
        return true;
      }

      onUnload();

      onReload(reloadURI.spec);

      return false;
    };

    try {
      nav.sessionHistory.addSHistoryListener(listener);
    } catch (e) {
      return;
    }

    arEvent.oneshot(targetWindow, targetWindow, "unload", onUnload, true);
  },

  /**
   * エラーページのリトライボタンをフックする
   *
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   */
  hookRetryButton: function(browser) {
    let location = browser.currentURI.spec;
    let targetWindow = browser.contentWindow;
    let targetDocument = browser.contentDocument;

    let retryButton = targetDocument.getElementById("unmht-retry-button");
    if (!retryButton) {
      return;
    }

    arEvent.add(targetWindow, retryButton, "click", () => {
      try {
        targetWindow.location.href = UnMHTScheme.deErrorURI(location);
      } catch (e) {
      }
    }, false);

    arEvent.oneshot(targetWindow, targetWindow, "unload", () => {
      arScopeManager.leave(targetWindow);
    }, true);
  }
});

module.UnMHTFileLoader_content = UnMHTFileLoader_content;

})(this);
