"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTFileSaver",
                         "UnMHTFileSaverResult",
                         "UnMHTFileSaverParam"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arAlertUtils } = Cu.import("resource://unmht/modules/base/arAlertUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTSave, UnMHTSaveResult } = Cu.import("resource://unmht/modules/UnMHTSave.jsm", {});
let { UnMHTSaveState, UnMHTSaveStateMask, UnMHTSaveStateUtils } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arFilePicker",
                                  "resource://unmht/modules/base/arFilePicker.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arMIMEDecoder",
                                  "resource://unmht/modules/base/arMIMEDecoder.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arNetFileLoader",
                                  "resource://unmht/modules/base/arNetFileLoader.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arNetFileLoaderStatus",
                                  "resource://unmht/modules/base/arNetFileLoader.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTMail",
                                  "resource://unmht/modules/UnMHTMail.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTAdditionalDialog",
                                  "resource://unmht/modules/UnMHTAdditionalDialog.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTConfirmDialog",
                                  "resource://unmht/modules/UnMHTConfirmDialog.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSaveDialog",
                                  "resource://unmht/modules/UnMHTSaveDialog.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSaveOptionsDialog",
                                  "resource://unmht/modules/UnMHTSaveOptionsDialog.jsm");
XPCOMUtils.defineLazyServiceGetter(lazy, "MsgMailSession",
                                   "@mozilla.org/messenger/services/session;1",
                                   "nsIMsgMailSession");

/**
 * 保存結果 (エラー)
 * @enum {number}
 */
const UnMHTFileSaverResult = Object.freeze({
  /** window が無い */
  NO_WINDOW:        0x20000101,
  /** browser も scrapBookData も無い */
  NO_TARGET:        0x20000102,
  /** ブラウザが無い */
  NO_BROWSER:       0x20000105,

  /** targetFilePath がファイルではない */
  NOT_A_FILE:       0x20000201,
  /** ドキュメントがローカルファイルではない
   *  (targetFilePath が UnMHTFileSaverParam.OVERWRITE の場合) */
  NOT_A_LOCAL_FILE: 0x20000202,
  /** リンクに載っていない */
  NOT_A_LINK:       0x20000203,
  /** ドキュメントがメールではない */
  NOT_A_MAIL:       0x20000204,
  /** 保存可能ではない */
  NOT_SAVABLE:      0x20000205,
});

/**
 * saveWithParam のパラメータ
 *
 * @class
 */
const UnMHTFileSaverParam = Object.freeze({
  CHOOSE:    {},
  CURRENT:   {},
  MAIL:      {},
  OVERWRITE: {},
});

/**
 * MHT で保存のラッパー
 * UI からの要求に応じてデータを取得して MHT での保存を開始する
 *
 * @class
 */
let UnMHTFileSaver = Object.freeze({
  /**
   * 現在のドキュメントを MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Object=} paramOverride
   *          追加のパラメータ
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  save: function(window, paramOverride=null) {
    let param = {
      window:         window,
      browser:        UnMHTFileSaverParam.CURRENT,
    };
    if (paramOverride) {
      for (let name of Object.keys(paramOverride)) {
        param[name] = paramOverride[name];
      }
    }
    return this.saveWithParam(param);
  },

  /**
   * 対象のドキュメントを MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  saveDocument: function(window, browser, targetDocument) {
    let param = {
      window:         window,
      browser:        browser,
      targetDocument: targetDocument,
    };
    return this.saveWithParam(param);
  },

  /**
   * 現在のドキュメントの選択範囲を保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  saveSelection: function(window) {
    let param = {
      window:       window,
      browser:      UnMHTFileSaverParam.CURRENT,
      useSelection: true
    };
    return this.saveWithParam(param);
  },

  /**
   * 対象のドキュメントの現在の選択範囲を MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  saveSelectionInDocument: function(window, browser, targetDocument) {
    let param = {
      window:         window,
      browser:        browser,
      targetDocument: targetDocument,
      useSelection:   true
    };
    return this.saveWithParam(param);
  },

  /**
   * 現在のドキュメントを MHT に上書き保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  overwrite: function(window) {
    let param = {
      window:          window,
      browser:         UnMHTFileSaverParam.CURRENT,
      targetFilePath:  UnMHTFileSaverParam.OVERWRITE,
    };
    return this.saveWithParam(param);
  },

  /**
   * リンク先を MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  saveLink: function(window) {
    let href = this._getLinkURL(window);
    if (!href) {
      let arg = {
        title:    TEXT("name"),
        message:  TEXT("save.link.error.not_link"),
        accept:   TEXT("ok"),
        focus:    "yes"
      };
      lazy.UnMHTConfirmDialog.open(window, arg);

      return Promise.reject(UnMHTFileSaverResult.NOT_A_LINK);
    }

    let tabbrowser = arWindowUtils.getTabBrowser(window);
    if (!tabbrowser) {
      return Promise.reject(UnMHTFileSaverResult.NO_BROWSER);
    }

    return new Promise((resolve, reject) => {
      let browser = window.document.createElement("browser");
      browser.collapsed = true;
      browser.setAttribute("src", href);
      browser.setAttribute("type", "content");
      let onload = () => {
        browser.removeEventListener("load", onload, true);

        let param = {
          window:            window,
          browser:           browser,
          noHandleRejection: true
        };
        let p = Task.spawn(function*() {
          let result = yield this.saveWithParam(param);
          browser.remove();
          resolve(result);
        }.bind(this));
        this._handleRejection(param, p);
      };
      browser.addEventListener("load", onload, true);

      tabbrowser.parentNode.insertBefore(browser, tabbrowser.nextSibling);
    });
  },

  /**
   * QuickSave
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {boolean=} useSelection
   *          選択範囲を使用するか
   * @param   {Object=} paramOverride
   *          追加のパラメータ
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  quicksave: function(window, useSelection=false, paramOverride=null) {
    let targetFilePathFunc = Task.async(function*(param) {
      let parentPath = UnMHTPref.getPref("save.quick.dir");

      if (parentPath) {
        let exists = yield OS.File.exists(parentPath);
        if (!exists) {
          try {
            yield OS.File.makeDir(parentPath);
          } catch (e) {
            parentPath = null;
          }
        }
      }

      if (UnMHTPref.getPref("save.local.same_dir")) {
        let href = param.browser.currentURI.spec;
        let localFilePath = { value: null };
        if (UnMHTDocumentCategory.isLocalFile(href, localFilePath)) {
          parentPath = OS.Path.dirname(localFilePath.value);
        }
      }

      if (!parentPath) {
        parentPath = yield arFileUtils.getDefaultDownloadsDirectory();
      }

      let targetFilePath = "";
      let dupAction = UnMHTPref.getPref("save.quick.dup.action");
      if (dupAction == "rename") {
        targetFilePath = yield UnMHTSave.getUniqueFilePath(parentPath,
                                                           param.filenameBase);
      } else {
        let suffix = "." + UnMHTFilename.defaultExtension;
        let options = { allowExists: true, noCreate: true };
        let result = yield arFileUtils.prepareFile(parentPath,
                                                   param.filenameBase,
                                                   suffix, options);
        targetFilePath = result.path;
        if (result.exists) {
          if (arAppInfo.isFennec && dupAction == "ask") {
            dupAction = "rename";
          }

          if (dupAction == "ask") {
            let oResult = yield UnMHTSave.showOverwriteConfirm(
              param.window, targetFilePath, param.filenameBase,
              false, false);
            if (oResult.action == "abort" || oResult.action == "no") {
              throw UnMHTSaveResult.CANCEL;
            }

            if (oResult.action == "rename") {
              targetFilePath = oResult.renamePath;
            }
          }
        }
      }

      return targetFilePath;
    });

    let param = {
      window:            window,
      browser:           UnMHTFileSaverParam.CURRENT,
      useSelection:      (useSelection ||
                          UnMHTPref.getPref("save.quick.use_selection")),
      targetFilePath:    targetFilePathFunc,
      additionalURIs:    null,
      showOptionDialog:  false,
      saveState:         UnMHTSaveStateUtils.stringToSaveState(UnMHTPref.getPref("save.quick.state")),
      noHandleRejection: true
    };
    let p = Task.spawn(function*() {
      if (paramOverride) {
        for (let name of Object.keys(paramOverride)) {
          param[name] = paramOverride[name];
        }
      }

      let ret = yield this.saveWithParam(param);

      return ret;
    }.bind(this));
    this._handleRejection(param, p);
    return p;
  },

  /**
   * 現在のドキュメントを MHT としてメールで送信
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  sendmail: function(window) {
    let param = {
      window:         window,
      browser:        UnMHTFileSaverParam.CURRENT,
      targetFilePath: UnMHTFileSaverParam.MAIL,
    };
    return this.saveWithParam(param);
  },

  /**
   * 対象のドキュメントを MHT としてメールで送信
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  sendmailDocument: function(window, browser, targetDocument) {
    let param = {
      window:         window,
      browser:        browser,
      targetDocument: targetDocument,
      targetFilePath: UnMHTFileSaverParam.MAIL,
    };
    return this.saveWithParam(param);
  },

  /**
   * ScrapBook を MHT で保存する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Object} scrapBookData
   *          ScrapBook のデータ
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  saveScrapBook: function(window, scrapBookData) {
    let param = {
      window:           window,
      scrapBookData:    scrapBookData,
      showOptionDialog: false,
    };
    return this.saveWithParam(param);
  },

  /**
   * パラメータに従ってドキュメントを保存する
   *
   * @param   {Object} param
   *          保存パラメータ
   *            {
   *              window: {nsIDOMWindow}
   *                      実行中のウィンドウ
   *              browser: {nsIXULElement=}
   *                       対象のドキュメント
   *                       (scrapBookData とどちらか必須)
   *                         * UnMHTFileSaverParam.CURRENT
   *                             現在のブラウザ
   *                         * nsIXULElement
   *                             指定したブラウザ
   *              targetDocument: {nsIDOMDocument=}
   *                              ブラウザ内のドキュメント
   *              scrapBookData: {Object=}
   *                             ScrapBook のデータ
   *                             (browser とどちらか必須)
   *              date: {Date=}
   *                    Dateフィールドの値
   *                      * UnMHTFileSaverParam.CURRENT
   *                          現在時刻
   *                      * UnMHTFileSaverParam.MAIL (デフォルト)
   *                          メールならばメールの Date フィールド
   *                          そうでないなら現在時刻
   *                      * Date
   *                          指定した時刻
   *              useSelection: {boolean=}
   *                            選択範囲を使用する
   *              showOptionDialog: {boolean=}
   *                                オプションダイアログを表示するか
   *                                指定しない場合設定による
   *              targetFilePath: {string=}
   *                              保存先
   *                                * UnMHTFileSaverParam.CHOOSE (デフォルト)
   *                                    ファイル選択ダイアログを表示
   *                                * UnMHTFileSaverParam.OVERWRITE
   *                                    上書き
   *                                * UnMHTFileSaverParam.MAIL
   *                                    メールで送信
   *                                * (function(Object):Promise.<string>)
   *                                    ファイルを返す関数
   *                                      function(param)
   *                                        *param   {Object} param
   *                                        *returns {Promise.<string>}
   *                                * string
   *                                    指定したファイル
   *              filenameBase: {string=}
   *                            ファイル名の基本部分
   *                              * string
   *                                  指定した文字列
   *                              * null (デフォルト)
   *                                  自動生成
   *              filenameType: {string=}
   *                            ファイル名の種類
   *                              * string
   *                                  title/leafname/custom
   *                              * null (デフォルト)
   *                                  設定による
   *              additionalURIs: {Array.<string>=}
   *                              追加のファイル
   *                                * UnMHTFileSaverParam.CHOOSE (デフォルト)
   *                                    ダイアログを表示
   *                                * Array.<string>
   *                                    指定したファイル
   *                                * null
   *                                    なし
   *              saveState: {number=}
   *                         保存状態
   *                         指定しない場合は選択する
   *              downloadListener: {nsIDownloadProgressListener=}
   *                                 ダウンロードのイベントリスナ
   *              button: {string=}
   *                      プログレスを表示するツールバーの ID
   *            }
   * @param   {function(Object,number)} f
   *          終了時コールバック
   *            function(param, result)
   *              *param   {Object} param
   *              *param   {number} result
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {(UnMHTSaveResult|UnMHTFileSaverResult)}
   */
  saveWithParam: function(param) {
    let p = Task.spawn(function*() {
      if (!("window" in param)) {
        throw UnMHTFileSaverResult.NO_WINDOW;
      }

      /* デフォルト値の設定 */
      if (!("browser" in param)) {
        param.browser = null;
      }
      if (!("targetDocument" in param)) {
        param.targetDocument = null;
      }
      if (!("scrapBookData" in param)) {
        param.scrapBookData = null;
      }
      if (!("date" in param)) {
        param.date = UnMHTFileSaverParam.MAIL;
      }
      if (!("useSelection" in param)) {
        param.useSelection = false;
      }
      if (!("showOptionDialog" in param)) {
        param.showOptionDialog
          = (UnMHTPref.getPref("save.show_option_dialog") ||
             param.targetFilePath == UnMHTFileSaverParam.OVERWRITE);
      }
      if (!("additionalURIs" in param)) {
        param.additionalURIs
          = (param.targetFilePath == UnMHTFileSaverParam.MAIL)
          ? null
          : UnMHTFileSaverParam.CHOOSE;
      }
      if (!param.additionalURIs) {
        param.additionalURIs = [];
      }
      if (!("filenameType" in param)) {
        param.filenameType = null;
      }
      if (!("filenameBase" in param)) {
        param.filenameBase = null;
      }
      if (!("targetFilePath" in param)) {
        param.targetFilePath = null;
      }
      if (!("downloadListener" in param)) {
        param.downloadListener = null;
      }
      if (!("saveState" in param)) {
        param.saveState
          = (param.targetFilePath == UnMHTFileSaverParam.MAIL)
          ? UnMHTSaveStateUtils.stringToSaveState(UnMHTPref.getPref("save.mail.state"))
          : null;
      }

      /* 保存対象 */
      if (!param.browser &&
          !param.scrapBookData) {
        param.browser = UnMHTFileSaverParam.CURRENT;
      }
      if (param.browser == UnMHTFileSaverParam.CURRENT) {
        param.browser = arWindowUtils.getCurrentBrowser(param.window);
        if (!UnMHTDocumentCategory.isSavable(param.window, param.browser, false)) {
          throw UnMHTFileSaverResult.NOT_SAVABLE;
        }
      }
      if (param.browser && !param.targetDocument) {
        param.targetDocument = param.browser.contentDocument;
      }
      if (!(param.targetDocument || param.scrapBookData)) {
        throw UnMHTFileSaverResult.NO_TARGET;
      }

      /* Date フィールド */
      if (param.date == UnMHTFileSaverParam.MAIL) {
        if (this._needMailDate(param.window)) {
          param.date = yield* this._getMailDateTask(param);
        } else {
          param.date = new Date();
        }
      }

      /* ファイル名の基本部分 */
      if (!param.filenameBase) {
        param.filenameBase
          = UnMHTSave.getFilenameBase(param.targetDocument,
                                      param.scrapBookData,
                                      param.filenameType,
                                      param.date);
      }

      /* オプションダイアログ */
      if (param.showOptionDialog) {
        param.showOptionDialog = false;
        let saveParam = yield this._getSaveOption(param);

        param.filenameBase = saveParam.filenameBase;
        param.saveState = saveParam.saveState;
        param.additionalURIs = saveParam.additionalURIs;
      }

      /* 保存先 */
      if (!param.targetFilePath ||
          param.targetFilePath == UnMHTFileSaverParam.CHOOSE) {
        yield* this._showFilePickerTask(param);
      }
      let isMail = false;
      if (typeof param.targetFilePath == "function") {
        param.targetFilePath = yield param.targetFilePath(param);
      } else if (param.targetFilePath == UnMHTFileSaverParam.OVERWRITE) {
        if (!param.targetDocument) {
          throw UnMHTFileSaverResult.NO_TARGET;
        }

        let loc = param.targetDocument.documentURI;
        let origURLSpec = UnMHTScheme.deUnMHTURI(loc);
        param.targetFilePath = arFileUtils.URLSpecToPath(origURLSpec);
        if (!param.targetFilePath) {
          throw UnMHTFileSaverResult.NOT_A_LOCAL_FILE;
        }
      } else if (param.targetFilePath == UnMHTFileSaverParam.MAIL) {
        isMail = true;

        let tempDirPath = OS.Path.join(OS.Constants.Path.tmpDir,
                                       Date.now() + "_" + Math.random());
        yield OS.File.makeDir(tempDirPath);

        let suffix = "." + UnMHTFilename.defaultExtension;
        let options = { allowExists: true, noCreate: true };
        let result = yield arFileUtils.prepareFile(tempDirPath,
                                                   param.filenameBase,
                                                   suffix, options);
        param.targetFilePath = result.path;
      }

      if (typeof param.targetFilePath != "string") {
        throw UnMHTFileSaverResult.NOT_A_FILE;
      }

      /* 追加ファイル */
      if (param.additionalURIs == UnMHTFileSaverParam.CHOOSE) {
        param.additionalURIs = yield this._getAdditionalURIs(param);
      }

      /* 保存開始 */
      let context = yield* this._createContextTask(param);

      /* リスナ作成 */
      let listener;
      if (arAppInfo.isFennec) {
        listener = this._getFennecListener(param, context);
      } else {
        switch(UnMHTPref.getPref("save.progress.display")) {
          case "dialog": {
            listener = yield this._getSaveDialogListener(param, context);
            break;
          }
          case "toolbar": {
            listener = yield this._getToolbarListener(param, context);
            break;
          }
          case "none": {
            listener = this._getNoneListener(param, context);
            break;
          }
        }
      }

      context.setLisener(listener);

      let saveResult = yield context.start();

      if (isMail) {
        /* メール送信 */
        let ret = yield lazy.UnMHTMail.send(param.window, param.targetFilePath,
                                            param.targetDocument.title);
        if (!ret) {
          let arg = {
            title:    TEXT("name"),
            message:  (TEXT("save.mail.fail") + "\n"
                       + TEXT("save.mail.fail2")),
            accept:   TEXT("ok"),
            focus:    "yes"
          };
          lazy.UnMHTConfirmDialog.open(param.window, arg);
        }
      }

      return saveResult;
    }.bind(this));
    if (!("noHandleRejection" in param && param.noHandleRejection)) {
      this._handleRejection(param, p);
    }
    return p;
  },

  /**
   * reject をキャッチして対応する
   *
   * @param   {Object} param
   *          保存パラメータ
   * @param   {Promise.<Tuple.<string,Array.<string>>>} p
   *          saveWithParam の返り値
   */
  _handleRejection: function(param, p) {
    p.catch(result => {
      if (result == UnMHTSaveResult.CANCEL ||
          result == UnMHTSaveResult.CANCEL) {
        return;
      }

      if ((typeof result) == 'number') {
        let name = "";
        for (let n of Object.keys(UnMHTFileSaverResult)) {
          if (result == UnMHTFileSaverResult[n]) {
            name = "UnMHTFileSaverResult." + n;
          }
        }
        for (let n of Object.keys(UnMHTSaveResult)) {
          if (result == UnMHTSaveResult[n]) {
            name = "UnMHTSaveResult." + n;
          }
        }
        Cu.reportError("onFail: " + name + "(" + result.toString(16) + ")");
      } else {
        let title = TEXT("error.title");
        let message = "";
        if (param.targetDocument &&
            !UnMHTDocumentCategory.isDocumentAlive(param.targetDocument)) {
          message += TEXT("save.error.document_closed");
        } else {
          message += result;
          let reportMessage = result;
          if ("operation" in result && result.operation){
            reportMessage += "\n" + result.operation;
          }
          if ("fileName" in result && result.fileName) {
            reportMessage += "\n" + result.fileName + ":" + result.lineNumber;
          }
          if ("stack" in result && result.stack) {
            reportMessage += result.stack;
          }
          Cu.reportError(reportMessage);
        }

        if (arAppInfo.isFennec) {
          arAlertUtils.alert(param.window, title, message);
        } else {
          let arg = {
            title: title,
            message: message,
            accept: TEXT("ok"),
            focus: "yes"
          };
          lazy.UnMHTConfirmDialog.open(param.window, arg);
        }
      }
    });
  },

  /**
   * オプションダイアログを表示する
   *
   * @param   {Object} param
   *          保存パラメータ
   * @returns {Promise.<Object>}
   * @resolves 保存オプション
   * @rejects  {number}
   *           失敗時の結果
   */
  _getSaveOption: function(param) {
    let saveStatePref
      = (param.targetFilePath == UnMHTFileSaverParam.MAIL)
      ? "save.mail.state"
      : "save.single.state";

    let saveState
      = UnMHTSaveStateUtils.stringToSaveState(UnMHTPref.getPref(saveStatePref));
    let saveStateMask = UnMHTSaveStateMask.ALL;

    if ("targetFilePath" in param &&
        param.targetFilePath == UnMHTFileSaverParam.OVERWRITE) {
      saveStateMask = UnMHTSaveStateMask.CURRENT;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSaveState.CURRENT;
      }
    } else if (param.scrapBookData) {
      saveStateMask = UnMHTSaveStateMask.ORIGINAL;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSaveState.ORIGINAL;
      }
    } else if (param.useSelection) {
      saveStateMask = UnMHTSaveStateMask.CURRENT;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSaveState.CURRENT;
      }
    }

    let arg = {
      filenameBase:   param.filenameBase,
      showFilename:   param.targetFilePath == UnMHTFileSaverParam.MAIL,
      saveState:      saveState,
      saveStateMask:  saveStateMask,
      saveStatePref:  saveStatePref,
      showInsertLink: param.targetFilePath == UnMHTFileSaverParam.OVERWRITE,
    };
    return lazy.UnMHTSaveOptionsDialog.open(param.window, arg);
  },

  /**
   * ファイル選択ダイアログを表示する
   *
   * @param   {Object} param
   *          保存パラメータ
   * @yields  {Promise.<*>}
   * @returns {number}
   *          成功時の結果
   * @throws  {number}
   *          失敗時の結果
   */
  _showFilePickerTask: function*(param) {
    let displayDirectory = null;
    let localFilePath = { value: null };
    if (param.targetDocument &&
        UnMHTPref.getPref("save.local.same_dir") &&
        UnMHTDocumentCategory.isLocalFile(param.targetDocument.documentURI,
                                          localFilePath)) {
      displayDirectory = localFilePath.value;
    }

    let types = [];
    if (param.saveState &&
        param.additionalURIs != UnMHTFileSaverParam.CHOOSE) {
      types = [[param.saveState, param.additionalURIs.length > 0]];
    } else {
      let saveStateMask = UnMHTSaveStateMask.ALL;

      if (param.scrapBookData) {
        saveStateMask = UnMHTSaveStateMask.ORIGINAL;
      }
      if (param.useSelection) {
        saveStateMask = UnMHTSaveStateMask.CURRENT;
      }

      let stateList = UnMHTSaveStateUtils.getList(saveStateMask);

      types = [...stateList.map(state => [state, false]),
               ...stateList.map(state => [state, true ])];
    }

    let lastStateStr = UnMHTPref.getPref("save.single.state");
    let lastAdd = UnMHTPref.getPref("save.single.add");

    let filterIndex = -1;
    let filter = [];
    types.forEach(([saveState, useAdd], i) => {
      filter.push([UnMHTSaveStateUtils.makeFiletype(saveState, useAdd),
                   "*.mht"]);

      if (UnMHTSaveStateUtils.saveStateToString(saveState) == lastStateStr &&
          useAdd == lastAdd) {
        filterIndex = i;
      }
    });

    let params = {
      title:            TEXT("save.title"),
      useLastDir:       true,
      displayDirectory: displayDirectory,
      defaultString:    (param.filenameBase
                         + "." + UnMHTFilename.defaultExtension),
      defaultExtension: UnMHTFilename.defaultExtension,
      filter:           filter,
      filterIndex:      filterIndex,
      uri:              (param.targetDocument
                         ? param.targetDocument.documentURI
                         : null)
    };

    let [targetFilePath, filePicker] = yield lazy.arFilePicker.getSaveFile(param.window, params);
    if (!targetFilePath) {
      throw UnMHTSaveResult.CANCEL;
    }

    if (!(param.saveState &&
          param.additionalURIs != UnMHTFileSaverParam.CHOOSE)) {
      let [saveState, useAdd] = types[filePicker.filterIndex];

      if (UnMHTPref.getPref("save.single.state.remember")) {
        UnMHTPref.setPref("save.single.state",
                          UnMHTSaveStateUtils.saveStateToString(saveState));
        UnMHTPref.setPref("save.single.add", useAdd);
      }

      param.saveState = saveState;

      if (!useAdd) {
        param.additionalURIs = [];
      }
    }

    param.targetFilePath = yield UnMHTSave.getValidFilePath(param.window,
                                                            targetFilePath);

    return UnMHTSaveResult.OK;
  },

  /**
   * 追加ファイルのダイアログを表示する
   *
   * @param   {Object} param
   *          保存パラメータ
   * @returns {Promise.<Array.<string>>}
   * @resolves 追加ファイル
   * @rejects  {number}
   *           失敗時の結果
   */
  _getAdditionalURIs: function(param) {
    return lazy.UnMHTAdditionalDialog.open(param.window, null);
  },

  /**
   * メールの日時が必要かを返す
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {boolean}
   *          メールの日時が必要か
   */
  _needMailDate: function(window) {
    return (UnMHTPref.getPref("save.filename.custom.mail_date") &&
            window &&
            arWindowUtils.isMailer(window) &&
            "gFolderDisplay" in window &&
            "selectedMessageUris" in window.gFolderDisplay &&
            window.gFolderDisplay.selectedMessageUris.length == 1);
  },

  /**
   * メールの日時を取得する
   *
   * @param   {Object} param
   *          保存パラメータ
   * @yields  {Promise.<*>}
   * @returns {Date}
   *          メールの日時
   * @throws  {number}
   *          失敗時の結果
   */
  _getMailDateTask: function*(param) {
    let window = param.window;

    let messageURL
      = lazy.MsgMailSession.ConvertMsgURIToMsgURL(window.gFolderDisplay
                                                  .selectedMessageUris[0],
                                                  window.msgWindow);

    let uri;
    try {
      uri = Services.io.newURI(messageURL, null, null);
    } catch (e) {
    }

    if (!uri) {
      throw UnMHTFileSaverResult.NOT_A_MAIL;
    }

    let result = yield lazy.arNetFileLoader.loadData(uri, null, true, window);

    let date = new Date();
    if (result.status != lazy.arNetFileLoaderStatus.NG) {
      let mailPart
        = lazy.arMIMEDecoder.decodeMessage(result.data.replace(/\r\n|\r|\n/g, "\r\n"), true);
      if (mailPart &&
          mailPart.date) {
        date = mailPart.date;
      }
    }

    return date;
  },

  /**
   * Fennec 用のイベントリスナを返す
   *
   * @param   {Object} param
   *          保存パラメータ
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @returns {IUnMHTSaveListener}
   *          イベントリスナ
   */
  _getFennecListener: function(param, context) {
    let listener = {
      /**
       * キャンセルが完了したイベント
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       */
      onCancel: function(context) {
      },

      /**
       * 保存失敗イベント
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       * @param   {UnMHTSaveErrorInfo} errorInfo
       *          エラー情報
       */
      onError: function(context, errorInfo) {
        let title = TEXT("error.title");
        let message
          = errorInfo.message
          + (errorInfo.errorList
             ? (errorInfo.errorList
                .map(([url, reason]) => reason + ": " + url)
                .join("\n"))
             : "")
          + (errorInfo.e ? ("\n" + errorInfo.e.toString()) : "");
        arAlertUtils.alert(param.window, title, message);
      },

      /**
       * 進捗状況変化イベント
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       * @param   {string} currentFileName
       *          現在取得のファイル名
       * @param   {string} currentSize
       *          取得したバイト数 / ファイルサイズ
       * @param   {number} ok
       *          取得完了した数
       * @param   {number} na
       *          取得中の数
       * @param   {number} ng
       *          取得失敗した数
       */
      onProgress: function(context, currentFileName, currentSize, ok, na, ng) {
      },

      /**
       * 保存完了
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       * @param   {Array.<string>} failedList
       *          失敗したファイル名
       */
      onSave: function(context, failedList) {
        if (UnMHTPref.getPref("save.show_errors") && failedList.length > 0) {
          let message
            = TEXT("save.error.failed_files") + "\n" + failedList.join("\n");
          let buttons = [{
            label: TEXT("ok"),
            callback: function() {
            }
          }];

          arAlertUtils.doorhanger(param.window, message, "unmht", buttons);
        }

        arAlertUtils.toast(param.window, TEXT("save.done"), "short");
      },

      /**
       * 保存コンテキストの破棄
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       */
      onDestruct: function(context) {
      }
    };
    return listener;
  },

  /**
   * 何も表示しないイベントリスナを返す
   *
   * @param   {Object} param
   *          保存パラメータ
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @returns {IUnMHTSaveListener}
   *          イベントリスナ
   */
  _getNoneListener: function(param, context) {
    return this._getSimpleListener(param, context, {});
  },

  /**
   * ツールバーに表示するイベントリスナを返す
   *
   * @param   {Object} param
   *          保存パラメータ
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @returns {IUnMHTSaveListener}
   *          イベントリスナ
   */
  _getToolbarListener: function(param, context) {
    let document = param.window.document;
    let $ = id => document.getElementById(id);
    let button = null;
    if ("button" in param && param.button) {
      button = $(param.button);
    }
    if (!button) {
      button = $("unmht-toolbarbutton-preferences");
      if (!button) {
        return this._getNoneListener(param, context);
      }
    }

    /* XBL はキャッシュされるのでハッシュを付加する */
    let progressXMLHash = "3c5da6cd7bc522d10d655f89bef00fbc07cc4645";
    button.style.MozBinding = "url('chrome://unmht/content/progress.xml?" + progressXMLHash + "#toolbar')";

    return this._getSimpleListener(param, context, {
      onProgress: function(context, currentFileName, currentSize, ok, na, ng) {
        button.setAttribute("progress", Math.floor(context.progress * 100).toString(10));
      },

      onDone: function() {
        button.style.MozBinding = "";
      }
    });
  },

  /**
   * シンプルなイベントリスナを返す
   *
   * @param   {Object} param
   *          保存パラメータ
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Object} additionalListener
   *          追加のリスナ
   * @returns {IUnMHTSaveListener}
   *          イベントリスナ
   */
  _getSimpleListener: function(param, context, additionalListener) {
    let onProgress = "onProgress" in additionalListener ? additionalListener.onProgress : null;
    let onDone = "onDone" in additionalListener ? additionalListener.onDone : null;

    let listener = {
      /**
       * キャンセルが完了したイベント
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       */
      onCancel: function(context) {
        if (onDone) {
          onDone(context);
        }
      },

      /**
       * 保存失敗イベント
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       * @param   {UnMHTSaveErrorInfo} errorInfo
       *          エラー情報
       */
      onError: function(context, errorInfo) {
        if (onDone) {
          onDone(context);
        }

        let arg = {
          title: TEXT("error.title"),
          message: errorInfo.message
            + (errorInfo.e ? ("\n" + errorInfo.e.toString()) : ""),
          log: (errorInfo.errorList
                ? (errorInfo.errorList
                   .map(([url, reason]) => reason + ": " + url)
                   .join("\n"))
                : ""),
          accept: TEXT("ok"),
          focus: "yes"
        };
        lazy.UnMHTConfirmDialog.open(param.window, arg);
      },

      /**
       * 進捗状況変化イベント
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       * @param   {string} currentFileName
       *          現在取得のファイル名
       * @param   {string} currentSize
       *          取得したバイト数 / ファイルサイズ
       * @param   {number} ok
       *          取得完了した数
       * @param   {number} na
       *          取得中の数
       * @param   {number} ng
       *          取得失敗した数
       */
      onProgress: function(context, currentFileName, currentSize, ok, na, ng) {
        if (onProgress) {
          onProgress(context, currentFileName, currentSize, ok, na, ng);
        }
      },

      /**
       * 保存完了
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       * @param   {Array.<string>} failedList
       *          失敗したファイル名
       */
      onSave: function(context, failedList) {
        if (onDone) {
          onDone(context);
        }

        if (UnMHTPref.getPref("save.show_errors") && failedList.length > 0) {
          let arg = {
            title: TEXT("save.title"),
            log: (TEXT("save.error.failed_files") + "\n"
                  + failedList.join("\n")),
            accept: TEXT("ok"),
            focus: "yes"
          };
          lazy.UnMHTConfirmDialog.open(param.window, arg);
        }
      },

      /**
       * 保存コンテキストの破棄
       *
       * @param   {UnMHTSaveContext} context
       *          対象の保存コンテキスト
       */
      onDestruct: function(context) {
      }
    };
    return listener;
  },

  /**
   * ダイアログで保存を開始する
   *
   * @param   {Object} param
   *          保存パラメータ
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @returns {Promise.<IUnMHTSaveListener>}
   *          イベントリスナ
   */
  _getSaveDialogListener: function(param, context) {
    let arg = {
      saveState:         param.saveState,
      saveContext:       context
    };
    return lazy.UnMHTSaveDialog.open(param.window, arg);
  },

  /**
   * 保存コンテキストを作成する
   *
   * @param   {Object} param
   *          保存パラメータ
   * @yields  {Promise.<*>}
   * @returns {UnMHTSaveContext}
   *          保存コンテキスト
   * @throws  UnMHTSaveResult.ERROR_ADD_DOCUMENT
   */
  _createContextTask: function*(param) {
    let [location, subject]
      = UnMHTSave.getLocationAndTitle(param.targetDocument, param.scrapBookData);

    let saveContext = UnMHTSave.createContext(param.window,
                                              param.targetFilePath,
                                              location, subject,
                                              param.downloadListener,
                                              param.saveState);

    let arg = {
      browser:         param.browser,
      targetDocument:  param.targetDocument,
      useSelection:    param.useSelection,
      scrapBookData:   param.scrapBookData,
      saveState:       param.saveState,
      location:        location,
      isHTML:          true,
      isStartPart:     true,
      window:          param.window
    };
    try {
      yield saveContext.addDocument(arg);
    } catch (e) {
      saveContext.destruct();
      throw e;
    }

    if (param.additionalURIs) {
      for (let file of param.additionalURIs) {
        saveContext.addFile(null, null, file, false);
      }
    }

    return saveContext;
  },

  /**
   * リンク先の URI を取得する
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @param   {nsIDOMNode} node
   *          対象のノード
   * @returns {?string}
   *          リンク先の URI
   *          なければ null
   */
  _getLinkURL: function(window) {
    if ("gContextMenu" in window &&
        window.gContextMenu) {
      return window.gContextMenu.linkURL;
    }
    return null;
  }
});

module.UnMHTFileSaver = UnMHTFileSaver;
module.UnMHTFileSaverResult = UnMHTFileSaverResult;
module.UnMHTFileSaverParam = UnMHTFileSaverParam;

})(this);
