"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTFilename"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});

/**
 * mht ファイルのファイル名操作
 *
 * @class
 */
let UnMHTFilename = Object.freeze({
  /**
   * MHT ファイルの拡張子
   * @type {string}
   */
  get defaultExtension() {
    return "mht";
  },

  /**
   * ファイル名から MHT の拡張子を削除する
   *
   * @param   {string} filename
   *          ファイル名
   * @returns {string}
   *          ファイル名 (拡張子以外)
   */
  removeExtension: function(filename) {
    return filename.replace(/\.mht(ml)?$/i, "");
  },


  /**
   * ファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @returns {string}
   *          変換したファイル名
   */
  validateFileName: function(filename) {
    return this.validateFileNameWithOS(filename, Services.appinfo.OS);
  },

  /**
   * 指定した OS でファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @param   {string} OS
   *          対象の OS
   * @returns {string}
   *          変換したファイル名
   */
  validateFileNameWithOS: function(filename, OS) {
    let [fromRE, toMap] = this._loadReplaceList();
    if (fromRE && toMap) {
      filename = filename.replace(fromRE, matched => toMap.get(matched));
    }

    if (OS == "WINNT" &&
        !UnMHTPref.getPref("save.filename.replace.ie_bug.disable")) {
      /*
       * ファイル名に使用すると IE が MHT ファイルを開けなくなる文字
       *   #
       *   !
       */
      filename = filename.replace(/[#\!]/g, "_");
    }

    return arFileUtils.validateFileNameWithOS(filename, OS);
  },

  /**
   * 文字の置換元の正規表現と置換先のマップを返す
   *
   * @returns {Tuple.<RegExp,Map.<string,string>>}
   *          置換元の正規表現と置換先のマップ]
   *          正しく読み込めなければどちらも null
   */
  _loadReplaceList: function() {
    let datas;
    try {
      datas = JSON.parse(UnMHTPref.getPref("save.filename.replace.list"));
      if (!datas.every(data => "from" in data && "to" in data)) {
        return [null, null];
      }
    } catch (e) {
      return [null, null];
    }

    if (!datas.length) {
      return [null, null];
    }

    let validDatas = datas.filter(data => data.from !== "");
    if (!validDatas.length) {
      return [null, null];
    }

    let escapeRegExp
      = text => text.replace(/([\\\[\]\(\)\^\,\.\{\}\|\?\!\-\*\+\$])/g, "\\$1");

    let toMap = new Map();
    for (let data of datas) {
      toMap.set(data.from, data.to);
    }

    return [new RegExp(validDatas
                       .map(data => escapeRegExp(data.from))
                       .join("|"), "g"),
            toMap];
  }
});

module.UnMHTFilename = UnMHTFilename;

})(this);
