"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTKeyboardShortcut"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arScoped } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileSaver",
                                  "resource://unmht/modules/UnMHTFileSaver.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSaveMultiDialog",
                                  "resource://unmht/modules/UnMHTSaveMultiDialog.jsm");

/**
 * キーボードショートカット情報
 *
 * @constructor
 * @param   {string} pref
 *          設定名
 * @param   {function(nsIDOMWindow)} func
 *          実行する関数
 */
function UnMHTKeyboardShortcutItem(pref, func) {
  /**
   * 設定名
   * @type {string}
   */
  this._pref = pref;

  /**
   * 実行する関数
   * @type {function(nsIDOMWindow)}
   */
  this._func = func;

  /**
   * 有効
   * @type {boolean}
   */
  this._enabled = false;

  /**
   * キー名
   * @type {string}
   */
  this._key = "";

  /**
   * キーコード
   * @type {?number}
   */
  this._keycode = null;

  /**
   * Alt
   * @type {boolean}
   */
  this._modsAlt = false;

  /**
   * Ctrl
   * @type {boolean}
   */
  this._modsCtrl = false;

  /**
   * Meta
   * @type {boolean}
   */
  this._modsMeta = false;

  /**
   * Shift
   * @type {boolean}
   */
  this._modsShift = false;

  Object.seal(this);
}
UnMHTKeyboardShortcutItem.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    this._func = null;
  },

  /**
   * 設定を読み込む
   */
  loadPref: function() {
    this._enabled = UnMHTPref.getPref(this._pref);
    this._key = UnMHTPref.getPref(this._pref + ".keycode");
    this._keycode = null;
    this._modsAlt = UnMHTPref.getPref(this._pref + ".mods.alt");
    this._modsCtrl = UnMHTPref.getPref(this._pref + ".mods.ctrl");
    this._modsMeta = UnMHTPref.getPref(this._pref + ".mods.meta");
    this._modsShift = UnMHTPref.getPref(this._pref + ".mods.shift");
  },

  /**
   * 有効かどうか
   * @type {boolean}
   */
  get enabled() {
    return this._enabled;
  },

  /**
   * ショートカットキーを調べる
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {Event} event
   *          対象のイベント
   */
  check: function(window, event) {
    if (this._keycode === null) {
      this._keycode = window.KeyboardEvent["DOM_" + this._key];
    }

    return (this._keycode == event.keyCode &&
            this._modsAlt == event.altKey &&
            this._modsCtrl == event.ctrlKey &&
            this._modsMeta == event.metaKey &&
            this._modsShift == event.shiftKey);
  },

  /**
   * 関数を実行する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  fire: function(window) {
    return this._func(window);
  }
});

/**
 * ショートカットのフック
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @implements arIScopedClass
 */
function UnMHTKeyboardShortcutHook(window) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 現在イベントをフックしているか
   * @type {boolean}
   */
  this._hooked = false;

  /**
   * キーダウンイベントハンドラ
   * @type {function(event)}
   */
  this.onKeyDown = null;

  Object.seal(this);
}
UnMHTKeyboardShortcutHook.prototype = {
  /**
   * フックの状態を更新する
   *
   * @param   {boolean} enable
   *          有効にするか
   */
  update: function(enable) {
    if (enable) {
      if (!this._hooked) {
        arEvent.add(this._window,
                    this._window, "keydown", this.onKeyDown, true);
        this._hooked = true;
      }
    } else {
      if (this._hooked) {
        arEvent.remove(this._window,
                       this._window, "keydown", this.onKeyDown, true);
        this._hooked = false;
      }
    }
  },

  /* ==== arIScopedClass ==== */

  /**
   * 破棄する
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  destruct: function(isShutdown, data, reason) {
    this._window = null;
    this.onKeyDown = null;
  }
};

/**
 * ウィンドウ監視
 *
 * @class
 * @implements arIShutdownEventListener
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
let UnMHTKeyboardShortcut = Object.freeze({
  /* ---- UI イベント - キーボード ---- */

  /**
   * キーが押されたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {Event} event
   *          対象のイベント
   */
  onKeyDown: function(window, event) {
    for (let item of this._enabledItems) {
      if (item.check(window, event) &&
          item.fire(window)) {
        event.preventDefault();
        event.stopPropagation();
        return;
      }
    }
  },

  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    this._initItems();

    UnMHTPref.addListener("shortcut.save", () => {
      this._enabledItems.clear();
      for (let item of this._items) {
        item.loadPref();
        if (item.enabled) {
          this._enabledItems.add(item);
        }
      }
      this._updateAllHook();
    }, true);

    arWindowWatcher.addListener(this);
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!arWindowUtils.isBrowser(window)) {
      return;
    }

    this._hookShortcut(window, null);
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
  },

  /* ==== private ==== */

  /**
   * ショートカット
   * @type {Set.<UnMHTKeyboardShortcutItem>}
   */
  _items: new Set(),

  /**
   * 有効なショートカット
   * @type {Set.<UnMHTKeyboardShortcutItem>}
   */
  _enabledItems: new Set(),

  /**
   * フックしたウィンドウごとのフック
   * @type {arScoped.<UnMHTKeyboardShortcutHook>}
   */
  _hooks: new arScoped(UnMHTKeyboardShortcutHook),

  /**
   * 全てのウィンドウで設定に合わせてキーイベントをフックする
   */
  _updateAllHook: function() {
    this._hooks.forEach((window, hook) => {
      hook.update(this._enabledItems.size > 0);
    });
  },

  /**
   * 設定に合わせてキーイベントをフックする
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _hookShortcut: function(window) {
    let hook = this._hooks.create(window);
    hook.onKeyDown = this.onKeyDown.bind(this, window);
    hook.update(this._enabledItems.size > 0);
  },

  /**
   * キーボードショートカット情報を作成する
   */
  _initItems: function() {
    this._items.add(new UnMHTKeyboardShortcutItem("shortcut.save.single", window => {
      let browser = arWindowUtils.getCurrentBrowser(window);
      if (!UnMHTDocumentCategory.isSavable(window, browser, false)) {
        return false;
      }

      lazy.UnMHTFileSaver.save(window);
      return true;
    }));
    this._items.add(new UnMHTKeyboardShortcutItem("shortcut.save.multi", window => {
      let arg = {
        targetWindow: window
      };
      lazy.UnMHTSaveMultiDialog.open(window, arg);
      return true;
    }));
    this._items.add(new UnMHTKeyboardShortcutItem("shortcut.save.quick", window => {
      let browser = arWindowUtils.getCurrentBrowser(window);
      if (!UnMHTDocumentCategory.isSavable(window, browser, false)) {
        return false;
      }

      lazy.UnMHTFileSaver.quicksave(window);
      return true;
    }));
    this._items.add(new UnMHTKeyboardShortcutItem("shortcut.save.overwrite", window => {
      let browser = arWindowUtils.getCurrentBrowser(window);
      if (!(UnMHTDocumentCategory.isSavable(window, browser, false) &&
            UnMHTDocumentCategory.isLocalMHT(window, browser))) {
        return false;
      }

      lazy.UnMHTFileSaver.overwrite(window);
      return true;
    }));

    arShutdownEvent.addListener(this);
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    for (let item of this._items) {
      item.destruct();
    }
    this._items.clear();
    this._enabledItems.clear();
  },
});

module.UnMHTKeyboardShortcut = UnMHTKeyboardShortcut;

})(this);
