"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTLoadModifier"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});
let { arPathUtils } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { UnMHTCache } = Cu.import("resource://unmht/modules/UnMHTCache.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

/**
 * 画像読み込み設定名
 * @type {string}
 */
const PERM_IMAGE_PREF = "permissions.default.image";

/**
 * 画像読み込み設定値
 * @enum {number}
 */
const PermissionsImage = Object.freeze({
  /** 全て許可 */
  ALLOW_ALL: 1,

  /** 全てブロック */
  BLOCK_ALL: 2,

  /** サードパーティのみブロック */
  BLOCK_THIRD_PIRTY: 3
});

/**
 * 読み込み設定
 */
function UnMHTLoadOptionsClass() {
  /**
   * Firefox の画像読み込み設定 (permissions.default.image の値)
   * @type {number}
   */
  this.loadImageType = PermissionsImage.ALLOW_ALL;

  /**
   * オフラインモードかどうか
   * @type {boolean}
   */
  this.enableOffline = false;

  this.load();

  Object.seal(this);
}
UnMHTLoadOptionsClass.prototype = Object.freeze({
  /**
   * 初期化処理
   */
  load: function() {
    UnMHTPref.addListener("open.offline", () => {
      this.enableOffline = UnMHTPref.getPref("open.offline");
    }, true);

    UnMHTPref.addGlobalListener(PERM_IMAGE_PREF, () => {
      this.loadImageType
        = Services.prefs.getIntPref(PERM_IMAGE_PREF);
    }, true);
  }
});
let UnMHTLoadOptions = new UnMHTLoadOptionsClass();

/**
 * 動的なリソースの読み込みを修正する
 *
 * @class
 */
let UnMHTLoadModifier = Object.freeze({
  /**
   * unmht 形式の URI からの相対パスを展開する
   *
   * @param   {string} uri
   *          unmht: 形式のファイルの URI
   * @param   {string} spec
   *          相対パス
   * @returns {?string}
   *          ファイルの URI
   *          見付からなかった場合は null
   */
  resolveURI: function(uri, spec) {
    if (!spec) {
      return uri;
    }

    let [eFileInfo, part] = UnMHTCache.getPart(uri);
    if (!(eFileInfo && part)) {
      return null;
    }

    let param = UnMHTScheme.getUnMHTURIParam(uri);
    let refName = param.refName;

    let [targetPart, fragment]
      = UnMHTCache.findPart(eFileInfo, part, spec, false);
    if (targetPart) {
      if (refName === "" &&
          targetPart == eFileInfo.startPart) {
        /* 開始パート中のフラグメントのみのリンクでは参照名は付けない
         * これは CSS の filter 等で使用される */
        return eFileInfo.baseURI + fragment;
      }

      return eFileInfo.baseURI + targetPart.eParam.refName + fragment;
    }

    if (part.eParam.location.startsWith("file:")) {
      return null;
    }

    try {
      let resolved = arPathUtils.resolve(part.eParam.baseDir, spec);
      return Services.io.newURI(resolved, null, null).spec;
    } catch (e) {
    }

    return null;
  },

  /**
   * mht への参照のチェックおよび
   * mht 内の URI を置き換える
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {nsIURI} requestOrigin
   *          参照元の URI
   * @param   {nsIDOMElement} context
   *          ロード先
   * @returns {boolean}
   *          許可するかどうか
   */
  shouldLoad: function(contentType, contentLocation, requestOrigin, context) {
    if (UnMHTLoadOptions.loadImageType == PermissionsImage.BLOCK_ALL &&
        contentType == Ci.nsIContentPolicy.TYPE_IMAGE) {
      /* 画像の読み込みの拒否は Firefox の設定に従う */
      return false;
    }

    if (context instanceof Ci.nsIDOMHTMLDocument ||
        context instanceof Ci.nsIDOMHTMLStyleElement) {
      /* CSS によるロードの場合 */
      return this._checkExternal(contentLocation);
    }

    /* ---- ここから HTML 要素へのロード ---- */

    let [eFileInfo, originPart] = UnMHTCache.getPart(requestOrigin.spec);
    if (!(eFileInfo && originPart)) {
      /* 参照元のファイルが未展開、もしくは存在しないので拒否 */
      return false;
    }

    if (UnMHTScheme.isUnMHTURI(contentLocation.spec)) {
      let param = UnMHTScheme.getUnMHTURIParam(contentLocation.spec);
      if (param.original == eFileInfo.original) {
        /* 既に同じ mht ファイル内への参照になっているので許可 */
        return true;
      } else {
        return false;
      }
    }

    /**
     * 属性を変更する関数
     *
     * @param   {nsIDOMElement} node
     *          対象の要素
     * @param   {string} attr
     *          属性名
     * @param   {string} value
     *          属性値
     */
    let setAttr = (node, attr, value) => {
      if (node instanceof Ci.nsIDOMHTMLEmbedElement ||
          node instanceof Ci.nsIDOMHTMLObjectElement) {
        /* EMBED, OBJECT は属性を変えても効果が無いので
         * 要素を置き換える */
        setTimeout(() => {
          try {
            let newNode = node.cloneNode(true);
            newNode.setAttribute(attr, value);
            node.parentNode.replaceChild(newNode, node);
          } catch (e) {
            /* ノードかドキュメントが生きていない */
          }
        }, 100);
      } else {
        setTimeout(() => {
          try {
            node.setAttribute(attr, value);
            if (attr == "src" && node.hasAttribute("srcset")) {
              node.removeAttribute("srcset");
            }
            if (node.hasAttribute("integrity")) {
              node.removeAttribute("integrity");
            }
          } catch (e) {
            /* ノードかドキュメントが生きていない */
          }
        }, 100);
      }
    };

    /**
     * 属性を削除する関数
     *
     * @param   {nsIDOMElement} node
     *          対象の要素
     * @param   {string} attr
     *          属性名
     */
    let removeAttr = (node, attr) => {
      setTimeout(() => {
        try {
          node.removeAttribute(attr);
        } catch (e) {
          /* ノードかドキュメントが生きていない */
        }
      }, 100);
    };

    let attr = "src";
    if (context instanceof Ci.nsIDOMHTMLObjectElement) {
      attr = "data";
    }

    let [targetPart, fragment]
      = UnMHTCache.findPart(eFileInfo, originPart, contentLocation.spec,
                            false);
    if (targetPart) {
      setAttr(context, attr,
              eFileInfo.baseURI + targetPart.eParam.refName + fragment);
      return false;
    }

    if (contentLocation.scheme == "file") {
      /* file への読み込みは削除 */
      removeAttr(context, attr);
      return false;
    }

    if (contentLocation.spec.startsWith(requestOrigin.spec)) {
      /* JavaScript 等で生成された参照元からの相対アドレス */
      /* **** FIXME: パスを遡るものはチェック不可 **** */
      let path = contentLocation.spec.slice(requestOrigin.spec.length);

      let origSpec;
      if (path[0] == "?") {
        /* クエリのみ */
        origSpec
          = originPart.eParam.location.replace(/\?.*$/, "") + path;
      } else {
        origSpec
          = arPathUtils.resolve(originPart.eParam.baseDir, path);
      }

      let origContentLocation;
      try {
        origContentLocation = Services.io.newURI(origSpec, null, null);
      } catch (e) {
        /* 正しい URI を生成できなかった */
        return this._checkExternal(contentLocation);
      }

      if (origContentLocation.spec != contentLocation.spec) {
        /* 本来のアドレスでチェックし直す */
        return this.shouldLoad(contentType, origContentLocation,
                               requestOrigin, context);
      }
    }

    return this._checkExternal(contentLocation);
  },

  /**
   * 外部読み込みをチェックする
   *
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @returns {boolean}
   *          許可するかどうか
   */
  _checkExternal: function(contentLocation) {
    if (contentLocation.scheme == "unmht") {
      return true;
    }

    return !UnMHTLoadOptions.enableOffline;
  }
});

module.UnMHTLoadModifier = UnMHTLoadModifier;

})(this);
