"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTMAFIntegration"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arScoped } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileSaver",
                                  "resource://unmht/modules/UnMHTFileSaver.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileSaverParam",
                                  "resource://unmht/modules/UnMHTFileSaver.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSave",
                                  "resource://unmht/modules/UnMHTSave.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSaveStateUtils",
                                  "resource://unmht/modules/UnMHTSaveState.jsm");

/**
 * MAF のインターフェースがあるかどうか確認する
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @returns {boolean}
 *          MAF のインターフェースがあるか
 */
function ensureMAFInterface(window) {
  return ("InternalSaveBehavior" in window &&
          "gInternalSaveBehaviors" in window);
}

/**
 * MAF 組み込みオブジェクト
 *
 * @constructor
 * @param   {number} saveState
 *          保存状態
 * @param   {boolean} useAdditionalURIs
 *          追加の URI を使用するか
 * @implements nsICancelable
 * @implements nsIWebBrowserPersist
 */
function UnMHTMAFSavePersist(saveState, useAdditionalURIs) {
  /**
   * 状態
   * @type {number}
   */
  this.currentState = Ci.nsIWebBrowserPersist.PERSIST_STATE_READY;

  /**
   * データ取得方法フラグ
   * @type {number}
   */
  this.persistFlags = 0;

  /**
   * 進捗状況リスナ
   * @type {nsIWebProgressListener}
   */
  this.progressListener = null;

  /**
   * 理由
   * @type {NSRESULT}
   */
  this.result = Cr.NS_OK;

  /**
   * 保存状態
   * @type {number}
   */
  this._saveState = saveState;

  /**
   * 追加ファイルを使用するか
   * @type {boolean}
   */
  this._useAdditionalURIs = useAdditionalURIs;

  Object.seal(this);
}
UnMHTMAFSavePersist.prototype = Object.freeze({
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIWebBrowserPersist,
                                         Ci.nsICancelable]),

  /* ==== nsICancelable ==== */

  /**
   * 保存キャンセル
   *
   * @param   {nsresult} reason
   *          理由
   */
  cancel: function(reason) {
    this.result = reason;
  },

  /* ==== nsIWebBrowserPersist ==== */

  /**
   * 保存キャンセルイベント
   */
  cancelSave: function() {
    this.cancel(Cr.NS_BINDING_ABORTED);
  },

  /**
   * 保存開始イベント
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {nsISupports} targetFileURI
   *          保存先のファイル (nsIFileURI)
   * @param   {nsISupports} targetDataDir
   *          保存先のディレクトリ? (nsIFile/nsIURI)
   * @param   {string} outputContentType
   *          保存先のファイルの MIME-Type
   * @param   {number} encodingFlags
   *          エンコーダに渡すのフラグ
   * @param   {number} wrapColumn
   *          折り返し文字数
   */
  saveDocument: function(targetDocument, targetFileURI,
                         targetDataDir, outputContentType,
                         encodingFlags, wrapColumn) {
    Task.spawn(function*() {
      this.currentState = Ci.nsIWebBrowserPersist
        .PERSIST_STATE_SAVING;
      let chromeWindow = Services.wm.getMostRecentWindow("navigator:browser");

      let targetFilePath = targetFileURI.QueryInterface(Ci.nsIFileURL).file.path;
      try {
        targetFilePath = yield lazy.UnMHTSave.getValidFilePath(chromeWindow,
                                                               targetFilePath);
      } catch (e) {
        return Cr.NS_OK;
      }

      let param = {
        window:            chromeWindow,
        targetDocument:    targetDocument,
        saveState:         this._saveState,
        additionalURIs:    (this._useAdditionalURIs
                            ? lazy.UnMHTFileSaverParam.CHOOSE
                            : null),
        showOptionDialog:  false,
        targetFilePath:    targetFilePath,
        downloadListener:  this.progressListener
      };

      yield lazy.UnMHTFileSaver.saveWithParam(param);

      return Cr.NS_OK;
    }.bind(this)).then(_result => {
      this.result = _result;
      this.currentState = Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED;
    }, e => {
      if (e instanceof Ci.nsIXPCException) {
        this.result = e.result;
      } else {
        this.result = Cr.NS_ERROR_FAILURE;
      }

      this.currentState = Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED;
    });
  },

  /**
   * チャネルを保存する
   *
   * @param   {nsIChannel} channel
   *          対象のチャネル
   * @param   {nsIFile} file
   *          対象のファイル
   * @throws  {number} Cr.NS_ERROR_NOT_IMPLEMENTED
   */
  saveChannel: function(channel, file) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  /**
   * URI を保存する
   *
   * @param   {nsIURI} URI
   *          対象の URI
   * @param   {nsISupports} cacheKey
   *          キャッシュオブジェクト
   * @param   {nsIURI} referrer
   *          リファラ
   * @param   {nsIInputStream} postData
   *          POST データ
   * @param   {string} extraHeaders
   *          拡張ヘッダ
   * @param   {nsIFile} file
   *          対象のファイル
   * @param   {nsILoadContext} privacyContext
   *          プライバシーコンテキスト
   * @throws  {number} Cr.NS_ERROR_NOT_IMPLEMENTED
   */
  saveURI: function(URI, cacheKey, referrer, postData, extraHeaders, file,
                    privacyContext) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  /**
   * プライバシーを意識して URI を保存する
   *
   * @param   {nsIURI} URI
   *          対象の URI
   * @param   {nsISupports} cacheKey
   *          キャッシュオブジェクト
   * @param   {nsIURI} referrer
   *          リファラ
   * @param   {nsIInputStream} postData
   *          POST データ
   * @param   {string} extraHeaders
   *          拡張ヘッダ
   * @param   {nsIFile} file
   *          対象のファイル
   * @param   {boolean} isPrivate
   *          プライベートモードか
   * @throws  {number} Cr.NS_ERROR_NOT_IMPLEMENTED
   */
  savePrivacyAwareURI: function(URI, cacheKey, referrer, postData,
                                extraHeaders, file, isPrivate) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  }
});

/**
 * ビヘイビアのリスト
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @implements arIScopedClass
 */
function UnMHTMAFWindowIntegration(window) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 登録したビヘイビア
   * @type {Set.<InternalSaveBehavior>}
   */
  this._behaviors = new Set();

  Object.seal(this);

  this._register();
}
UnMHTMAFWindowIntegration.prototype = {
  /* ==== arIScopedClass ==== */

  /**
   * 破棄する
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  destruct: function(isShutdown, data, reason) {
    /* 登録時にあったからといって破棄時にもあるとは限らないので
     * 改めてチェックする */
    if (ensureMAFInterface(this._window)) {
      let behaviors = this._behaviors;

      try {
        this._window.gInternalSaveBehaviors
          = this._window.gInternalSaveBehaviors.filter(b => !behaviors.has(b));
      } catch (e) {
        /* インターフェースが変わった */
      }
    }

    this._behaviors.clear();
    this._behaviors = null;
    this._window = null;
  },

  /* ==== private ==== */

  /**
   * ビヘイビアを登録する
   */
  _register: function() {
    let stateList = lazy.UnMHTSaveStateUtils.getList();

    let makeType = (state, useAdditionalURIs) => ({
      label:             lazy.UnMHTSaveStateUtils.makeFiletype(state,
                                                               useAdditionalURIs),
      ext:               "*.mht",
      saveState:         state,
      useAdditionalURIs: useAdditionalURIs
    });

    for (let type of stateList.map(state => makeType(state, false))) {
      this._add(type);
    }
    for (let type of stateList.map(state => makeType(state, true))) {
      this._add(type);
    }
  },

  /**
   * 個別のビヘイビアを作成し追加する
   *
   * @param   {object} type
   *          ビヘイビアの種類
   */
  _add: function(type) {
    let newSaveBehavior;

    try {
      newSaveBehavior = new this._window.InternalSaveBehavior();
    } catch (e) {
      /* インターフェースが変わった */
      return;
    }

    newSaveBehavior.isComplete = true;
    newSaveBehavior.mandatoryExtension = true;
    newSaveBehavior.isValidForSaveMode = saveMode => {
      return saveMode & 0x01;
    };
    newSaveBehavior.getFileFilter = (contentType, fileExtension) => {
      return {
        title:           type.label,
        extensionstring: type.ext
      };
    };
    newSaveBehavior.getPersistObject = saveBrowsers => {
      return new UnMHTMAFSavePersist(type.saveState, type.useAdditionalURIs);
    };

    try {
      /* 末尾は *.* なのでその前に挿入する */
      this._window.gInternalSaveBehaviors
        .splice(this._window.gInternalSaveBehaviors.length - 1, 0,
                newSaveBehavior);
    } catch (e) {
      /* インターフェースが変わった */
      return;
    }

    this._behaviors.add(newSaveBehavior);
  }
};

/**
 * ウィンドウごとの MAF 組み込み
 *
 * @class
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
let UnMHTMAFIntegration = Object.freeze({
  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    arWindowWatcher.addListener(this);
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!arWindowUtils.isBrowser(window)) {
      return;
    }

    this._register(window);
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
  },

  /* ==== private ==== */

  /**
   * ウィンドウごとの MAF 組み込み
   * @type {arScoped.<UnMHTMAFWindowIntegration>}
   */
  _integs: new arScoped(UnMHTMAFWindowIntegration),

  /**
   * ウィンドウに組み込む
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _register: function(window) {
    if (!ensureMAFInterface(window)) {
      return;
    }

    this._integs.create(window);
  }
});

module.UnMHTMAFIntegration = UnMHTMAFIntegration;

})(this);

