"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTMail"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arAlertUtils } = Cu.import("resource://unmht/modules/base/arAlertUtils.jsm", {});
let { arArrayUtils } = Cu.import("resource://unmht/modules/base/arArrayUtils.jsm", {});
let { arFileUtils, arFilePermission } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { UnMHTPref, UnMHTMailArg } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arWinRegUtils",
                                  "resource://unmht/modules/base/arWinRegUtils.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSystemDirectory",
                                  "resource://unmht/modules/UnMHTSystemDirectory.jsm");
XPCOMUtils.defineLazyServiceGetter(lazy, "MsgComposeService",
                                   "@mozilla.org/messengercompose;1",
                                   "nsIMsgComposeService");

/**
 * メール送信
 *
 * @class
 */
let UnMHTMail = Object.freeze({
  /**
   * メールを送信する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} targetFilePath
   *          添付ファイルのパス
   * @param   {string} title
   *          ページのタイトル
   * @returns {Promise.<boolean>}
   * @resolves 成功したか
   */
  send: Task.async(function*(window, targetFilePath, title) {
    if (UnMHTPref.getPref("save.mail.zip")) {
      targetFilePath = this._createZipFile(targetFilePath);
    }

    if (UnMHTPref.getPref("save.mail.app.type") == "seamonkey" &&
        arAppInfo.isSeaMonkey) {
      return this._openComposeWindow(targetFilePath, title);
    }

    let [appPath, arg] = yield* this._getApplicationAndArgumentTask();
    arg = this._replaceVars(arg, targetFilePath, title);
    let args = this.argToArray(arg);

    if (UnMHTPref.getPref("save.mail.debug")) {
      let atitle = TEXT("name");
      let message =
          "[Debug]\n" +
          "Command line for mail\n" +
          "\n" +
          (appPath ? appPath : "???") + " " + arg + "\n" +
          "\n" +
          "Arguments will be passed as following array\n" +
          "\n" +
          args.join("\n");
      yield arAlertUtils.alert(window, atitle, message);
    }

    try {
      let process
        = Cc["@mozilla.org/process/util;1"].createInstance(Ci.nsIProcess);
      process.init(arFileUtils.getFile(appPath));
      process.runw(false, args, args.length);
    } catch (e) {
      return false;
    }

    return true;
  }),

  /* ==== private ==== */

  /**
   * スペース区切りの引数を引数配列に変換する
   *
   * @param   {string} argString
   *          スペース区切りの引数
   * @returns {Array.<string>}
   *          引数配列
   */
  argToArray: function(argString) {
    let STATE_NORMAL          = 0;
    let STATE_ESCAPED         = 1;
    let STATE_DQUOTED         = 10;
    let STATE_DQUOTED_ESCAPED = 11;
    let STATE_QUOTED          = 20;
    let STATE_QUOTED_ESCAPED  = 21;

    let args = [];
    let currentArg = "";
    let state = STATE_NORMAL;

    for (let c of Array.from(argString)) {
      switch (state) {
        case STATE_NORMAL: {
          if (c === "\\") {
            state = STATE_ESCAPED;
          } else if (currentArg === "" && c === "\"") {
            state = STATE_DQUOTED;
          } else if (currentArg === "" && c === "\'") {
            state = STATE_QUOTED;
          } else if (c === " ") {
            if (currentArg !== "") {
              args.push(currentArg);
            }
            currentArg = "";
          } else {
            currentArg += c;
          }
          break;
        }
        case STATE_ESCAPED: {
          currentArg += c;
          state = STATE_NORMAL;
          break;
        }
        case STATE_DQUOTED: {
          if (c === "\"") {
            state = STATE_NORMAL;
          } else if (c === "\\") {
            state = STATE_DQUOTED_ESCAPED;
          } else {
            currentArg += c;
          }
          break;
        }
        case STATE_DQUOTED_ESCAPED: {
          currentArg += c;
          state = STATE_DQUOTED;
          break;
        }
        case STATE_QUOTED: {
          if (c === "\'") {
            state = STATE_NORMAL;
          } else if (c === "\\") {
            state = STATE_QUOTED_ESCAPED;
          } else {
            currentArg += c;
          }
          break;
        }
        case STATE_QUOTED_ESCAPED: {
          currentArg += c;
          state = STATE_QUOTED;
          break;
        }
      }
    }
    if (currentArg !== "") {
      args.push(currentArg);
    }

    return args;
  },

  /**
   * SeaMonkey でメール作成画面を開く
   *
   * @param   {string} targetFilePath
   *          添付ファイルのパス
   * @param   {string} title
   *          ページのタイトル
   * @returns {boolean}
   *          成功したか
   */
  _openComposeWindow: function(targetFilePath, title) {
    let params
      = Cc["@mozilla.org/messengercompose/composeparams;1"]
      .createInstance(Ci.nsIMsgComposeParams);

    params.composeFields
      = Cc["@mozilla.org/messengercompose/composefields;1"]
      .createInstance(Ci.nsIMsgCompFields);
    params.composeFields.subject = title;

    let attachmentData
      = Cc["@mozilla.org/messengercompose/attachment;1"]
      .createInstance(Ci.nsIMsgAttachment);
    attachmentData.url = arFileUtils.pathToURLSpec(targetFilePath);

    params.composeFields.addAttachment(attachmentData);

    try {
      params.identity = lazy.MsgComposeService.defaultIdentity;
    } catch (e) {
      params.identity = null;
    }

    lazy.MsgComposeService.OpenComposeWindowWithParams(null, params);

    return true;
  },

  /**
   * 使用するアプリケーションと引数を取得する
   *
   * @yields  {Promise.<*>}
   * @returns {Tuple.<?string,string>}
   *          アプリケーションのパスと引数
   */
  _getApplicationAndArgumentTask: function*() {
    let arg;
    switch (UnMHTPref.getPref("save.mail.arg.type")) {
    case "thunderbird": {
      arg = UnMHTMailArg.thunderbird;
      break;
    }
    case "seamonkey": {
      arg = UnMHTMailArg.seamonkey;
      break;
    }
    case "macmail": {
      arg = UnMHTMailArg.macmail;
      break;
    }
    default: {
      arg = UnMHTPref.getPref("save.mail.arg.other");
    }
    }

    let appPath;
    switch (UnMHTPref.getPref("save.mail.app.type")) {
    case "thunderbird": {
      appPath = yield* this._getThunderbirdTask();
      break;
    }
    case "seamonkey": {
      appPath = yield* this._getSeamonkeyTask();
      break;
    }
    case "macmail": {
      appPath = yield* this._generateMacMailScriptTask();
      break;
    }
    default: {
      let path = UnMHTPref.getPref("save.mail.app");
      if (path) {
        appPath = path;
      } else {
        appPath = yield* this._getThunderbirdTask();
        if (appPath) {
          arg = UnMHTMailArg.thunderbird;
        }
      }
    }
    }

    return [appPath, arg];
  },

  /**
   * 引数中の変数を置換する
   *
   * @param   {string} arg
   *          引数
   * @param   {string} targetFilePath
   *          添付ファイルのパス
   * @param   {string} title
   *          ページのタイトル
   * @returns {string}
   *          置換した引数
   */
  _replaceVars: function(arg, targetFilePath, title) {
    let targetURI
      = arFileUtils.pathToURLSpec(targetFilePath)
      .replace(/,/g, "%2C");

    let vars = {
      "FILE":       targetFilePath.replace(/\\/g, "\\\\"),
      "URL":        targetURI.replace(/\\/g, "\\\\"),
      /* %URI% から %URL% に変更した */
      "URI":        targetURI.replace(/\\/g, "\\\\"),
      "TITLE":      (title
                     .replace(/\\/g, "\\\\")
                     .replace(/[\'\"]/g, "")),
      "TITLE_NOSP": (title
                     .replace(/\\/g, "\\\\")
                     .replace(/[\'\"]/g, "")
                     .replace(/[ \t\r\n]/g, "_"))
    };

    let rexp = new RegExp("%(" + Object.keys(vars).join("|") + ")%", "g");
    return arg.replace(rexp, (matched, name) => vars[name]);
  },

  /**
   * Thunderbird の実行ファイルを探す
   *
   * @yields  {Promise.<*>}
   * @returns {?string}
   *          Thunderbird の実行ファイルのパス
   *          見付からない場合は null
   */
  _getThunderbirdTask: function*() {
    let list = [];

    switch (Services.appinfo.OS) {
    case "Darwin": {
      list = [
        [OS.Constants.Path.macLocalApplicationsDir,
         ["Thunderbird.app", "Contents", "MacOS", "thunderbird-bin"]],
        [arFileUtils.userApplicationsDirectory,
         ["Thunderbird.app", "Contents", "MacOS", "thunderbird-bin"]]
      ];
      break;
    }
    case "WINNT": {
      let appPath = null;

      let currentVersion
        = lazy.arWinRegUtils.getHKLMValue("SOFTWARE\\Mozilla\\Mozilla Thunderbird\\CurrentVersion");
      if (currentVersion) {
        let PathToExe
          = lazy.arWinRegUtils.getHKLMValue("SOFTWARE\\Mozilla\\Mozilla Thunderbird\\" + currentVersion + "\\Main", "PathToExe");
        if (PathToExe) {
          appPath = PathToExe;
        }
      }

      list = [
        [appPath, []],
        [arFileUtils.programFilesDirectory,
         ["Mozilla Thunderbird", "thunderbird.exe"]],
        [arFileUtils.programFilesDirectory,
         ["thunderbird", "thunderbird.exe"]]
      ];
      break;
    }
    }

    let appPath = yield this._getApplicationTask(list);

    return appPath;
  },

  /**
   * SeaMonkey の実行ファイルを探す
   *
   * @yields  {Promise.<*>}
   * @returns {?string}
   *          SeaMonkey の実行ファイルのパス
   *          見付からない場合は null
   */
  _getSeamonkeyTask: function*() {
    let list = [];

    switch (Services.appinfo.OS) {
    case "Darwin": {
      list = [
        [OS.Constants.Path.macLocalApplicationsDir,
         ["SeaMonkey.app", "Contents", "MacOS", "seamonkey-bin"]],
        [arFileUtils.userApplicationsDirectory,
         ["SeaMonkey.app", "Contents", "MacOS", "seamonkey-bin"]]
      ];
      break;
    }
    case "WINNT": {
      let appPath = null;

      let currentVersion
        = lazy.arWinRegUtils.getHKLMValue("SOFTWARE\\Mozilla\\SeaMonkey\\CurrentVersion");
      if (currentVersion) {
        let PathToExe
          = lazy.arWinRegUtils.getHKLMValue("SOFTWARE\\Mozilla\\SeaMonkey\\" + currentVersion + "\\Main", "PathToExe");
        if (PathToExe) {
          appPath = PathToExe;
        }
      }

      list = [
        [appPath, []],
        [arFileUtils.programFilesDirectory,
         ["mozilla.org", "SeaMonkey", "seamonkey.exe"]],
        [arFileUtils.programFilesDirectory,
         ["SeaMonkey", "seamonkey.exe"]]
      ];
      break;
    }
    }

    let appPath = yield this._getApplicationTask(list);

    return appPath;
  },

  /**
   * 存在するアプリケーションを返す
   *
   * @param   {Array.<Tuple.<string,Array.<string>>>} list
   *          アプリケーションのパスの候補
   *          基本となるパスとパスコンポーネントのリストのリスト
   * @yields  {Promise.<*>}
   * @returns {?string}
   *          実行ファイルのパス
   *          見付からない場合は null
   */
  _getApplicationTask: function*(list) {
    for (let [root, cs] of list) {
      let appPath = root;
      for (let c of cs) {
        appPath = OS.Path.join(appPath, c);
      }

      let exists = yield OS.File.exists(appPath);
      if (exists) {
        return appPath;
      }
    }

    return null;
  },

  /**
   * Mac で Mail を起動するスクリプトを作成する
   *
   * @yields  {Promise.<*>}
   * @returns {?string}
   *          実行ファイルのパス
   */
  _generateMacMailScriptTask: function*() {
    let data
      = "#!/usr/bin/osascript" + "\n"
      + "on run argv" + "\n"
      + "  tell application \"Mail\"" + "\n"
      + "    tell content of (make new outgoing message with properties {visible:true, subject:(item 1 of argv)})" + "\n"
      + "      make new attachment with properties {file name:(item 2 of argv)} at after the last paragraph" + "\n"
      + "    end tell" + "\n"
      + "    activate" + "\n"
      + "  end tell" + "\n"
      + "end run" + "\n";

    let macmailFilePath = yield lazy.UnMHTSystemDirectory.getCacheFilePath("macmailscript");

    yield arFileUtils.createUTF8(macmailFilePath, data, true);

    return macmailFilePath;
  },

  /**
   * ファイルを ZIP で圧縮する
   *
   * @param   {string} targetFilePath
   *          圧縮するファイルのパス
   * @returns {string}
   *          圧縮したファイルのパス
   */
  _createZipFile: function(targetFilePath) {
    let parent = OS.Path.dirname(targetFilePath);
    let basename = OS.Path.basename(targetFilePath);
    let zipFilePath = OS.Path.join(parent,
                                   UnMHTFilename.removeExtension(basename)
                                   + ".zip");
    let targetFile = arFileUtils.getFile(targetFilePath);

    let zip = arFileUtils.openZipWriter(zipFilePath);

    zip.addEntryFile(basename,
                     Ci.nsIZipWriter.COMPRESSION_DEFAULT,
                     targetFile,
                     false);

    zip.close();

    return zipFilePath;
  }
});

module.UnMHTMail = UnMHTMail;

})(this);
