"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTMenu"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arLocaleUtils",
                                  "resource://unmht/modules/base/arLocaleUtils.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arWindowOverlay",
                                  "resource://unmht/modules/base/arWindowOverlay.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCache",
                                  "resource://unmht/modules/UnMHTCache.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTConfirmDialog",
                                  "resource://unmht/modules/UnMHTConfirmDialog.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileLoader",
                                  "resource://unmht/modules/UnMHTFileLoader.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileSaver",
                                  "resource://unmht/modules/UnMHTFileSaver.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTPrefWindow",
                                  "resource://unmht/modules/UnMHTPrefWindow.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSaveMultiDialog",
                                  "resource://unmht/modules/UnMHTSaveMultiDialog.jsm");

/**
 * メニューの設定
 */
function UnMHTMenuOptionsClass() {
  /**
   * MHT で保存
   * @type {boolean}
   */
  this.enableContextMenuSave = false;

  /**
   * QuickSave
   * @type {boolean}
   */
  this.enableContextMenuQuickSave = false;

  /**
   * メール送信
   * @type {boolean}
   */
  this.enableContextMenuSendmail = false;

  /**
   * キャッシュを表示/キャッシュをクリア
   * @type {boolean}
   */
  this.enableMenuAllCache = false;

  this.load();

  Object.seal(this);
}
UnMHTMenuOptionsClass.prototype = Object.freeze({
  /**
   * 読み込む
   */
  load: function() {
    UnMHTPref.addListener("menu.context", () => {
      this.enableContextMenuSave
        = UnMHTPref.getPref("menu.context.save");
      this.enableContextMenuQuickSave
        = UnMHTPref.getPref("menu.context.save.quick");
      this.enableContextMenuSendmail
        = UnMHTPref.getPref("menu.context.save.mail");
    }, true);

    UnMHTPref.addListener("menu.all", () => {
      this.enableMenuAllCache
        = UnMHTPref.getPref("menu.all.cache");
    }, true);
  }
});
let UnMHTMenuOptions = new UnMHTMenuOptionsClass();

/**
 * メニュー管理
 *
 * @class
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
let UnMHTMenu = Object.freeze({
  /**
   * メニューの項目を作成する
   *
   * @param   {boolean} isMailer
   *          メーラか
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @returns {arOverlayMenuPopupElement} menu
   *          メニュー
   * @param   {string} name
   *          メニューの名前
   * @param   {string=} className
   *          メニュー項目のクラス名
   */
  createMenuItems: function(isMailer, window,
                            menu, name, className=null) {
    menu.addItem({
      id: "unmht-" + name + "-preferences",
      label: TEXT("menuitem.pref"),
      "class": className,
      oncommand: function(event) {
        lazy.UnMHTPrefWindow.open(window, null);
      }
    });
    menu.addSeparator({
      id: "unmht-" + name + "-menuitem-separator-save"
    });
    menu.addItem({
      id: "unmht-" + name + "-menuitem-save",
      label: TEXT("menuitem.save.single"),
      "class": className,
      oncommand: function(event) {
        lazy.UnMHTFileSaver.save(window);
      }
    });

    if (!isMailer) {
      menu.addItem({
        id: "unmht-" + name + "-menuitem-overwrite",
        label: TEXT("menuitem.save.overwrite"),
        "class": className,
        oncommand: function(event) {
          lazy.UnMHTFileSaver.overwrite(window);
        }
      });
    }

    menu.addItem({
      id: "unmht-" + name + "-menuitem-quicksave",
      label: TEXT("menuitem.save.quick"),
      "class": className,
      oncommand: function(event) {
        lazy.UnMHTFileSaver.quicksave(window);
      }
    });

    menu.addItem({
      id: "unmht-" + name + "-menuitem-save-selection",
      label: TEXT("menuitem.save.selection"),
      "class": className,
      oncommand: function(event) {
        lazy.UnMHTFileSaver.saveSelection(window);
      }
    });

    if (!isMailer) {
      menu.addItem({
        id: "unmht-" + name + "-menuitem-save-multi",
        label: TEXT("menuitem.save.multi"),
        "class": className,
        oncommand: function(event) {
          let arg = {
            targetWindow: window
          };
          lazy.UnMHTSaveMultiDialog.open(window, arg);
        }
      });
      menu.addItem({
        id: "unmht-" + name + "-menuitem-sendmail",
        label: TEXT("menuitem.save.mail"),
        "class": className,
        oncommand: function(event) {
          lazy.UnMHTFileSaver.sendmail(window);
        }
      });
      menu.addSeparator({
        id: "unmht-" + name + "-menuitem-separator-cache"
      });
      menu.addItem({
        id: "unmht-" + name + "-menuitem-cache-show",
        label: TEXT("menuitem.cache.show"),
        "class": className,
        oncommand: function(event) {
          let tabbrowser = arWindowUtils.getTabBrowser(window);
          tabbrowser.selectedTab = tabbrowser.addTab(UnMHTScheme.createCacheURI());
        }
      });
      menu.addItem({
        id: "unmht-" + name + "-menuitem-cache-clear",
        label: TEXT("menuitem.cache.clear"),
        "class": className,
        oncommand: function(event) {
          Task.spawn(function*() {
            let arg = {
              title:    TEXT("name"),
              message:  TEXT("cache.clear.confirm"),
              accept:   TEXT("ok"),
              cancel:   TEXT("cancel"),
              focus:    "yes"
            };
            let result = yield lazy.UnMHTConfirmDialog.open(window, arg);
            if (result.action == "no") {
              return;
            }

            lazy.UnMHTCache.removeOld(0);
          }).catch(e => {
            Cu.reportError(e + "\n" + e.stack);
          });
        }
      });
      menu.addSeparator({
        id: "unmht-" + name + "-menuitem-separator-extract"
      });
      menu.addItem({
        id: "unmht-" + name + "-menuitem-extract",
        label: TEXT("menuitem.extract"),
        "class": className,
        oncommand: function(event) {
          let browser = arWindowUtils.getCurrentBrowser(window);
          lazy.UnMHTFileLoader.openWithExtract(window, browser);
        }
      });
      menu.addItem({
        id: "unmht-" + name + "-menuitem-view-source",
        label: TEXT("menuitem.show_source"),
        "class": className,
        oncommand: function(event) {
          let browser = arWindowUtils.getCurrentBrowser(window);
          lazy.UnMHTFileLoader.openWithViewSource(window, browser);
        }
      });
      menu.addSeparator({
        id: "unmht-" + name + "-menuitem-separator-website"
      });
      menu.addItem({
        id: "unmht-" + name + "-menuitem-openwebsite",
        label: TEXT("menuitem.openwebsite"),
        "class": className,
        oncommand: function(event) {
          let url = lazy.arLocaleUtils.isJapanese()
            ? "http://www.unmht.org/unmht/"
            : "http://www.unmht.org/unmht/en_index.html";

          let tabbrowser = arWindowUtils.getTabBrowser(window);
          tabbrowser.selectedTab = tabbrowser.addTab(url);
        }
      });
    }
  },

  /**
   * メニューを設定する
   *
   * @param   {Event} event
   *          対象のイベント
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {string} type
   *          メニューの種類
   * @param   {boolean} isContext
   *          コンテキストメニューか
   * @param   {nsIDOMWindow=} targetDocument
   *          対象のドキュメント
   *          null の場合現在のドキュメント
   */
  onMenuShowing: function(event, window, type, isContext, targetDocument) {
    let document = window.document;
    let $ = id => document.getElementById(id);

    let savable, isLocalMHT, location;

    if (!targetDocument) {
      let browser = arWindowUtils.getCurrentBrowser(window);

      try {
        savable = UnMHTDocumentCategory.isSavable(window, browser, false);
        isLocalMHT = UnMHTDocumentCategory.isLocalMHT(window, browser);
        location = browser.currentURI.spec;
      } catch (e) {
        /* ブラウザが生きてない
         * タブを閉じる直前にメニューを開いた場合など */
        return;
      }
    } else {
      try {
        savable = UnMHTDocumentCategory.isDocumentSavable(window, targetDocument, false);
        isLocalMHT = UnMHTDocumentCategory.isDocumentLocalMHT(window, targetDocument);
        location = targetDocument.documentURI;
      } catch (e) {
        /* ドキュメントが生きてない
         * ページ移動直前にメニューを開いた場合など */
        return;
      }
    }

    let showMenu = (subtype, shown) => {
      let menuitem = $("unmht-" + type + "-menuitem-" + subtype);
      if (menuitem) {
        menuitem.hidden = !shown;
      }

      return shown;
    };

    showMenu("cache-show",      UnMHTMenuOptions.enableMenuAllCache);
    showMenu("cache-clear",     UnMHTMenuOptions.enableMenuAllCache);
    showMenu("separator-cache", UnMHTMenuOptions.enableMenuAllCache);

    let [eFileInfo, part] = lazy.UnMHTCache.getPart(location);
    if (eFileInfo && part) {
      showMenu("view-source",     true);
      showMenu("extract",         false);
    } else {
      showMenu("extract",         true);
      showMenu("view-source",     false);
    }

    showMenu("separator-extract", true);

    let onPage = isContext ? this._onPage(window) : true;
    let isContentSelected = this._isContentSelected(window, isContext);
    let onSavableLink = isContext ? this._onSavableLink(window) : false;

    showMenu("save-selection",
             savable &&
             isContentSelected &&
             (!isContext ||
              (onPage && UnMHTMenuOptions.enableContextMenuSave)));
    showMenu("quicksave-selection",
             savable &&
             isContentSelected &&
             (!isContext ||
              (onPage && UnMHTMenuOptions.enableContextMenuQuickSave)));

    let shown = [
      showMenu("save",
               savable &&
               (!isContext ||
                (!isContentSelected &&
                 onPage && UnMHTMenuOptions.enableContextMenuSave))),
      showMenu("overwrite",
               savable &&
               isLocalMHT &&
               (!isContext ||
                (!isContentSelected &&
                 onPage && UnMHTMenuOptions.enableContextMenuSave))),
      showMenu("save-link",
               onSavableLink &&
               (!isContext ||
                UnMHTMenuOptions.enableContextMenuSave)),
      showMenu("quicksave",
               savable &&
               (!isContext ||
                (!isContentSelected &&
                 onPage && UnMHTMenuOptions.enableContextMenuQuickSave))),
      showMenu("sendmail",
               savable &&
               (!isContext ||
                (!isContentSelected &&
                 onPage && UnMHTMenuOptions.enableContextMenuSendmail)))
    ];

    showMenu("separator", shown.some(b => b));
  },

  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    arWindowWatcher.addListener(this);
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!arWindowUtils.isBrowser(window) && !arWindowUtils.isMailer(window)) {
      return;
    }

    this._createUI(window);
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
  },

  /* ==== private ==== */

  /**
   * UI を作る
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _createUI: function(window) {
    let overlay = lazy.arWindowOverlay.get(window);

    let isMailer = arWindowUtils.isMailer(window);

    overlay.fileMenu.addEventListener("popupshowing", event => {
      UnMHTMenu.onMenuShowing(event, window, "filemenu");
    });

    overlay.fileMenu.addItem({
      id: "unmht-filemenu-menuitem-save",
      insertafter: "menu_sendLink,menu_saveAs",
      label: TEXT("menuitem.save.single"),
      oncommand: event => {
        lazy.UnMHTFileSaver.save(window);
      }
    });

    if (!isMailer) {
      overlay.fileMenu.addItem({
        id: "unmht-filemenu-menuitem-overwrite",
        insertafter: "unmht-filemenu-menuitem-save",
        label: TEXT("menuitem.save.overwrite"),
        oncommand: event => {
          lazy.UnMHTFileSaver.overwrite(window);
        }
      });
      overlay.fileMenu.addItem({
        id: "unmht-filemenu-menuitem-save-multi",
        insertafter: "unmht-filemenu-menuitem-overwrite",
        label: TEXT("menuitem.save.multi"),
        oncommand: event => {
          let arg = {
            targetWindow: window
          };
          lazy.UnMHTSaveMultiDialog.open(window, arg);
        }
      });
    }

    let menu = overlay.toolsMenu.addMenu({
      id: "unmht-tools-preferences",
      insertafter: "javascriptConsole,devToolsSeparator",
      "class": "menu-iconic",
      image: "resource://unmht/images/icon_small.png",
      label: TEXT("name"),
      onpopupshowing: event => {
        UnMHTMenu.onMenuShowing(event, window, "toolmenu");
      }
    });
    this.createMenuItems(isMailer, window, menu, "toolmenu");

    if (!isMailer) {
      overlay.contentMenu.addEventListener("popupshowing", event => {
        UnMHTMenu.onMenuShowing(event, window, "context", true);

        let popupNode = this._getPopupNode(window);
        if (!popupNode) {
          return;
        }
        UnMHTMenu.onMenuShowing(event, window, "frame-context", true,
                                popupNode.ownerDocument);
      });

      overlay.contentMenu.addItem({
        id: "unmht-context-menuitem-save",
        insertafter: "context-savepage",
        label: TEXT("menuitem.save.single"),
        oncommand: event => {
          lazy.UnMHTFileSaver.save(window);
        }
      });
      overlay.contentMenu.addItem({
        id: "unmht-context-menuitem-overwrite",
        insertafter: "unmht-context-menuitem-save",
        label: TEXT("menuitem.save.overwrite"),
        oncommand: event => {
          lazy.UnMHTFileSaver.overwrite(window);
        }
      });
      overlay.contentMenu.addItem({
        id: "unmht-context-menuitem-quicksave",
        insertafter: "unmht-context-menuitem-overwrite",
        label: TEXT("menuitem.save.quick"),
        oncommand: event => {
          lazy.UnMHTFileSaver.quicksave(window);
        }
      });
      overlay.contentMenu.addItem({
        id: "unmht-context-menuitem-sendmail",
        insertafter: "unmht-context-menuitem-quicksave",
        label: TEXT("menuitem.save.mail"),
        oncommand: event => {
          lazy.UnMHTFileSaver.sendmail(window);
        }
      });
      overlay.contentMenu.addItem({
        id: "unmht-context-menuitem-save-link",
        insertafter: "unmht-context-menuitem-sendmail",
        label: TEXT("menuitem.save.link"),
        oncommand: event => {
          lazy.UnMHTFileSaver.saveLink(window);
        }
      });

      overlay.contentMenu.addItem({
        id: "unmht-context-menuitem-save-selection",
        insertafter: "context-copy,unmht-context-menuitem-sendmail",
        label: TEXT("menuitem.save.selection"),
        oncommand: event => {
          let popupNode = this._getPopupNode(window);
          if (!popupNode) {
            return;
          }
          let browser = arWindowUtils.getCurrentBrowser(window);
          let frameDocument = popupNode.ownerDocument;
          lazy.UnMHTFileSaver.saveSelectionInDocument(window, browser, frameDocument);
        }
      });
      overlay.contentMenu.addItem({
        id: "unmht-context-menuitem-quicksave-selection",
        insertafter: "unmht-context-menuitem-save-selection",
        label: TEXT("menuitem.save.quick.selection"),
        oncommand: event => {
          lazy.UnMHTFileSaver.quicksave(window, true);
        }
      });



      overlay.frameContentMenu.addItem({
        id: "unmht-frame-context-menuitem-save",
        insertafter: "context-saveframe",
        label: TEXT("menuitem.save.single.frame"),
        oncommand: event => {
          let popupNode = this._getPopupNode(window);
          if (!popupNode) {
            return;
          }
          let browser = arWindowUtils.getCurrentBrowser(window);
          let frameDocument = popupNode.ownerDocument;
          lazy.UnMHTFileSaver.saveDocument(window, browser, frameDocument);
        }
      });
      overlay.frameContentMenu.addItem({
        id: "unmht-frame-context-menuitem-sendmail",
        insertafter: "unmht-frame-context-menuitem-save",
        label: TEXT("menuitem.save.mail.frame"),
        oncommand: event => {
          let popupNode = this._getPopupNode(window);
          if (!popupNode) {
            return;
          }
          let browser = arWindowUtils.getCurrentBrowser(window);
          let frameDocument = popupNode.ownerDocument;
          lazy.UnMHTFileSaver.sendmailDocument(window, browser, frameDocument);
        }
      });
    }
  },

  /**
   * ポップアップを開いた要素を取得する
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @returns {?nsIDOMHTMLElement}
   *          ポップアップを開いた要素
   */
  _getPopupNode: function(window) {
    if (window.document.popupNode) {
      return window.document.popupNode;
    }

    if ("gContextMenu" in window && window.gContextMenu &&
        "target" in window.gContextMenu &&
        window.gContextMenu.target) {
      return window.gContextMenu.target;
    }

    return null;
  },

  /**
   * コンテンツが選択されているかどうかを返す
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @param   {boolean} usePopupNode
   *          コンテキストメニューから情報を取得するか
   * @returns {boolean}
   *          コンテンツが選択されているかどうか
   */
  _isContentSelected: function(window, usePopupNode) {
    if (usePopupNode &&
        "gContextMenu" in window &&
        window.gContextMenu) {
      return window.gContextMenu.isContentSelected;
    }

    let targetDocument = null;
    let popupNode = null;
    if (usePopupNode) {
      popupNode = this._getPopupNode(window);
    }
    if (popupNode) {
      targetDocument = popupNode.ownerDocument;
    } else {
      let browser = arWindowUtils.getCurrentBrowser(window);
      targetDocument = browser.contentDocument;
    }

    if (!targetDocument) {
      return false;
    }
    if (!targetDocument.defaultView) {
      return false;
    }

    try {
      return !targetDocument.defaultView.getSelection().isCollapsed;
    } catch (e) {
    }

    return false;
  },

  /**
   * 保存可能なリンクの上に乗っているかどうかを返す
   * コンテキストメニューを表示してからでないと正しい値を返さない
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @returns {boolean}
   *          リンクの上に乗っているかどうか
   */
  _onSavableLink: function(window) {
    if ("gContextMenu" in window &&
        window.gContextMenu) {
      return window.gContextMenu.onSaveableLink;
    }

    return false;
  },

  /**
   * ページ上の何でもない所に乗っているかどうかを返す
   *   (リンクや画像、入力フィールド等でない事)
   * コンテキストメニューを表示してからでないと正しい値を返さない
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @returns {boolean}
   *          ページ上の何でもない所に乗っているかどうか
   */
  _onPage: function(window) {
    if ("gContextMenu" in window &&
        window.gContextMenu) {
      return !(window.gContextMenu.onImage ||
               window.gContextMenu.onCanvas ||
               window.gContextMenu.onVideo ||
               window.gContextMenu.onAudio ||
               window.gContextMenu.onLink ||
               window.gContextMenu.onTextInput);
    }

    return true;
  }
});

module.UnMHTMenu = UnMHTMenu;

})(this);
