"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTPageInfo"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});
let { UnMHTSaveState, UnMHTSaveStateUtils } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arFormatter",
                                  "resource://unmht/modules/base/arFormatter.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arWindowOverlay",
                                  "resource://unmht/modules/base/arWindowOverlay.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCache",
                                  "resource://unmht/modules/UnMHTCache.jsm");

/**
 * ページ情報に MHT ファイルの情報を表示する
 *
 * @class
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
let UnMHTPageInfo = Object.freeze({
  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    arWindowWatcher.addListener(this);
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!arWindowUtils.isPageInfo(window)) {
      return;
    }

    this._createUI(window);
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
  },

  /* ==== private ==== */

  /**
   * UI を作る
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _createUI: function(window) {
    let overlay = lazy.arWindowOverlay.get(window);

    let document = window.document;
    let $ = id => document.getElementById(id);

    Task.spawn(function*() {
      let i = 0, TRY_MAX = 10;
      let url;
      for (;;) {
        if ("gDocument" in window &&
            window.gDocument) {
          url = window.gDocument.documentURI;
          break;
        }
        if ("gDocInfo" in window &&
            window.gDocInfo) {
          url = window.gDocInfo.location;
          break;
        }

        /* 本体のスクリプトよりも先に動いた、または未知の対象の場合
         * とりあえずしばらく待つ */
        yield arTaskUtils.wait(100);

        i += 1;
        if (i == TRY_MAX) {
          return;
        }
      }

      if (!UnMHTScheme.isUnMHTURI(url)) {
        return;
      }

      let urlRow = $("generalURLRow");
      if (!urlRow) {
        let urltext = $("urltext");
        if (!urltext) {
          return;
        }

        urlRow = arDOMUtils.findParentNode(urltext, "row");
        if (!urlRow) {
          return;
        }
      }

      let [eFileInfo, part] = lazy.UnMHTCache.getPart(url);
      if (!(eFileInfo && part)) {
        return;
      }

      let nextSibling = urlRow.nextSibling;

      let addRow = (labelText, valueText, id) => {
        let row = overlay.pageInfoGeneralGrid.addRow({
          POSITION: { before: nextSibling }
        });

        row.addLabel({
          control: "unmht-original-url",
          value: labelText
        });
        row.addSeparator({});
        row.addTextBox({
          readonly: "true",
          id: id,
          value: valueText
        });
      };

      addRow(TEXT("info.original_address"),
             eFileInfo.original,
             "unmht-original-url");

      if (eFileInfo.startPart &&
          eFileInfo.startPart.contentLocation) {
        addRow(TEXT("info.saved_from"),
               part.contentLocation ||
               eFileInfo.startPart.contentLocation,
               "unmht-saved-from");
      }

      if (eFileInfo.date) {
        let savedTime = lazy.arFormatter.date(eFileInfo.date);
        addRow(TEXT("info.saved_date"),
               savedTime,
               "unmht-saved-date");
      }

      if (eFileInfo.topPart.hasField("X-UnMHT-Save-State")) {
        let state = UnMHTSaveState.UNKNOWN;

        switch (eFileInfo.topPart.getField("X-UnMHT-Save-State")) {
          case "Current-State": {
            state = UnMHTSaveState.CURRENT;
            break;
          }
          case "Current-State-Without-Script": {
            state = UnMHTSaveState.CURRENT_NOJS;
            break;
          }
          case "Original-File": {
            state = UnMHTSaveState.ORIGINAL;
            break;
          }
        }

        let saveState = UnMHTSaveStateUtils.saveStateToReadableString(state);

        if (saveState) {
          addRow(TEXT("info.save_state"),
                 saveState,
                 "unmht-save_state");
        }
      }
    }).catch(e => {
      Cu.reportError(e + "\n" + e.stack);
    });
  }
});

module.UnMHTPageInfo = UnMHTPageInfo;

})(this);
