"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTPrefWindow"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { DeferredTask } = Cu.import("resource://gre/modules/DeferredTask.jsm", {});
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arAlertUtils } = Cu.import("resource://unmht/modules/base/arAlertUtils.jsm", {});
let { arDialog } = Cu.import("resource://unmht/modules/base/arDialog.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arArrayUtils } = Cu.import("resource://unmht/modules/base/arArrayUtils.jsm", {});
let { arCustomizableUI } = Cu.import("resource://unmht/modules/base/arCustomizableUI.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});

let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});
let { UnMHTPref, UnMHTMailArg, UnMHTPrefJSONNames, UnMHTSaveMultiIndexStyleFileName } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { UnMHTSystemDirectory } = Cu.import("resource://unmht/modules/UnMHTSystemDirectory.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arFilePicker",
                                  "resource://unmht/modules/base/arFilePicker.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arListView",
                                  "resource://unmht/modules/base/arListView.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arLocaleUtils",
                                  "resource://unmht/modules/base/arLocaleUtils.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCache",
                                  "resource://unmht/modules/UnMHTCache.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCustomFilename",
                                  "resource://unmht/modules/UnMHTCustomFilename.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileConverter",
                                  "resource://unmht/modules/UnMHTFileConverter.jsm");
XPCOMUtils.defineLazyServiceGetter(lazy, "DragService",
                                   "@mozilla.org/widget/dragservice;1",
                                   "nsIDragService");

/**
 * HTML ファイルから MHT への変換器
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @implements IUnMHTFileConverterListener
 */
function UnMHTPrefWindowFileConverter(window) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 実行中の変換コンテキスト
   * @type {Set.<UnMHTFileConvertContext>}
   */
  this._contexts = new Set();

  /**
   * 検索する予定のフォルダのパスとフォルダ名の末尾
   * @type {Array.<Tuple.<string,string>>}
   */
  this._dirPaths = [];

  /**
   * リソースフォルダのパス
   * この中は検索しない
   * @type {Set.<string>}
   */
  this._resDirPaths = new Set();

  /**
   * 変換したファイル数の合計
   * @type {number}
   */
  this._total = 0;

  /**
   * 失敗したファイル数の合計
   * @type {number}
   */
  this._failed = 0;

  /**
   * リソースファイルが無いファイルの合計
   * @type {number}
   */
  this._no_res = 0;

  /**
   * 無視されたファイルの合計
   * @type {number}
   */
  this._ignored = 0;

  Object.seal(this);
}
UnMHTPrefWindowFileConverter.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;
    if (this._contexts) {
      for (let context of this._contexts) {
        context.abort();
      }
      this._contexts = null;
    }
  },

  /**
   * ファイルを変換する
   *
   * @param   {Array.<string>} paths
   *          対象のファイルのパス
   * @param   {string} suffix
   *          フォルダ名の末尾
   * @param   {boolean} searchSub
   *          サブフォルダを検索するか
   */
  convertFiles: Task.async(function*(paths, suffix, searchSub) {
    let filePaths = [];
    for (let path of paths) {
      let exists = yield OS.File.exists(path);
      if (!exists) {
        continue;
      }
      let stat = yield OS.File.stat(path);
      if (stat.isDir) {
        if (searchSub) {
          this._dirPaths.push([path, suffix]);
        }
      } else {
        filePaths.push(path);
      }
    }
    let context = lazy.UnMHTFileConverter.createContext(filePaths, suffix,
                                                        this);
    this._contexts.add(context);
    this._updateButton();
    context.start();
  }),

  /**
   * 全ての変換を中断する
   */
  abort: function() {
    this._dirPaths = [];
    this._resDirPaths = new Set();
    for (let context of this._contexts) {
      context.abort();
    }
  },

  /**
   * ログを追加する
   *
   * @param   {string} s
   *          ログ
   */
  log: function(s) {
    let logBox = this.$("convert_log");
    logBox.value += s;
    logBox.inputField.scrollTop = logBox.inputField.scrollTopMax;
  },

  /**
   * ログを改行付きで追加する
   *
   * @param   {string} s
   *          ログ
   */
  logln: function(s) {
    this.log(s + "\n");
  },

  /**
   * 変換ログを消去
   */
  clearLog: function() {
    this.$("convert_log").value = "";
  },

  /* ==== IUnMHTFileConverterListener ==== */

  /**
   * ファイル変換開始イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onStart: function(context, info, current, total) {
    this.log(TEXT("convert.log.start", [info.htmlPath]));
    this.$("convert_progressbar").value
      = Math.floor((this._total + current) / (this._total + total) * 100);
    this.$("convert_progressbar").style.visibility = "visible";
  },

  /**
   * ファイル変換完了イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onFinish: function(context, info, current, total) {
    this._resDirPaths.add(info.resDirPath);
    this.logln(TEXT("convert.log.finish"));
    this.logln("  " + TEXT("convert.log.input", [info.htmlPath]));
    this.logln("  " + TEXT("convert.log.res", [info.resDirPath]));
    this.logln("  " + TEXT("convert.log.output", [info.mhtPath]));
  },

  /**
   * ファイル変換失敗イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onFail: function(context, info, current, total) {
    this.logln(TEXT("convert.log.failed"));
    this._failed += 1;
  },

  /**
   * 無視されたファイル
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onIgnored: function(context, info, current, total) {
    this._ignored += 1;
  },

  /**
   * リソースファイルが無いファイル
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {UnMHTFileConvertInfo} info
   *          変換情報
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onResDirNotFound: function(context, info, current, total) {
    this.logln(TEXT("convert.log.no_res", [info.htmlPath]));
    this._no_res += 1;
  },

  /**
   * 全ファイル変換完了イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {number} count
   *          変換したファイル数
   */
  onComplete: Task.async(function*(context, count) {
    this._contexts.delete(context);
    this._updateButton();

    this._total += count;

    while (this._dirPaths.length) {
      let [dirPath, suffix] = this._dirPaths.shift();
      if (!this._resDirPaths.has(dirPath)) {
        let resIter = new OS.File.DirectoryIterator(dirPath);
        let resEntries = yield resIter.nextBatch();
        let paths = resEntries.map(entry => entry.path);
        this.convertFiles(paths, suffix, true);
        return;
      }
    }

    this._updateButton();

    this.logln(TEXT("convert.log.complete"));
    this.logln("  " + TEXT("convert.log.summary.ok", [this._total]));
    this.logln("  " + TEXT("convert.log.summary.failed", [this._failed]));
    this.logln("  " + TEXT("convert.log.summary.no_res", [this._no_res]));
    this.logln("  " + TEXT("convert.log.summary.ignored", [this._ignored]));
    this.$("convert_progressbar").value = 100;
    this.$("convert_progressbar").style.visibility = "hidden";

    this._total = 0;
    this._failed = 0;
    this._no_res = 0;
    this._ignored = 0;
  }),

  /**
   * 中断イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   */
  onAbort: function(context) {
    this._contexts.delete(context);
    this._updateButton();
    this.logln(TEXT("convert.log.abort"));
    this.$("convert_progressbar").style.visibility = "hidden";
  },

  /* ==== private ==== */

  /**
   * 中断ボタンの状態を更新する
   */
  _updateButton: function() {
    this.$("convert_abort").disabled
      = (this._contexts.size === 0 && this._dirPaths.length === 0);
  },

  /**
   * getElementById の省略
   *
   * @param   {string} id
   *          要素の ID
   * @returns {?nsIDOMElement}
   *          id に対応する要素
   *          見付からなければ null
   */
  $: function(id) {
    return this._window.document.getElementById(id);
  }
});

/**
 * 設定ウィンドウクラス
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 */
function UnMHTPrefWindowClass(window) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 変換器
   * @type {UnMHTPrefWindowFileConverter}
   */
  this._converter = new UnMHTPrefWindowFileConverter(window);

  /**
   * 設定を手動で更新中か
   * @type {boolean}
   */
  this._updating = false;

  /**
   * インデックスページの CSS のテキストボックスにデータをロードしたか
   * @type {boolean}
   */
  this._indexStyleLoaded = false;

  /**
   * インデックスページの CSS のテキストボックスの内容
   * @type {string}
   */
  this._indexStyle = "";

  /**
   * ファイル名の置換のリストのドラッグアンドドロップをフックしたか
   * @type {boolean}
   */
  this._replaceHooked = false;

  /**
   * キャッシュのボタンをフックしたか
   * @type {boolean}
   */
  this._cacheHooked = false;

  /**
   * prefwindow.instantApply の値
   * @type {boolean}
   */
  this._instantApply = false;

  /**
   * ファイル名の文字の置換のリスト
   * @type {arListView}
   */
  this._filenameReplaceListView = null;

  /**
   * スタイルシートの保存タスク
   * @type {DeferredTask}
   */
  this._saveTask = new DeferredTask(() => {
    return Task.spawn(function*() {
      try {
        yield* this._saveIndexStyleTask();
      } catch (e) {
        Cu.reportError(e + "\n" + e.stack);
      }
    }.bind(this));
  }, 3000);

  Object.seal(this);
}
UnMHTPrefWindowClass.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;
    if (this._converter) {
      this._converter.destruct();
      this._converter = null;
    }
    if (this._filenameReplaceListView) {
      this._filenameReplaceListView.destruct();
      this._filenameReplaceListView = null;
    }
  },

  /* ---- UI 関係イベント ---- */

  /**
   * ロードイベント
   */
  onLoad: function() {
    this._instantApply = this.$("unmht_pref_window").instantApply;

    this._disableItems();

    this._updateUI();

    this._window.sizeToContent();
  },

  /**
   * アンロードイベント
   */
  onUnload: function() {
    this._saveTask.finalize();
  },

  /**
   * 各パネルがロードされたイベント
   */
  onPaneLoad: function() {
    this._disableItems();

    this._updateUI();
  },

  /**
   * ヘルブボタンがクリックされたイベント
   */
  onHelp: function(event) {
    this._openURI((lazy.arLocaleUtils.isJapanese()
                   ? "http://www.unmht.org/unmht/usage-8.0.0.html"
                   : "http://www.unmht.org/unmht/en_usage-8.0.0.html")
                  + "#" + this.$("unmht_pref_window").currentPane.helpTopic);
  },

  /**
   * サイトへのリンクがクリックされたイベント
   * 現在未使用
   */
  onClickWebsiteLink: function() {
    this._openURI(lazy.arLocaleUtils.isJapanese()
                  ? "http://www.unmht.org/unmht/"
                  : "http://www.unmht.org/unmht/en_index.html");
  },

  /**
   * ショートカットキー入力欄でキーを押したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onShortcutKeyDown: function(event, id) {
    let elem = type => this.$("shortcut.save." + id + "." + type);

    let keycode = "";
    let KeyboardEvent = this._window.KeyboardEvent;
    for (let attribute of Object.keys(KeyboardEvent)) {
      let m = attribute.match(/^DOM_(VK_.+)$/);
      if (!m) {
        continue;
      }
      let code = m[1];
      if (event.keyCode == KeyboardEvent[attribute]) {
        keycode = code;
      }
    }
    if (keycode == "VK_SHIFT" ||
        keycode == "VK_CONTROL" ||
        keycode == "VK_ALT" ||
        keycode == "VK_META") {
      return;
    }

    if (keycode !== "") {
      elem("keycode").value = keycode;
      elem("mods.alt").value = event.altKey;
      elem("mods.ctrl").value = event.ctrlKey;
      elem("mods.meta").value = event.metaKey;
      elem("mods.shift").value = event.shiftKey;
    }

    event.preventDefault();
    event.stopPropagation();
  },

  /**
   * メール送信アプリケーションの変更ボタンを押したイベント
   */
  onClickSelectSendMailApp: Task.async(function*() {
    let node = this.$("save.mail.app");

    let params = {
      title:            TEXT("save.mail.app.title"),
      useLastDir:       false,
      displayDirectory: arFileUtils.getFile(node.value),
      uri:              null
    };

    let [path] = yield lazy.arFilePicker.getOpenFile(this._window, params);
    if (!path) {
      return;
    }
    node.value = path;
  }),

  /**
   * メール送信アプリケーションの引数を変更したイベント
   */
  onChangeSendMailArgType: function() {
    let textbox = this.$("ui-save.mail.arg.other");
    let value = "";

    switch (this.$("save.mail.arg.type").value) {
      case "thunderbird": {
        value = UnMHTMailArg.thunderbird;
        textbox.disabled = true;
        break;
      }
      case "seamonkey": {
        value = UnMHTMailArg.seamonkey;
        textbox.disabled = true;
        break;
      }
      case "macmail": {
        value = UnMHTMailArg.macmail;
        textbox.disabled = true;
        break;
      }
      default: {
        textbox.disabled = false;
        return;
      }
    }

    this.$("save.mail.arg.other").value = value;
  },

  /**
   * ディレクトリの変更ボタンを押したイベント
   *
   * @param   {string} id
   *          対象の設定
   */
  onClickSelectBase: Task.async(function*(id) {
    let node = this.$("save." + id + ".dir");

    let params = {
      title:            TEXT("save." + id + ".base.title"),
      useLastDir:       false,
      displayDirectory: arFileUtils.getFile(node.value),
      uri:              null
    };
    let [dirPath] = yield lazy.arFilePicker.getFolder(this._window, params);
    if (!dirPath) {
      return;
    }
    node.value = dirPath;
  }),

  /**
   * カスタムファイル名のサンプルの更新が必要になったイベント
   */
  onNeedUpdateCustomFilename: function() {
    this.$("save_filename_custom_sample").value
      = UnMHTFilename.validateFileName(
        lazy.UnMHTCustomFilename.formatWith(
          "no title",
          "index.html",
          "http://localhost/index.html?query#fragment",
          new Date(),
          this.$("save.filename.custom").value,
          this.$("save.filename.custom.am").value,
          this.$("save.filename.custom.pm").value));
  },

  /**
   * ファイル名の文字の置換のリストの設定値が変更されたイベント
   */
  onFilenameReplaceListPrefChange: function() {
    if (this._updating) {
      return;
    }

    let node = this.$("save.filename.replace.list");

    let datas;
    try {
      datas = JSON.parse(node.value);
      if (!datas.every(data => "from" in data && "to" in data)) {
        datas = [];
      }
    } catch (e) {
      datas = [];
    }

    this._filenameReplaceListView.setDatas(datas);
  },

  /**
   * HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertFiles: Task.async(function*() {
    let params = {
      title:            TEXT("convert.title"),
      useLastDir:       true,
      displayDirectory: null,
      uri:              null
    };
    let [paths] = yield lazy.arFilePicker.getOpenFiles(this._window, params);
    if (!paths) {
      return;
    }
    this._converter.convertFiles(paths,
                                 this.$("convert.dir.suffix").value,
                                 this.$("ui-convert.dir.sub").checked);
  }),

  /**
   * ディレクトリ内の HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertDir: Task.async(function*() {
    let params = {
      title:            TEXT("convert.title"),
      useLastDir:       true,
      displayDirectory: null,
      uri:              null
    };
    let [dirPath] = yield lazy.arFilePicker.getFolder(this._window, params);
    if (!dirPath) {
      return;
    }

    let resIter = new OS.File.DirectoryIterator(dirPath);
    let resEntries = yield resIter.nextBatch();
    let paths = resEntries.map(entry => entry.path);

    this._converter.convertFiles(paths,
                                 this.$("convert.dir.suffix").value,
                                 this.$("ui-convert.dir.sub").checked);

  }),

  /**
   * 変換ログを消去するボタンを押したイベント
   */
  onClickClearConvertLog: function() {
    this._converter.clearLog();
  },

  /**
   * 変換パネルにファイルがドラッグされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDragOverConvertFiles: function(event) {
    if (lazy.DragService.getCurrentSession()
        .isDataFlavorSupported("application/x-moz-file")) {
      event.stopPropagation();
      event.preventDefault();
    }
  },

  /**
   * 変換パネルにファイルがドロップされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDragDropConvertFiles: Task.async(function*(event) {
    let dataTransfer = event.dataTransfer;

    let paths = [];
    for (let i of arArrayUtils.range(0, dataTransfer.mozItemCount)) {
      let types = dataTransfer.mozTypesAt(i);
      if (Array.indexOf(types, "application/x-moz-file") == -1) {
        continue;
      }

      let data = dataTransfer.mozGetDataAt("application/x-moz-file", i);
      try {
        let path = data.QueryInterface(Ci.nsIFile).path;

        let stat = yield OS.File.stat(path);
        if (stat.isDir) {
          let resIter = new OS.File.DirectoryIterator(path);
          let resEntries = yield resIter.nextBatch();
          paths = paths.concat(resEntries.map(entry => entry.path));
        } else {
          paths.push(path);
        }
      } catch (e) {
      }
    }

    this._converter.convertFiles(paths,
                                 this.$("convert.dir.suffix").value,
                                 this.$("ui-convert.dir.sub").checked);
  }),

  /**
   * ディレクトリ名の末尾を初期化するボタンを押したイベント
   */
  onClickInitConvertDirSuffix: function() {
    this.$("convert.dir.suffix").value = "_files";
  },

  /**
   * 変換中断ボタンを押したイベント
   */
  onClickAbortConvert: function() {
    this._converter.abort();
  },

  /**
   * 設定のインポートボタンを押したイベント
   */
  onClickImportPrefs: Task.async(function*() {
    this._saveTask.disarm();
    yield* this._saveIndexStyleTask();

    let setStatus = type => {
      this.$("import_export_message").value = TEXT("pref.import." + type);
      this.$("import_export_message").setAttribute("message_type", "import_" + type);
    };

    let params = {
      title:            TEXT("pref.import.title"),
      useLastDir:       false,
      displayDirectory: null,
      defaultString:    "unmhtConfig.txt",
      filter:           [["Text", "*.txt"], [Ci.nsIFilePicker.filterAll]],
      uri:              null
    };

    let [path] = yield lazy.arFilePicker.getOpenFile(this._window, params);
    if (!path) {
      setStatus("abort");
      return;
    }

    let input;
    try {
      input = yield UnMHTPref.initPrefFromFile(path);
    } catch (e) {
      setStatus("illegal");
      Cu.reportError("File format error\n"
                     + e + "\n" + e.stack);
      return;
    }

    if (!this._instantApply) {
      yield this._showAllPanes();
    }

    try {
      this._loadPrefs(input);
      yield UnMHTPref.closePrefFromFile(input);
      yield* this._loadIndexStyleTask();
      setStatus("done");
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      setStatus("fail");
    }
  }),

  /**
   * 設定のエクスポートボタンを押したイベント
   */
  onClickExportPrefs: Task.async(function*() {
    this._saveTask.disarm();
    yield* this._saveIndexStyleTask();

    let setStatus = type => {
      this.$("import_export_message").value = TEXT("pref.export." + type);
      this.$("import_export_message").setAttribute("message_type", "export_" + type);
    };

    let params = {
      title:            TEXT("pref.export.title"),
      useLastDir:       false,
      displayDirectory: null,
      defaultString:    "unmhtConfig.txt",
      filter:           [["Text", "*.txt"], [Ci.nsIFilePicker.filterAll]],
      uri:              null
    };

    let [path] = yield lazy.arFilePicker.getSaveFile(this._window, params);
    if (!path) {
      setStatus("abort");
      return;
    }

    let output = yield UnMHTPref.initPrefToFile(path);
    if (!output) {
      setStatus("fail");
      return;
    }

    if (!this._instantApply) {
      yield this._showAllPanes();
    }

    try {
      this._savePrefs(output);
      yield UnMHTPref.closePrefToFile(output);
      setStatus("done");
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      setStatus("fail");
    }
  }),

  /**
   * 設定の初期化ボタンを押したイベント
   */
  onClickInitPrefs: Task.async(function*() {
    this._saveTask.disarm();
    yield* this._saveIndexStyleTask();

    let title = TEXT("name");
    let message = TEXT("pref.reset.confirm");
    let ret = yield arAlertUtils.confirm(this._window, title, message);
    if (!ret) {
      return;
    }

    let setStatus = type => {
      this.$("import_export_message").value = TEXT("pref.reset." + type);
      this.$("import_export_message").setAttribute("message_type", "reset_" + type);
    };

    let input = yield UnMHTPref.initPrefFromFile(null);

    if (!this._instantApply) {
      yield this._showAllPanes();
    }

    this._loadPrefs(input);
    yield* this._removeIndexStyleTask();
    yield UnMHTPref.closePrefFromFile(input);
    setStatus("done");
  }),

  /**
   * カスタムファイル名に変数を挿入する
   *
   * @param   {string} text
   *          挿入するテキスト
   */
  insertCustom: function(text) {
    let textbox = this.$("ui-save.filename.custom");
    let value = textbox.value;
    let start = textbox.selectionStart;
    let end = textbox.selectionEnd;

    textbox.value = value.slice(0, start) + text + value.slice(end);
    this.$("save.filename.custom").value = textbox.value;

    textbox.focus();

    textbox.selectionStart = textbox.selectionEnd = start + text.length;

    this.onNeedUpdateCustomFilename();
  },

  /**
   * カスタムファイル名のサンプルを表示する
   *
   * @param   {string} text
   *          挿入するテキスト
   */
  hoverCustom: function(text) {
    let samplebox = this.$("custom_var_sample");

    while (samplebox.firstChild) {
      samplebox.firstChild.remove();
    }

    let desc = {
      "%YYYY%": TEXT("filename.custom.desc.YYYY"),
      "%YY%": TEXT("filename.custom.desc.YY"),
      "%MM%": TEXT("filename.custom.desc.MM"),
      "%M%": TEXT("filename.custom.desc.M"),
      "%DD%": TEXT("filename.custom.desc.DD"),
      "%D%": TEXT("filename.custom.desc.D"),
      "%MMM%": TEXT("filename.custom.desc.MMM"),
      "%MMMM%": TEXT("filename.custom.desc.MMMM"),
      "%HH%": TEXT("filename.custom.desc.HH"),
      "%H%": TEXT("filename.custom.desc.H"),
      "%hh%": TEXT("filename.custom.desc.hh"),
      "%h%": TEXT("filename.custom.desc.h"),
      "%mm%": TEXT("filename.custom.desc.mm"),
      "%m%": TEXT("filename.custom.desc.m"),
      "%ss%": TEXT("filename.custom.desc.ss"),
      "%s%": TEXT("filename.custom.desc.s"),
      "%ampm%": TEXT("filename.custom.desc.ampm"),
      "%TITLE%": TEXT("filename.custom.desc.TITLE"),
      "%FILENAME%": ["http://unmht.org/", ["sample.cgi"], "?page=2#chapter2"],
      "%URL%": [["http://unmht.org/sample.cgi?page=2#chapter2"]],
      "%URL2%": ["http://", ["unmht.org/sample.cgi?page=2#chapter2"]],
      "%SCHEME%": [["http"], "://unmht.org/sample.cgi?page=2#chapter2"],
      "%HOST%": ["http://", ["unmht.org"], "/sample.cgi?page=2#chapter2"],
      "%PORT%": ["http://unmht.org:", ["80"], "/sample.cgi?page=2#chapter2"],
      "%FULLPATH%": ["http://unmht.org", ["/sample.cgi?page=2#chapter2"]],
      "%PATH%": ["http://unmht.org", ["/sample.cgi"], "?page=2#chapter2"],
      "%QUERY%": ["http://unmht.org/sample.cgi?", ["page=2"], "#chapter2"],
      "%FRAGMENT%": ["http://unmht.org/sample.cgi?page=2#", ["chapter2"]],
    };

    if (!(text in desc)) {
      return;
    }

    for (let t of desc[text]) {
      let label = this._window.document.createElement("label");
      if (typeof t == "string") {
        label.setAttribute("value", t);
      } else {
        label.style.color = "red";
        label.style.textDecoration = "underline";
        label.setAttribute("value", t[0]);
      }
      samplebox.appendChild(label);
    }
  },

  /**
   * スピンボタンで数値を上げる
   *
   * @param   {string} id
   *          対象の要素の ID
   * @param   {number} max
   *          数値の最大値
   * @param   {number=} step
   *          ステップ
   */
  onSpinUp: function(id, max, step=1) {
    let node = this.$(id);

    let n = Number.parseInt(node.value, 10) + step;
    if (n > max) {
      n = max;
    }

    node.value = n;
  },

  /**
   * スピンボタンで数値を下げる
   *
   * @param   {string} id
   *          対象の要素の ID
   * @param   {number} max
   *          数値の最小値
   * @param   {number=} step
   *          ステップ
   */
  onSpinDown: function(id, min, step=1) {
    let node = this.$(id);

    let n = Number.parseInt(node.value, 10) - step;
    if (n < min) {
      n = min;
    }

    node.value = n;
  },

  /**
   * インデックスページのスタイルを変更したイベント
   * ファイルを更新する
   */
  onChangeSaveMultiIndexStyleContent: function() {
    this._indexStyle = this.$("ui-save.multi.index.style.content").value;
    this._saveTask.disarm();
    this._saveTask.arm();
  },

  /* ---- UI イベント - 依存関係 ---- */

  onChangeSaveLocalSameDir: function() {
    this._setDisabled(this.$("box_save_local_same_dir"),
                      !this.$("save.local.same_dir").value);
  },
  onChangeSaveInsertLink: function() {
    this._setDisabled(this.$("box_save_insert_link_1"),
                      !this.$("save.insert.link").value);
    this._setDisabled(this.$("box_save_insert_link_2"),
                      !this.$("save.insert.link").value);
  },
  onChangeSaveMultiIndexStyle: function() {
    this._setDisabled(this.$("ui-save.multi.index.style.content"),
                      !this.$("save.multi.index.style").value);
  },
  onChangeSaveMailAppType: function() {
    this._setDisabled(this.$("box_save_mail_app"),
                      this.$("save.mail.app.type").value != "other");

    if (this.$("save.mail.app.type").value != "other") {
      let argtype = this.$("save.mail.arg.type");

      argtype.value = this.$("save.mail.app.type").value;

      this.onChangeSendMailArgType();
    }
  },
  onChangeShortcut: function(id) {
    this._setDisabled(this.$("box_shortcut_" + id.replace(/\./g, "_")),
                      !this.$("shortcut." + id).value);
    this._setDisabled(this.$("field_shortcut_" + id.replace(/\./g, "_")),
                      !this.$("shortcut." + id).value);
  },

  /* ==== private ==== */

  /**
   * 設定を読み込む
   *
   * @param   {UnMHTPrefFileInputObject} input
   *          設定入力オブジェクト
   */
  _loadPrefs: function(input) {
    for (let name of UnMHTPref.names) {
      let value = UnMHTPref.getPrefFromFile(input, name);

      if (UnMHTPrefJSONNames.indexOf(name) != -1) {
        value = JSON.stringify(value);
      }

      if (this._instantApply) {
        UnMHTPref.setPref(name, value);
      } else {
        let node = this.$(name);
        if (!node) {
          continue;
        }
        node.value = value;
      }
    }

    this._updateUI();
  },

  /**
   * 全ての設定を保存する
   *
   * @param   {UnMHTPrefFileOutputObject} output
   *          設定出力オブジェクト
   */
  _savePrefs: function(output) {
    for (let name of UnMHTPref.names) {
      let value;
      if (this._instantApply) {
        value = UnMHTPref.getPref(name);
      } else {
        let node = this.$(name);
        if (!node) {
          continue;
        }

        value = node.value;
      }

      if (UnMHTPrefJSONNames.indexOf(name) != -1) {
        value = JSON.parse(value);
      }

      UnMHTPref.setPrefToFile(output, name, value);
    }
  },

  /**
   * 全てのパネルを表示する
   * パネルは選択されるまでロードされないので
   * インポート、エクスポート、初期化時に必要
   *
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  _showAllPanes: Task.async(function*() {
    let panes = this._window.document.getElementsByTagName("prefpane");
    let defaultPane = this.$("pane_general");
    let prefWindow = this.$("unmht_pref_window");
    for (let pane of Array.from(panes)) {
      while (!pane.loaded) {
        prefWindow.showPane(pane);

        yield arTaskUtils.wait(100);
      }
    }

    prefWindow.showPane(defaultPane);
  }),

  /**
   * インデックスページのスタイルをロードする
   *
   * @yields  {Promise.<*>}
   */
  _loadIndexStyleTask: function*() {
    let node = this.$("ui-save.multi.index.style.content");
    if (node) {
      let filePath = yield UnMHTSystemDirectory.getSystemFilePath(UnMHTSaveMultiIndexStyleFileName);
      let content = "";
      try {
        content = yield arFileUtils.readUTF8(filePath);
      } catch (e) {
      }
      node.value = content;
    }
  },

  /**
   * インデックスページのスタイルをセーブする
   *
   * @yields  {Promise.<*>}
   */
  _saveIndexStyleTask: function*() {
    let filePath = yield UnMHTSystemDirectory.getSystemFilePath(UnMHTSaveMultiIndexStyleFileName);

    yield arFileUtils.createUTF8(filePath, this._indexStyle);
  },

  /**
   * インデックスページのスタイルを削除する
   *
   * @yields  {Promise.<*>}
   */
  _removeIndexStyleTask: function*() {
    let prefix = UnMHTPref.prefix;
    let name = "save.multi.index.style.content";
    if (Services.prefs.getPrefType(prefix + name)
        == Services.prefs.PREF_STRING) {
      Services.prefs.clearUserPref(prefix + name);
    }

    let filePath = yield UnMHTSystemDirectory.getSystemFilePath(UnMHTSaveMultiIndexStyleFileName);
    yield OS.File.remove(filePath);

    let node = this.$("ui-save.multi.index.style.content");
    if (node) {
      node.value = "";
    }
  },

  /**
   * チェックボックスの状態等を更新する
   */
  _updateUI: function() {
    if (this.$("pane_save").loaded) {
      this.onChangeSaveLocalSameDir();
      this.onChangeSaveInsertLink();
    }
    if (this.$("pane_savemulti").loaded) {
      this.onChangeSaveMultiIndexStyle();

      if (!this._indexStyleLoaded) {
        this._indexStyleLoaded = true;

        Task.spawn(function*() {
          yield* this._loadIndexStyleTask();
        }.bind(this));
      }
    }
    if (this.$("pane_mail").loaded) {
      this.onChangeSaveMailAppType();
      this.onChangeSendMailArgType();
    }
    if (this.$("pane_filename").loaded) {
      this.onNeedUpdateCustomFilename();

      if (!this._replaceHooked) {
        this._replaceHooked = true;
        this._hookReplace();
      }

      this.onFilenameReplaceListPrefChange();
    }
    if (this.$("pane_keyboard").loaded) {
      if (Services.appinfo.OS == "Darwin") {
        this.$("ui-shortcut.save.single.mods.alt").label
          = this.$("ui-shortcut.save.multi.mods.alt").label
          = this.$("ui-shortcut.save.quick.mods.alt").label
          = this.$("ui-shortcut.save.overwrite.mods.alt").label
          = this.$("keyboard_mods_option").value;

        this.$("ui-shortcut.save.single.mods.meta").label
          = this.$("ui-shortcut.save.multi.mods.meta").label
          = this.$("ui-shortcut.save.quick.mods.meta").label
          = this.$("ui-shortcut.save.overwrite.mods.meta").label
          = this.$("keyboard_mods_command").value;
      }

      this.onChangeShortcut("save.single");
      this.onChangeShortcut("save.multi");
      this.onChangeShortcut("save.quick");
      this.onChangeShortcut("save.overwrite");
    }
    if (this.$("pane_advanced").loaded) {
      if (!this._cacheHooked) {
        this._cacheHooked = true;

        let messageNode = this.$("cache_message");
        let prefWindow = this._window;
        let clear = this.$("ui-open.cache.clear");
        arEvent.add(this._window, clear, "command", Task.async(function*() {
          messageNode.value = "";

          let title = TEXT("name");
          let message = TEXT("cache.clear.confirm");
          let result = yield arAlertUtils.confirm(prefWindow, title, message);

          if (!result) {
            return;
          }

          lazy.UnMHTCache.removeOld(0);

          messageNode.value = TEXT("cache.clear.done");
        }), true);

        let show = this.$("ui-open.cache.show");
        arEvent.add(this._window, show, "command", () => {
          let window = Services.wm.getMostRecentWindow("navigator:browser");
          if (window) {
            let tabbrowser = arWindowUtils.getTabBrowser(window);
            tabbrowser.selectedTab = tabbrowser.addTab(UnMHTScheme.createCacheURI());
          }
        }, true);
      }
    }
  },

  /**
   * ファイル名の文字の置換のリストを初期化する
   */
  _hookReplace: function() {
    let dataHandler = {
      /**
       * データを作成する
       *
       * @returns {DataType}
       *          datas に追加するデータ
       */
      newData: () => {
        return { from: "/", to: "_" };
      },

      /**
       * データを複製する
       *
       * @param   {DataType} origItem
       *          複製元のデータ
       * @returns {DataType}
       *          複製したデータ
       */
      dupData: origItem => {
        return { from: origItem.from, to: origItem.to };
      },

      /**
       * データが更新されたイベント
       *
       * @param   {Array.<DataType>} datas
       *          更新されたデータ
       */
      onModify: datas => {
        let node = this.$("save.filename.replace.list");

        this._updating = true;
        node.value = JSON.stringify(datas);
        this._updating = false;

        this.onNeedUpdateCustomFilename();
      },

      /**
       * データをリストの行に変換する
       *
       * @param   {DataType} item
       *          対象のデータ
       * @returns {Array.<Object>}
       *          セルのデータのリスト
       */
      dataToRow: item => {
        let row = [];

        row.push({ label: item.from });
        row.push({ label: item.to });

        return row;
      },

      /**
       * リストの行をデータに変換する
       *
       * @param   {Array.<Object>} row
       *          行のラベル
       *            [{ label: string ラベル }, ...]
       * @param   {DataType} oldItem
       *          過去のデータ
       * @returns {DataType}
       *          対象のデータ
       */
      rowToData: (row, oldItem) => {
        let item = {
          from: row[0].label,
          to: row[1].label
        };

        return item;
      },

      /**
       * リストのデータを選択したイベント
       *
       * @param   {Array.<Tuple.<DataType,number>>} items
       *          選択したデータと行のインデックスのリスト
       */
      onSelect: items => {
        /* 直接編集するので何もしない */
      },

      /**
       * ソートのためのデータを返す
       *
       * @param   {DataType} item
       *          対象のデータ
       * @param   {number} pos
       *          カラムの位置
       * @returns {comparable}
       *          ソート時の比較に使うデータ
       */
      getSortKeyFor: (item, pos) => {
        /* ソートは不要なので何もしない */
      }
    };

    this._filenameReplaceListView
      = new lazy.arListView(this.$("filename_replace_list"),
                            null,
                            this.$("filename_replace_list_children"),
                            this.$("filename_replace_list_new"),
                            this.$("filename_replace_list_dup"),
                            this.$("filename_replace_list_del"),
                            "application/x-unmht-filename-replace",
                            [],
                            true,
                            [],
                            dataHandler);

  },

  /**
   * ノードの状態を変更する
   *
   * @param   {nsIDOMXULElement} node
   *          対象のノード
   * @param   {boolean} disabled
   *          無効か
   * @param   {boolean} uncheck
   *          チェックを外すか
   */
  _setDisabled: function(node, disabled, uncheck) {
    try {
      node.disabled = disabled;
    } catch (e) {
    }

    if (uncheck) {
      try {
        node.checked = false;
      } catch (e) {
      }
    }

    if ("childNodes" in node) {
      for (let n of Array.from(node.childNodes)) {
        this._setDisabled(n, disabled, uncheck);
      }
    }
  },

  /**
   * 環境によって使用できない設定を無効にする
   */
  _disableItems: function() {
    let disableByIds = ids => {
      for (let id of ids) {
        let node = this.$(id);
        if (node) {
          this._setDisabled(node, true, true);
        }
      }
    };

    if (arAppInfo.isThunderbird) {
      disableByIds([
        "group_open",

        "group_addonbar",

        "box_insert",
        "box_insert_url",
        "ui-save.local.same_dir",
        "box_save_local_same_dir",

        "group_save_multi",
        "group_save_multi_index",
        "group_save_mail",

        "save_filename_leafname",

        "group_menu_context",
        "group_shortcut",

        "group_menu_all",
        "group_cache",
        "group_advanced_mail",
        "group_other"
      ]);
    }

    if (arAppInfo.isFirefox) {
      disableByIds([
        "ui-save.header.mail.original_field",
        "ui-save.filename.custom.mail_date",
        "ui-save.header.mail.show",
        "ui-save.header.mail.show-desc"
      ]);
    }

    if (!arCustomizableUI.supported) {
      disableByIds([
        "save_progress_display_toolbar"
      ]);
    }

    if (Services.appinfo.OS != "Darwin") {
      disableByIds([
        "save_mail_app_type_macmail",
        "save_mail_arg_type_macmail"
      ]);
    }
    if (Services.appinfo.OS != "WINNT") {
      disableByIds([
        "group_advanced_filename",
      ]);
    }
  },

  /**
   * ページをタブか新しいウィンドウに開く
   *
   * @param   {string} uri
   *          アドレス
   */
  _openURI: function(uri) {
    let chromeWindow = Services.wm.getMostRecentWindow("navigator:browser");
    if (chromeWindow) {
      let tabbrowser = arWindowUtils.getTabBrowser(chromeWindow);
      tabbrowser.selectedTab = tabbrowser.addTab(uri);
    } else {
      this._window.open(uri);
    }
  },

  /**
   * getElementById の省略
   *
   * @param   {string} id
   *          要素の ID
   * @returns {?nsIDOMElement}
   *          id に対応する要素
   *          見付からなければ null
   */
  $: function(id) {
    return this._window.document.getElementById(id);
  }
});

/**
 * 設定ウィンドウ
 * @type {arDialog}
 */
let UnMHTPrefWindow = new arDialog(UnMHTPrefWindowClass, {
  url: "chrome://unmht/content/pref/pref.xul",
  features: "chrome,toolbar",
  single: true,
  usePromise: false
});

module.UnMHTPrefWindow = UnMHTPrefWindow;

})(this);
