"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTProtocolHandler"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { toUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCache",
                                  "resource://unmht/modules/UnMHTCache.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTLoadModifier",
                                  "resource://unmht/modules/UnMHTLoadModifier.jsm");

/**
 * unmht プロトコルからチャネルを生成する
 *
 * @constructor
 * @implements nsIProtocolHandler
 */
function UnMHTProtocolHandler() {
  Object.seal(this);
}
UnMHTProtocolHandler.prototype = Object.freeze({
  classID:          Components.ID("{d605b959-0688-4846-887c-f6658725f7f1}"),
  contractID:       "@mozilla.org/network/protocol;1?name=unmht",
  classDescription: "UnMHT Protocol Handler JS Component",
  QueryInterface:   XPCOMUtils.generateQI([Ci.nsIProtocolHandler]),

  /* ==== nsIProtocolHandler ==== */

  /**
   * プロトコルスキーム
   * @type {string}
   */
  scheme: "unmht",

  /**
   * ポート番号
   * ネットワークベースではないので -1
   * @type {number}
   */
  defaultPort: -1,

  /**
   * プロトコルの詳細
   * file と同様のプロトコル
   * @type {number}
   */
  protocolFlags: (Ci.nsIProtocolHandler.URI_NOAUTH
                  | Ci.nsIProtocolHandler.URI_LOADABLE_BY_ANYONE),

  /**
   * ブラックリストのポートを上書きするか
   *
   * @param   {number} port
   *          ポート番号
   * @param   {string} scheme
   *          プロトコルスキーム
   * @returns {boolean}
   *          上書きするか
   */
  allowPort: function(port, scheme) {
    return false;
  },

  /**
   * URI を作成する
   *
   * @param   {string} spec
   *          対象の URI
   * @param   {string} charset
   *          対象の文字コード
   * @param   {nsIURI} baseURI
   *          読み込み元の URI
   * @returns {nsIURI}
   *          作成した URI
   */
  newURI: function(spec, charset, baseURI) {
    let uri
      = Cc["@mozilla.org/network/standard-url;1"].createInstance(Ci.nsIURI);

    if (!baseURI) {
      uri.spec = spec;
      return uri;
    }

    if (UnMHTScheme.isUnMHTURI(baseURI.spec)) {
      let resolved = lazy.UnMHTLoadModifier.resolveURI(baseURI.spec, spec);
      if (resolved) {
        uri.spec = resolved;
        return uri;
      }
    }

    uri.spec = baseURI.resolve(spec);
    return uri;
  },

  /**
   * チャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  newChannel: function(uri) {
    return this.newChannel2(uri, null);
  },

  /**
   * チャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  newChannel2: function(uri, loadInfo) {
    let param = UnMHTScheme.getUnMHTURIParam(uri.spec);

    try {
      if (!param.original) {
        /* 不正な URI */
        return this._createIllegalChannel(uri, loadInfo);
      } else if (param.scheme == "unmht-cache") {
        return this._createCacheChannel(uri, loadInfo);
      } else if (param.scheme == "unmht-empty") {
        return this._createEmptyChannel(uri, loadInfo);
      } else if (param.scheme == "unmht-loading") {
        return this._createLoadingChannel(uri, loadInfo);
      } else if (param.scheme == "unmht-error") {
        return this._createNotFoundChannel(uri, loadInfo);
      } else if (param.scheme == "unmht-view-source") {
        return this._createViewSourceChannel(uri, param, loadInfo);
      } else {
        /* mht 内のファイル */
        return this._createFileChannel(uri, param, loadInfo);
      }
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }

    return this._createEmptyChannel(uri, loadInfo);
  },

  /* ==== private ==== */

  /**
   * mht 内のファイルのチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI (umht:// 形式)
   * @param   {UnMHTURIParam} param
   *          対象の URI の情報
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createFileChannel: function(uri, param, loadInfo) {
    let [eFileInfo, part] = lazy.UnMHTCache.getPart(uri.spec);
    if (!eFileInfo) {
      try {
        /* URI が不正でないかチェック */
        Services.io.newURI(param.original, null, null);
      } catch (e) {
        return this._createIllegalChannel(uri, loadInfo);
      }

      /* 未展開 */
      return this._createLoadingChannel(uri, loadInfo);
    }

    if (!part) {
      /* 存在しないパート */
      return this._createEmptyChannel(uri, loadInfo);
    }

    return this._createStringInputStreamChannel(uri, part.eParam.charset,
                                                part.eParam.mimetype,
                                                part.eParam.content,
                                                part.eParam.content.length,
                                                loadInfo);
  },

  /**
   * mht のソースを表示するチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI (umht:// 形式)
   * @param   {UnMHTURIParam} param
   *          対象の URI の情報
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createViewSourceChannel: function(uri, param, loadInfo) {
    let origURISpec = UnMHTScheme.deViewSourceURI(uri.spec);

    let origParam = UnMHTScheme.getUnMHTURIParam(origURISpec);
    let eFileInfo = lazy.UnMHTCache.get(origParam.baseURI);
    if (!eFileInfo) {
      try {
        /* URI が不正でないかチェック */
        Services.io.newURI(param.original, null, null);
      } catch (e) {
        return this._createIllegalChannel(uri, loadInfo);
      }

      /* 未展開 */
      return this._createEmptyChannel(uri, loadInfo);
    }

    return this._createStringInputStreamChannel(uri, "utf-8",
                                                "text/plain",
                                                eFileInfo.source,
                                                eFileInfo.source.length,
                                                loadInfo);
  },

  /**
   * キャッシュ情報のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createCacheChannel: function(uri, loadInfo) {
    let data = lazy.UnMHTCache.createDocument();

    return this._createStringInputStreamChannel(uri, "utf-8", "text/html",
                                                data, data.length,
                                                loadInfo);
  },

  /**
   * ファイルが無い情報のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createNotFoundChannel: function(uri, loadInfo) {
    let ETEXT = name => arDOMUtils.escapeEntity(TEXT(name));

    let message = arDOMUtils.escapeEntity(UnMHTScheme.deErrorURI(uri.spec));
    return this._createErrorChannel(uri, ETEXT("error.file_not_found"),
                                    message, true,
                                    loadInfo);
  },

  /**
   * 不正な URI 情報のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createIllegalChannel: function(uri, loadInfo) {
    let ETEXT = name => arDOMUtils.escapeEntity(TEXT(name));

    let message = arDOMUtils.escapeEntity(uri.spec);
    return this._createErrorChannel(uri, ETEXT("error.illegal_address"),
                                    message, false,
                                    loadInfo);
  },

  /**
   * エラーメッセージを表示するチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {string} title
   *          タイトル (HTML)
   * @param   {string} message
   *          表示するメッセージ (HTML)
   * @param   {boolean} showRetryButton
   *          リトライボタンを表示する
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createErrorChannel: function(uri, title, message, showRetryButton,
                                loadInfo) {
    let ETEXT = name => arDOMUtils.escapeEntity(TEXT(name));

    let data
      = "<!DOCTYPE html>" + "\n"
      + "<html>" + "\n"
      + "<head>" + "\n"
      + "  <meta charset=\"UTF-8\">" + "\n"
      + "  <title>" + ETEXT("error.title") + "</title>" + "\n"
      + "  <link rel=\"stylesheet\" type=\"text/css\" href=\"resource://unmht/css/error.css\">" + "\n"
      + "</head>" + "\n"
      + "<body>" + "\n"
      + "  <div id=\"errorPageContainer\">" + "\n"
      + "    <h1>" + ETEXT("error.title") + "</h1>" + "\n"
      + "    <h2>" + title + "</h2>" + "\n"
      + "    <div>" + "\n"
      + "      <div id=\"errorDesc\">" + "\n"
      + "        <p>" + message + "</p>" + "\n"
      + "      </div>" + "\n"
      + "    </div>" + "\n"
      + (showRetryButton
         ? ("    <button id=\"unmht-retry-button\" autocomplete=\"off\" autofocus=\"true\">"
            + ETEXT("error.retry") + "</button>" + "\n")
         : "")
      + "  </div>" + "\n"
      + "</body>" + "\n"
      + "</html>" + "\n";

    data = toUTF8(data);

    return this._createStringInputStreamChannel(uri, "utf-8", "text/html",
                                                data, data.length,
                                                loadInfo);
  },

  /**
   * 空のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createEmptyChannel: function(uri, loadInfo) {
    return this._createStringInputStreamChannel(uri, "utf-8", "text/html",
                                                "", 0,
                                                loadInfo);
  },

  /**
   * ロード中のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createLoadingChannel: function(uri, loadInfo) {
    let data
      = "<!DOCTYPE html>" + "\n"
      + "<html>" + "\n"
      + "<head>" + "\n"
      + "  <meta charset=\"UTF-8\">" + "\n"
      + "  <title>Loading...</title>" + "\n"
      + "  <link rel=\"stylesheet\" type=\"text/css\" href=\"resource://unmht/css/loading.css\">" + "\n"
      + "  </head>" + "\n"
      + "  <body>" + "\n"
      + "    <div id=\"unmht-loading-container\"><div id=\"unmht-loading\"></div></div>" + "\n"
      + "  </body>" + "\n"
      + "</html>" + "\n";

    return this._createStringInputStreamChannel(uri, "utf-8", "text/html",
                                                data, data.length,
                                                loadInfo);
  },

  /**
   * 文字列を返すチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {string} charset
   *          文字セット
   * @param   {string} type
   *          MIME-Type
   * @param   {string} data
   *          文字列
   * @param   {number} length
   *          文字列の長さ
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createStringInputStreamChannel: function(uri, charset, type,
                                            data, length,
                                            loadInfo) {
    let sstream
      = Cc["@mozilla.org/io/string-input-stream;1"]
      .createInstance(Ci.nsIStringInputStream);
    sstream.setData(data, length);

    return this._createInputStreamChannel(uri, charset, type,
                                          sstream, length,
                                          loadInfo);
  },

  /**
   * ストリームの内容を返すチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {string} charset
   *          文字セット
   * @param   {string} type
   *          MIME-Type
   * @param   {nsIInputStream} stream
   *          ストリーム
   * @param   {number} length
   *          ストリームの長さ
   * @param   {nsILoadInfo} loadInfo
   *          ロード開始時の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createInputStreamChannel: function(uri, charset, type, stream, length,
                                      loadInfo) {
    let inputStreamChannel
      = Cc["@mozilla.org/network/input-stream-channel;1"]
      .createInstance(Ci.nsIInputStreamChannel)
      .QueryInterface(Ci.nsIChannel);

    inputStreamChannel.setURI(uri);
    inputStreamChannel.contentCharset = charset;
    inputStreamChannel.contentType    = type;
    inputStreamChannel.contentStream  = stream;
    inputStreamChannel.contentLength  = length;
    if (loadInfo) {
      try {
        inputStreamChannel.loadInfo = loadInfo;
      } catch (e) {
      }
    }

    return inputStreamChannel;
  }
});

module.UnMHTProtocolHandler = UnMHTProtocolHandler;

})(this);
