"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTSaveDialog"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { arDialog, arDialogClass } = Cu.import("resource://unmht/modules/base/arDialog.jsm", {});
let { arPrototype } = Cu.import("resource://unmht/modules/base/arPrototype.jsm", {});
let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTSaveState } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});
let { UnMHTSaveProgress } = Cu.import("resource://unmht/modules/UnMHTSaveProgress.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

/**
 * 保存ダイアログクラス
 *
 * ウィンドウの引数 {Object}
 *   {
 *     saveState:       {number}
 *                      保存状態
 *     saveContext:     {UnMHTSaveContext}
 *                      対象の保存コンテキスト
 *   }
 *
 * 結果 {Promise.<IUnMHTSaveListener>}
 * @resolves イベントリスナ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialogClass
 * @implements arIDialog
 * @implements IUnMHTSaveProgressOwner
 */
function UnMHTSaveDialogClass(window) {
  /**
   * 保存コンテキスト
   * @type {UnMHTSaveContext}
   */
  this._saveContext = null;

  arDialogClass.call(this, window);
}
UnMHTSaveDialogClass.prototype = Object.freeze({
  /* ==== arIDialog ==== */

  /**
   * 破棄
   */
  destruct: function() {
    this._saveContext = null;
    try {
      arDialogClass.prototype.destruct.call(this);
    } catch (e) {
      /* モジュールのアンロード後 */
    }
  },

  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this.arg;

    let title = TEXT("save.title");
    switch (arg.saveState) {
      case UnMHTSaveState.CURRENT: {
        title += " " + TEXT("save.state.current");
        break;
      }
      case UnMHTSaveState.CURRENT_NOJS: {
        title += " " + TEXT("save.state.current")
          + " (" + TEXT("save.nojs") + ")";
        break;
      }
      case UnMHTSaveState.ORIGINAL: {
        title += " " + TEXT("save.state.original");
        break;
      }
    }

    let dialog = this.$("unmht_save_dialog");
    dialog.setAttribute("title", title);

    dialog.getButton("accept").disabled = true;

    this.resolve(new UnMHTSaveProgress(this._window, this));

    this._saveContext = arg.saveContext;

    dialog.getButton("cancel").focus();
  },

  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function(context) {
    if (!this._saveContext) {
      return true;
    }
    if (this._saveContext) {
      this._saveContext.cancel();
      this._saveContext = null;

      return false;
    }

    return true;
  },

  /**
   * ファイルをスキップボタンをクリックしたイベント
   */
  onSkipFileClick: function() {
    if (this._saveContext) {
      this._saveContext.skipFile();
    }
  },

  /* ==== IUnMHTSaveProgressOwner ==== */

  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(progress, context) {
    if (this._window) {
      this._window.close();
    }
  },

  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {UnMHTSaveErrorInfo} errorInfo
   *          エラー情報
   */
  onError: function(progress, context, errorInfo) {
    let dialog = this.$("unmht_save_dialog");

    dialog.getButton("accept").disabled = false;
    dialog.getButton("accept").focus();
    dialog.getButton("cancel").disabled = true;
    dialog.getButton("extra1").disabled = true;

    progress.showFail(null, [errorInfo]);
  },

  /**
   * 保存完了
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   */
  onSave: function(progress, context, failedList) {
    let dialog = this.$("unmht_save_dialog");

    dialog.getButton("accept").disabled = false;
    dialog.getButton("accept").focus();
    dialog.getButton("cancel").disabled = true;
    dialog.getButton("extra1").disabled = true;

    if (!progress.showFail(failedList, null)) {
      Task.spawn(function*(window) {
        yield arTaskUtils.wait(UnMHTPref.getPref("save.dialog.wait"));
        window.close();
      }.bind(null, this._window));
    }
  }
});
arPrototype.extend(UnMHTSaveDialogClass, arDialogClass);

/**
 * 保存ダイアログ
 * @type {arDialog}
 */
let UnMHTSaveDialog = new arDialog(UnMHTSaveDialogClass, {
  url: "chrome://unmht/content/save.xul",
  features: "chrome,titlebar,centerscreen",
  single: false,
  usePromise: true
});

module.UnMHTSaveDialog = UnMHTSaveDialog;

})(this);
