"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTSaveMulti",
                         "UnMHTSaveMultiStatus",
                         "UnMHTSaveMultiFormat",
                         "UnMHTSaveMultiError"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arPseudoID } = Cu.import("resource://unmht/modules/base/arPseudoID.jsm", {});
let { toUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});
let { UnMHTPref, UnMHTSaveMultiIndexStyleFileName } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTSave, UnMHTSaveResult } = Cu.import("resource://unmht/modules/UnMHTSave.jsm", {});
let { UnMHTSaveProgress } = Cu.import("resource://unmht/modules/UnMHTSaveProgress.jsm", {});
let { UnMHTSystemDirectory } = Cu.import("resource://unmht/modules/UnMHTSystemDirectory.jsm", {});

/**
 * 保存対象の状況
 *
 * @enum {number}
 */
const UnMHTSaveMultiStatus = Object.freeze({
  /** 未実行 */
  NA:      0x0001,
  /** 保存中 */
  CURRENT: 0x0002,
  /** 保存完了 */
  OK:      0x0003,
  /** 保存失敗 */
  NG:      0x0004
});

/**
 * 保存形式
 *
 * @enum {number}
 */
const UnMHTSaveMultiFormat = Object.freeze({
  /** 別々のファイル */
  SEPARATED:      0x0001,
  /** 単一のファイル */
  SINGLE:         0x0011,
  /** 単一のファイル (インデックスなし) */
  SINGLE_NOINDEX: 0x0012
});

/**
 * 保存エラー
 *
 * @enum {number}
 */
const UnMHTSaveMultiError = Object.freeze({
  /** ディレクトリが異常 */
  ERROR_ILLEGAL_DIRECTORY: 0x0001
});

/**
 * 保存ユーティリティ
 *
 * @class
 */
let UnMHTSaveMultiUtils = Object.freeze({
  /**
   * HTML かどうかを調べる
   *
   * @param   {nsIXULWindow} browser
   *          対象のブラウザ
   * @returns {boolean}
   *          HTML かどうか
   */
  isHTML: function(browser) {
    let targetDocument = browser.contentDocument;
    return (targetDocument.contentType.startsWith("text/") ||
            /\/x(ht)?ml(\+.+)$/.test(targetDocument.contentType));
  },

  /**
   * インデックスページの内容を生成する
   *
   * @param   {Array.<Tuple.<string,string,string>>} tabs
   *          タブ情報(Content-ID, タイトル, 説明)のリスト
   * @param   {Array.<string>} additionalURIs
   *          追加 URI のリスト
   * @param   {string} cssFilename
   *          CSS ファイルのファイル名
   * @returns {string}
   *          インデックスページの内容 (UTF-8)
   */
  createIndexDocument: function(tabs, additionalURIs, cssFilename) {
    let content
      = "<!DOCTYPE html>"
      + "<html>\n"
      + "<head>\n"
      + "<meta charset=\"UTF-8\">\n"
      + "<title>index</title>\n"
      + "<meta http-equiv=\"Content-Style-Type\" content=\"text/css\">"
      + "<link rel=\"stylesheet\" type=\"text/css\" href=\""
      + cssFilename
      + "\">"
      + "</head>\n"
      + "<body>\n"
      + "<ul class=\"unmht-tabs-list\">\n"
      + (tabs
         .map(([cid, title, desc]) =>
              "<li class=\"unmht-tabs-item\">"
              + "<a class=\"unmht-tabs\" href=\""
              + arDOMUtils.escapeEntity("cid:" + cid)
              + "\">"
              + arDOMUtils.escapeEntity(title)
              + "</a>"
              + (desc ? ("<div class=\"unmht-tabs-desc\">"
                         + arDOMUtils.escapeEntity(desc)
                         + "</div>") : "")
              + "</li>\n")
         .join(""))
      + (additionalURIs
         .map(uri =>
              "<li class=\"unmht-additional-item\">"
              + "<a class=\"unmht-additional\" href=\""
              + arDOMUtils.escapeEntity(uri)
              + "\">"
              + arDOMUtils.escapeEntity(uri)
              + "</a></li>\n")
         .join(""))
      + "</ul>\n"
      + "</body>\n"
      + "</html>\n";

    return toUTF8(content);
  },

  /**
   * インデックスページの CSS を生成する
   *
   * @param   {boolean} useCustom
   *          カスタマイズを使用するか
   * @returns {Promise.<string>}
   * @resolves インデックスページの CSS (UTF-8)
   */
  createIndexCSS: Task.async(function*(useCustom) {
    let content
        = ".unmht-tabs-desc {\n"
        + "  margin-left: 1em;\n"
        + "}\n";

    if (useCustom && UnMHTPref.getPref("save.multi.index.style")) {
      let filePath = yield UnMHTSystemDirectory.getSystemFilePath(UnMHTSaveMultiIndexStyleFileName);

      try {
        let data = yield arFileUtils.readUTF8(filePath);
        content += toUTF8(data);
      } catch (e) {
      }
    }

    return content;
  })
});

/**
 * 複数のタブの保存コンテキスト
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @param   {IUnMHTSaveMultiListener} listener
 *          イベントリスナ
 * @implements IUnMHTSaveProgressOwner
 */
function UnMHTSaveMultiContext(window, listener) {
  /**
   * 保蔵形式
   * @type {UnMHTSaveMultiFormat}
   */
  this.saveFormat = 0;

  /**
   * 保存先フォルダのパス
   * @type {string}
   */
  this.parentPath = null;

  /**
   * 保存状態
   * @type {UnMHTSaveState}
   */
  this.saveState = 0;

  /**
   * 保存先フォルダ
   * @type {string}
   */
  this.filenameBase = "";

  /**
   * 単一ファイルの場合の URL
   * @type {string}
   */
  this.location = "";

  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * イベントリスナ
   * @type {IUnMHTSaveMultiListener}
   */
  this._listener = listener;

  /**
   * 対象のドキュメントと ID とパラメータのリスト
   * @type {Array.<Tuple.<nsIDOMDocument,number,Object>>}
   */
  this._targets = [];

  /**
   * 追加 URI
   * @type {Array.<string>}
   */
  this._additionalURIs = [];

  /**
   * 同名のファイルがあった場合にどうするか
   *   "replace":  上書きする
   *   "rename":   名前を変える
   *   "ask":      ダイアログを表示する
   * @type {string}
   */
  this._dupAction = UnMHTPref.getPref("save.multi.dup.action");

  /**
   * 現在保存中のドキュメントの位置
   * @type {number}
   */
  this._currentIndex = 0;

  /**
   * 保存コンテキスト
   * @type {UnMHTSaveContext}
   */
  this._saveContext = null;

  /**
   * 失敗したファイル名
   * @type {Array.<string>}
   */
  this._failedList = [];

  /**
   * エラー情報のリスト
   * @type {Array.<UnMHTSaveErrorInfo>}
   */
  this._errorInfoList = [];

  /**
   * キャンセルされたか
   * @type {boolean}
   */
  this._isCanceled = false;

  Object.seal(this);
}
UnMHTSaveMultiContext.prototype = Object.freeze({
  /**
   * 破棄
   * start 後は自分で呼ぶ
   */
  destruct: function() {
    this._saveContext = null;
    this._failedList = null;
    this._errorInfoList = null;
  },

  /**
   * 保存対象を追加する
   *
   * @param   {nsIXULWindow} browsewr
   *          対象のブラウザ
   * @param   {number} id
   *          保存対象の ID
   * @param   {Object} param
   *          パラメータ
   *            {
   *              filenameBase: {string}
   *                            ファイル名の拡張子以外
   *              title:        {string}
   *                            インデックスページのタイトル
   *              desc:         {string}
   *                            インデックスページの説明
   *            }
   */
  addTarget: function(browser, id, param) {
    this._targets.push([browser, id, param]);
  },

  /**
   * ファイルを追加する
   *
   * @param   {string} url
   *          ファイルのアドレス
   */
  addFile: function(url) {
    this._additionalURIs.push(url);
  },

  /* ==== IUnMHTSaveProgressOwner ==== */

  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(progress, context) {
    if (this._saveContext == context) {
      this._saveContext = null;
    }
    if (this._isCanceled) {
      this._listener.onCancel(this);
      this.destruct();
    }
  },

  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {UnMHTSaveErrorInfo} errorInfo
   *          エラー情報
   */
  onError: function(progress, context, errorInfo) {
    let [, id] = this._targets[this._currentIndex];
    this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.NG);

    this._errorInfoList.push(errorInfo);

    this._saveContext = null;

    if (this.saveFormat == UnMHTSaveMultiFormat.SEPARATED) {
      this._saveNext();
    } else {
      this.destruct();
    }
  },

  /**
   * 保存完了
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   */
  onSave: function(progress, context, failedList) {
    this._saveContext = null;

    this._failedList = this._failedList.concat(failedList);

    if (this.saveFormat == UnMHTSaveMultiFormat.SEPARATED) {
      let [, id] = this._targets[this._currentIndex];
      this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.OK);

      this._saveNext();
    } else {
      this._listener.onSave(this, this._failedList, this._errorInfoList);
      this.destruct();
    }
  },

  /**
   * 保存を開始する
   */
  start: function() {
    if (this.saveFormat == UnMHTSaveMultiFormat.SEPARATED) {
      this._currentIndex = -1;
      this._saveNext();
    } else {
      this._currentIndex = 0;
      this._saveSingle();
    }
  },

  /**
   * 現在のタブをスキップする
   *
   * @param   {boolean} skipOnNoContext
   */
  skip: function(skipOnNoContext) {
    if (!skipOnNoContext && !this._saveContext) {
      return;
    }

    if (this._saveContext) {
      this._saveContext.cancel();
      this._saveContext = null;
    }

    if (this.saveFormat == UnMHTSaveMultiFormat.SEPARATED) {
      if (this._currentIndex < this._targets.length) {
        let [, id] = this._targets[this._currentIndex];
        this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.NG);
      }

      this._saveNext();
    } else {
      this._listener.onAborted(this);
      this.destruct();
    }
  },

  /**
   * 現在のファイルをスキップする
   */
  skipFile: function() {
    if (this._saveContext) {
      this._saveContext.skipFile();
    }
  },

  /**
   * 保存をキャンセルする
   */
  cancel: function() {
    this._isCanceled = true;

    if (this._saveContext) {
      this._saveContext.cancel();
      this._saveContext = null;
    } else {
      this._listener.onCancel(this);
      this.destruct();
    }
  },

  /* ==== private ==== */

  _reportError: function(result) {
    if (result == UnMHTSaveResult.CANCEL) {
      return;
    }
    if ((typeof result) == 'number') {
      let name = "";
      for (let n of Object.keys(UnMHTSaveResult)) {
        if (result == UnMHTSaveResult[n]) {
          name = "UnMHTSaveResult." + n;
        }
      }
      Cu.reportError("onFail: " + name + "(" + result.toString(16) + ")");
    } else {
      Cu.reportError(result + "\n" + result.stack);
    }
  },

  /**
   * 次のタブを保存する
   * 別々のファイル用
   */
  _saveNext: function() {
    Task.spawn(function*() {
      this._currentIndex += 1;

      if (this._currentIndex >= this._targets.length) {
        this._listener.onSave(this, this._failedList, this._errorInfoList);
        return;
      }

      let [browser, id, param] = this._targets[this._currentIndex];

      if (!UnMHTDocumentCategory.isAlive(browser)) {
        this.skip(true);
        return;
      }

      let [path, isAborted] = yield* this._getFilePathTask(param.filenameBase);
      if (!path) {
        if (isAborted) {
          this._listener.onAborted(this);
          this.destruct();
        } else {
          this.skip(true);
        }
        return;
      }

      let location;
      let title;

      try {
        location = browser.currentURI.spec;
        title = browser.contentTitle;
      } catch (e) {
        if (UnMHTDocumentCategory.isAlive(browser)) {
          Cu.reportError(e + "\n" + e.stack);
        }
        this.skip(true);
        return;
      }

      let saveContext = UnMHTSave.createContext(this._window,
                                                path, location,
                                                title,
                                                null, this.saveState);
      saveContext.setLisener(new UnMHTSaveProgress(this._window, this));


      let targetDocument = browser.contentDocument;

      let arg = {
        browser:        browser,
        targetDocument: targetDocument,
        saveState:      this.saveState,
        location:       location,
        isHTML:         true,
        isStartPart:    true
      };

      try {
        yield saveContext.addDocument(arg);
      } catch (e) {
        saveContext.destruct();
        this.skip(true);
        return;
      }

      this._saveContext = saveContext;

      this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.CURRENT);

      this._listener.onStart();
      yield this._saveContext.start();
    }.bind(this)).catch(this._reportError);
  },

  /**
   * 単一のファイルとして保存する
   */
  _saveSingle: function() {
    Task.spawn(function*() {
      let [path] = yield* this._getFilePathTask(this.filenameBase);
      if (!path) {
        this._listener.onAborted(this);
        this.destruct();
        return;
      }

      let saveContext = UnMHTSave.createContext(this._window,
                                                path, this.location,
                                                this.filenameBase,
                                                null, this.saveState);
      saveContext.setLisener(new UnMHTSaveProgress(this._window, this));

      let isStartPart = (this.saveFormat == UnMHTSaveMultiFormat.SINGLE_NOINDEX);

      let tabs = [];
      for (let i = 0, len = this._targets.length; i < len; i ++) {
        let [browser, id, param] = this._targets[i];

        if (!UnMHTDocumentCategory.isAlive(browser)) {
          this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.NG);
          continue;
        }

        let savedFrom;
        let isHTML;
        try {
          savedFrom = browser.currentURI.spec;
          isHTML = UnMHTSaveMultiUtils.isHTML(browser);
        } catch (e) {
          if (UnMHTDocumentCategory.isAlive(browser)) {
            Cu.reportError(e + "\n" + e.stack);
          }

          this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.NG);
          continue;
        }

        let targetDocument = browser.contentDocument;

        let arg = {
          browser:        param.browser,
          targetDocument: targetDocument,
          saveState:      this.saveState,
          location:       savedFrom,
          savedFrom:      savedFrom,
          isHTML:         isHTML,
          isStartPart:    isStartPart,
          needContentID:  true
        };

        let part = null;
        try {
          part = yield saveContext.addDocument(arg);
        } catch (e) {
          this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.NG);
          continue;
        }

        let cid = "";

        if (part.sParam) {
          if (!part.sParam.cid) {
            part.sParam.cid
              = "tab." + i + "." + arPseudoID.gen() + "@unmht.org";
            part.sParam.storeContentID = true;
          }

          cid = part.sParam.cid;
        } else {
          cid = part.contentID;
        }

        isStartPart = false;

        let title
            = ("title" in param)
          ? param.title
          : (browser.contentTitle || browser.currentURI.spec);

        let desc
          = ("desc" in param)
          ? param.desc
          : "";

        this._listener.onStatusChange(this, id, UnMHTSaveMultiStatus.OK);

        tabs.push([cid, title, desc]);
      }

      if (this.saveFormat == UnMHTSaveMultiFormat.SINGLE) {
        let n = Math.floor(Math.random() * 1000000);
        let indexFilename = "unmht-index-" + n + ".html";
        let cssFilename = "unmht-index-" + n + ".css";

        saveContext.addDataDocument(
          UnMHTSaveMultiUtils.createIndexDocument(tabs, this._additionalURIs,
                                                  cssFilename),
          indexFilename, "UTF-8", "text/html", true);

        let cssContent = yield UnMHTSaveMultiUtils.createIndexCSS(true);

        saveContext.addDataDocument(cssContent, cssFilename,
                                    "UTF-8", "text/css", false);

        saveContext.setTabs(true);
      }
      if (this.saveFormat == UnMHTSaveMultiFormat.SINGLE ||
          this.saveFormat == UnMHTSaveMultiFormat.SINGLE_NOINDEX) {
        saveContext.setTabList(tabs.map(([cid, title, desc]) => cid));
      }

      for (let uri of this._additionalURIs) {
        saveContext.addFile(null, null, uri, false);
      }

      this._saveContext = saveContext;
      this._listener.onStart();
      yield this._saveContext.start();
    }.bind(this)).catch(this._reportError);
  },

  /**
   * 保存先のファイルを返す
   *
   * @param   {string} filenameBase
   *          ファイル名の拡張子以外
   * @yields  {Promise.<*>}
   * @returns {Tuple.<string,boolean>}
   *          保存先のファイルのパスと中断されたか
   */
  _getFilePathTask: function*(filenameBase) {
    if (!this.parentPath) {
      this._listener.onError(this, UnMHTSaveMultiError.ILLEGAL_DIRECTORY);

      return [null, true];
    }

    let exists = yield OS.File.exists(this.parentPath);
    if (!exists) {
      try {
        yield OS.File.makeDir(this.parentPath);
      } catch (e) {
        this._listener.onError(this, UnMHTSaveMultiError.ILLEGAL_DIRECTORY);

        return [null, true];
      }
    }

    let targetFilePath = "";
    if (this._dupAction == "rename") {
      targetFilePath = yield UnMHTSave.getUniqueFilePath(this.parentPath,
                                                         filenameBase);
    } else {
      let suffix = "." + UnMHTFilename.defaultExtension;
      let options = { allowExists: true, noCreate: true };
      let result = yield arFileUtils.prepareFile(this.parentPath,
                                                 filenameBase,
                                                 suffix, options);
      targetFilePath = result.path;
      if (result.exists) {
        if (this._dupAction == "ask") {
          let oResult
            = yield UnMHTSave.showOverwriteConfirm(
              this._window, targetFilePath, filenameBase,
              this.saveFormat == UnMHTSaveMultiFormat.SEPARATED,
              this.saveFormat == UnMHTSaveMultiFormat.SEPARATED);

          if (oResult.action == "skip") {
            if (oResult.applyToAll) {
              this._dupAction = oResult.action;
            }
            return [null, false];
          } else if (oResult.action == "abort" || oResult.action == "no") {
            return [null, true];
          }

          if (oResult.applyToAll) {
            this._dupAction = oResult.action;
          }

          if (oResult.action == "rename") {
            targetFilePath = oResult.renamePath;
          }
        } else if (this._dupAction == "skip") {
          return [null, false];
        }
      }
    }

    return [targetFilePath, false];
  }
});

/**
 * 複数のタブを MHT で保存
 *
 * @class
 */
let UnMHTSaveMulti = Object.freeze({
  /**
   * 保存コンテキストを作成する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {IUnMHTSaveMultiListener} listener
   *          イベントリスナ
   * @returns {UnMHTSaveMultiContext}
   *          保存コンテキスト
   */
  createContext: function(window, listener) {
    return new UnMHTSaveMultiContext(window, listener);
  }
});

module.UnMHTSaveMulti = UnMHTSaveMulti;
module.UnMHTSaveMultiStatus = UnMHTSaveMultiStatus;
module.UnMHTSaveMultiFormat = UnMHTSaveMultiFormat;
module.UnMHTSaveMultiError = UnMHTSaveMultiError;

})(this);
