"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTSaveMultiDialog"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arDialog, arDialogClass } = Cu.import("resource://unmht/modules/base/arDialog.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arPrototype } = Cu.import("resource://unmht/modules/base/arPrototype.jsm", {});
let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTSave } = Cu.import("resource://unmht/modules/UnMHTSave.jsm", {});
let { UnMHTSaveStateUtils } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});
let { UnMHTSaveMulti, UnMHTSaveMultiStatus, UnMHTSaveMultiFormat, UnMHTSaveMultiError } = Cu.import("resource://unmht/modules/UnMHTSaveMulti.jsm", {});
let { UnMHTSaveProgress } = Cu.import("resource://unmht/modules/UnMHTSaveProgress.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arFilePicker",
                                  "resource://unmht/modules/base/arFilePicker.jsm");

/**
 * 複数のタブを MHT で保存ダイアログクラス
 *
 * ウィンドウの引数
 *   {
 *     targetWindow: {nsIDOMWindow}
 *                   現在のウィンドウ
 *   }
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialogClass
 * @implements arIDialog
 * @implements IUnMHTSaveMultiListener
 */
function UnMHTSaveMultiDialogClass(window) {
  /**
   * 現在のウィンドウ
   * @type {nsIWindow}
   */
  this._targetWindow = null;

  /**
   * 対象のブラウザ
   * @type {Array.<nsIXULElement>}
   */
  this._browsers = null;

  /**
   * ドラッグ開始した項目の選択の状態
   * @type {boolean}
   */
  this._lastDragType = false;

  /**
   * ドラッグ開始した項目
   * @type {nsIDOMXULElement}
   */
  this._lastDragNode = null;

  /**
   * ファイル名のフィールドが変更されたか
   * @type {boolean}
   */
  this._isFilenameChanged = false;

  /**
   * ファイル名のフィールドのフォーカス前の値
   * @type {string}
   */
  this._lastFilename = "";

  /**
   * 保存コンテキスト
   * @type {UnMHTSaveMultiContext}
   */
  this._savemultiContext = null;

  arDialogClass.call(this, window);
}
UnMHTSaveMultiDialogClass.prototype = Object.freeze({
  /* ---- UI イベント - 一般 ---- */

  /**
   * テキストフィールドでキーが押されたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTextFieldKeyPress: function(event) {
    if (event.keyCode == 13) {
      this.onDialogAccept();
    }
  },

  /**
   * 対象を変更したイベント
   */
  onTargetChange: function() {
    this._updateTargetList();
  },

  /**
   * すべてを選択/解除ボタンがクリックされたイベント
   */
  onCheckAllClick: function() {
    if (!this._browsers.length) {
      return;
    }

    let save = !this._isSave(0);
    this._browsers.forEach((browser, i) => this._changeSave(i, save));

    this._updateCount();
    this._updateSingleFilename();
  },

  /**
   * インデックスページのタイトルと説明を変更するのチェックを変更したイベント
   */
  onChangeTitleChange: function() {
    this._updateTargetListState();
  },

  /**
   * ファイルの URI を追加のチェックを変更したイベント
   */
  onAddFilesChange: function() {
    this.$("add_uris").hidden = !this.$("use_add_uris").checked;
  },

  /**
   * ファイル名のドロップダウンメニューを変更したイベント
   */
  onFilenameTypeChange: function() {
    this._updateFilenameBase();
  },

  /**
   * 保存形式を変更したイベント
   */
  onFormatChange: function() {
    this.$("changetitle").disabled
      = this.$("saveformat").value != "single";

    this.$("filename_sequential").disabled
      = this.$("saveformat").value != "separated";

    this.$("use_add_uris").disabled
      = this.$("saveformat").value != "single" &&
      this.$("saveformat").value != "single_noindex";

    if (this.$("saveformat").value == "separated") {
      this.$("use_add_uris").checked = false;
      this.onAddFilesChange();
    }

    this._updateFilenameBase();

    this._updateTargetList();
  },

  /**
   * ディレクトリ選択ボタンがクリックされたイベント
   */
  onBaseDirSelectClick: Task.async(function*() {
    let baseDir = arFileUtils.getFile(this.$("basedir").value);

    let params = {
      title:            "",
      useLastDir:       false,
      displayDirectory: baseDir,
      uri:              null
    };
    let [dirPath] = yield lazy.arFilePicker.getFolder(this._window, params);
    if (!dirPath) {
      return;
    }
    this.$("basedir").value = dirPath;
  }),

  /* ---- UI イベント - 保存制御 ---- */

  /**
   * ファイルをスキップボタンをクリックしたイベント
   */
  onSkipFileClick: function() {
    if (this._savemultiContext) {
      this._savemultiContext.skipFile();
    }
  },

  /**
   * タブをスキップボタンをクリックしたイベント
   */
  onSkipTabClick: function() {
    if (this._savemultiContext) {
      this._savemultiContext.skip(false);
    }
  },

  /**
   * 戻るボタンのイベント
   */
  onBackClick: function() {
    let dialog = this.$("unmht_savemulti_dialog");
    dialog.getButton("accept").disabled = false;
    dialog.getButton("cancel").label = TEXT("save.multi.cancel");

    this.$("status_box").selectedPanel = this.$("status_box_control");
  },

  /* ---- UI イベント - 対象リスト ---- */

  /**
   * マウスが出たイベント
   * 対象リストのドラッグを終了する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDialogMouseOut: function(event) {
    if (event.explicitOriginalTarget == this.dialog) {
      this._lastDragNode = null;
    }
  },

  /**
   * マウスのボタンを開放したイベント
   * 対象リストのドラッグを終了する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDialogMouseUp: function(event) {
    this._lastDragNode = null;
  },

  /**
   * 対象リストが変更されたイベント
   * タイトルの変更をチェックする
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTargetListChange: function(event) {
    if (this.$("filename").value == "title") {
      this._updateSingleFilename();
    }
  },

  /**
   * 対象リストでキーを押したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTargetListKeyPress: function(event) {
    if (event.key == " ") {
      let index = this.$("target_list").currentIndex;

      this._changeSaveNode(this.$("savecheck_" + index), !this._isSave(index));

      this._updateCount();
      this._updateSingleFilename();
    }
  },

  /**
   * 対象リストでマウスを動かしたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTargetListMouseMove: function(event) {
    let checkNode = this._getCheckNode(event);
    if (!checkNode) {
      return;
    }

    this.onTargetListCheckMouseMove(checkNode);
  },

  /**
   * 対象リストでマウスを押下したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTargetListMouseDown: function(event) {
    let checkNode = this._getCheckNode(event);
    if (!checkNode) {
      return;
    }

    this.onTargetListCheckMouseDown(checkNode);
  },

  /**
   * 対象リストで対象のチェック上でマウスを押下したイベント
   * チェックを変更する
   *
   * @param   {nsIDOMXULElement} checkNode
   *          対象の要素
   */
  onTargetListCheckMouseDown: function(checkNode) {
    let save = this._isSaveNode(checkNode);

    this._lastDragType = save;
    this._lastDragNode = checkNode;

    this._changeSaveNode(checkNode, !save);

    this._updateCount();
    this._updateSingleFilename();
  },

  /**
   * 対象リストで対象のチェック上でマウスを動かしたイベント
   * ドラッグ時にチェックを変更する
   *
   * @param   {nsIDOMXULElement} checkNode
   *          対象の要素
   */
  onTargetListCheckMouseMove: function(checkNode) {
    if (!this._lastDragNode) {
      return;
    }

    if (this._lastDragNode == checkNode) {
      return;
    }

    this._changeSaveNode(checkNode, !this._lastDragType);

    this._updateCount();
    this._updateSingleFilename();
  },

  /* ---- UI イベント - ファイル名フィールド ---- */

  /**
   * ファイル名のフィールドがフォーカスを得たイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onFilenameBaseFocus: function(event) {
    this._lastFilename = this.$("filename_base").value;
  },

  /**
   * ファイル名のフィールドがフォーカスを失なったイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onFilenameBaseBlur: function(event) {
    if (this._lastFilename != this.$("filename_base").value) {
      this._isFilenameChanged = true;
    }
  },

  /* ==== arIDialog ==== */

  /**
   * 破棄
   */
  destruct: function() {
    this._targetWindow = null;

    this._browsers = null;

    this._lastDragNode = null;

    this._savemultiContext = null;

    try {
      arDialogClass.prototype.destruct.call(this);
    } catch (e) {
      /* モジュールのアンロード後 */
    }
  },

  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this.arg;

    this._targetWindow = arg.targetWindow;

    this._browsers = [];

    let dialog = this.$("unmht_savemulti_dialog");
    dialog.getButton("accept").setAttribute("accesskey", "s");
    dialog.getButton("accept").label = TEXT("save.multi.save");
    dialog.getButton("cancel").label = TEXT("save.multi.cancel");

    let basedir = UnMHTPref.getPref("save.multi.dir");
    if (basedir) {
      this.$("basedir").value = basedir;
    } else {
      arFileUtils.getDefaultDownloadsDirectory().then(dirPath => {
        this.$("basedir").value = dirPath;
      });
    }

    this.$("target").value = "this";
    this.$("filename").value = UnMHTPref.getPref("save.filename");
    this.$("saveformat").value = UnMHTPref.getPref("save.multi.format");
    this.$("savestate").value = UnMHTPref.getPref("save.multi.state");
    this.$("use_add_uris").checked = false;

    this.onFormatChange();
    this.onAddFilesChange();
    this._updateSingleFilename();
  },

  /**
   * ダイアログを OK したイベント
   * 保存を開始する
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    try {
      this._startSave();
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }

    return false;
  },

  /**
   * ダイアログをキャンセルしたイベント
   * 保存中であればキャンセルする
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    try {
      if (this._savemultiContext) {
        this._savemultiContext.cancel();

        return false;
      }

      return true;
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }
    return false;
  },

  /* ==== IUnMHTSaveMultiListener ==== */

  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(context) {
    this._savemultiContext = null;

    let dialog = this.$("unmht_savemulti_dialog");
    dialog.getButton("accept").disabled = false;

    this.$("status_box").selectedPanel = this.$("status_box_control");
  },

  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          対象の保存コンテキスト
   * @param   {number} reason
   *          エラーの理由
   */
  onError: function(context, reason) {
    switch (reason) {
      case UnMHTSaveMultiError.ILLEGAL_DIRECTORY: {
        this.failedLabel1.value = TEXT("save.error.illegal_directory");
        this.failedLabel1.hidden = false;
        break;
      }
    }
    this.$("skip").disabled = true;
    this._savemultiContext = null;
  },

  /**
   * 実際に保存が開始したイベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          対象の保存コンテキスト
   */
  onStart: function(context) {
    this.$("skip").disabled = false;
  },

  /**
   * 保存状況が変わったイベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          保存コンテキスト
   * @param   {number} id
   *          保存対象の ID
   * @param   {UnMHTSaveMultiStatus} status
   *          状況
   */
  onStatusChange: function(context, id, status) {
    switch (status) {
      case UnMHTSaveMultiStatus.NA: {
        this._changeStatus(id, "");
        break;
      }
      case UnMHTSaveMultiStatus.OK: {
        this._changeStatus(id, "o");
        break;
      }
      case UnMHTSaveMultiStatus.CURRENT: {
        this._changeStatus(id, "current");
        break;
      }
      case UnMHTSaveMultiStatus.NG: {
        this._changeStatus(id, "x");
        break;
      }
    }
  },

  /**
   * 全ファイル保存完了イベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   * @param   {Array.<UnMHTSaveErrorInfo>} errorInfoList
   *          エラー情報のリスト
   */
  onSave: function(context, failedList, errorInfoList) {
    this.$("saved_label").value = TEXT("save.done");
    let isErrorShown = this._showResult(failedList, errorInfoList);
    this._savemultiContext = null;

    this.$("status_box").selectedPanel = this.$("status_box_result");

    if (!isErrorShown && UnMHTPref.getPref("save.multi.tabs.autoclose")) {
      this._browsers.forEach((browser, i) => {
        if (!this._getStatus(i)) {
          return;
        }

        for (let window of this._getTargetWindows()) {
          arWindowUtils.closeTabForBrowser(window, browser);
        }
      });
    }

    if (!isErrorShown &&
        UnMHTPref.getPref("save.multi.autoclose")) {
      Task.spawn(function*(window) {
        yield arTaskUtils.wait(UnMHTPref.getPref("save.dialog.wait"));
        window.close();
      }.bind(null, this._window));
    }
  },

  /**
   * 保存中断イベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          保存コンテキスト
   */
  onAborted: function(context) {
    this.$("saved_label").value = TEXT("save.aborted");
    this._showResult(null, null);
    this._savemultiContext = null;

    this.$("status_box").selectedPanel = this.$("status_box_result");
  },

  /* ==== private ==== */

  /* ---- 保存関係メソッド ---- */

  /**
   * 保存を開始する
   *
   * @returns {boolean}
   *          開始したか
   */
  _startSave: function() {
    if (!this._browsers
        .some((browser, i) => (this._isSave(i) &&
                               UnMHTDocumentCategory.isAlive(browser)))) {
      /* 保存できるものが 1 つも無いので中断 */
      this._browsers
        .filter(browser => !UnMHTDocumentCategory.isAlive(browser))
        .forEach((browser, i) => this._changeStatus(i, "x"));

      return false;
    }

    UnMHTPref.setPref("save.multi.dir",    this.$("basedir").value);
    UnMHTPref.setPref("save.multi.format", this.$("saveformat").value);
    if (UnMHTPref.getPref("save.multi.state.remember")) {
      UnMHTPref.setPref("save.multi.state",  this.$("savestate").value);
    }

    this.$("status_box").selectedPanel = this.$("status_box_progress");

    if (this.$("saveformat").value == "single" ||
        this.$("saveformat").value == "single_noindex") {
      if (this.$("filename_base").value === "") {
        this.$("status_box").selectedPanel = this.$("status_box_control");
        this.$("filename_base").focus();

        return false;
      }
    }

    let dialog = this.$("unmht_savemulti_dialog");
    dialog.getButton("accept").disabled = true;

    let filenameBase
      = UnMHTFilename.validateFileName(this.$("filename_base").value);

    let savemultiContext
      = UnMHTSaveMulti.createContext(this._window, this);

    let isSingle = false;
    switch (this.$("saveformat").value) {
      case "separated": {
        savemultiContext.saveFormat = UnMHTSaveMultiFormat.SEPARATED;
        break;
      }
      case "single": {
        savemultiContext.saveFormat = UnMHTSaveMultiFormat.SINGLE;
        isSingle = true;
        break;
      }
      case "single_noindex": {
        savemultiContext.saveFormat = UnMHTSaveMultiFormat.SINGLE_NOINDEX;
        isSingle = true;
        break;
      }
    }

    savemultiContext.parentPath = this.$("basedir").value;
    savemultiContext.saveState
      = UnMHTSaveStateUtils.stringToSaveState(this.$("savestate").value);
    savemultiContext.filenameBase = filenameBase;

    if (isSingle) {
      for (let uri of UnMHTSave.parseAdditionalURIs(this.$("add_uris").value)) {
        savemultiContext.addFile(uri);
      }
    }

    let isChangeTitle
      = (savemultiContext.saveFormat == UnMHTSaveMultiFormat.SINGLE &&
         this.$("changetitle").checked);

    let first = true;
    this._browsers.forEach((browser, i) => {
      if (!this._isSave(i)) {
        return;
      }

      if (!UnMHTDocumentCategory.isAlive(browser)) {
        return;
      }

      let param = {};

      if (isSingle) {
        if (first) {
          first = false;
          savemultiContext.location = browser.currentURI.spec;
        }
      } else {
        if (this.$("filename").value == "leafname" ||
            this.$("filename").value == "title" ||
            this.$("filename").value == "custom") {
          param.filenameBase = this._getFilenameBase(browser);
        } else if (this.$("filename").value == "sequential") {
          let pad = (width, n) => {
            let s = n.toString(10);
            return "0".repeat(Math.max(width - s.length, 0)) + s;
          };
          let digits = UnMHTPref.getPref("save.filename.sequential.digits");
          param.filenameBase = filenameBase + "-" + pad(digits, i + 1);
        }
      }

      if (isChangeTitle) {
        param.title = this._getTitle(i);
        param.desc = this._getDesc(i);
      }

      savemultiContext.addTarget(browser, i, param);
    });

    this._savemultiContext = savemultiContext;

    this._savemultiContext.start();

    return true;
  },

  /**
   * 保存結果を表示する
   *
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   * @param   {Array.<UnMHTSaveErrorInfo>} errorInfoList
   *          エラー情報のリスト
   * @returns {boolean}
   *          エラーが表示されたか
   */
  _showResult: function(failedList, errorInfoList) {
    let dialog = this.$("unmht_savemulti_dialog");
    dialog.getButton("accept").disabled = true;
    dialog.getButton("cancel").label = TEXT("save.multi.close");
    dialog.getButton("cancel").focus ();

    let progress = new UnMHTSaveProgress(this._window, this);
    let isErrorShown = progress.showFail(failedList, errorInfoList);
    progress.destruct();

    return isErrorShown;
  },

  /**
   * 保存する際のファイル名を取得する
   *
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @returns {string}
   *          ファイル名 (拡張子以外)
   */
  _getFilenameBase: function(browser) {
    let targetDocument = browser.contentDocument;
    return UnMHTSave.getFilenameBase(targetDocument, null,
                                     this.$("filename").value, new Date());
  },

  /* ---- 対象リスト関連メソッド ---- */

  /**
   * 対象リストの選択している数を更新する
   */
  _updateCount: function() {
    this.$("count_label").value
      = this._browsers.filter((browser, i) => this._isSave(i)).length
      + "/" + this._browsers.length;
  },

  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {boolean}
   *          対象のタブが保存対象かどうか
   */
  _isSave: function(index) {
    return this._isSaveNode(this.$("savecheck_" + index));
  },

  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @returns {boolean}
   *          対象のタブが保存対象かどうか
   */
  _isSaveNode: function(node) {
    return node.getAttribute("value") == "save";
  },

  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @param   {boolean} save
   *          対象のタブが保存対象かどうか
   */
  _changeSave: function(index, save) {
    this._changeSaveNode(this.$("savecheck_" + index), save);
  },

  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @param   {boolean} save
   *          対象のタブが保存対象かどうか
   */
  _changeSaveNode: function(node, save) {
    node.setAttribute("value", save ? "save" : "ignore");
    node.setAttribute("src",
                      save
                      ? "resource://unmht/images/check_v.png"
                      : "resource://unmht/images/check_dot.png");
  },

  /**
   * 対象のタブの保存状況を取得する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {boolean} status
   *          対象のタブの保存に成功したか
   */
  _getStatus: function(index) {
    return this._getStatusNode(this.$("savestatus_" + index));
  },

  /**
   * 対象のタブの保存状況を取得する
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @returns {boolean} status
   *          対象のタブの保存に成功したか
   */
  _getStatusNode: function(node) {
    return node.getAttribute("value") == "o";
  },

  /**
   * 対象のタブの保存状況を変更する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @param   {string} status
   *          対象のタブの保存状況
   */
  _changeStatus: function(index, status) {
    return this._changeStatusNode(this.$("savestatus_" + index), status);
  },

  /**
   * 対象のタブの保存状況を変更する
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @param   {string} status
   *          対象のタブの保存状況
   */
  _changeStatusNode: function(node, status) {
    node.setAttribute("src",
                      status ?
                      ("resource://unmht/images/check_" + status + ".png")
                      : "");
    node.setAttribute("value", status);
  },

  /**
   * イベントの対象のチェックボックスを取得する
   *
   * @param   {Event} event
   *          対象のイベント
   * @returns {nsIDOMXULElement}
   *          イベントの対象のチェックボックス
   */
  _getCheckNode: function(event) {
    let row = {}, col = {}, child = {};

    this.$("target_list").treeBoxObject
      .getCellAt(event.clientX, event.clientY, row, col, child);

    if (row.value == -1) {
      return null;
    }

    if (!col.value) {
      return null;
    }

    if (col.value.index !== 0) {
      return null;
    }

    let listbox = this.$("target_list_children");

    return listbox.childNodes[row.value].firstChild.firstChild;
  },

  /**
   * 対象リストを空にする
   */
  _clearTargetList: function() {
    let listbox = this.$("target_list_children");
    while (listbox.firstChild) {
      listbox.removeChild(listbox.firstChild);
    }
  },

  /**
   * 対象リストのアイテムを追加する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @param   {string} title
   *          タブのタイトル
   */
  _addTargetListItem: function(index, title) {
    let listbox = this.$("target_list_children");

    let listitem = this._window.document.createElement("treeitem");
    listbox.appendChild(listitem);

    let listrow = this._window.document.createElement("treerow");
    listitem.appendChild(listrow);

    let checkNode = this._window.document.createElement("treecell");
    listrow.appendChild(checkNode);
    checkNode.id = "savecheck_" + index;
    checkNode.className = "save-iconic";
    checkNode.setAttribute("editable", "false");
    this._changeSaveNode(checkNode,
                         !UnMHTPref.getPref("save.multi.check_none"));

    let statusNode = this._window.document.createElement("treecell");
    listrow.appendChild(statusNode);
    statusNode.id = "savestatus_" + index;
    statusNode.setAttribute("editable", "false");
    statusNode.className = "save-iconic";

    let titleNode = this._window.document.createElement("treecell");
    listrow.appendChild(titleNode);
    titleNode.id = "savetitle_" + index;
    titleNode.setAttribute("editable", "true");
    titleNode.setAttribute("label", title);
    titleNode.setAttribute("value", title);
    titleNode.flex = 1;

    let descNode = this._window.document.createElement("treecell");
    listrow.appendChild(descNode);
    descNode.id = "savedesc_" + index;
    descNode.setAttribute("editable", "true");
    descNode.flex = 1;

    this._updateTargetListItemState(index);
  },

  /**
   * 対象リストのアイテムの状態を更新する
   *
   * @param   {number} index
   *          対象のアイテムのインデックス
   */
  _updateTargetListItemState: function(index) {
    let titleNode = this.$("savetitle_" + index);
    let descNode = this.$("savedesc_" + index);

    let isEditable
      = (this.$("saveformat").value == "single" &&
         this.$("changetitle").checked);

    if (!isEditable) {
      titleNode.setAttribute("label", titleNode.getAttribute("value"));
    }

    titleNode.setAttribute("editable", isEditable);
    descNode.setAttribute("editable", isEditable);
    descNode.hidden = !isEditable;
  },

  /**
   * 対象のドキュメントリストを更新する
   */
  _updateTargetList: function() {
    let asSingle
      = (this.$("saveformat").value == "single" ||
         this.$("saveformat").value == "single_noindex");

    this._browsers = [];
    for (let window of this._getTargetWindows()) {
      for (let browser of arWindowUtils.getBrowsers(window)) {
        if (UnMHTDocumentCategory.isSavable(null, browser, asSingle)) {
          this._browsers.push(browser);
        }
      }
    }

    this._clearTargetList();
    this._browsers.forEach((browser, i) => {
      this._addTargetListItem(i,
                              browser.contentTitle || browser.currentURI.spec);
    });

    this._updateTargetListState();
    this._updateCount();
  },

  /**
   * 対象リストの状態を更新する
   */
  _updateTargetListState: function() {
    this._browsers.forEach((browser, i) => this._updateTargetListItemState(i));

    this.$("target_list_desc").hidden
      = !(this.$("saveformat").value == "single" &&
          this.$("changetitle").checked);

    this._updateSingleFilename();
  },

  /**
   * 対象になるウィンドウのリストを返す
   *
   * @returns {Array.<nsIDOMWindow>}
   */
  _getTargetWindows: function() {
    if (this.$("target").value == "this") {
      return [this._targetWindow];
    } else {
      return arWindowUtils.getBrowserWindows();
    }
  },

  /**
   * 対象のタブのタイトルを取得する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {string}
   *          対象のタブのタイトル
   */
  _getTitle: function(index) {
    return this.$("savetitle_" + index).getAttribute("label");
  },

  /**
   * 対象のタブのタイトルを説明する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {string}
   *          対象のタブの説明
   */
  _getDesc: function(index) {
    return this.$("savedesc_" + index).getAttribute("label");
  },

  /* ---- ファイル名フィールド関連メソッド ---- */

  /**
   * 必要であれば単一のファイルのファイル名を更新する
   */
  _updateSingleFilename: function() {
    if (!(this.$("saveformat").value == "single" ||
          this.$("saveformat").value == "single_noindex")) {
      return;
    }

    if (this._isFilenameChanged && this.$("filename_base").value !== "") {
      return;
    }
    let [browser, i]
        = (this._browsers
           .map((browser, i) => [browser, i])
           .find(([browser, i]) => (this._isSave(i) &&
                                    UnMHTDocumentCategory.isAlive(browser))) ||
           [null, -1]);
    if (!browser) {
      return;
    }

    let filenameBase = "";
    if (this.$("changetitle").checked &&
        this.$("filename").value == "title") {
      filenameBase = this._getTitle(i);
    } else {
      filenameBase = this._getFilenameBase(browser);
    }

    this.$("filename_base").value = filenameBase;
    this._lastFilename = this.$("filename_base").value;
    this._isFilenameChanged = false;
  },

  /**
   * ファイル名フィールドの状態を更新する
   */
  _updateFilenameBase: function() {
    this.$("filename_base").disabled
      = !(this.$("filename").value == "sequential" ||
          this.$("saveformat").value == "single" ||
          this.$("saveformat").value == "single_noindex");

    if (this.$("filename").value == "sequential") {
      this.$("filename_base").setAttribute("placeholder",
                                            TEXT("save.multi.sequential.placeholder"));
      this.$("filename_base").value = "";
      this._lastFilename = this.$("filename_base").value;
      this._isFilenameChanged = false;
    } else {
      this.$("filename_base").setAttribute("placeholder", "");
    }

    this._updateSingleFilename();
  }
});
arPrototype.extend(UnMHTSaveMultiDialogClass, arDialogClass);

/**
 * 複数のタブを MHT で保存ダイアログ
 * @type {arDialog}
 */
let UnMHTSaveMultiDialog = new arDialog(UnMHTSaveMultiDialogClass, {
  url: "chrome://unmht/content/savemulti.xul",
  features: "chrome,titlebar,centerscreen,resizable",
  single: true,
  usePromise: false
});

module.UnMHTSaveMultiDialog = UnMHTSaveMultiDialog;

})(this);
