"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTSaveOptionsDialog"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arDialog, arDialogClass } = Cu.import("resource://unmht/modules/base/arDialog.jsm", {});
let { arPrototype } = Cu.import("resource://unmht/modules/base/arPrototype.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTSave, UnMHTSaveResult } = Cu.import("resource://unmht/modules/UnMHTSave.jsm", {});
let { UnMHTSaveStateUtils } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});

/**
 * 保存オプションクラス
 *
 * ウィンドウの引数 {Object}
 *   {
 *     showFilename:   {boolean}
 *                     ファイル名を表示し、修正可能にするか
 *     filenameBase:   {string}
 *                     ファイル名 (拡張子以外)
 *
 *     saveStateMask:  {number}
 *                     リスト表示する保存状態
 *     saveState:      {number}
 *                     デフォルトの保存状態
 *     saveStatePref:  {string}
 *                     選択された保存状態を格納する設定名
 *     showInsertLink: {boolean}
 *                     元ページへのリンクを挿入を表示し、修正可能にするか
 *   }
 *
 * 結果 {Promise.<Object>}
 * @resolves {
 *     filenameBase:   {string}
 *                     ファイル名のベース
 *     saveState:      {number}
 *                     保存状態
 *     additionalURIs: {Array.<string>}
 *                     追加の URI
 *   }
 * @rejects {UnMHTSaveResult}
 *            UnMHTSaveResult.CANCEL
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialogClass
 * @implements arIDialog
 */
function UnMHTSaveOptionsDialogClass(window) {
  /**
   * ファイル名を表示し、修正可能にするか
   * @type {boolean}
   */
  this._showFilename = false;

  /**
   * ファイル名 (拡張子以外)
   * @type {string}
   */
  this._filenameBase = "";

  /**
   * 選択された保存状態を格納する設定名
   * @type {string}
   */
  this._saveStatePref = "";

  /**
   * 元ページへのリンクを挿入を表示し、修正可能にするか
   * @type {boolean}
   */
  this._showInsertLink = false;

  arDialogClass.call(this, window);
}
UnMHTSaveOptionsDialogClass.prototype = Object.freeze({
  /* ==== arIDialog ==== */

  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this.arg;

    this._showFilename = arg.showFilename;
    this._filenameBase = arg.filenameBase;
    if (this._showFilename) {
      this.$("filename").value
        = this._filenameBase + "." + UnMHTFilename.defaultExtension;
      this.$("filename_box").hidden = false;
    }

    for (let state of UnMHTSaveStateUtils.getList(arg.saveStateMask)) {
      this.$("savestate_" + UnMHTSaveStateUtils.saveStateToString(state)).hidden = false;
    }
    this.$("savestate").value
      = UnMHTSaveStateUtils.saveStateToString(arg.saveState);

    this._saveStatePref = arg.saveStatePref;

    this._showInsertLink = arg.showInsertLink;
    if (this._showInsertLink) {
      this.$("save_insertlink").checked
        = UnMHTPref.getPref("save.insert.link");
    } else {
      this.$("save_insertlink").hidden = true;
    }
  },

  /**
   * ダイアログを OK したイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    try {
      if (UnMHTPref.getPref(this._saveStatePref + ".remember")) {
        UnMHTPref.setPref(this._saveStatePref, this.$("savestate").value);
      }
      if (this._showInsertLink) {
        UnMHTPref.setPref("save.insert.link",
                          this.$("save_insertlink").checked);
      }

      let resultParam = {
        filenameBase:   (this._showFilename
                         ? this.$("filename").value.replace(/\.mht/i, "")
                         : this._filenameBase),
        saveState:      UnMHTSaveStateUtils.stringToSaveState(this.$("savestate").value),
        additionalURIs: (UnMHTSave.parseAdditionalURIs(this.$("add_uris").value))
      };

      this.resolve(resultParam);

      return true;
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }
    return false;
  },

  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    try {
      this.reject(UnMHTSaveResult.CANCEL);

      return true;
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }
    return false;
  }
});
arPrototype.extend(UnMHTSaveOptionsDialogClass, arDialogClass);

/**
 * 保存オプション
 * @type {arDialog}
 */
let UnMHTSaveOptionsDialog = new arDialog(UnMHTSaveOptionsDialogClass, {
  url: "chrome://unmht/content/saveoptions.xul",
  features: "chrome,titlebar,centerscreen,resizable",
  single: false,
  usePromise: true
});

module.UnMHTSaveOptionsDialog = UnMHTSaveOptionsDialog;

})(this);
