"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTSaveProgress"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

/**
 * 保存の進捗状況表示
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @param   {IUnMHTSaveProgressOwner} owner
 *          オーナー
 * @implements IUnMHTSaveListener
 */
function UnMHTSaveProgress(window, owner) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * オーナー
   * @type {IUnMHTSaveProgressOwner}
   */
  this._owner = owner;

  let document = window.document;
  let $ = id => document.getElementById(id);

  /**
   * 取得状況のラベル
   * @type {nsIDOMXULElement}
   */
  this._progressLabel = $("progress_label");
  this._progressLabel.value = TEXT("save.begin");

  /**
   * 現在取得しているファイル
   * @type {nsIDOMXULElement}
   */
  this._progressCurrent = $("progress_current");

  /**
   * 現在取得しているファイルのサイズ
   * @type {nsIDOMXULElement}
   */
  this._progressCurrentSize = $("progress_current_size");

  /**
   * 取得状況のバー
   * @type {nsIDOMXULElement}
   */
  this._progress = $("progress");
  this._progress.value = 0;

  /**
   * 失敗したファイル
   * @type {nsIDOMXULElement}
   */
  this._failedList = $("failed");
  this._failedList.hidden = true;

  Object.seal(this);
}
UnMHTSaveProgress.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;

    this._progressLabel = null;
    this._progressCurrent = null;
    this._progressCurrentSize = null;
    this._progress = null;

    this._failedList = null;

    this._owner = null;
  },

  /* ==== IUnMHTSaveListener ==== */

  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(context) {
    this._owner.onCancel(this, context);
  },

  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {UnMHTSaveErrorInfo} errorInfo
   *          エラー情報
   */
  onError: function(context, errorInfo) {
    let message = errorInfo.message;
    if (message) {
      this._progressLabel.value = message;
    }

    if (errorInfo.e) {
      this._failedList.value = errorInfo.e.toString();
      if (errorInfo.e.stack) {
        this._failedList.value += "\n\n" + errorInfo.e.stack;
      }
      this._failedList.hidden = false;

      this._window.sizeToContent();
    }
    this._progress.mode = "determined";
    this._progress.value = 0;

    this._owner.onError(this, context, errorInfo);
  },

  /**
   * 進捗状況変化イベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {string} currentFileName
   *          現在取得のファイル名
   * @param   {string} currentSize
   *          取得したバイト数 / ファイルサイズ
   * @param   {number} ok
   *          取得完了した数
   * @param   {number} na
   *          取得中の数
   * @param   {number} ng
   *          取得失敗した数
   */
  onProgress: function(context, currentFileName, currentSize, ok, na, ng) {
    this._progressCurrent.value = currentFileName;
    this._progressCurrentSize.value = currentSize;

    this._progressLabel.value
      = TEXT("save.begin") + " " + (ok + ng) + "/" + (ok + na + ng);

    if (context.progress == 1) {
      this._progress.mode = "undetermined";
    } else {
      this._progress.mode = "determined";
      this._progress.value = Math.floor(context.progress * 100);
    }
  },

  /**
   * 保存完了
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   */
  onSave: function(context, failedList) {
    this._progress.mode = "determined";
    this._progress.value = 100;

    this._progressLabel.value = TEXT("save.done");

    this._owner.onSave(this, context, failedList);
  },

  /**
   * 保存コンテキストの破棄
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onDestruct: function(context) {
    this.destruct();
  },

  /**
   * 必要であればエラーを表示する
   *
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   * @param   {Array.<UnMHTSaveErrorInfo>} errorInfoList
   *          エラー情報のリスト
   * @returns {boolean}
   *          エラーが表示されたか
   */
  showFail: function(failedList, errorInfoList) {
    let texts = [];

    if (errorInfoList) {
      for (let errorInfo of errorInfoList) {
        if (!errorInfo.errorList ||
            errorInfo.errorList.length === 0) {
          continue;
        }
        if (texts.length > 0) {
          texts.push("");
        }
        let message = errorInfo.message;
        if (message) {
          texts.push(message);
        }
        texts = texts.concat(errorInfo.errorList
                             .map(([url, reason]) => reason + ": " + url));
        if (errorInfo.e) {
          texts.push(errorInfo.e.toString());
        }
      }
    }

    if (UnMHTPref.getPref("save.show_errors") &&
        failedList && failedList.length > 0) {
      if (texts.length > 0) {
        texts.push("");
      }
      texts.push(TEXT("save.error.failed_files"));
      texts = texts.concat(failedList);
    }

    if (texts.length > 0) {
      this._failedList.value = texts.join("\n");
      this._failedList.hidden = false;

      this._window.sizeToContent();

      return true;
    }

    return false;
  }
});

module.UnMHTSaveProgress = UnMHTSaveProgress;

})(this);
