"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTSaveState",
                         "UnMHTSaveStateMask",
                         "UnMHTSaveStateUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

/**
 * 保存状態
 * @enum {number}
 */
const UnMHTSaveState = Object.freeze({
  /** 元のファイル */
  ORIGINAL:       0x0001,
  /** 現在の状態 */
  CURRENT:        0x0010,
  /** 現在の状態 (スクリプト削除) */
  CURRENT_NOJS:   0x0020,
  /** 不明 */
  UNKNOWN:        0x0000
});

/**
 * 保存状態のマスク
 * @enum {number}
 */
const UnMHTSaveStateMask = Object.freeze({
  /** 全て */
  ALL:       0xffff,
  /** 元のファイル */
  ORIGINAL:  0x000f,
  /** 現在の状態 */
  CURRENT:   0x00f0
});

/**
 * 保存状態ユーティリティ
 *
 * @class
 */
let UnMHTSaveStateUtils = Object.freeze({
  /**
   * 保存状態の数値表現と文字列表現のリスト
   * @type {Array.<Tuple.<number,string>>}
   */
  saveStateList: [
    [UnMHTSaveState.CURRENT,      "current"],
    [UnMHTSaveState.CURRENT_NOJS, "curnojs"],
    [UnMHTSaveState.ORIGINAL,     "original"]
  ],

  /**
   * 保存状態を文字列表現から数値表現に変換する
   *
   * @param   {string} _str
   *          文字列表現
   * @returns {UnMHTSaveState}
   *          数値表現
   */
  stringToSaveState: function(_str) {
    return (this.saveStateList
            .find(([state, str]) => str == _str) ||
            [UnMHTSaveState.UNKNOWN, "unknown"])[0];
  },

  /**
   * 保存状態を数値表現から文字列表現に変換する
   *
   * @param   {UnMHTSaveState} _state
   *          数値表現
   * @returns {string}
   *          文字列表現
   */
  saveStateToString: function(_state) {
    return (this.saveStateList
            .find(([state, str]) => state == _state) ||
            [UnMHTSaveState.UNKNOWN, "unknown"])[1];
  },

  /**
   * マスクでフィルタした保存状態の数値表現のリストを返す
   *
   * @param   {UnMHTSaveStateMask=} mask
   *          保存状態のマスク
   * @returns {Array.<number>}
   *          保存状態の数値表現のリスト
   */
  getList: function(mask=UnMHTSaveStateMask.ALL) {
    return this.saveStateList
      .filter(([state, str]) => state & mask)
      .map(([state, str]) => state);
  },

  /**
   * 保存状態の数値表現と追加ファイルの有無から
   * ファイルタイプの文字列を作成する
   *
   * @param   {UnMHTSaveState} saveState
   *          保存状態の数値表現
   * @param   {boolean} useAdd
   *          追加ファイルの有無
   * @returns {string}
   *          ファイルタイプ
   */
  makeFiletype: function(saveState, useAdd) {
    let filetype
      = TEXT("save.filetype") + " " + this.saveStateToReadableString(saveState);
    if (useAdd) {
      filetype += " + " + TEXT("save.add");
    }

    return filetype;
  },

  /**
   * 保存状態の数値表現を可読な文字列に変換する
   *
   * @param   {UnMHTSaveState} saveState
   *          保存状態の数値表現
   * @returns {string}
   *          可読な文字列
   */
  saveStateToReadableString: function(saveState) {
    switch (saveState) {
      case UnMHTSaveState.CURRENT: {
        return TEXT("save.state.current");
      }
      case UnMHTSaveState.CURRENT_NOJS: {
        return TEXT("save.state.current")
          + " (" + TEXT("save.nojs") + ")";
      }
      case UnMHTSaveState.ORIGINAL: {
        return TEXT("save.state.original");
      }
    }

    return "";
  }
});

module.UnMHTSaveState = UnMHTSaveState;
module.UnMHTSaveStateMask = UnMHTSaveStateMask;
module.UnMHTSaveStateUtils = UnMHTSaveStateUtils;

})(this);
