"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTScheme"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyServiceGetter(lazy, "IDNService",
                                   "@mozilla.org/network/idn-service;1",
                                   "nsIIDNService");

/**
 * URI の情報
 *
 * @constructor
 */
function UnMHTURIParam() {
  /**
   * 本来の URI
   * @type {string}
   */
  this.original = "";

  /**
   * 本来の URI の scheme
   * @type {string}
   */
  this.scheme = "";

  /**
   * ホスト名
   * @type {string}
   */
  this.host = "";

  /**
   * ポート番号
   * @type {string}
   */
  this.port = "";

  /**
   * パス
   * @type {string}
   */
  this.path = "";

  /**
   * 本来の URI のフラグメント
   * @type {string}
   */
  this.fragment = "";

  /**
   * 参照名
   * @type {string}
   */
  this.refName = "";

  /**
   * 開始パートの URI
   * @type {string}
   */
  this.baseURI = "";

  Object.seal(this);
}
UnMHTURIParam.prototype = Object.freeze({
});

/**
 * unmht: スキーム操作
 *
 * @class
 */
let UnMHTScheme = Object.freeze({
  /**
   * unmht: スキームの URI にする
   *
   * @param   {string} uri
   *          URI
   * @returns {string}
   *          unmht: スキームの URI
   */
  enUnMHTURI: function(uri) {
    let m = uri.match(/^([A-Za-z0-9\-]+):(\/\/)?([^\/]*)(\/\/)?(\/)?(.*)$/);
    if (m) {
      let [, scheme, sep1, host, sep2, sep3, path] = m;

      let sep = 0;
      if (sep1) {
        sep |= 1;
      }
      if (sep2) {
        sep |= 2;
      }
      if (sep3) {
        sep |= 4;
      }

      if (scheme.match(/^data$/i) && sep == 4) {
        path = host + "/" + path;
        host = "";
        sep = 0;
      }

      let port = "";
      m = host.match(/^([^:]+):([0-9]+)$/);
      if (m) {
        [, host, port] = m;
      }

      uri
        = "unmht://" + host + (host ? "." : "") + "unmht"
        + (port ? (":" + port) : "")
        + "/"
        + scheme + "." + sep + "/" + path + "/";
    }

    return uri;
  },

  /**
   * unmht: スキームの URI かどうかを返す
   *
   * @param   {string} uri
   *          URI
   * @returns {boolean}
   *          unmht: スキームの URI かどうか
   */
  isUnMHTURI: function(uri) {
    let param = UnMHTScheme.getUnMHTURIParam(uri);
    if ("original" in param &&
        param.original) {
      return true;
    }

    return false;
  },

  /**
   * unmht: スキームの URI を元に戻す
   *
   * @param   {string} uri
   *          unmht: スキームの URI
   * @returns {string}
   *          URI
   */
  deUnMHTURI: function(uri) {
    let param = UnMHTScheme.getUnMHTURIParam(uri);
    if ("original" in param &&
        param.original) {
      return param.original;
    }

    return uri;
  },

  /**
   * unmht: スキームの URI の情報を取得する
   *
   * @param   {string} uri
   *          unmht: スキームの URI
   * @returns {UnMHTURIParam}
   *          URI の情報
   */
  getUnMHTURIParam: function(uri) {
    let param = new UnMHTURIParam();

    let m = uri.match(/^unmht:\/\/(?:([^\/]+)\.)?unmht(?::([0-9]+))?\/([A-Za-z0-9\-]+)\.([0-9]+)\/([^#]*)\/((?:&#|[^\/#])*)(?:#(.*))?$/);
    if (!m) {
      return param;
    }

    param.host     = m[1] ? (lazy.IDNService.convertUTF8toACE(m[1])) : "";
    param.port     = m[2];
    param.scheme   = m[3].toLowerCase();
    let sep        = Number.parseInt(m[4], 10);
    param.path     = m[5];
    param.refName  = m[6];
    param.fragment = m[7];

    let sep1 = (sep & 1) ? "//" : "";
    let sep2 = (sep & 2) ? "//" : "";
    let sep3 = (sep & 4) ? "/" : "";

    param.original
      = param.scheme + ":" + sep1 + param.host
      + (param.port ? (":" + param.port) : "")
      + sep2 + sep3
      + param.path;

    param.baseURI
      = "unmht://" + param.host + (param.host ? "." : "") + "unmht"
      + (param.port ? (":" + param.port) : "")
      + "/"
      + param.scheme + "." + sep + "/" + param.path + "/";

    return param;
  },

  /**
   * エラーページの URI を返す
   *
   * @param   {string} uri
   *          URI
   * @returns {string}
   *          エラーページの URI
   */
  enErrorURI: function(uri) {
    uri = this.enUnMHTURI(uri);
    uri = uri.replace(/^unmht:/, "unmht-error:");
    uri = this.enUnMHTURI(uri);

    return uri;
  },

  /**
   * エラーページの URI を元に戻す
   *
   * @param   {string} uri
   *          エラーページの URI
   * @returns {string}
   *          URI
   */
  deErrorURI: function(uri) {
    uri = this.deUnMHTURI(uri);
    uri = uri.replace(/^unmht-error:/, "unmht:");
    uri = this.deUnMHTURI(uri);

    return uri;
  },

  /**
   * ソース表示ページの URI を返す
   *
   * @param   {string} uri
   *          URI
   * @returns {string}
   *          ソース表示ページの URI
   */
  enViewSourceURI: function(uri) {
    uri = this.enUnMHTURI(uri);
    uri = uri.replace(/^unmht:/, "unmht-view-source:");
    uri = this.enUnMHTURI(uri);

    return uri;
  },

  /**
   * ソース表示ページの URI を元に戻す
   *
   * @param   {string} uri
   *          ソース表示ページの URI
   * @returns {string}
   *          URI
   */
  deViewSourceURI: function(uri) {
    uri = this.deUnMHTURI(uri);
    uri = uri.replace(/^unmht-view-source:/, "unmht:");
    uri = this.deUnMHTURI(uri);

    return uri;
  },

  /**
   * 空のページの URI を返す
   *
   * @returns {string}
   *          空のページの URI
   */
  createEmptyURI: function() {
    return this.enUnMHTURI("unmht-empty:");
  },

  /**
   * ロード中のページの URI を返す
   *
   * @returns {string}
   *          ロード中のページの URI
   */
  createLoadingURI: function() {
    return this.enUnMHTURI("unmht-loading:");
  },

  /**
   * キャッシュページの URI を返す
   *
   * @returns {string}
   *          キャッシュページの URI
   */
  createCacheURI: function() {
    return this.enUnMHTURI("unmht-cache:");
  }
});

module.UnMHTScheme = UnMHTScheme;

})(this);
