"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTScrapBookIntegration"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arScoped, arScopeManager } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileSaver",
                                  "resource://unmht/modules/UnMHTFileSaver.jsm");

const ScrapBookURI = "chrome://scrapbook/content/scrapbook.xul";

/**
 * ScrapBook サイトバー組み込み
 *
 * @constructor
 * @param   {nsIDocument} sidebarDocument
 *          対象のサイドバーのドキュメント
 * @implements arIScopedClass
 */
function UnMHTScrapBookSidebarIntegration(sidebarDocument) {
  /**
   * 対象のサイドバーのドキュメント
   * @type {nsIDocument}
   */
  this._sidebarDocument = sidebarDocument;

  /**
   * ScrapBook のポップアップ
   * @type {nsIDOMXULElement}
   */
  this._popup = null;

  /**
   * 追加したメニューアイテム
   * @type {nsIDOMXULElement}
   */
  this._menuitem = null;

  /**
   * ScrapBook のデータ
   * @type {Object}
   */
  this._scrapBookData = null;

  Object.seal(this);

  this._createMenuItem();
}
UnMHTScrapBookSidebarIntegration.prototype = Object.freeze({
  /**
   * ScrapBook のポップアップを設定する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onPopupShowing: function(event) {
    let sidebarWindow = this._sidebarDocument.defaultView;

    let sbData;
    let sbUtils;
    let sbResource;
    if ("ScrapBookData" in sidebarWindow) {
      sbData = sidebarWindow.ScrapBookData;
    } else if ("sbDataSource" in sidebarWindow) {
      sbData = sidebarWindow.sbDataSource;
    } else {
      return;
    }
    if ("ScrapBookUtils" in sidebarWindow) {
      sbUtils = sidebarWindow.ScrapBookUtils;
    } else if ("sbCommonUtils" in sidebarWindow) {
      sbUtils = sidebarWindow.sbCommonUtils;
    } else {
      return;
    }
    if ("sbTreeUI" in sidebarWindow) {
      sbResource = sidebarWindow.sbTreeUI.resource;
    } else if ("sbTreeHandler" in sidebarWindow) {
      sbResource = sidebarWindow.sbTreeHandler.resource;
    } else if ("sbListHandler" in sidebarWindow) {
      sbResource = sidebarWindow.sbListHandler.resource;
    } else {
      return;
    }

    let id = sbData.getProperty(sbResource, "id");

    let scrapBookData = {
      source:  sbData.getProperty(sbResource, "source"),
      title:   sbData.getProperty(sbResource, "title"),
      id:      id,
      dir:     sbUtils.getContentDir(id).path,
      charset: sbData.getProperty(sbResource, "chars")
    };

    if (scrapBookData["source"]) {
      this._menuitem.hidden = false;
    } else {
      this._menuitem.hidden = true;
    }

    this._scrapBookData = scrapBookData;
  },

  /**
   * 保存
   */
  onSave: function() {
    if (!this._scrapBookData) {
      return;
    }

    let window = this._sidebarDocument.defaultView;
    if (window.parent) {
      window = window.parent;
    }

    lazy.UnMHTFileSaver.saveScrapBook(window, this._scrapBookData);
  },

  /* ==== arIScopedClass ==== */

  /**
   * 破棄
   */
  destruct: function() {
    if (this._menuitem) {
      try {
        this._menuitem.remove();
      } catch (e) {
      }
      this._menuitem = null;
    }

    this._popup = null;
    this._menuitem = null;
    this._sidebarDocument = null;

    this._scrapBookData = null;
  },

  /* ==== private ==== */

  /**
   * メニューアイテムを作成する
   */
  _createMenuItem: function() {
    let popup = this._sidebarDocument.getElementById("sbPopup");
    let pos = this._sidebarDocument.getElementById("sbPopupTools");
    if (!(popup && pos)) {
      return;
    }

    let item = this._sidebarDocument.createElement("menuitem");
    item.id = "unmht-scrapbook-save";
    item.className = "menuitem-iconic";
    item.setAttribute("label", TEXT("save.title"));
    item.setAttribute("image", "resource://unmht/images/icon_small.png");
    arEvent.add(this._sidebarDocument, item, "command", this.onSave.bind(this), false);
    pos.parentNode.insertBefore(item, pos);

    arEvent.add(this._sidebarDocument, popup, "popupshowing", this.onPopupShowing.bind(this), false);

    this._popup = popup;
    this._menuitem = item;
  }
});

/**
 * ScrapBook 組み込み
 *
 * @class
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
let UnMHTScrapBookIntegration = Object.freeze({
  /**
   * サイドバーの内容が ScrapBook かどうかチェック
   * ScrapBook ならばメニューの項目を追加する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {Event} event
   *          対象のイベント
   * @param   {nsIDocument} sidebarDocument
   *          対象のサイドバーのドキュメント
   */
  onSidebarLoad: function(window, event, sidebarDocument) {
    if (!sidebarDocument) {
      sidebarDocument = event.target.defaultView.document;
    }

    if (sidebarDocument.documentURI != ScrapBookURI) {
      return;
    }

    this._integs.create(sidebarDocument);

    arEvent.oneshot(sidebarDocument, sidebarDocument.defaultView, "beforeunload", () => {
      arScopeManager.leave(sidebarDocument);
    }, false);
  },

  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    arWindowWatcher.addListener(this);
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!arWindowUtils.isBrowser(window)) {
      return;
    }

    this._hookSidebar(window);
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
  },

  /* ==== private ==== */

  /**
   * ウィンドウごとの組み込み
   * @type {arScoped.<UnMHTScrapBookSidebarIntegration>}
   */
  _integs: new arScoped(UnMHTScrapBookSidebarIntegration),

  /**
   * サイドバーのロードをフックする
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _hookSidebar: function(window) {
    let sidebar = window.document.getElementById("sidebar");
    if (!sidebar) {
      return;
    }

    arEvent.add(window, sidebar, "DOMContentLoaded",
                this.onSidebarLoad.bind(this, window), false);

    this.onSidebarLoad(window, null, sidebar.contentDocument);
  }
});

module.UnMHTScrapBookIntegration = UnMHTScrapBookIntegration;

})(this);
