"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTToolbar"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arCustomizableUI } = Cu.import("resource://unmht/modules/base/arCustomizableUI.jsm", {});
let { arWindowOverlay } = Cu.import("resource://unmht/modules/base/arWindowOverlay.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTMenu } = Cu.import("resource://unmht/modules/UnMHTMenu.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileSaver",
                                  "resource://unmht/modules/UnMHTFileSaver.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTSaveMultiDialog",
                                  "resource://unmht/modules/UnMHTSaveMultiDialog.jsm");

/**
 * ツールバーのボタン管理
 *
 * @constructor
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
function UnMHTToolbarClass() {
  Object.seal(this);
}
UnMHTToolbarClass.prototype = Object.freeze({
  /**
   * ツールバーのボタンの状況に合わせて切り替える
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  setState: function(window) {
    let browser = arWindowUtils.getCurrentBrowser(window);

    let savable
      = UnMHTDocumentCategory.isSavable(window, browser, false);

    let document = window.document;
    let $ = id => document.getElementById(id);

    let enableButton = (type, enabled) => {
      let button = $("unmht-toolbarbutton-" + type);
      if (button) {
        button.disabled = !enabled;
      }
    };

    enableButton("save", savable);
    enableButton("quicksave", savable);
  },

  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    arWindowWatcher.addListener(this);
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
    if (!arWindowUtils.isBrowser(window)) {
      return;
    }

    /* CustomizableUI の view の参照先の作成は onWindowLoad では遅い */
    this._createPanel(window);
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!(arWindowUtils.isBrowser(window) || arWindowUtils.isMailer(window))) {
      return;
    }

    this._createUI(window);

    let toolbar = this;
    arWindowUtils.addProgressListener(window, {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsISupportsWeakReference,
                                             Ci.nsIWebProgressListener]),

      onLocationChange: function(webProgress, request, location) {
        toolbar.setState(window);
      },

      onProgressChange: function(webProgress, request,
                                 curSelfProgress, maxSelfProgress,
                                 curTotalProgress, maxTotalProgress) {},
      onSecurityChange: function(webProgress, request, state) {},
      onStateChange: function(webProgress, request, flags, status) {},
      onStatusChange: function(webProgress, request, status, message) {}
    });
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
  },

  /* ==== private ==== */

  /**
   * メニューパネルを作成する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _createPanel: function(window) {
    if (!arCustomizableUI.supported) {
      return;
    }

    let overlay = arWindowOverlay.get(window);

    let panel = overlay.panelUIMultiView.addPanelView({
      id: "unmht-menu-panel",
      flex: "1",
    }, TEXT("name"));
    UnMHTMenu.createMenuItems(false, window, panel, "panel", "subviewbutton");
  },

  /**
   * CustomizableUI で UI を作る
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _createNewUI: function(window) {
    arCustomizableUI.batch(() => {
      arCustomizableUI.createWidget({
        id: "unmht-toolbarbutton-preferences",
        type: "view",
        viewId: "unmht-menu-panel",
        label: TEXT("name"),
        tooltiptext: TEXT("name"),
        removable: true,
        defaultArea: arCustomizableUI.CustomizableUI.AREA_NAVBAR,
        onViewShowing: function(event) {
          let window = event.target.ownerDocument.defaultView;
          UnMHTMenu.onMenuShowing(event, window, "panel");
        }
      });

      arCustomizableUI.createWidget({
        id: "unmht-toolbarbutton-save",
        type: "button",
        label: TEXT("menuitem.save.single.instant"),
        tooltiptext: TEXT("menuitem.save.single.instant"),
        removable: true,
        onCommand: function(event) {
          let window = event.target.ownerDocument.defaultView;
          lazy.UnMHTFileSaver.save(window, {
            button: "unmht-toolbarbutton-save"
          });
        }
      });

      arCustomizableUI.createWidget({
        id: "unmht-toolbarbutton-quicksave",
        type: "button",
        label: TEXT("menuitem.save.quick"),
        tooltiptext: TEXT("menuitem.save.quick"),
        removable: true,
        onCommand: function(event) {
          let window = event.target.ownerDocument.defaultView;
          lazy.UnMHTFileSaver.quicksave(window, null, {
            button: "unmht-toolbarbutton-quicksave"
          });
        }
      });

      arCustomizableUI.createWidget({
        id: "unmht-toolbarbutton-savemulti",
        type: "button",
        label: TEXT("menuitem.save.multi.toolbar"),
        tooltiptext: TEXT("menuitem.save.multi.toolbar"),
        removable: true,
        onCommand: function(event) {
          let window = event.target.ownerDocument.defaultView;
          let arg = {
            targetWindow: window
          };
          lazy.UnMHTSaveMultiDialog.open(window, arg);
        }
      });
    });
  },

  /**
   * UI を作る
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _createUI: function(window) {
    if (arCustomizableUI.supported) {
      this._createNewUI(window);
      return;
    }

    let overlay = arWindowOverlay.get(window);

    let isMailer = arWindowUtils.isMailer(window);

    let button = overlay.toolbar.addButton({
      id: "unmht-toolbarbutton-preferences",
      type: "menu",
      "class": "toolbarbutton-1 chromeclass-toolbar-additional",
      style: "-moz-box-orient: horizontal; "
        + "list-style-image: url('resource://unmht/images/all.png'); "
        + "-moz-image-region: rect(32px, 24px, 56px, 0px);",
      label: TEXT("name"),
      tooltiptext: TEXT("name")
    });

    let menu = button.addMenuPopup({
      id: 'unmht-toolbar-popup',
      onpopupshowing: function(event) {
        UnMHTMenu.onMenuShowing(event, window, "toolbar");
      }
    });
    UnMHTMenu.createMenuItems(isMailer, window, menu, "toolbar");

    overlay.toolbar.addButton({
      id: "unmht-toolbarbutton-save",
      "class": "toolbarbutton-1 chromeclass-toolbar-additional",
      style: "list-style-image: url('resource://unmht/images/all.png'); "
      + "-moz-image-region: rect(32px, 72px, 56px, 48px);",
      label: TEXT("menuitem.save.single.instant"),
      tooltiptext: TEXT("menuitem.save.single.instant"),
      oncommand: function(event) {
        lazy.UnMHTFileSaver.save(window);
      }
    });

    overlay.toolbar.addButton({
      id: "unmht-toolbarbutton-quicksave",
      "class": "toolbarbutton-1 chromeclass-toolbar-additional",
      style: "list-style-image: url('resource://unmht/images/all.png'); "
        + "-moz-image-region: rect(32px, 48px, 56px, 24px);",
      label: TEXT("menuitem.save.quick"),
      tooltiptext: TEXT("menuitem.save.quick"),
      oncommand: function(event) {
        lazy.UnMHTFileSaver.quicksave(window);
      }
    });

    if (!isMailer) {
      overlay.toolbar.addButton({
        id: "unmht-toolbarbutton-savemulti",
        "class": "toolbarbutton-1 chromeclass-toolbar-additional",
        style: "list-style-image: url('resource://unmht/images/all.png'); "
          + "-moz-image-region: rect(32px, 96px, 56px, 72px);",
        label: TEXT("menuitem.save.multi.toolbar"),
        tooltiptext: TEXT("menuitem.save.multi.toolbar"),
        oncommand: function(event) {
          let arg = {
            targetWindow: window
          };
          lazy.UnMHTSaveMultiDialog.open(window, arg);
        }
      });
    }
  }
});

/**
 * ツールバーのボタン管理
 * @type {UnMHTToolbarClass}
 */
let UnMHTToolbar = new UnMHTToolbarClass();

module.UnMHTToolbar = UnMHTToolbar;

})(this);
