"use strict";

this.EXPORTED_SYMBOLS = ["arComponentLoader"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu,
      manager: Cm } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});

/**
 * XPCOM のロードとアンロード
 *
 * @class
 * @implements arIShutdownEventListener
 */
let arComponentLoader = Object.freeze({
  /**
   * XPCOM を登録してアンロード時のリスナの登録も行う
   *
   * @param   {Constructor.<nsISupport>} component
   *          登録するコンポーネントのコンストラクタ
   * @returns {Promise.<boolean>}
   * @resolves ロードした
   * @rejects {boolean} ロード失敗
   */
  loadComponent: function(component) {
    let factory = Object.freeze({
      /* ==== nsIFactory ==== */

      /**
       * インスタンスを生成する
       *
       * @param   {nsISupport} outer
       *          統合する対象
       * @param   {nsIIDRef} iid
       *          生成する対象のインターフェース ID
       * @returns {nsISupports}
       *          生成したインスタンス
       * @throws  {Cr.NS_ERROR_NO_AGGREGATION}
       */
      createInstance: function(outer, iid) {
        if (outer) {
          throw Cr.NS_ERROR_NO_AGGREGATION;
        }
        return (new component()).QueryInterface(iid);
      }
    });

    return this._registerComponent(component, factory).catch(e => {
      Cu.reportError(e + "\n" + e.stack);
    });
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    this._unregisterComponents();
  },

  /* ==== private ==== */

  /**
   * コンポーネントのコンストラクタとファクトリーのリスト
   * @type {Array.<Tuple.<Constructor.<nsISupports>,nsIFactory>>}
   */
  _components: [],

  /**
   * 初期化処理
   */
  init: function() {
    arShutdownEvent.addListener(this);
  },

  /**
   * XPCOM を登録する
   *
   * @param   {Constructor.<nsISupport>} component
   *          登録するコンポーネントのコンストラクタ
   * @param   {nsIFactory} factory
   *          登録するコンポーネントの Factory
   * @returns {Promise.<void>}
   *          結果
   * @resolves 登録完了
   * @rejects  {Error}
   *           登録失敗
   */
  _registerComponent: Task.async(function*(component, factory) {
    let registrar = Cm.QueryInterface(Ci.nsIComponentRegistrar);

    let i = 0, TRY_MAX = 10;
    for (;;) {
      if (!registrar.isCIDRegistered(component.prototype.classID)) {
        break;
      }

      /* 更新時には unregisterFactory より先に動く場合があるので
       * その場合にはしばらく待つ */

      yield arTaskUtils.wait(i * 100);

      i += 1;
      if (i == TRY_MAX) {
        throw new Error("CID " + component.prototype.classID + " is registered. Abort registration.");
      }
    }

    registrar.registerFactory(
      component.prototype.classID,
      component.prototype.classDescription,
      component.prototype.contractID,
      factory);

    if ("_xpcom_categories" in component.prototype) {
      for (let cat of component.prototype._xpcom_categories) {
        XPCOMUtils.categoryManager.addCategoryEntry(
          cat.category,
          component.prototype.contractID,
          component.prototype.contractID,
          false, true);
      }
    }

    this._components.push([component, factory]);

    return true;
  }),

  /**
   * 全ての XPCOM の登録を解除する
   */
  _unregisterComponents: function() {
    let factoryInfos
      = this._components.map(([component, factory]) => {
        if ("_xpcom_categories" in component.prototype) {
          for (let cat of component.prototype._xpcom_categories) {
            XPCOMUtils.categoryManager.deleteCategoryEntry(
              cat.category,
              component.prototype.contractID,
              false);
          }
        }
        return [component.prototype.classID, factory];
      });
    this._components.length = 0;

    /* SeaMonkey で同期的に実行すると正しく登録解除されない場合があるので
     * 別スレッドで行う */
    let registrar = Cm.QueryInterface(Ci.nsIComponentRegistrar);
    Services.tm.mainThread.dispatch(() => {
      /* ここはモジュールのアンロード後に動く別スレッドなので
       * 必要な変数を解決してからキャプチャする */
      for (let [classID, factory] of factoryInfos) {
        registrar.unregisterFactory(classID, factory);
      }
    }, Ci.nsIEventTarget.DISPATCH_NORMAL);
  }
});
arComponentLoader.init();

module.arComponentLoader = arComponentLoader;

})(this);
