"use strict";

this.EXPORTED_SYMBOLS = ["arDLFRegisterer"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyServiceGetter(lazy, "MIMEService",
                                   "@mozilla.org/mime;1",
                                   "nsIMIMEService");

/**
 * This object takes care of registering the specified Document Loader Factory
 * in the "Gecko-Content-Viewers" category, for specific MIME types or file
 * extensions.
 *
 * To use this object, firstly call the addMimeType or addFileExtension methods,
 * then call the register method.
 *
 * To unregister all entries, call unregister method.
 *
 * All the changes made by this object are valid only for the current
 * application session.
 *
 * For more information on the document loading process, see
 *  <http://www.mozilla.org/newlayout/doc/webwidget.html> and
 *  <https://developer.mozilla.org/En/The_life_of_an_HTML_HTTP_request>
 *  (retrieved 2008-12-27).
 *
 * @constructor
 * @param   {string} aDLFContractId
 *          Document Loader Factory の Contract ID
 */
function arDLFRegisterer(aDLFContractId) {
  /**
   * 登録した MIME-Type と isRecommended のマップ
   * This object will be filled in with properties whose name is the MIME type
   * we need to handle, and whose value is true if this is the DLF that
   * should preferably handle that MIME type.
   * @type {Map.<string,boolean>}
   */
  this._mimeMap = new Map();

  /**
   * 拡張子と MIME-Type のマップ
   * @type {Map.<string,string>}
   */
  this._extMap = new Map();

  /**
   * Contract ID of the Document Loader Factory to be registered
   * @type {string}
   */
  this._dLFContractId = aDLFContractId;

  Object.seal(this);
}
arDLFRegisterer.prototype = Object.freeze({
  /**
   * Prepare for registering the DLF with the specified MIME type.
   *
   * @param   {string} aMimeType
   *          The MIME type to be managed.
   * @param   {boolean} aIsRecommended
   *          True if this DLF is the one that should
   *          preferably handle the associated type.
   * @returns {arDLFRegisterer}
   *          Reference to this object, for fluent interface support.
   */
  addMimeType: function(aMimeType, aIsRecommended) {
    /* If the the type was already added with aIsRecommended set to true,
     * leave it alone even if aIsRecommended is now false. */
    if (!this._mimeMap.has(aMimeType) ||
        !this._mimeMap.get(aMimeType)) {
      this._mimeMap.set(aMimeType, aIsRecommended);
    }

    return this;
  },

  /**
   * Prepare for registering the DLF with the MIME type actually associated
   * with the specified file extension.
   *
   * @param   {string} aExtension
   *          The file extension to be managed.
   * @param   {string} aPossibleMimeType
   *          Suggestion for a possible MIME type to be
   *          associated with the file extension.
   * @param   {boolean} aIsRecommended
   *          True if this DLF is the one that should
   *          preferably handle the associated type.
   * @returns {arDLFRegisterer}
   *          Reference to this object, for fluent interface support.
   */
  addFileExtension: function(aExtension, aPossibleMimeType, aIsRecommended) {
    /* First of all, for the specified file extensions, we must ensure that
     * the application is able to determine a specific MIME type.
     * If there are no other means to determine the MIME type,
     * the last resort is the "ext-to-type-mapping" category,
     * so we set an entry there, replacing an existing entry only for
     * the file extensions that are preferably managed by this DLF. */
    this._addCategoryEntryForSession("ext-to-type-mapping",
                                     aExtension, aPossibleMimeType,
                                     aIsRecommended);
    this._extMap.set(aExtension, aPossibleMimeType);

    /* Find out the actual MIME type that will be used for the file
     * extension, and register the DLF for handling it. See also
     * <https://developer.mozilla.org/En/How_Mozilla_determines_MIME_Types>
     * (retrieved 2008-11-21). */
    let realMimeType
      = this._getTypeFromExtensionSafely(aExtension, aPossibleMimeType);
    this.addMimeType(realMimeType, aIsRecommended);

    return this;
  },

  /**
   * Perform the actual registration.
   */
  register: function() {
    for (let [mimeType, isRecommended] of this._mimeMap) {
      if (arAppInfo.isSeaMonkey && mimeType == "message/rfc822") {
        continue;
      }

      /* Add the appropriate category entry, without persisting it,
       * and replacing an existing entry only for the MIME types
       * that are preferably managed by this DLF. */
      this._addCategoryEntryForSession("Gecko-Content-Viewers",
                                       mimeType, this._dLFContractId,
                                       isRecommended);
    }
  },

  /**
   * unregister all entries
   */
  unregister: function() {
    for (let mimeType of this._mimeMap.keys()) {
      if (arAppInfo.isSeaMonkey && mimeType == "message/rfc822") {
        continue;
      }

      this._deleteCategoryEntryForSession("Gecko-Content-Viewers",
                                          mimeType, this._dLFContractId);
    }

    for (let [extension, possibleMimeType] of this._extMap) {
      this._deleteCategoryEntryForSession("ext-to-type-mapping",
                                          extension, possibleMimeType);
    }
  },

  /* ==== private ==== */

  /**
   * Calls nsICategoryManager.addCategoryEntry with aPersist set to false, but
   * if aReplace is false and the category entry already has a value, no
   * exception is thrown.
   *
   * @param   {string} aCategory
   *          The name of the category being modified.
   * @param   {string} aEntry
   *          The name of the category entry being modified.
   * @param   {string} aValue
   *          The value for the category entry.
   * @param   {boolean} aReplace
   *          A flag indicating whether or not to overwrite the
   *          value of an existing category entry.
   */
  _addCategoryEntryForSession: function(aCategory, aEntry, aValue, aReplace) {
    try {
      XPCOMUtils.categoryManager.addCategoryEntry(aCategory, aEntry, aValue,
                                                  false, aReplace);
    } catch (e) {
      if (e instanceof Ci.nsIXPCException &&
          e.result != Cr.NS_ERROR_INVALID_ARG) {
        /* Ignore only the NS_ERROR_INVALID_ARG result */
      } else {
        throw e;
      }
    }
  },

  /**
   * カテゴリが設定した値のままであれば削除する
   *
   * @param   {string} aCategory
   *          The name of the category being modified.
   * @param   {string} aEntry
   *          The name of the category entry being modified.
   * @param   {string} aValue
   *          The value for the category entry.
   */
  _deleteCategoryEntryForSession: function(aCategory, aEntry, aValue) {
    try {
      if (XPCOMUtils.categoryManager.getCategoryEntry(aCategory, aEntry) == aValue) {
        XPCOMUtils.categoryManager.deleteCategoryEntry(aCategory, aEntry, false);
      }
    } catch (e) {
      if (e instanceof Ci.nsIXPCException &&
          e.result != Cr.NS_ERROR_INVALID_ARG) {
        /* Ignore only the NS_ERROR_INVALID_ARG result */
      } else {
        throw e;
      }
    }
  },

  /**
   * Calls nsIMIMEService.getTypeFromExtension, and if the call fails
   * unexpectedly, returns the specified MIME type as a fallback.
   *
   * @param   {string} aExtension
   *          拡張子
   * @param   {string} aFallbackMimeType
   *          失敗した場合の MIME-Type
   * @returns {string}
   *          拡張子に対応する MIME-Type
   */
  _getTypeFromExtensionSafely: function(aExtension, aFallbackMimeType) {
    try {
      return lazy.MIMEService.getTypeFromExtension(aExtension);
    } catch (e) {
      if (e instanceof Ci.nsIException &&
          e.result == Cr.NS_ERROR_NOT_INITIALIZED) {
        /* The getTypeFromExtension call may throw NS_ERROR_NOT_INITIALIZED
         *  because of Mozilla bug 484579. In this case, return an arbitrary
         *  MIME type to mitigate the problem. */
        return aFallbackMimeType;
      } else {
        throw e;
      }
    }
  }
});

module.arDLFRegisterer = arDLFRegisterer;

})(this);
