"use strict";

this.EXPORTED_SYMBOLS = ["arDOMUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * DOM 操作ユーティリティ
 *
 * @class
 */
let arDOMUtils = Object.freeze({
  /**
   * ノードが選択範囲内かどうかチェックする
   *
   * @param   {nsIDOMNode} node
   *          対象のノード
   * @param   {nsISelection} selection
   *          対象の選択範囲
   * @returns {boolean}
   *          範囲内かどうか
   */
  isNodeInSelection: function(node, selection) {
    let nodeRange = node.ownerDocument.createRange();
    nodeRange.selectNode(node);

    for (let i = 0, len = selection.rangeCount; i < len; i += 1) {
      let range = selection.getRangeAt(i);

      let compareSE = 0;
      let compareES = 0;
      try {
        compareSE = range.compareBoundaryPoints(range.START_TO_END, nodeRange);
        compareES = range.compareBoundaryPoints(range.END_TO_START, nodeRange);
      } catch (e) {
        /* 選択範囲が別のドキュメント */
        return false;
      }

      if (compareSE == -compareES) {
        return true;
      }
    }

    return false;
  },

  /**
   * ノードを親方向に探す
   * 対象のノードも含む
   *
   * @param   {nsIDOMNode} node
   *          対象のノード
   * @param   {string} nodeName
   *          探すノード名
   * @returns {?nsIDOMNode}
   *          見付かったノード
   *          見付からなかった場合 null
   */
  findParentNode: function(node, nodeName) {
    while (node) {
      if (node.nodeName.toLowerCase() == nodeName) {
        return node;
      }
      node = node.parentNode;
    }

    return null;
  },

  /**
   * HTML に使えない文字をエスケープする
   *
   * @param   {string} text
   *          エスケープする文字列
   * @returns {string}
   *          エスケープした文字列
   */
  escapeEntity: function(text) {
    return (text
            .replace(/&/g, "&amp;")
            .replace(/\"/g, "&quot;")
            .replace(/\'/g, "&#x27;")
            .replace(/</g, "&lt;")
            .replace(/>/g, "&gt;")
            .replace(/\xa0/g, "&#xa0;"));
  },

  /**
   * HTML に使えない文字のエスケープを解除する
   *
   * @param   {string} text
   *          エスケープを解除する文字列
   * @returns {string}
   *          エスケープを解除した文字列
   */
  unescapeEntity: function(text) {
    return (text
            .replace(/&gt;/g, ">")
            .replace(/&lt;/g, "<")
            .replace(/&quot;/g, "\"")
            .replace(/&#x27;/g, "\'")
            .replace(/&#xa0;/g, "\xa0")
            .replace(/&nbsp;/g, "\xa0")
            .replace(/&amp;/g, "&"));
  }
});

module.arDOMUtils = arDOMUtils;

})(this);
