"use strict";

this.EXPORTED_SYMBOLS = ["arDialog", "arDialogClass"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});

/**
 * ダイアログを開くためのクラス
 *
 * @constructor
 * @param   {Constructor.<arIDialog>} _class
 *          ダイアログのクラスのコンストラクタ
 * @param   {Object} params
 *          ダイアログの情報
 *            {
 *              url:        {string}
 *                          ダイアログの URL,
 *              features:   {string}
 *                          ダイアログのオプション,
 *              single:     {boolean}
 *                          開いているダイアログにフォーカスするか
 *              usePromise: {boolean}
 *                          Promise を返すか
 *            }
 */
function arDialog(_class, params) {
  /**
   * ダイアログのクラスコンストラクタ
   * @type {Constructor.<arIDialog>}
   */
  this._class = _class;

  /**
   * ダイアログの URL
   * @type {string}
   */
  this._url = params.url;

  /**
   * ダイアログのオプション
   * @type {string}
   */
  this._features = params.features;

  /**
   * 開いているダイアログにフォーカスするか
   * @type {boolean}
   */
  this._single = params.single;

  /**
   * Promise を返すか
   * @type {boolean}
   */
  this._usePromise = params.usePromise;

  Object.seal(this);
}
arDialog.prototype = Object.freeze({
  /**
   * ダイアログのインスタンスを作成する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {arIDialog}
   *          ダイアログのインスタンス
   */
  create: function(window) {
    return new this._class(window);
  },

  /**
   * ダイアログを開く
   *
   * @param    {nsIDOMWindow} window
   *           実行中のウィンドウ
   * @param    {nsISupport=} arg
   *           ダイアログの引数
   * @returns  {?Promise.<nsISupport>}
   *           usePromise が false なら null
   * @resolves ダイアログの結果
   * @rejects  {*}
   */
  open: function(window, arg) {
    if (this._single) {
      if (arWindowUtils.focus(this._url)) {
        return null;
      }
    }

    let executer = (resolve, reject) => {
      window.openDialog(this._url, "", this._features, arg, resolve, reject);
    };

    if (this._usePromise) {
      return new Promise(executer);
    }
    executer(null, null);
    return null;
  }
});

/**
 * ダイアログ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @implements arIDialog
 */
function arDialogClass(window) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  Object.seal(this);
}
arDialogClass.prototype = Object.freeze({
  /* ==== arIDialog ==== */

  /**
   * ダイアログの引数
   * @type {*}
   */
  get arg() {
    return this._window.arguments[0];
  },

  /**
   * ダイアログの結果を解決する
   *
   * @param   {*} value
   *          ダイアログの結果
   */
  resolve: function(value) {
    let resolve = this._window.arguments[1];
    if (!resolve) {
      return;
    }
    setTimeout(() => {
      resolve(value);
    }, 0);
  },

  /**
   * ダイアログの結果を reject する
   *
   * @param   {*} value
   *          ダイアログの結果
   */
  reject: function(value) {
    let reject = this._window.arguments[2];
    if (!reject) {
      return;
    }
    setTimeout(() => {
      reject(value);
    }, 0);
  },

  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;
  },

  /**
   * ロードイベント
   */
  onLoad: function() {
  },

  /**
   * アンロードイベント
   */
  onUnload: function() {
  },

  /**
   * ダイアログを OK したイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    return true;
  },

  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    return true;
  },

  /* ==== private ==== */

  /**
   * getElementById の省略
   *
   * @param   {string} id
   *          要素の ID
   * @returns {?nsIDOMElement}
   *          id に対応する要素
   *          見付からなければ null
   */
  $: function(id) {
    return this._window.document.getElementById(id);
  }
});

module.arDialog = arDialog;
module.arDialogClass = arDialogClass;

})(this);
