"use strict";

this.EXPORTED_SYMBOLS = ["arEvent"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arScoped } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});

/**
 * スコープ対象ごとのイベント
 *
 * @constructor
 * @template ScopeTargetType
 * @param   {ScopeTargetType} scopeTarget
 *          スコープ対象
 * @implements arIScopedClass
 */
function arEventPerScope(scopeTarget) {
  /**
   * 追加したイベント
   * @type {Array.<Tuple.<nsIDOMElement,string,function(Event),boolean>>}
   */
  this.events = [];

  Object.seal(this);
}
arEventPerScope.prototype = Object.freeze({
  /**
   * イベントハンドラを登録する
   *
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} type
   *          イベントの種類
   * @param   {function(Event)} func
   *          イベントハンドラ
   * @param   {boolean} useCapture
   *          捕捉フェーズを使用する
   */
  add: function(node, type, func, useCapture) {
    node.addEventListener(type, func, useCapture);

    this.events.push([node, type, func, useCapture]);
  },

  /**
   * イベントハンドラを削除する
   *
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} type
   *          イベントの種類
   * @param   {function(Event)} func
   *          イベントハンドラ
   * @param   {boolean} useCapture
   *          捕捉フェーズを使用する
   */
  remove: function(node, type, func, useCapture) {
    node.removeEventListener(type, func, useCapture);

    this.events = this.events.filter(([n, t, f, u]) => !(n == node &&
                                                         t == type &&
                                                         f == func &&
                                                         u == useCapture));
  },

  /* ==== arIScopedClass ==== */

  /**
   * 破棄する
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  destruct: function(isShutdown, data, reason) {
    for (let [node, type, func, useCapture] of this.events) {
      node.removeEventListener(type, func, useCapture);
    }
    this.events.length = 0;
  }
});

/**
 * イベント管理
 *
 * @class
 * @template ScopeTargetType
 */
let arEvent = Object.freeze({
  /**
   * イベントハンドラを登録する
   * スコープを抜けると削除される
   *
   * @param   {ScopeTargetType} scopeTarget
   *          スコープ対象
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} type
   *          イベントの種類
   * @param   {function(Event)} func
   *          イベントハンドラ
   * @param   {boolean} useCapture
   *          捕捉フェーズを使用する
   */
  add: function(scopeTarget, node, type, func, useCapture) {
    this._events.create(scopeTarget).add(node, type, func, useCapture);
  },

  /**
   * 一回限りのイベントハンドラを登録する
   * スコープを抜けると削除される
   *
   * @param   {ScopeTargetType} scopeTarget
   *          スコープ対象
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} type
   *          イベントの種類
   * @param   {function(Event)} func
   *          イベントハンドラ
   * @param   {boolean} useCapture
   *          捕捉フェーズを使用する
   */
  oneshot: function(scopeTarget, node, type, func, useCapture) {
    let wrapped = event => {
      this.remove(scopeTarget, node, type, wrapped);

      return func(event);
    };

    this._events.create(scopeTarget).add(node, type, wrapped, useCapture);
  },

  /**
   * イベントハンドラを削除する
   *
   * @param   {ScopeTargetType} scopeTarget
   *          スコープ対象
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} type
   *          イベントの種類
   * @param   {function(Event)} func
   *          イベントハンドラ
   * @param   {boolean} useCapture
   *          捕捉フェーズを使用する
   */
  remove: function(scopeTarget, node, type, func, useCapture) {
    /* 本来ここは create ではなく get すべきだが
     * 基本的に存在するはずであり、また作成してしまっても問題ないので
     * コードをシンプルにするために create を呼ぶ */
    this._events.create(scopeTarget).remove(node, type, func, useCapture);
  },

  /* ==== private ==== */

  /**
   * 追加したイベント
   * @type {arScoped.<arEventPerScope>}
   */
  _events: new arScoped(arEventPerScope)
});

module.arEvent = arEvent;

})(this);
