"use strict";

this.EXPORTED_SYMBOLS = ["arFilePicker"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Downloads } = Cu.import("resource://gre/modules/Downloads.jsm");
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "PrivateBrowsingUtils",
                                  "resource://gre/modules/PrivateBrowsingUtils.jsm");

const OPEN_LAST_DIR_PREF = "browser.open.lastDir";

/**
 * ファイル選択ダイアログ
 *
 * @class
 */
let arFilePicker = Object.freeze({
  /**
   * 開くファイルを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {Object} params
   *          パラメータ
   * @returns {Promise.<Tuple.<?string,nsIFilePicker>>}
   * @resolves 開くファイルのパスとファイルピッカー
   *           キャンセルの場合はファイルが null
   */
  getOpenFile: function(window, params) {
    let getResult = filePicker => {
      let path = filePicker.file.path;

      return Promise.resolve([path, OS.Path.dirname(path)]);
    };
    return this._getFileHelper(window, params, Ci.nsIFilePicker.modeOpen,
                               getResult);
  },

  /**
   * 保存するファイルを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {Object} params
   *          パラメータ
   * @returns {Promise.<Tuple.<?string,nsIFilePicker>>}
   * @resolves 保存するファイルとファイルピッカー
   *           キャンセルの場合はファイルが null
   */
  getSaveFile: function(window, params) {
    let getResult = filePicker => {
      let path = filePicker.file.path;

      return Promise.resolve([path, OS.Path.dirname(path)]);
    };
    return this._getFileHelper(window, params, Ci.nsIFilePicker.modeSave,
                               getResult);
  },

  /**
   * フォルダを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {Object} params
   *          パラメータ
   * @returns {Promise.<Tuple.<?string,nsIFilePicker>>}
   * @resolves フォルダとファイルピッカー
   *           キャンセルの場合はフォルダが null
   */
  getFolder: function(window, params) {
    let getResult = filePicker => {
      let dirPath = filePicker.file.path;

      return Promise.resolve([dirPath, dirPath]);
    };
    return this._getFileHelper(window, params, Ci.nsIFilePicker.modeGetFolder,
                               getResult);
  },

  /**
   * 複数の開くファイルを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {Object} params
   *          パラメータ
   * @returns {Promise.<Tuple.<?Array.<string>,nsIFilePicker>>}
   * @resolves 開くファイルの配列とファイルピッカー
   *           キャンセルの場合はファイルの配列が null
   */
  getOpenFiles: function(window, params) {
    let getResult = filePicker => {
      let paths = [...XPCOMUtils.IterSimpleEnumerator(filePicker.files,
                                                      Ci.nsIFile)]
        .map(file => file.path);

      let dir = null;
      if (paths.length) {
        dir = OS.Path.dirname(paths[0]);
      }

      return Promise.resolve([paths, dir]);
    };
    return this._getFileHelper(window, params,
                               Ci.nsIFilePicker.modeOpenMultiple,
                               getResult);
  },

  /**
   * 開くのに最後に使用したディレクトリを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @returns {?string}
   *          最後に使用したディレクトリのパス
   *          無い場合は null
   */
  getLastOpenDirPath: function(window) {
    try {
      /* window.gLastOpenDirectory.path は nsIFile */
      return window.gLastOpenDirectory.path.path;
    } catch (e) {
      /* COMPAT: SeaMonkey, Thunderbird */
      if (!lazy.PrivateBrowsingUtils.isWindowPrivate(window)) {
        if (Services.prefs.getPrefType(OPEN_LAST_DIR_PREF) !== 0) {
          return Services.prefs.getComplexValue(OPEN_LAST_DIR_PREF,
                                                Ci.nsILocalFile).path;
        }
      }
    }

    return null;
  },

  /**
   * 開くのに最後に使用したディレクトリを設定する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {string} dir
   *          最後に使用したディレクトリのパス
   */
  setLastOpenDirPath: function(window, dir) {
    try {
      /* window.gLastOpenDirectory.path は nsIFile */
      window.gLastOpenDirectory.path = arFileUtils.getFile(dir);
    } catch (e) {
      /* COMPAT: SeaMonkey, Thunderbird */
      if (!lazy.PrivateBrowsingUtils.isWindowPrivate(window)) {
        if (dir) {
          Services.prefs.setComplexValue(OPEN_LAST_DIR_PREF, Ci.nsILocalFile,
                                         arFileUtils.getFile(dir));
        } else {
          Services.prefs.clearUserPref(OPEN_LAST_DIR_PREF);
        }
      }
    }
  },

  /**
   * 保存するのに最後に使用したディレクトリを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {?string} uri
   *          保存対象の URI
   *          不要ならば null
   * @param   {DownloadLastDir} _downloadLastDir
   *          DownloadLastDir オブジェクト
   * @returns {Promise.<string>}
   * @resolves 最後に使用したディレクトリのパス
   */
  getLastSaveDirPath: Task.async(function*(window, uri, _downloadLastDir) {
    let dirPath = null;
    let path = yield Downloads.getPreferredDownloadsDirectory();
    let exists = yield OS.File.exists(path);
    if (exists) {
      dirPath = path;
    }

    path = yield new Promise((resolve, reject) => {
      _downloadLastDir.getFileAsync(uri, file => {
        if (file) {
          resolve(file.path);
        } else {
          resolve(null);
        }
      });
    });
    if (path) {
      let exists = yield OS.File.exists(dirPath);
      if (exists) {
        dirPath = path;
      }
    }

    if (!dirPath) {
      dirPath = OS.Constants.Path.desktopDir;
    }

    return dirPath;
  }),

  /**
   * 保存するのに最後に使用したディレクトリを設定する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {?string} uri
   *          保存対象の URI
   *          不要ならば null
   * @param   {string} dirPath
   *          最後に使用したディレクトリのパス
   * @param   {DownloadLastDir} _downloadLastDir
   *          DownloadLastDir オブジェクト
   */
  setLastSaveDirPath: function(window, uri, dirPath, _downloadLastDir) {
    _downloadLastDir.setFile(uri, arFileUtils.getFile(dirPath));
  },

  /* ==== private ==== */

  /**
   * ファイル選択ダイアログを初期化する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {string=} title
   *          ダイアログのタイトル (オプショナル)
   * @param   {string=} displayDirectoryPath
   *           useLastDir が false の際に表示するディレクトリ (オプショナル)
   *           ファイルを指している場合は親ディレクトリになる
   * @param   {string=} defaultString
   *          デフォルトファイル名 (オプショナル),
   * @param   {string=} defaultextension
   *          デフォルト拡張子 (オプショナル),
   * @param   {Array.<(Tuple.<string,string>|number)>=} filter
   *          ファイルタイプ (オプショナル)
   *             [type, ...]
   *               type: [string 名前, string 拡張子]
   *                     または
   *                     number filterMask
   * @param   {number=} filterIndex
   *          デフォルトのファイルタイプのインデックス (オプショナル)
   * @param   {number} mode
   *          nsIFilePicker のモード
   *            Ci.nsIFilePicker.mode*
   * @param   {nsIURI} uri
   *          保存対象の URI (保存の場合のみ)
   * @returns {nsIFilePicker}
   *          ファイルピッカー
   */
  _initFilePicker: function(window,
                            title,
                            displayDirectoryPath,
                            defaultString, defaultExtension,
                            filter, filterIndex,
                            mode, uri) {
    let filePicker
      = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);

    filePicker.init(window, title, mode);

    if (displayDirectoryPath) {
      filePicker.displayDirectory = arFileUtils.getFile(displayDirectoryPath);
    }

    if (filter) {
      for (let type of filter) {
        if (type.length == 1) {
          let [mask] = type;
          filePicker.appendFilters(mask);
        } else {
          let [name, ext] = type;
          filePicker.appendFilter(name, ext);
        }
      }
    }
    if (filterIndex >= 0 && filterIndex < filter.length) {
      filePicker.filterIndex = filterIndex;
    }

    if (defaultString) {
      filePicker.defaultString = defaultString;
    }
    if (defaultExtension) {
      filePicker.defaultExtension = defaultExtension;
    }

    return filePicker;
  },

  /**
   * ファイル選択ダイアログを初期化する
   *
   * @template ResultType
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {Object} params
   *          パラメータ
   *            {
   *              title: {string=}
   *                     ダイアログのタイトル (オプショナル),
   *              useLastDir: {boolean=}
   *                      最後に使用したディレクトリを使用/保存する
   *                      (オプショナル)
   *                      デフォルトは false,
   *              displayDirectoryPath: {string=}
   *                      useLastDir が false の際に表示するディレクトリ
   *                      (オプショナル)
   *                      ファイルを指している場合は親ディレクトリになる,
   *              uri: {string=}
   *                   保存対象の URI (オプショナル)
   *                   サイトごとに記憶する際に使用する
   *                   デフォルトは null,
   *              defaultString: {string=}
   *                             デフォルトファイル名 (オプショナル),
   *              defaultextension: {string=}
   *                                デフォルト拡張子 (オプショナル),
   *              filter: {Array.<(Tuple.<string,string>|number)>=}
   *                      ファイルタイプ (オプショナル)
   *                        [type, ...]
   *                          type: [string 名前, string 拡張子]
   *                                または
   *                                number filterMask
   *              filterIndex: {number=}
   *                           デフォルトのファイルタイプの
   *                           インデックス (オプショナル),
   *            }
   * @param   {number} mode
   *          nsIFilePicker のモード
   *            Ci.nsIFilePicker.mode*
   * @param   {function(nsIFilePicker):Promise.<ResultType>} getResult
   *          結果を返す関数
   *            function(filePicker, uri)
   *              *param   {nsIFilePicker} filePicker
   *                       使用したファイル選択ダイアログ
   *              *param   {?nsIURI} uri
   *                       保存対象の URI (保存の場合のみ)
   *              *returns {ResultType}
   *                       結果
   * @returns {Promise.<Tuple.<?ResultType,nsIFilePicker>>}
   * @resolves 結果とファイルピッカー
   *           キャンセルの場合は結果が null
   */
  _getFileHelper: Task.async(function*(window, params, mode, getResult) {
    let uri = null;
    if ("uri" in params) {
      try {
        uri = Services.io.newURI(params.uri, null, null);
      } catch (e) {
      }
    }

    let getParam = (name, defValue) => {
      return (name in params) ? params[name] : defValue;
    };

    let useLastDir           = getParam("useLastDir", false);
    let displayDirectoryPath = getParam("displayDirectoryPath", null);

    let title                = getParam("title", "");
    let defaultString        = getParam("defaultString", null);
    let defaultExtension     = getParam("defaultExtension", null);
    let filter               = getParam("filter", null);
    let filterIndex          = getParam("filterIndex", -1);

    let dirPath = null;
    let _downloadLastDir = null;

    if (mode == Ci.nsIFilePicker.modeSave) {
      let { DownloadLastDir } = Cu.import("resource://gre/modules/DownloadLastDir.jsm", {});
      _downloadLastDir = new DownloadLastDir(window);
    }

    if (useLastDir) {
      if (mode == Ci.nsIFilePicker.modeSave) {
        dirPath = yield this.getLastSaveDirPath(window, uri,
                                                _downloadLastDir);
      } else {
        dirPath = this.getLastOpenDirPath(window);
      }
    } else if (displayDirectoryPath) {
      let exists = yield OS.File.exists(displayDirectoryPath);
      if (exists) {
        let stat = yield OS.File.stat(displayDirectoryPath);
        if (stat.isDir) {
          dirPath = displayDirectoryPath;
        } else {
          dirPath = OS.Path.dirname(displayDirectoryPath);
        }
      } else {
        dirPath = null;
      }
    }

    let filePicker = this._initFilePicker(window,
                                          title,
                                          dirPath,
                                          defaultString,
                                          defaultExtension,
                                          filter, filterIndex,
                                          mode, uri);

    let ret = yield new Promise((resolve, reject) => {
      let callback = {
        QueryInterface: XPCOMUtils.generateQI([Ci.nsIFilePickerShownCallback]),

        done: Task.async(function*(pickerResult) {
          if (!(pickerResult == Ci.nsIFilePicker.returnOK ||
                pickerResult == Ci.nsIFilePicker.returnReplace)) {
            resolve([null, filePicker]);
            return;
          }

          let [result, dir] = yield getResult(filePicker);
          if (mode == Ci.nsIFilePicker.modeSave) {
            this.setLastSaveDirPath(window, uri, dir, _downloadLastDir);
          } else {
            this.setLastOpenDirPath(window, dir);
          }

          resolve([result, filePicker]);
        }.bind(this))
      };
      filePicker.open(callback);

    });

    return ret;
  })
});

module.arFilePicker = arFilePicker;

})(this);
