"use strict";

this.EXPORTED_SYMBOLS = ["arFileUtils",
                         "arFilePermission"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Downloads } = Cu.import("resource://gre/modules/Downloads.jsm", {});
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { fromUint8Array, toUint8Array } = Cu.import("resource://unmht/modules/base/arArrayBufferUtils.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyServiceGetter(lazy, "FileProtocolHandler",
                                   "@mozilla.org/network/protocol;1?name=file",
                                   "nsIFileProtocolHandler");

/**
 * PR_Open の定数
 * @enum {number}
 */
const FileOpenMode = Object.freeze({
  PR_RDONLY:      0x01,
  PR_WRONLY:      0x02,
  PR_CREATE_FILE: 0x08,
  PR_APPEND:      0x10,
  PR_TRUNCATE:    0x20,
  PR_SYNC:        0x40,
  PR_EXCL:        0x80
});

/**
 * パーミッション
 * @enum {number}
 */
const arFilePermission = Object.freeze({
  /** 通常のファイル */
  NORMAL_FILE: 0o0644,
  /** 実行ファイル */
  EXEC_FILE:   0o0755,
  /** ディレクトリ */
  DIRECTORY:   0o0755,
});

/**
 * ファイル名が長すぎる事によるエラーかどうかを返す
 * Windows では非対応
 *
 * @param   {OS.File.Error} e
 *          エラー
 * @returns {boolean}
 *          結果
 */
function becauseTooLong(e) {
  if ("unixErrno" in e && e.unixErrno == OS.Constants.libc.ENAMETOOLONG) {
    return true;
  }
  return false;
}

/**
 * ファイル操作ユーティリティ
 *
 * @class
 */
let arFileUtils = Object.freeze({
  /**
   * ファイルを取得する
   *
   * @param   {string} path
   *          ファイルのパス
   * @returns {?nsIFile}
   *          指定されたパスのファイル
   *          失敗した場合は null
   */
  getFile: function(path) {
    let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
    try {
      file.initWithPath(path);
    } catch (e) {
      return null;
    }

    return file;
  },

  /**
   * ファイルを用意する
   * ファイル名が長すぎる場合切り詰め、同名のファイルが存在する場合
   * 必要であればユニークなファイル名を算出する
   *
   * @param   {string} parent
   *          親ディレクトリのパス
   * @param   {string} base
   *          ファイル名のベース部分
   *          ファイル名が長すぎる場合省略される
   * @param   {string} suffix
   *          拡張子など、ファイル名の末尾に必ず付加する文字列
   *          ファイル名が長すぎても省略されない
   * @param   {object=} options
   *          {
   *            allowExists: {boolean}
   *                         同名のファイルが存在してもそのまま返す
   *            noCreate:    {boolean}
   *                         ファイルを作成した場合は削除する
   *            indexInit:   {Number}
   *                         同名のファイルが存在した場合のインデックスの初期値
   *            indexDigits: {Number}
   *                         同名のファイルが存在した場合のインデックスの桁数
   *          }
   * @returns {Promise.<object>}
   * @resolves
   *          {
   *            path:    {string}
   *                     ファイルのパス
   *            exists:  {boolean}
   *                     存在したか
   *            created: {boolean}
   *                     作成したか
   *                     作成したが即座に削除した場合は false
   *          }
   */
  prepareFile: Task.async(function*(parent, base, suffix, options) {
    let allowExists = true;
    if (options && 'allowExists' in options) {
      allowExists = options.allowExists;
    }
    let noCreate = false;
    if (options && 'noCreate' in options) {
      noCreate = options.noCreate;
    }

    let pad = (width, n) => {
      let s = n.toString(10);
      return "0".repeat(Math.max(width - s.length, 0)) + s;
    };

    let MAX_FILENAME_LENGTH = 255;
    let MAX_WIN_PATH_LENGTH = 259;

    let suffixLen = suffix.length;
    let origLen = Math.min(base.length + suffixLen, MAX_FILENAME_LENGTH);

    let len = origLen;
    let min = 0;
    let max = origLen;

    let tooLong = false;
    let exists = false;
    let created = false;

    let resultPath = "";

    let basename = base.slice(0, len - suffixLen) + suffix;
    let path = OS.Path.join(parent, basename);

    if (Services.appinfo.OS == "WINNT" &&
        path.length > MAX_WIN_PATH_LENGTH) {
      max -= path.length - MAX_WIN_PATH_LENGTH;
      len = max;

      basename = base.slice(0, len - suffixLen) + suffix;
      path = OS.Path.join(parent, basename);
    }

    try {
      /* まずは元のファイル名で作成を試す */
      let file = yield OS.File.open(path, { create: true });
      yield file.close();
      created = true;
      resultPath = path;
    } catch (e) {
      if (e instanceof OS.File.Error) {
        if (becauseTooLong(e)) {
          tooLong = true;
        } else if (e.becauseExists) {
          exists = true;
          if (allowExists) {
            resultPath = path;
          }
        } else {
          throw e;
        }
      } else {
        throw e;
      }
    }

    if (tooLong || (!allowExists && exists)) {
      /* ファイル名を探すための一時ファイルを作成する */
      let tmpPath = "";
      for (let i = 0; i < 10; i += 1) {
        let tmp = "." + Date.now() + "-" + Math.round(Math.random() * 1000000);
        tmpPath = OS.Path.join(parent, tmp);
        try {
          let file = yield OS.File.open(tmpPath, { create: true });
          yield file.close();
          created = true;
          break;
        } catch (e) {
          if (e instanceof OS.File.Error) {
            if (e.becauseExists) {
            } else {
              throw e;
            }
          } else {
            throw e;
          }
        }
      }
      if (!created) {
        throw new Error("Failed to create temporary file.");
      }

      if (tooLong) {
        /* 元のファイル名が長すぎる場合
         * 許容される最長のファイル名を算出する */
        for (;;) {
          let ok = false;

          let basename = base.slice(0, len - suffixLen) + suffix;
          let path = OS.Path.join(parent, basename);
          try {
            if (tmpPath != path) {
              yield OS.File.move(tmpPath, path, { noOverwrite: true });
            }

            tmpPath = path;
            exists = false;
            ok = true;
          } catch (e) {
            if (e instanceof OS.File.Error) {
              if (becauseTooLong(e)) {
                max = len;
                len = Math.floor((len + min) / 2);
              } else if (e.becauseExists) {
                exists = true;
                ok = true;
              } else {
                throw e;
              }
            } else {
              throw e;
            }
          }

          if (ok) {
            if (len == min && min == max - 1) {
              if (exists && allowExists) {
                /* 使わなくなった一時ファイルを削除する */
                yield OS.File.remove(tmpPath);
                tmpPath = null;
                created = false;

                resultPath = path;
              } else if (!exists) {
                tmpPath = null;
                created = true;

                resultPath = path;
              }
              break;
            }

            min = len;
            len = Math.floor((len + max) / 2);
          }
        }
      }

      if (exists && !allowExists) {
        /* 存在しないファイル名を探す */
        let indexInit = 2;
        let indexDigits = 1;
        if (options && "indexInit" in options) {
          indexInit = options.indexInit;
        }
        if (options && "indexDigits" in options) {
          indexDigits = options.indexDigits;
        }

        let index = indexInit;
        for (;;) {
          let basename = base.slice(0, len - suffixLen)
            + "-" + pad(indexDigits, index) + suffix;
          let over = basename.length - MAX_FILENAME_LENGTH;
          if (over > 0) {
            len -= over;
            if (len <= 0) {
              throw new Error("Failed to generate filename");
            }
            basename = base.slice(0, len - suffixLen)
              + "-" + pad(indexDigits, index) + suffix;
          }
          let path = OS.Path.join(parent, basename);

          try {
            if (tmpPath != path) {
              yield OS.File.move(tmpPath, path, { noOverwrite: true });
            }
            exists = false;
            resultPath = path;
            break;
          } catch (e) {
            if (e instanceof OS.File.Error) {
              if (becauseTooLong(e)) {
                /* エンコーディングやノーマライズの結果
                 * ファイル名中の 1 文字と添字の 1 文字が同じ文字数で
                 * カウントされるとも限らないので 1 文字ずつ減らす */
                len -= 1;
                if (len <= 0) {
                  throw new Error("Failed to generate filename");
                }
              } else if (e.becauseExists) {
                index += 1;
              } else {
                throw e;
              }
            } else {
              throw e;
            }
          }
        }
      }
    }

    if (created && noCreate) {
      /* 作成したファイルを削除する */
      created = false;
      yield OS.File.remove(resultPath);
    }

    return {
      path: resultPath,
      exists: exists,
      created: created
    };
  }),

  /**
   * ファイルを作成する
   *
   * @param   {string} path
   *          対象のファイルのパス
   * @param   {string} data
   *          ファイルの内容
   * @param   {boolean=} executable
   *          実行可能か
   *          undefined の場合は false
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  createBinary: function(path, data, executable) {
    return this.createFromArray(path, toUint8Array(data), executable);
  },

  /**
   * UTF8 のファイルを作成する
   *
   * @param   {string} path
   *          対象のファイルのパス
   * @param   {string} data
   *          ファイルの内容
   * @param   {boolean=} executable
   *          実行可能か
   *          undefined の場合は false
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  createUTF8: function(path, data, executable) {
    let encoder = new TextEncoder("utf-8");
    let arr = encoder.encode(data);
    return this.createFromArray(path, arr, executable);
  },

  /**
   * ファイルを作成する
   *
   * @param   {string} path
   *          対象のファイルのパス
   * @param   {Uint8Array} data
   *          ファイルの内容
   * @param   {boolean=} executable
   *          実行可能か
   *          undefined の場合は false
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  createFromArray: Task.async(function*(path, data, executable) {
    let options;
    if (executable) {
      options = {
        unixMode: OS.Constants.libc.S_IRWXU
      };
    }
    let file = yield OS.File.open(path, { write: true, trunc: true },
                                  options);

    if (data.length > 0) {
      yield file.write(data);
    }

    yield file.close();
  }),

  /**
   * ファイルからデータを読み込む
   *
   * @param   {string} path
   *          対象のファイルのパス
   * @returns {Promise.<string>}
   * @resolves ファイルのデータ
   */
  readBinary: Task.async(function*(path) {
    let data = yield this.readAsArray(path);
    return fromUint8Array(data);
  }),

  /**
   * ファイルからデータを読み込む
   *
   * @param   {string} path
   *          対象のファイルのパス
   * @returns {Promise.<Uint8Array>}
   * @resolves ファイルのデータ
   */
  readAsArray: function(path) {
    return OS.File.read(path);
  },

  /**
   * ファイルから UTF8 のデータを読み込む
   *
   * @param   {string} path
   *          対象のファイルのパス
   * @returns {Promise.<string>}
   * @resolves ファイルのデータ
   */
  readUTF8: Task.async(function*(path) {
    let arr = yield this.readAsArray(path);

    let decoder = new TextDecoder("utf-8");
    return decoder.decode(arr);
  }),

  /**
   * Zip 書き込みを作成する
   *
   * @param   {string} path
   *          対象のファイルのパス
   * @returns {?nsIZipWriter}
   *          ZIP 書き込み
   *          失敗した場合は null
   */
  openZipWriter: function(path) {
    let zip = Cc["@mozilla.org/zipwriter;1"].createInstance(Ci.nsIZipWriter);

    try {
      zip.open(this.getFile(path),
               FileOpenMode.PR_WRONLY |
               FileOpenMode.PR_CREATE_FILE |
               FileOpenMode.PR_TRUNCATE);
    } catch (e) {
      return null;
    }

    return zip;
  },

  /**
   * 現在の OS でファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @returns {string}
   *          変換したファイル名
   */
  validateFileName: function(filename) {
    return this.validateFileNameWithOS(filename, Services.appinfo.OS);
  },

  /**
   * 指定した OS でファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @param   {string} OS
   *          対象の OS
   * @returns {string}
   *          変換したファイル名
   */
  validateFileNameWithOS: function(filename, OS) {
    let sepRe;
    if (OS == "WINNT") {
      sepRe = /[\\\/]/g;

      /* ファイル名に使えない文字
       *   |
       *   *
       *   :
       *   ?
       *   "
       *   <
       *   >
       */

      filename = filename.replace(/\"/g, "'")
        .replace(/[\*\:\?\|]/g, "_")
        .replace(/</g, "(")
        .replace(/>/g, ")");
    } else if (OS == "Darwin") {
      /* ファイル名に使えない文字
       *   :
       *   /
       */
      sepRe = /[\:\/]/g;
    } else {
      /* ファイル名に使えない文字
       *   /
       */
      sepRe = /\//g;
    }

    return filename.replace(sepRe, "_");
  },

  /**
   * file プロトコルの URL をファイルのパスに変換する
   *
   * @param   {string} uri
   *          file プロトコル
   * @returns {string}
   *          ファイルのパス
   */
  URLSpecToPath: function(uri) {
    return lazy.FileProtocolHandler.getFileFromURLSpec(uri).path;
  },

  /**
   * ファイルのパスを file プロトコルの URL に変換する
   *
   * @param   {string} path
   *          ファイル
   * @returns {string}
   *          file プロトコル
   */
  pathToURLSpec: function(file) {
    return lazy.FileProtocolHandler.getURLSpecFromFile(this.getFile(file));
  },

  /* ---- ディレクトリ取得 ---- */

  /**
   * ユーザの Applications ディレクトリのパス
   * Darwin 専用
   * @type {?string}
   */
  get userApplicationsDirectory() {
    try {
      return Services.dirsvc.get("UsrApp", Ci.nsIFile).path;
    } catch (e) {
    }

    return null;
  },

  /**
   * Program Files ディレクトリのパス
   * WINNT 専用
   * @type {?string}
   */
  get programFilesDirectory() {
    try {
      return Services.dirsvc.get("ProgF", Ci.nsIFile).path;
    } catch (e) {
    }

    return null;
  },

  /**
   * 現在のユーザのダウンロードディレクトリを取得する
   *
   * @return {Promise.<string>}
   * @resolves 現在のユーザのダウンロードディレクトリのパス
   */
  getDefaultDownloadsDirectory: function() {
    return Downloads.getPreferredDownloadsDirectory();
  }
});

module.arFileUtils = arFileUtils;
module.arFilePermission = arFilePermission;

})(this);
