"use strict";

this.EXPORTED_SYMBOLS = ["arListView"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arArrayUtils } = Cu.import("resource://unmht/modules/base/arArrayUtils.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arScopeManager } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});

/**
 * ドラッグアンドドロップ時のマーカー
 *
 * @constructor
 * @param   {nsIDOMXULElement} node
 *          tree 要素
 */
function arListViewDnDMarker(tree) {
  let document = tree.ownerDocument;

  /**
   * マーカー要素
   * @type {nsIDOMXULElement}
   */
  this._marker = document.createElement("panel");

  tree.parentNode.appendChild(this._marker);

  this._marker.style.mozAappearance = "none";
  this._marker.style.mozWindowShadow = "none";
  this._marker.style.width = "200px";
  this._marker.style.height = "1px";
  this._marker.style.border = "none";
  this._marker.style.margin = "0";
  this._marker.style.padding = "0";
  this._marker.setAttribute("level", "top");
  this._marker.setAttribute("noautofocus", "true");
  this._marker.setAttribute("noautohide", "true");

  let ns = "http://www.w3.org/1999/xhtml";
  let sep = document.createElementNS(ns, "div");
  sep.style.width = "100%";
  sep.style.height = "1px";
  sep.style.backgroundColor = "#000000";
  this._marker.appendChild(sep);

  Object.seal(this);
}
arListViewDnDMarker.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    if (this._marker) {
      this._marker.remove();
      this._marker = null;
    }
  },

  /**
   * マーカーを表示する
   *
   * @param   {number} x
   *          左上の X 座標 (スクリーン座標) [px]
   * @param   {number} y
   *          左上の Y 座標 (スクリーン座標) [px]
   * @param   {number} width
   *          幅 [px]
   */
  show: function(x, y, width) {
    this._marker.style.width = width + "px";
    this._marker.openPopupAtScreen(x, y, false);
    this._marker.moveTo(x, y);
  },

  /**
   * マーカーを消す
   */
  hide: function() {
    this._marker.hidePopup();
  }
});

/**
 * データのリストビューによる操作
 *
 * @constructor
 * @template DataType
 * @param   {nsIDOMXULElement} tree
 *          tree 要素
 * @param   {?nsIDOMXULElement} cols
 *          treecols 要素
 *          ソートを必要としない場合は null
 * @param   {nsIDOMXULElement} treechildren
 *          treechildren 要素
 * @param   {nsIDOMXULElement} newButton
 *          追加ボタン
 * @param   {?nsIDOMXULElement} dupButton
 *          複製ボタン
 *          無い場合 null
 * @param   {nsIDOMXULElement} delButton
 *          削除ボタン
 * @param   {string} type
 *          ドラッグ中の行の MIME-Type
 * @param   {Array.<number>} ignoreColumns
 *          ドラッグアンドドロップの対象にしないカラムの位置
 * @param   {boolean} isEditable
 *          tree が直接編集可能か
 * @param   {Array.<DataType>} datas
 *          表示するデータ
 * @param   {arIListViewDataHandler} dataHandler
 *          データハンドラ
 */
function arListView(tree, cols, treechildren,
                    newButton, dupButton, delButton,
                    type, ignoreColumns, isEditable,
                    datas,
                    dataHandler) {
  /**
   * tree 要素
   * @type {nsIDOMXULElement}
   */
  this._tree = tree;

  /**
   * treecols 要素
   * @type {?nsIDOMXULElement}
   */
  this._cols = cols;

  /**
   * treechildren 要素
   * @type {nsIDOMXULElement}
   */
  this._treechildren = treechildren;

  /**
   * 追加ボタン
   * @type {nsIDOMXULElement}
   */
  this._newButton = newButton;

  /**
   * 複製ボタン
   * @type {?nsIDOMXULElement}
   */
  this._dupButton = dupButton;

  /**
   * 削除ボタン
   * @type {nsIDOMXULElement}
   */
  this._delButton = delButton;

  /**
   * ドラッグ中の行の MIME-Type
   * @type {string}
   */
  this._type = type;

  /**
   * ドラッグアンドドロップの対象にしないカラムの位置
   * @type {Array.<number>}
   */
  this._ignoreColumns = ignoreColumns;

  /**
   * tree が直接編集可能か
   * @type {boolean}
   */
  this._isEditable = isEditable;

  /**
   * 表示するデータ
   * @type {Array.<DataType>}
   */
  this._datas = datas;

  /**
   * データハンドラ
   * @type {arIListViewDataHandler}
   */
  this._dataHandler = dataHandler;

  /**
   * 各カラムについてソートが昇順であるか
   * @type {Array.<boolean>}
   */
  this._sortAsc = [];

  /**
   * ドラッグアンドドロップ時のマーカー
   * @type {arListViewDnDMarker}
   */
  this._marker = new arListViewDnDMarker(this._tree);

  /**
   * onselect イベントをキャッチしない
   * @type {boolean}
   */
  this._doNotCatchSelected = false;

  /**
   * ドラッグ先の行
   * @type {number}
   */
  this._currentRow = -1;

  /**
   * ドラッグ先の行に対する位置
   * @type {number}
   */
  this._orientation = 0;

  this._init();

  Object.seal(this);
}
arListView.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    arScopeManager.leave(this);

    this._tree = null;
    this._cols = null;
    this._treechildren = null;
    this._newButton = null;
    this._dupButton = null;
    this._delButton = null;
    this._type = null;
    this._ignoreColumns = null;
    this._dataHandler = null;
    this._datas = null;

    this._marker.destruct();
    this._marker = null;
  },

  /**
   * 選択中のデータを取得する
   *
   * @returns {Array.<Tuple.<DataType,number>>}
   *          選択中のデータと行のインデックスのリスト
   */
  getSelectedItems: function() {
    return this._datas
      .map((item, i) => [item, i])
      .filter(([item, i]) => this._tree.view.selection.isSelected(i));
  },

  /**
   * 非選択中のデータを取得する
   *
   * @returns {Array.<Tuple.<DataType,number>>}
   *          非選択中のデータと行のインデックスのリスト
   */
  getUnselectedItems: function() {
    return this._datas
      .map((item, i) => [item, i])
      .filter(([item, i]) => !this._tree.view.selection.isSelected(i));
  },

  /**
   * 全てのデータを更新する
   *
   * @param   {Array.<DataType>} datas
   *          表示するデータ
   */
  setDatas: function(datas) {
    this._datas = datas;

    /* tree を直接編集中かもしれないので不用意に要素を削除してはいけない */

    this._doNotCatchSelected = true;
    if (this._treechildren.childNodes.length > this._datas.length) {
      for (let count = this._treechildren.childNodes.length - datas.length;
           count > 0; count -= 1) {
        this._treechildren.removeChild(this._treechildren.lastChild);
      }
    }
    this._doNotCatchSelected = false;

    this._datas.forEach((item, i) => {
      if (i < this._treechildren.childNodes.length) {
        this._changeRow(item, i);
      } else {
        this._addRow(item, -1, false);
      }
    });

    if (this._datas.length) {
      this.select(0);
    }
  },

  /**
   * データを追加する
   *
   * @param   {DataType} item
   *          対応するデータ
   * @param   {number} pos
   *          追加する位置
   *          -1 にすると末尾
   * @param   {boolean} _select
   *          追加した行を選択するか
   */
  addData: function(item, pos, _select) {
    if (pos < 0 || pos > this._datas.length) {
      pos = this._datas.length;
    }

    this._datas
      = [...this._datas.slice(0, pos), item, ...this._datas.slice(pos)];

    this._addRow(item, pos, _select);

    try {
      this._dataHandler.onModify(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /**
   * データを選択する
   *
   * @param   {number} pos
   *          行のインデックス
   */
  select: function(pos) {
    if (pos < 0 && pos >= this._treechildren.childNodes.length) {
      return;
    }

    this._tree.view.selection.select(pos);
  },

  /**
   * 複数データを選択する
   *
   * @param   {Array.<number>} selectedIndices
   *          行のインデックス
   */
  selectRange: function(selectedIndices) {
    selectedIndices.forEach((pos, i) => {
      if (pos < 0 && pos >= this._treechildren.childNodes.length) {
        return;
      }

      this._doNotCatchSelected = true;
      if (i === 0) {
        this._tree.view.selection.select(pos);
      } else {
        this._tree.view.selection.rangedSelect(pos, pos, true);
      }
      this._doNotCatchSelected = false;
    });

    let items = this.getSelectedItems();

    try {
      this._dataHandler.onSelect(items);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /**
   * 選択イベントを強制的に発行する
   * イベント中の場合発行されないため
   */
  dispatchSelect: function() {
    let items = this.getSelectedItems();

    try {
      this._dataHandler.onSelect(items);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /**
   * 選択中のデータを更新する
   *
   * @param   {function(DataType)} modifyFunc
   *          選択中のデータを更新する関数
   *            function(item)
   *              *param   {DataType} item
   *                       選択中のデータ
   * @returns {?DataType}
   *          選択中のデータ
   *          選択していない、または複数選択している場合は null
   */
  modifySelectedData: function(modifyFunc) {
    if (this._tree.view.selection.count != 1) {
      return null;
    }

    return this.modifyDataAt(this._tree.view.selection.currentIndex, modifyFunc);
  },

  /**
   * データを更新する
   *
   * @param   {function(DataType)} modifyFunc
   *          選択中のデータを更新する関数
   *            function(item)
   *              *param   {DataType} item
   *                       選択中のデータ
   * @param   {number} index
   *          行のインデックス
   * @returns {?DataType}
   *          対象のデータ
   *          選択していない、または複数選択している場合は null
   */
  modifyDataAt: function(index, modifyFunc) {
    if (index < 0 || index >= this._datas.length) {
      return null;
    }

    let item = this._datas[index];
    modifyFunc(item);

    let treerow = this._treechildren.childNodes[index].firstChild;

    this._dataHandler.dataToRow(item)
      .forEach((cellData, i) => this._updateCell(treerow.childNodes[i], cellData));

    try {
      this._dataHandler.onModify(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }

    return item;
  },

  /* ---- UI イベント ---- */

  /**
   * 追加ボタンをクリックしたイベント
   * 選択中のデータの次にデータを追加する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onNewButtonClick: function(event) {
    let pos = this._tree.view.selection.currentIndex + 1;

    let item = this._dataHandler.newData();

    this.addData(item, pos, true);
  },

  /**
   * 追加ボタンをクリックしたイベント
   * 選択中のデータの次にデータを追加する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDupButtonClick: function(event) {
    let items = this.getSelectedItems();

    let n = 0;

    let selectedIndices
      = items.map(([origItem, origPos]) => {
        let item = this._dataHandler.dupData(origItem);
        let pos = n + origPos + 1;

        this.addData(item, pos, false);

        n += 1;

        return pos;
      });

    this.selectRange(selectedIndices);
  },

  /**
   * 削除ボタンをクリックしたイベント
   * 選択中のデータを削除する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDelButtonClick: function(event) {
    let selectedItems = this.getSelectedItems();
    let unselectedItems = this.getUnselectedItems();

    for (let row of selectedItems
         .map(([item, i]) => this._treechildren.childNodes[i])) {
      row.remove();
    }

    this._datas = unselectedItems.map(([item, i]) => item);

    try {
      this._dataHandler.onModify(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }

    if (selectedItems.length) {
      this.select(selectedItems[0][1]);
    }
  },

  /**
   * ツリーのカラムをクリックしたイベント
   * そのカラムでソートする
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeColClick: function(event) {
    if (!this._cols) {
      return;
    }

    if (this._sortAsc.length === 0) {
      this._sortAsc
        = [...arArrayUtils.range(0, this._cols.childNodes.length)]
        .map(() => false);
    }

    let pos = Array.indexOf(this._cols.getElementsByTagName("treecol"),
                            event.target);
    if (pos == -1) {
      return;
    }

    let descFunc = (a, b) => a[0] < b[0] ? 1 : -1;
    let ascFunc  = (a, b) => a[0] > b[0] ? 1 : -1;
    let sortFunc = descFunc;

    let asc = this._sortAsc[pos];
    Array.forEach(this._cols.childNodes, (col, i) => {
      if (i == pos) {
        col.setAttribute("sortDirection",
                         asc ? "ascending" : "descending");
      } else {
        col.setAttribute("sortDirection", "natural");
      }
    });
    if (this._sortAsc[pos]) {
      sortFunc = ascFunc;
    }
    this._sortAsc[pos] = !this._sortAsc[pos];

    this._doNotCatchSelected = true;
    this._datas
      = this._datas
      .map((item, i) => {
        let key;

        try {
          key = this._dataHandler.getSortKeyFor(item, pos);
        } catch (e) {
          Cu.reportError(e + "\n" + e.stack);
        }

        return [key, item, this._treechildren.childNodes[i]];
      })
      .sort(sortFunc)
      .map(([label, item, node]) => {
        this._treechildren.appendChild(node);
        return item;
      });
    this._doNotCatchSelected = false;

    try {
      this._dataHandler.onModify(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /**
   * ツリーの行をドラッグ開始したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeDragStart: function(event) {
    if (event.target != this._treechildren) {
      return;
    }

    let row = {}, col = {}, child = {};

    this._tree.treeBoxObject.getCellAt(event.clientX, event.clientY,
                                       row, col, child);

    if (col.value) {
      if (this._ignoreColumns.indexOf(col.value.index) != -1) {
        return;
      }
    }

    let items = this.getSelectedItems();
    if (items.length === 0) {
      return;
    }
    let firstRow = this._treechildren.childNodes[items[0][1]];

    event.dataTransfer.mozSetDataAt(this._type, JSON.stringify(items), 0);
    event.dataTransfer.mozSetDataAt(this._type + "-node", firstRow, 1);
    event.stopPropagation();
  },

  /**
   * ツリーの上でドラッグしているイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeDragOver: function(event) {
    if (event.target != this._treechildren) {
      return;
    }

    let draggingItems = this._getDraggingItems(event.dataTransfer);
    if (!draggingItems) {
      this._currentRow = -1;
      return;
    }

    let row = {}, col = {}, child = {};

    this._tree.treeBoxObject.getCellAt(event.clientX, event.clientY,
                                       row, col, child);

    this._currentRow = row.value;
    if (this._currentRow == -1) {
      this._currentRow = this._treechildren.childNodes.length - 1;
    }

    let tbo = this._tree.treeBoxObject;
    let rowHeight = tbo.rowHeight;
    let eventY
      = event.clientY - tbo.treeBody.boxObject.y
      - rowHeight * (this._currentRow - tbo.getFirstVisibleRow());

    let x = {}, y = {};
    tbo.getCoordsForCellItem(this._currentRow, tbo.columns[0], "cell",
                             x, y, {}, {});

    x.value += event.screenX - event.clientX + this._treechildren.boxObject.x;
    y.value += event.screenY - event.clientY + this._treechildren.boxObject.y;

    this._orientation = 0;
    if (row.value == -1 || eventY > rowHeight * 0.5) {
      y.value += rowHeight;
      this._orientation = 1;
    }
    this._marker.show(x.value, y.value, this._treechildren.boxObject.width);

    event.preventDefault();
    event.stopPropagation();
  },

  /**
   * ツリーの上にドロップしたイベント
   * ドロップした位置にデータを移動する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeDragEnd: function(event) {
    if (this._currentRow == -1) {
      return;
    }

    this._marker.hide();

    let draggingItems = this._getDraggingItems(event.dataTransfer);
    if (!draggingItems || draggingItems.length === 0) {
      this._currentRow = -1;
      return;
    }

    let markerPos = this._currentRow + this._orientation;

    let nextPos = markerPos;
    while (draggingItems.some(([item, pos]) => pos == nextPos)) {
      nextPos += 1;
    }
    let nodes = this._treechildren.childNodes;
    let nextRow = (nextPos < nodes.length) ? nodes[nextPos] : null;

    let insertPos
      = this._datas
      .filter((item, i) => (draggingItems.every(([item, pos]) => pos != i) &&
                            i < nextPos)).length;

    let tmpData
      = this._datas
      .filter((item, i) => draggingItems.every(([item, j]) => i != j));
    let nextData = tmpData.splice(insertPos);

    this._datas
      = [...tmpData,
         ...draggingItems.map(([item, i]) => item),
         ...nextData];

    this._doNotCatchSelected = true;
    for (let row of draggingItems
         .map(([item, i]) => this._treechildren.childNodes[i])) {
      row.parentNode.insertBefore(row, nextRow);
    }
    this._doNotCatchSelected = false;

    try {
      this._dataHandler.onModify(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }

    this.selectRange(draggingItems.map((_tmp, i) => insertPos + i));
  },

  /**
   * ツリーの行を選択したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeSelect: function(event) {
    if (this._doNotCatchSelected) {
      return;
    }

    let items = this.getSelectedItems();

    try {
      this._dataHandler.onSelect(items);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /**
   * ツリーのテキストが編集されたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeChange: function(event) {
    this._datas
      = this._datas.map((item, i) => {
        let row = [];
        for (let j = 0, len = this._tree.columns.length; j < len; j += 1) {
          let cell = {
            label: this._tree.view.getCellText(i, this._tree.columns[j])
          };
          row.push(cell);
        }

        return this._dataHandler.rowToData(row, item);
      });

    try {
      this._dataHandler.onModify(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /* ==== private ==== */

  /**
   * 初期化処理
   */
  _init: function() {
    if (this._cols) {
      for (let col of Array.from(this._cols.getElementsByTagName("treecol"))) {
        arEvent.add(this, col, "click", this.onTreeColClick.bind(this), false);
      }
    }

    arEvent.add(this, this._tree, "dragstart", this.onTreeDragStart.bind(this), false);
    arEvent.add(this, this._tree, "dragover", this.onTreeDragOver.bind(this), false);
    arEvent.add(this, this._tree, "dragend", this.onTreeDragEnd.bind(this), false);

    arEvent.add(this, this._tree, "select", this.onTreeSelect.bind(this), false);
    if (this._isEditable) {
      arEvent.add(this, this._tree, "change", this.onTreeChange.bind(this), false);
    }

    arEvent.add(this, this._newButton, "command", this.onNewButtonClick.bind(this), false);
    if (this._dupButton) {
      arEvent.add(this, this._dupButton, "command", this.onDupButtonClick.bind(this), false);
    }
    arEvent.add(this, this._delButton, "command", this.onDelButtonClick.bind(this), false);

    this.setDatas(this._datas);
  },

  /**
   * ドラッグ中の行データを取得する
   *
   * @param   {DataTransfer} dataTransfer
   *          ドラッグのデータ
   * @returns {Array.<Tuple.<DataType,number>>}
   *          データと行のインデックスの配列
   */
  _getDraggingItems: function(dataTransfer) {
    for (let i = 0, len = dataTransfer.mozItemCount; i < len; i += 1) {
      if (Array.indexOf(dataTransfer.mozTypesAt(i), this._type) != -1) {
        return JSON.parse(dataTransfer.mozGetDataAt(this._type, i));
      }
    }

    return null;
  },

  /**
   * 行を追加する
   *
   * @param   {DataType} item
   *          対応するデータ
   * @param   {number} pos
   *          追加する位置 (行のインデックス)
   *          -1 にすると末尾
   * @param   {boolean} _select
   *          追加した行を選択するか
   */
  _addRow: function(item, pos, _select) {
    let document = this._tree.ownerDocument;

    let nextSibling = null;
    if (pos >= 0 && pos < this._treechildren.childNodes.length) {
      nextSibling = this._treechildren.childNodes[pos];
    }

    let treeitem = document.createElement("treeitem");
    this._doNotCatchSelected = true;
    this._treechildren.insertBefore(treeitem, nextSibling);
    this._doNotCatchSelected = false;

    let treerow = document.createElement("treerow");
    treeitem.appendChild(treerow);

    for (let cellData of this._dataHandler.dataToRow(item)) {
      treerow.appendChild(this._updateCell(document.createElement("treecell"), cellData));
    }

    if (_select) {
      this.select(pos);
    }
  },

  /**
   * 行を修正する
   *
   * @param   {DataType} item
   *          対応するデータ
   * @param   {number} pos
   *          修正する位置 (行のインデックス)
   */
  _changeRow: function(item, pos) {
    let treeitem = null;
    if (pos >= 0 && pos < this._treechildren.childNodes.length) {
      treeitem = this._treechildren.childNodes[pos];
    } else {
      return;
    }

    let cells = treeitem.getElementsByTagName("treecell");

    this._dataHandler.dataToRow(item)
      .forEach((cellData, i) => this._updateCell(cells[i], cellData));
  },

  /**
   * セルの表示を更新する
   *
   * @param   {nsIDOMXULElement} cell
   *          セル
   * @param   {Object} cellData
   *          セルのデータ
   * @returns {nsIDOMXULElement}
   *          セル
   */
  _updateCell: function(cell, cellData) {
    if ("label" in cellData) {
      if (!cell.hasAttribute("label") ||
          cell.getAttribute("label") != cellData.label) {
        cell.setAttribute("label", cellData.label);
      }
    } else {
      if (cell.hasAttribute("label")) {
        cell.removeAttribute("label");
      }
    }

    if ("src" in cellData) {
      if (!cell.hasAttribute("src") ||
          cell.getAttribute("src") != cellData.src) {
        cell.setAttribute("src", cellData.src);
      }
    } else {
      if (cell.hasAttribute("src")) {
        cell.removeAttribute("src");
      }
    }

    if ("className" in cellData) {
      if (!("className" in cell) ||
          cell.className != cellData.className) {
        cell.className = cellData.className;
      }
    } else {
      if ("className" in cell &&
          cell.className !== "") {
        cell.className = "";
      }
    }

    return cell;
  }
});

module.arListView = arListView;

})(this);
