"use strict";

this.EXPORTED_SYMBOLS = ["arMIMEDecoder"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arMIMEParser, arMIMEParserError, safe_atob } = Cu.import("resource://unmht/modules/base/arMIMEParser.jsm", {});
let { arUconv, fromUTF8, toUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});

/**
 * MIME デコーダ
 *
 * @class
 */
let arMIMEDecoder = Object.freeze({
  /**
   * ext-octet[RFC2231] をデコードする
   *
   * @param   {string} text
   *          デコードする文字列
   * @returns {string}
   *          デコードした文字列
   */
  decodeExtOctet: function(text) {
    return text
      .replace(/%([A-Fa-f0-9]{2})/g,
               (matched, octet) => {
                 return String.fromCharCode(Number.parseInt(octet, 16));
               });
  },

  /**
   * quoted-printable[RFC2045] をデコードする
   *
   * @param   {string} text
   *          デコードする文字列
   * @returns {string}
   *          デコードした文字列
   */
  decodeQ: function(text) {
    /* quoted_printable never fail */
    let context = new arMIMEParser(text);
    return context.quoted_printable();
  },

  /**
   * encoded-word をデコードする
   *
   * @param   {string} text
   *          デコードする文字列
   * @returns {string}
   *          デコードした文字列 (Unicode)
   */
  decodeWord: function(text) {
    let ret = text;

    let context = new arMIMEParser(text);
    context._O(() => {
      let [, tmp] = context.encoded_word();
      context.__END();
      ret = tmp;
    });

    return ret;
  },

  /**
   * Base64 デコードする
   * 改行を含むものや破損したものにも対応
   *
   * @param   {string} text
   *          デコードする文字列
   * @returns {string}
   *          デコードされた文字列
   */
  decodeBase64: function(text) {
    return safe_atob(text);
  },

  /**
   * 日時を date-time[RFC5322] 形式等からデコードする
   * 実際は JavaScript の Date まかせなので色々デコードできる
   *
   * @param   {string} text
   *          エンコードされた日時
   * @returns {?Date}
   *          デコードされた日時
   *          失敗した場合は null
   */
  decodeDate: function(text) {
    let ret = null;

    let context = new arMIMEParser(text);
    context._O(() => {
      ret = context.orig_date__value();
      /* 不要な文字が含まれていても無視 */
    });

    if (!ret) {
      let tmp = new Date(text);
      if (!Number.isNaN(tmp.getFullYear())) {
        ret = tmp;
      }
    }

    return ret;
  },

  /**
   * Location ヘッダをデコードする
   *
   * @param   {string} text
   *          エンコードされた Location
   * @returns {?Date}
   *          デコードされた Location
   */
  decodeLocation: function(text) {
    let ret = null;

    let context = new arMIMEParser(text);
    context._O(() => {
      let tmp = context.content_location__value();
      context.__END();
      ret = tmp;
    });

    return ret;
  },

  /**
   * message [RFC5322] をデコードする
   *
   * @param   {string} text
   *          メッセージ
   *          改行コードは CR LF でなければならない
   * @returns {?arMIMEPart}
   *          トップレベルのパート
   *          データが不正ならば null
   */
  decodeMessage: function(text) {
    let part = null;

    let context = new arMIMEParser(text);
    context._O(() => {
      let tmp = context.message();
      context.__END();
      part = tmp;
    });

    if (!part) {
      return null;
    }

    this.decodeFields(part);

    if (part.isMultipart) {
      let isCorrupted = { value: false };
      part.parts = this._decodeMultipart(part.body, part.boundary, isCorrupted);
      part.isCorrupted = isCorrupted.value;
    } else {
      if (part.format == "flowed") {
        part.body = this.decodeFlowed(part.body, part.delsp);
      }

      if (part.contentTransferEncoding == "quoted-printable") {
        part.body = this.decodeQ(part.body);
      } else if (part.contentTransferEncoding == "base64") {
        part.body = this.decodeBase64(part.body);
      }
    }

    return part;
  },

  /**
   * フィールドをデコードする
   *
   * @param   {arMIMEPart} part
   *          パート
   */
  decodeFields: function(part) {
    if (part.hasField("Content-Type")) {
      let params = this.decodeParam(part.getField("Content-Type"));
      if (params) {
        part.contentTypeParams = params;
        part.contentType       = part.contentTypeParams.type.toLowerCase();
        part.contentSubType    = part.contentTypeParams.subtype.toLowerCase();
        part.mimetype          = part.contentType + "/" + part.contentSubType;
        if (part.contentTypeParams.hasParam("charset")) {
          part.charset = part.contentTypeParams.getParam("charset");
        }
        if (part.contentTypeParams.hasParam("format")) {
          part.format = part.contentTypeParams.getParam("format").toLowerCase();
        }
        if (part.contentTypeParams.hasParam("delsp")) {
          part.delsp = part.contentTypeParams.getParam("delsp").toLowerCase() == "yes";
        }
      }
    }

    if (part.hasField("Subject")) {
      let subject = null;

      let context = new arMIMEParser(part.getField("Subject"));
      context._O(() => {
        let tmp = context.subject__value();
        context.__END();
        subject = tmp;
      });

      if (subject) {
        if (!context.isUnicode) {
          subject = this._tryFromUTF8(subject);
        }
        part.subject = subject;
      }
    }

    if (part.hasField("Date")) {
      let date = this.decodeDate(part.getField("Date"));
      if (date) {
        part.date = date;
      }
    }

    if (part.hasField("Content-Location")) {
      let loc = part.getField("Content-Location");

      let context = new arMIMEParser(loc);
      context._O(() => {
        let tmp = context.content_location__value();
        context.__END();
        loc = tmp;
      });

      if (loc) {
        if (!context.isUnicode) {
          loc = this._tryFromUTF8(loc);
        }
        part.contentLocation = loc;
      }
    }

    if (part.hasField("Content-Disposition")) {
      let params = this.decodeParam(part.getField("Content-Disposition"));
      if (params) {
        part.contentDispositionParams = params;
        part.contentDispositionType
          = part.contentDispositionParams.type.toLowerCase();
        if (part.contentDispositionParams.hasParam("filename")) {
          let filename = part.contentDispositionParams.getParam("filename");
          part.contentDispositionFilename = this._tryFromUTF8(filename);
        }
      }
    }

    if (part.hasField("Content-ID")) {
      let id = part.getField("Content-ID");

      let context = new arMIMEParser(id);
      context._O(() => {
        let tmp = context.id__value();
        context.__END();
        id = tmp;
      });

      if (id) {
        part.contentID = id;
      }
    }

    if (part.contentType == "multipart") {
      part.isMultipart = true;

      if (!(part.contentSubType == "related" ||
            part.contentSubType == "alternative")) {
        part.isMixed = true;
      }

      if (part.contentTypeParams.hasParam("boundary")) {
        part.boundary = part.contentTypeParams.getParam("boundary");
      }

      if (part.contentTypeParams.hasParam("start")) {
        let id = part.contentTypeParams.getParam("start");

        let context = new arMIMEParser(id);
        context._O(() => {
          let tmp = context.related_param_start();
          context.__END();
          id = tmp;
        });

        part.start = id;
      }

      if (part.contentTypeParams.hasParam("type")) {
        part.startMimetype = part.contentTypeParams.getParam("type");
      }
    } else {
      if (part.hasField("Content-Transfer-Encoding")) {
        let encoding = null;

        let context = new arMIMEParser(part.getField("Content-Transfer-Encoding"));
        context._O(() => {
          let tmp = context.encoding__value();
          context.__END();
          encoding = tmp;
        });

        if (encoding) {
          part.contentTransferEncoding = encoding;
        }
      }
    }
  },

  /**
   * content[RFC2045] や disposition[RFC2183] の値を
   * parameter[RFC2231] に従ってデコードする
   *
   * ただし disposition-type[RFC2183] や type/subtype[RFC2045] の位置が
   * 先頭でない事例が存在するのでこれらのデコードも同時に行う
   *
   * @param   {string} value
   *          contentRFC2045] や disposition[RFC2183] の値
   * @returns {arMIMEParams}
   *          パラメータ
   */
  decodeParam: function(value) {
    let params = null;

    let context = new arMIMEParser(value);
    context._O(() => {
      params = context.parameter_list();
      /* パースに失敗しても最低限を返す */
    });

    return params;
  },

  /**
   * format=flowed をデコードする
   *
   * @param   {string} text
   *          デコードする文字列
   * @param   {boolean} delsp
   *          delsp の値
   * @returns {string}
   *          デコードされた文字列
   */
  decodeFlowed: function(text, delsp) {
    let ret = text;

    let context = new arMIMEParser(text);
    context._O(() => {
      let tmp = context.flowed_body(delsp);
      context.__END();
      ret = tmp;
    });

    return ret;
  },

  /* ==== private ==== */

  /**
   * multipart-body[RFC2046] をデコードする
   *
   * @param   {string} body
   *          multipart なメッセージボディ
   *          改行コードは CR LF でなければならない
   * @param   {string} boundary
   *          バウンダリ文字列
   * @param   {object} corrupted
   *          (オプショナル)
   *          {
   *            value: {boolean} 破損したかどうか
   *          }
   * @returns {Array.<arMIMEPart>}
   *          パートの配列
   */
  _decodeMultipart: function(body, boundary, currpted) {
    let ret = [];

    let context = new arMIMEParser(body);
    context._O(() => {
      let tmp = context.multipart_body(boundary, currpted);
      context.__END();
      ret = tmp;
    });

    return ret
      .map(data => this.decodeMessage(data))
      .filter(part => part);
  },

  /**
   * 正しくデコードできるならば UTF-8 からデコードする
   *
   * @param   {string} s
   *          入力文字列
   * @returns {string}
   *          UTF-16 の文字列
   */
  _tryFromUTF8: function(s) {
    let decoded = fromUTF8(s);
    let encoded = toUTF8(decoded);
    if (encoded == s) {
      return decoded;
    }
    return s;
  }
});

module.arMIMEDecoder = arMIMEDecoder;

})(this);
