"use strict";

this.EXPORTED_SYMBOLS = ["arMIMEEncoder"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { btoa } = Cu.import("resource://gre/modules/Services.jsm", {});
let { arFormatter } = Cu.import("resource://unmht/modules/base/arFormatter.jsm", {});
let { arUconv, toUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});

/**
 * date-time[RFC5322] の曜日
 * @type {Array.<string>}
 */
const weekdays = Object.freeze([
  "Sun",
  "Mon",
  "Tue",
  "Wed",
  "Thu",
  "Fri",
  "Sat"
]);

/**
 * date-time[RFC5322] の月
 * @type {Array.<string>}
 */
const months = Object.freeze([
  "Jan",
  "Feb",
  "Mar",
  "Apr",
  "May",
  "Jun",
  "Jul",
  "Aug",
  "Sep",
  "Oct",
  "Nov",
  "Dec"
]);

/**
 * フィールドの順番
 * @type {Array.<string>}
 */
const FieldOrder = Object.freeze([
  "from",
  "to",
  "subject",
  "date",
  "mime-version",
  "content-id",
  "content-location",
  "content-transfer-encoding"
]);

/**
 * フィールドの内容の種類
 * @enum {number}
 */
const FieldType = Object.freeze({
  /** 7bit */
  PLAIN:                1,
  /** 必要であれば encoded-word にする */
  ENCODED_WORD:         2,
  /** Unicode を含むので UTF-8 にし、必要であれば encoded-word にする */
  UNICODE_ENCODED_WORD: 3
});

/**
 * フィールドの種類
 * @type {Object.<stirng,number>}
 */
const FieldTypeMap = Object.freeze({
  "from":                      FieldType.PLAIN,
  "to":                        FieldType.PLAIN,
  "subject":                   FieldType.ENCODED_WORD,
  "date":                      FieldType.PLAIN,
  "mime-version":              FieldType.PLAIN,
  "content-transfer-encoding": FieldType.PLAIN,
  "content-location":          FieldType.UNICODE_ENCODED_WORD,
  "content-id":                FieldType.PLAIN
});

/**
 * MIME エンコーダ
 *
 * @class
 */
let arMIMEEncoder = Object.freeze({
  /**
   * quoted-printable[RFC2045] にエンコードする
   *
   * @param   {string} text
   *          エンコードする文字列
   * @param   {boolean} wrap
   *          76 文字で折り返す
   * @returns {string}
   *          エンコードした文字列
   */
  encodeQ: function(text, wrap) {
    let ret = "";
    let col = 0;
    let last = null;
    let lastCode = null;
    let lastWSP = false;
    function toHex(code) {
      return "=" + ("0" + code.toString(16).toUpperCase()).slice(-2);
    }
    function resolveLast(needEncode, hasNext) {
      if (!last) {
        return;
      }
      if (lastWSP && needEncode) {
        last = toHex(lastCode);
      }
      let lastLen = last.length;
      let max = 76;
      if (hasNext) {
        max = 75;
      }
      if (wrap && col + lastLen > max) {
        ret += "=\r\n";
        col = 0;
      }
      ret += last;
      col += lastLen;

      last = null;
      lastWSP = false;
    }
    let POS = 0;
    let INPUT = text;
    let INPUT_LEN = INPUT.length;
    while (POS < INPUT_LEN) {
      let c = INPUT.charAt(POS);
      POS += 1;
      let code = c.charCodeAt(0);

      if (code == 0x0d) {
        resolveLast(true, false);
        ret += "\r\n";
        col = 0;
        if (INPUT.charCodeAt(POS) == 0x0a) {
          POS += 1;
        }
        continue;
      }
      if (code == 0x0a) {
        resolveLast(true, false);
        ret += "\r\n";
        col = 0;
        continue;
      }

      resolveLast(false, true);

      if ((code >= 0x21 && code <= 0x3c) ||
          (code >= 0x3e && code <= 0x7e)) {
        last = c;
        lastWSP = false;
        continue;
      }
      if (code == 0x20 || code == 0x09) {
        last = c;
        lastWSP = true;
        lastCode = code;
        continue;
      }
      last = toHex(code);
      lastWSP = false;
    }
    resolveLast(true, false);

    return ret;
  },

  /**
   * encoded-word にエンコードする
   *
   * @param   {string} charset
   *          文字コード
   * @param   {string} type
   *          Q, B
   * @param   {string} text
   *          エンコードする文字列 (Unicode)
   * @returns {string}
   *          エンコードした文字列
   */
  encodeWord: function(charset, type, text) {
    if (text === "") {
      /* 空の文字列はエンコードできない */
      return "";
    }

    text = arUconv.fromUnicode(text, charset);

    if (type == "B") {
      text = this.encodeBase64(text, false);
    } else if (type == "Q") {
      text
        = this.encodeQ(text, false)
        .replace(/\?/g, "=3F")
        .replace(/ /g, "=20");
    }

    return ("=?" + charset + "?" + type + "?" + text + "?=");
  },

  /**
   * Base64 エンコードする
   *
   * @param   {string} text
   *          エンコードする文字列
   * @param   {boolean} wrap
   *          76 文字で折り返す
   * @returns {string}
   *          エンコードされた文字列
   */
  encodeBase64: function(text, wrap) {
    text = btoa(text);
    if (!wrap) {
      return text;
    }

    let s = "";
    let W = 76;
    s += text.slice(0, W);
    for (let i = W, len = text.length; i < len; i += W) {
      s += "\r\n" + text.slice(i, i + W);
    }
    return s;
  },

  /**
   * 日時を date-time[RFC5322] 形式にエンコードする
   *
   * @param   {Date} date
   *          対象の日時
   * @returns {string}
   *          エンコードされた日時
   */
  encodeDate: function(date) {
    let sign = "-";
    let zone = date.getTimezoneOffset() * 100 / 60;
    if (zone < 0) {
      zone = -zone;
      sign = "+";
    }

    return (weekdays[date.getDay()] + ", "
            + arFormatter.pad(2, date.getDate()) + " "
            + months[date.getMonth()] + " "
            + date.getFullYear() + " "
            + arFormatter.pad(2, date.getHours()) + ":"
            + arFormatter.pad(2, date.getMinutes()) + ":"
            + arFormatter.pad(2, date.getSeconds()) + " "
            + sign + arFormatter.pad(4, zone));
  },

  /**
   * message [RFC5322] にエンコードする
   *
   * @param   {arMIMEPart} part
   *          パート
   * @param   {boolean=} isTop
   *          トップレベルのパートかどうか
   * @returns {string}
   *          エンコード結果
   *          改行コードは CR LF である
   */
  encodeMessage: function(part, isTop=true) {
    let result = this.encodeFields(part, isTop);
    result += "\r\n";

    if (part.isMultipart) {
      for (let p of part.parts) {
        result += "--" + part.boundary + "\r\n";
        result += this.encodeMessage(p, false);
        result += "\r\n";
      }

      result += "--" + part.boundary + "--\r\n";

      return result;
    }

    if (part.contentTransferEncoding == "quoted-printable") {
      result += this.encodeQ(part.body, true);
    } else if (part.contentTransferEncoding == "base64") {
      result += this.encodeBase64(part.body, true) + "\r\n";
    } else {
      result += part.body;
    }

    return result;
  },

  /**
   * fields[RFC5322] にエンコードする
   *
   * @param   {arMIMEPart} part
   *          パート
   * @param   {boolean=} isTop
   *          トップレベルのパートかどうか
   * @returns {string}
   *          エンコード結果
   *          改行コードは CR LF である
   */
  encodeFields: function(part, isTop=true) {
    let result = "";

    let fieldNames;
    if (isTop) {
      fieldNames = part.fieldNames
        .filter(name => !(name.toLowerCase() == "content-type" ||
                          name.toLowerCase() == "content-disposition"))
        .sort((name1, name2) => {
          let nameL1 = name1.toLowerCase();
          let nameL2 = name2.toLowerCase();
          let p1 = FieldOrder.indexOf(nameL1);
          let p2 = FieldOrder.indexOf(nameL2);
          if (p1 >= 0) {
            if (p2 >= 0) {
              return p1 - p2;
            } else {
              return -1;
            }
          } else {
            if (p2 >= 0) {
              return 1;
            } else {
              return (nameL1 < nameL2) ? -1 : 1;
            }
          }
        })
        .concat("content-type", "content-disposition");
    } else {
      fieldNames = part.fieldNames
        .filter(name => !(name.toLowerCase() == "content-type" ||
                          name.toLowerCase() == "content-disposition"))
        .concat("content-type", "content-disposition")
        .sort((name1, name2) => {
          let nameL1 = name1.toLowerCase();
          let nameL2 = name2.toLowerCase();
          return (nameL1 < nameL2) ? -1 : 1;
        });
    }

    for (let name of fieldNames) {
      if (name == "content-type") {
        if (part.contentTypeParams) {
          let value = this.encodeParam(part.contentTypeParams, true, true);
          result += "Content-Type: " + value + "\r\n";
        }
        continue;
      }
      if (name == "content-disposition") {
        if (part.contentDispositionParams) {
          let value = this.encodeParam(part.contentDispositionParams, true, true);
          result += "Content-Disposition: " + value + "\r\n";
        }
        continue;
      }

      for (let [origName, value] of part.getFieldList(name)) {
        let type = FieldType.PLAIN;
        let nameL = origName.toLowerCase();
        if (nameL in FieldTypeMap) {
          type = FieldTypeMap[nameL];
        }

        if (type == FieldType.UNICODE_ENCODED_WORD) {
          let utf8 = toUTF8(value);
          if (!/^[\x20-\x7E]*$/.test(utf8)) {
            value = this.encodeWord("utf-8", "B", value);
          } else if (!/^[\x21-\x7E]*$/.test(utf8)) {
            value = this.encodeWord("us-ascii", "B", value);
          } else {
            value = utf8;
          }
        } else if (type == FieldType.ENCODED_WORD) {
          if (!/^[\x20-\x7E]*$/.test(value)) {
            value = this.encodeWord("utf-8", "B", value);
          }
        }

        result += origName + ": " + value + "\r\n";
      }
    }

    return result;
  },

  /**
   * arMIMEParams を文字列にエンコードする
   *
   * @param   {arMIMEParams} params
   *          パラメータ
   * @param   {boolean} wrap
   *          パラメータごとに改行するか
   * @param   {boolean} ignoreFormatDelSP
   *          format, delsp を無視するか
   * @returns {string}
   *          文字列表現
   */
  encodeParam: function(params, wrap, ignoreFormatDelSP) {
    let value = "";
    if (params.subtype) {
      value = params.type + "/" + params.subtype;
    } else {
      value = params.type;
    }

    let encodeSingleParam = (pName, pValue) => {
      if (/^[\x21-\x26\x28-\x7E]*$/.test(pValue)) {
        return pName + "=\"" + pValue.replace(/\"/g, "\\\"") + "\"";
      } else {
        return pName + "*=utf-8\'\'"
          + encodeURIComponent(pValue).replace(/\'/g, "%27");
      }
    };

    let ps
      = params.paramNames
      .filter(name => !(ignoreFormatDelSP &&
                        (name == "format" ||
                         name == "delsp")))
      .map(name => encodeSingleParam(name, params.getParam(name)));

    return [value, ...ps].join(wrap ? ";\r\n\t": ";");
  }
});

module.arMIMEEncoder = arMIMEEncoder;

})(this);
