"use strict";

this.EXPORTED_SYMBOLS = ["arMIMEParams"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * content[RFC2045] や disposition[RFC2183] の値
 *
 * @constructor
 * @param   {string} mimetype
 *          MIME-Type (オプショナル)
 */
function arMIMEParams(mimetype) {
  /**
   * disposition-type[RFC2183]/type[RFC2045]
   * @type {string}
   */
  this.type = "";

  /**
   * subtype[RFC2045]
   * @type {string}
   */
  this.subtype = "";

  if (mimetype) {
    let m = mimetype.match(/^([^\/]+)\/([^\/]+)$/);
    if (m) {
      this.type = m[1];
      this.subtype = m[2];
    } else {
      this.type = mimetype;
    }
  }

  /**
   * パラメータのマップ
   *   { 小文字の名前: [大文字の名前, 値] }
   * @type {Map.<string,Tuple.<string,string>>}
   */
  this._params = new Map();

  Object.seal(this);
}
arMIMEParams.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    this.type = "";
    this.subtype = "";
    this._params.clear();
  },

  /**
   * 複製する
   *
   * @returns {arMIMEParams}
   *          複製
   */
  clone: function() {
    let newParams = new arMIMEParams();

    newParams.type = this.type;
    newParams.subtype = this.subtype;
    for (let [name, value] of this._params.values()) {
      newParams.addParam(name, value);
    }

    return newParams;
  },

  /**
   * フィールド名の配列
   * @type {Array.<string>}
   */
  get paramNames() {
    return [...this._params.values()].map(([name, value]) => name);
  },

  /**
   * フィールドを持っているかを返す
   *
   * @param   {string} name
   *          フィールド名
   * @returns {boolean}
   *          フィールドを持っているか
   */
  hasParam: function(name) {
    let nameL = name.toLowerCase();
    return this._params.has(nameL);
  },

  /**
   * フィールドを持っているかを返す
   *
   * @param   {string} name
   *          フィールド名
   * @returns {string}
   *          フィールド値
   */
  getParam: function(name) {
    let nameL = name.toLowerCase();
    return this._params.get(nameL)[1];
  },

  /**
   * フィールドを追加する
   *
   * @param   {string} name
   *          フィールド名
   * @param   {string} value
   *          フィールド値
   */
  addParam: function(name, value) {
    let nameL = name.toLowerCase();
    this._params.set(nameL, [name, value]);
  },

  /**
   * フィールドを削除する
   *
   * @param   {string} name
   *          フィールド名
   */
  removeParam: function(name) {
    let nameL = name.toLowerCase();
    this._params.delete(nameL);
  }
});

module.arMIMEParams = arMIMEParams;

})(this);
