"use strict";

this.EXPORTED_SYMBOLS = ["arMIMEParser",
                         "arMIMEParserError",
                         "safe_atob"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { atob } = Cu.import("resource://gre/modules/Services.jsm", {});
let { arMIMEParams } = Cu.import("resource://unmht/modules/base/arMIMEParams.jsm");
let { arMIMEPart } = Cu.import("resource://unmht/modules/base/arMIMEPart.jsm");
let { arUconv, fromUTF8, toUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});

/**
 * Base64 デコードする
 * 改行を含むものや破損したものにも対応
 *
 * @param   {string} text
 *          デコードする文字列
 * @returns {string}
 *          デコードされた文字列
 */
function safe_atob(text) {
  try {
    /* まずは生のデータで試行 */
    return atob(text);
  } catch (e) {
    try {
      /* 不要なデータを削除する */
      text = text.replace(/[^A-Za-z0-9+\/=]+/g, "");
      return atob(text.slice(0, text.length - text.length % 4));
    } catch (e2) {
      /* 途中にパディングがある等、壊れているデータ */
      return "";
    }
  }
}

/**
 * パースエラー
 */
let arMIMEParserError = Object.freeze({
});

/**
 * メールボックス
 */
function arMIMEMailBox() {
  /**
   * グループか
   * @type {boolean}
   */
  this.isGroup = false;

  /**
   * 名前
   * @type {string}
   */
  this.name = null;

  /**
   * アドレス
   * @type {string}
   */
  this.addr = null;

  /**
   * メールボックスのリスト
   * isGroup が true の場合のみ
   * @type {Array.<arMIMEMailBox>}
   */
  this.list = null;

  Object.seal(this);
}

/**
 * パラメータの各セクション
 */
function arMIMEParamSection() {
  /**
   * 名前
   * @type {string}
   */
  this.name = null;

  /**
   * 値
   * @type {string}
   */
  this.value = null;

  /**
   * section[RFC 2231]
   * @type {number}
   */
  this.section = null;

  /**
   * charset[RFC 2231]
   * @type {string}
   */
  this.charset = null;

  /**
   * language[RFC 2231]
   * @type {string}
   */
  this.language = null;

  /**
   * extended-parameter か
   * @type {boolean}
   */
  this.extended = false;

  Object.seal(this);
}

/**
 * パースエラー以外を報告
 *
 * @param   {Error} e
 */
function check_error(e) {
  if (e != arMIMEParserError) {
    Cu.reportError(e + "\n" + e.stack);
  }
}

/**
 * パーサ
 *
 * @param   {string} _INPUT
 *          入力文字列
 */
function arMIMEParser(_INPUT) {
  this._INPUT = _INPUT;
  this._INPUT_LEN = _INPUT.length;
  this._POS = 0;

  /* 結果が UTF16 になっているか
   * (encoded-word をデコードしたかどうか, または RFC6532 かどうか) */
  this.isUnicode = false;

  Object.seal(this);
}
arMIMEParser.prototype = Object.freeze({
  /**
   * 正しくデコードできるならば UTF-8 からデコードする
   *
   * @param   {string} s
   *          入力文字列
   * @returns {string}
   *          UTF-16 の文字列
   */
  _tryFromUTF8: function(s) {
    if (/[\x80-\xff]/.test(s)) {
      let decoded = fromUTF8(s);
      let encoded = toUTF8(decoded);
      if (encoded == s) {
        this.isUnicode = true;
        return decoded;
      }
    }
    return s;
  },

  /**
   * 入力の終了にマッチする
   */
  __END: function() {
    if (this._POS < this._INPUT_LEN) {
      throw arMIMEParserError;
    }
  },

  /**
   * 文字列にマッチすれば消費する
   *
   * @param   {string}
   * @returns {string}
   */
  _C: function(c) {
    if (!this._INPUT.slice(this._POS).startsWith(c)) {
      throw arMIMEParserError;
    }
    this._POS += c.length;

    return c;
  },

  /**
   * 1 文字消費する
   *
   * @returns {string}
   */
  _SINGLE_CHAR: function() {
    if (this._POS >= this._INPUT_LEN) {
      throw arMIMEParserError;
    }
    let ret = this._INPUT[this._POS];
    this._POS += 1;

    return ret;
  },

  /**
   * 全て消費する
   *
   * @returns {string}
   */
  _ALL_CHARS: function() {
    let ret = this._INPUT.slice(this._POS);
    this._POS = this._INPUT_LEN;

    return ret;
  },

  /**
   * 全て破棄する
   */
  _DISCARD: function() {
    this._POS = this._INPUT_LEN;
  },

  /**
   * 指定文字数消費する
   *
   * @param   {number} len
   * @returns {string}
   */
  _SKIP: function(len) {
    let ret = this._INPUT.slice(this._POS, this._POS + len);
    this._POS += len;
    if (this._POS > this._INPUT_LEN) {
      throw arMIMEParserError;
    }

    return ret;
  },

  /**
   * 文字列にマッチするまで消費する
   *
   * @param   {string} c
   * @param   {boolean} consume_c
   *          マッチした文字列も消費するか
   * @returns {string}
   */
  _SKIP_TO: function(c, consume_c) {
    let p = this._INPUT.indexOf(c, this._POS);
    if (p == -1) {
      throw arMIMEParserError;
    }
    let s = this._INPUT.slice(this._POS, p);
    if (consume_c) {
      p += c.length;
    }
    this._POS = p;

    return s;
  },

  /**
   * 正規表現にマッチすれば消費する
   *
   * @param   {RegExp}
   * @returns {string}
   */
  _R: function(r) {
    let m = this._INPUT.slice(this._POS).match(r);

    if (!m) {
      throw arMIMEParserError;
    }
    this._POS += m[0].length;

    return m[0];
  },

  /**
   * 正規表現にマッチすれば消費する
   *
   * @param   {RegExp}
   * @returns {MatchObject}
   */
  _R_M: function(r) {
    let m = this._INPUT.slice(this._POS).match(r);

    if (!m) {
      throw arMIMEParserError;
    }
    this._POS += m[0].length;

    return m;
  },

  /**
   * 0 回もしくは 1 回の出現を消費する
   *
   * @param   {function} f
   */
  _O: function(f) {
    let p = this._POS;

    try {
      f();
    } catch (e) {
      check_error(e);
      this._POS = p;
    }
  },

  /**
   * min 回以上 max 回以下の出現を消費する
   *
   * @param   {number} min
   * @param   {number} max
   * @param   {function} f
   */
  _N: function(min, max, f) {
    let i = 0;

    for (; i < min; i += 1) {
      f();
    }

    for (; i < max; i += 1) {
      let p = this._POS;

      try {
        f();

        if (this._POS == p) {
          break;
        }
      } catch (e) {
        check_error(e);
        this._POS = p;
        break;
      }
    }
  },

  /**
   * 0 回以上の出現を消費する
   *
   * @param   {function} f
   */
  _0N: function(f) {
    for (;;) {
      let p = this._POS;

      try {
        f();

        if (this._POS == p) {
          break;
        }
      } catch (e) {
        check_error(e);
        this._POS = p;
        break;
      }
    }
  },

  /**
   * 1 回以上の出現を消費する
   *
   * @param   {function} f
   */
  _1N: function(f) {
    f();

    for (;;) {
      let p = this._POS;

      try {
        f();

        if (this._POS == p) {
          break;
        }
      } catch (e) {
        check_error(e);
        this._POS = p;
        break;
      }
    }
  },

  /**
   * 最初に出現したパターンを消費する
   *
   * @param   {...function} fs
   * @returns {*}
   */
  _ONEOF: function(...fs) {
    let p = this._POS;
    for (let f of fs) {
      try {
        return f();
      } catch (e) {
        this._POS = p;
        check_error(e);
      }
    }
    throw arMIMEParserError;
  },

  /* == RFC 2017: Definition of the URL MIME External-Body Access-Type == */

  /**
   * URL-parameter := <"> URL-word *(*LWSP-char URL-word) <">
   *
   * @returns {string}
   */
  RE_LWSP_char_0N: /^[\t ]*/,
  URL_parameter: function() {
    this._C("\"");
    let ret = [this.URL_word()];
    this._0N(() => {
      this._O(() => this.FWS());
      this._R(this.RE_LWSP_char_0N);
      this._O(() => this.FWS());
      ret.push(this.URL_word());
    });
    this._C("\"");

    return ret.join("");
  },

  /**
   * URL-word := token
   *             ; Must not exceed 40 characters in length
   *
   * @returns {string}
   */
  RE_URL_word: /^[^ \t\r\n\"\\]+/,
  URL_word: function() {
    return this._R(this.RE_URL_word);
  },

  /* == RFC 2045: Format of Internet Message Bodies == */

  /* #5.  Content-Type Header Field
   * #5.1.  Syntax of the Content-Type Header Field */

  /**
   * content := "Content-Type" ":" type "/" subtype
   *            *(";" parameter)
   *            ; Matching of media type and subtype
   *            ; is ALWAYS case-insensitive.
   * type := discrete-type / composite-type
   * discrete-type := "text" / "image" / "audio" / "video" /
   *                  "application" / extension-token
   * composite-type := "message" / "multipart" / extension-token
   * extension-token := ietf-token / x-token
   * ietf-token := <An extension token defined by a
   *                standards-track RFC and registered
   *                with IANA.>
   * x-token := <The two characters "X-" or "x-" followed, with
   *             no intervening white space, by any token>
   * subtype := extension-token / iana-token
   * iana-token := <A publicly-defined extension token. Tokens
   *                of this form must be registered with IANA
   *                as specified in RFC 2048.>
   *
   * @returns {arMIMEParam}
   */
  content__value: function() {
    return this.parameter_list();
  },

  /**
   * value := token / quoted-string
   *
   * @returns {string}
   */
  value: function() {
    return this._ONEOF(
      () => this.token(),
      () => this.quoted_string()
    );
  },

  /**
   * token := 1*<any (US-ASCII) CHAR except SPACE, CTLs,
   *             or tspecials>
   * tspecials :=  "(" / ")" / "<" / ">" / "@" /
   *               "," / ";" / ":" / "\" / <"> /
   *               "/" / "[" / "]" / "?" / "="
   *               ; Must be in quoted-string,
   *               ; to use within parameter values
   * Errata ID: 512
   *
   * @returns {string}
   */
  RE_token: /^[^\t\n\r \"\(\),\/:-@\[-\]]+/,
  token: function() {
    return this._R(this.RE_token);
  },

  /* #6.  Content-Transfer-Encoding Header Field
   * #6.1.  Content-Transfer-Encoding Syntax */

  /**
   * encoding := "Content-Transfer-Encoding" ":" mechanism
   * mechanism := "7bit" / "8bit" / "binary" /
   *              "quoted-printable" / "base64" /
   *              ietf-token / x-token
   *
   * @returns {string}
   *          小文字化した文字列
   */
  RE_encoding_7bit: /^7bit/i,
  RE_encoding_8bit: /^8bit/i,
  RE_encoding_binary: /^binary/i,
  RE_encoding_quoted_printable: /^quoted-printable/i,
  RE_encoding_base64: /^base64/i,
  encoding__value: function() {
    this._O(() => this.CFWS());
    let ret = this._ONEOF(
      () => this._R(this.RE_encoding_7bit),
      () => this._R(this.RE_encoding_8bit),
      () => this._R(this.RE_encoding_binary),
      () => this._R(this.RE_encoding_quoted_printable),
      () => this._R(this.RE_encoding_base64),
      () => this.token()
    ).toLowerCase();
    this._O(() => this.CFWS());

    return ret;
  },

  /* #6.7.  Quoted-Printable Content-Transfer-Encoding */

  /**
   * quoted-printable := qp-line *(CRLF qp-line)
   * qp-line := *(qp-segment transport-padding CRLF)
   *            qp-part transport-padding
   * qp-part := qp-section
   *            ; Maximum length of 76 characters
   * qp-segment := qp-section *(SPACE / TAB) "="
   *               ; Maximum length of 76 characters
   * qp-section := [*(ptext / SPACE / TAB) ptext]
   * ptext := hex-octet / safe-char
   * safe-char := <any octet with decimal value of 33 through
   *              60 inclusive, and 62 through 126>
   *              ; Characters not listed as "mail-safe" in
   *              ; RFC 2049 are also not recommended.
   * hex-octet := "=" 2(DIGIT / "A" / "B" / "C" / "D" / "E" / "F")
   *              ; Octet must be used for characters > 127, =,
   *              ; SPACEs or TABs at the ends of lines, and is
   *              ; recommended for any character not listed in
   *              ; RFC 2049 as "mail-safe".
   *
   * [simplified]
   *
   * quoted-printable := *((*(SPACE / TAB) "=" transport-padding CRLF) /
   *                       (transport-padding <end of input>) /
   *                       (transport-padding CRLF) /
   *                       "_" /
   *                       hex-octet /
   *                       (<any octet>
   *                        *<any octet except "=", "_", SPACE, HTAB,
   *                          CR and LF>))
   *
   * @returns {string}
   *          デコードした文字列
   *          未知の部分文字列はそのまま返す
   */
  quoted_printable: function(underscoreToSpace=false) {
    let ret = [];
    let POS = this._POS;
    let INPUT = this._INPUT;
    let INPUT_LEN = this._INPUT_LEN;
    let padding = [];
    let lastPadding = false;
    while (POS < INPUT_LEN) {
      let c = INPUT.charAt(POS);
      POS += 1;
      let code = c.charCodeAt(0);

      if (code == 0x20 || code == 0x09) {
        padding.push(c);
        lastPadding = true;
        continue;
      }
      if (code == 0x3d) {
        if (lastPadding) {
          ret.push(padding.join(""));
          padding.length = 0;
          lastPadding = false;
        }

        let c1 = INPUT.charAt(POS);
        let code1 = c1.charCodeAt(0);
        if ((code1 >= 0x30 && code1 <= 0x39) ||
            (code1 >= 0x41 && code1 <= 0x46) ||
             (code1 >= 0x61 && code1 <= 0x66)) {
          let c2 = INPUT.charAt(POS + 1);
          let code2 = c2.charCodeAt(0);
          if ((code2 >= 0x30 && code2 <= 0x39) ||
              (code2 >= 0x41 && code2 <= 0x46) ||
              (code2 >= 0x61 && code2 <= 0x66)) {
            POS += 2;
            ret.push(String.fromCharCode(Number.parseInt(c1 + c2, 16)));

            continue;
          }
        }

        let POS2 = POS;
        while (code1 == 0x20 || code1 == 0x09) {
          POS2 += 1;
          c1 = INPUT.charAt(POS2);
          code1 = c1.charCodeAt(0);
        }
        if (code1 == 0x0d) {
          POS = POS2 + 1;
          if (INPUT.charCodeAt(POS) == 0x0a) {
            POS += 1;
          }
          continue;
        }
        if (code1 == 0x0a) {
          POS = POS2 + 1;
          continue;
        }

        ret.push(c);
        continue;
      }
      if (code == 0x0d) {
        ret.push("\r");
        lastPadding = false;
        if (INPUT.charCodeAt(POS) == 0x0a) {
          ret.push("\n");
          POS += 1;
        }
        continue;
      }
      if (code == 0x0a) {
        ret.push("\n");
        lastPadding = false;
        continue;
      }

      if (lastPadding) {
        ret.push(padding.join(""));
        padding.length = 0;
        lastPadding = false;
      }

      if (code == 0x5f && underscoreToSpace) {
        ret.push(" ");
      } else {
        ret.push(c);
      }
    }

    this._POS = POS;

    return ret.join("");
  },

  /**
   * transport-padding := *LWSP-char
   *                      ; Composers MUST NOT generate
   *                      ; non-zero length transport
   *                      ; padding, but receivers MUST
   *                      ; be able to handle padding
   *                      ; added by message transports.
   * LWSP-char   =  SPACE / HTAB
   * (From RFC 822)
   */
  RE_transport_padding: /^[\t ]*/,
  transport_padding: function() {
    this._R(this.RE_transport_padding);
  },

  /* #7.  Content-ID Header Field */

  /**
   * id := "Content-ID" ":" msg-id
   *
   * msg-id の形式になっていないもの、angle が無いものも対応する
   *
   * @returns {string}
   */
  id__value: function() {
    return this.cid();
  },

  /* == RFC 2046: Media Types == */

  /* #5.  Composite Media Type Values
   * #5.1.  Multipart Media Type
   * #5.1.1.  Common Syntax */

  /**
   * boundary := 0*69<bchars> bcharsnospace
   * bchars := bcharsnospace / " "
   * bcharsnospace := DIGIT / ALPHA / "'" / "(" / ")" /
   *                  "+" / "_" / "," / "-" / "." /
   *                  "/" / ":" / "=" / "?"
   * dash-boundary := "--" boundary
   *                  ; boundary taken from the value of
   *                  ; boundary parameter of the
   *                  ; Content-Type field.
   * multipart-body := [preamble CRLF]
   *                   dash-boundary transport-padding CRLF
   *                   body-part *encapsulation
   *                   close-delimiter transport-padding
   *                   [CRLF epilogue]
   * encapsulation := delimiter transport-padding
   *                  CRLF body-part
   * delimiter := CRLF dash-boundary
   * close-delimiter := delimiter "--"
   * preamble := discard-text
   * epilogue := discard-text
   * discard-text := *(*text CRLF) *text
   *                 ; May be ignored or discarded.
   * body-part := MIME-part-headers [CRLF *OCTET]
   *              ; Lines in a body-part must not start
   *              ; with the specified dash-boundary and
   *              ; the delimiter must not appear anywhere
   *              ; in the body part.  Note that the
   *              ; semantics of a body-part differ from
   *              ; the semantics of a message, as
   *              ; described in the text.
   * OCTET := <any 0-255 octet value>
   *
   * [simplified]
   * multipart-body := [preamble CRLF]
   *                   dash-boundary transport-padding CRLF
   *                   *(body-part
   *                     delimiter transport-padding CRLF)
   *                   body-part
   *                   [delimiter "--" <discard>]
   *
   * 破損したファイルをサポートするために close-delimiter をオプショナルにする
   *
   * @param   {string} boundary
   *          boundary
   * @param   {object} isCorrupted
   *          (オプショナル)
   *          {
   *            value: {boolean} 破損したかどうか
   *          }
   * @returns {Array.<string>}
   */
  multipart_body: function(boundary, isCorrupted) {
    let dash_boundary = "--" + boundary;

    this._SKIP_TO(dash_boundary, true);
    this.transport_padding();
    this.CRLF();

    let datas = [];
    let delimiter = "\r\n--" + boundary;
    let delimiter_len = delimiter.length;
    let closed = false;
    while (!closed) {
      let done = false;
      let part = [];

      while (!done) {
        let hasDelimiter = false;
        this._O(() => {
          part.push(this._SKIP_TO(delimiter, false));
          hasDelimiter = true;
        });
        if (!hasDelimiter) {
          /* close-delimiter を含まない破損したファイル */
          part.push(this._ALL_CHARS());
          datas.push(part.join(""));
          closed = true;
          if (isCorrupted) {
            isCorrupted.value = true;
          }
          break;
        }

        this._ONEOF(
          () => {
            this._SKIP(delimiter_len);
            this._ONEOF(
              () => {
                this.transport_padding();
                this.CRLF();
              },
              () => {
                this._C("--");
                closed = true;
              }
            );

            datas.push(part.join(""));
            done = true;
          },
          () => {
            /* delimiter と前方一致するが delimiter でない物が含まれたので
             * 1 文字先まで飛ばす */
            part.push(this._SINGLE_CHAR());
          }
        );
      }
    }

    this._DISCARD();

    return datas;
  },

  /* == RFC 2047: Message Header Extensions for Non-ASCII Text == */

  /* #2. Syntax of encoded-words */

  /**
   * any 'linear-white-space' that separates a pair of
   * adjacent 'encoded-word's is ignored.
   *
   * ただし、デコードできなかった場合は無視しない
   *
   * [simplified]
   * encoded-word-seq = encoded-word *(FWS encoded-word)
   *
   * @returns {string}
   *          デコードした文字列
   */
  encoded_word_seq: function() {
    let [lastDecoded, tmp] = this.encoded_word();
    let ret = [tmp];
    this._0N(() => {
      let s = this.FWS();
      let [decoded, t] = this.encoded_word();
      if (!lastDecoded || !decoded) {
        ret.push(s);
      }
      ret.push(t);
      lastDecoded = decoded;
    });

    return ret.join("");
  },

  /**
   * token = 1*<Any CHAR except SPACE, CTLs, and especials>
   * especials = "(" / ")" / "<" / ">" / "@" / "," / ";" / ":" / "\" /
   *             <"> / "/" / "[" / "]" / "?" / "." / "="
   *
   * RFC 2045 の token と被るので名前を変える
   *
   * @returns {string}
   */
  RE_ew_token: /^[^\t\n\r \"\(\),\.\/:-@\[-\]]+/,
  ew_token: function() {
    return this._R(this.RE_ew_token);
  },

  RE_ew_token_noast: /^[^\t\n\r \"\(\),\.\/:-@\[-\]\*]+/,
  ew_token_noast: function() {
    return this._R(this.RE_ew_token_noast);
  },

  /* #5. Use of encoded-words in message headers */

  /**
   * RFC 2047 で encoded-word が RFC 822 の text 代わりに使えると書いてあるが
   * RFC 5322 で text は unstructured になっているので
   * unstructured 中の encoded-word をデコードできるようにする
   *
   * unstructured-ew =   *(*FWS
   *                       1*(encoded-word-seq / 1*obs-utext-ne / "=")) *WSP
   * obs-utext-ne    =   <obs-utext except "=">
   *
   * 前後の FWS は削除する
   *
   * unstructured-ew =   *FWS
   *                     *(*FWS
   *                       1*(encoded-word-seq / 1*obs-utext-ne / "=")) *FWS
   *
   * @returns {string}
   */
  RE_utext_ne_1n: /^[^\t\n\r \x3d]+/,
  unstructured_ew: function() {
    let ret = [];

    this._0N(() => this.FWS());
    this._0N(() => {
      let s = [];

      this._0N(() => {
        s.push(this.FWS());
      });
      this._1N(() => {
        s.push(this._ONEOF(
          () => this.encoded_word_seq(),
          () => this._R(this.RE_utext_ne_1n),
          () => this._C("=")
        ));
      });

      ret.push(s.join(""));
    });
    this._0N(() => this.FWS());

    return ret.join("");
  },

  /* == RFC 2183: The Content-Disposition Header Field == */

  /* 2.  The Content-Disposition Header Field */

  /**
   * disposition := "Content-Disposition" ":"
   *                disposition-type
   *                *(";" disposition-parm)
   * disposition-type := "inline"
   *                   / "attachment"
   *                   / extension-token
   *                   ; values are not case-sensitive
   * disposition-parm := filename-parm
   *                   / creation-date-parm
   *                   / modification-date-parm
   *                   / read-date-parm
   *                   / size-parm
   *                   / parameter
   * filename-parm := "filename" "=" value
   * creation-date-parm := "creation-date" "=" quoted-date-time
   * modification-date-parm := "modification-date" "=" quoted-date-time
   * read-date-parm := "read-date" "=" quoted-date-time
   * size-parm := "size" "=" 1*DIGIT
   * quoted-date-time := quoted-string
   *                  ; contents MUST be an RFC 822 `date-time'
   *                  ; numeric timezones (+HHMM or -HHMM) MUST be used
   *
   * [simplified]
   * disposition-type := token
   * disposition-parm := parameter
   *
   * @returns {arMIMEParam}
   */
  content_disposition__value: function() {
    return this.parameter_list();
  },

  /* == RFC 2231: Parameter Value and Encoded Word == */

  /**
   * parameter := regular-parameter / extended-parameter
   *
   * @return {arMIMEParamSection}
   */
  parameter: function() {
    return this._ONEOF(
      () => this.regular_parameter(),
      () => this.extended_parameter(),
      () => {
        /* クォートされるべき文字がクォートされていないものへの暫定対応
         * 以下の文字を許可する
         * ( ) < > @
         * , ; :
         * / [ ]
         */
        let ret = new arMIMEParamSection();

        ret.name = this.attribute();
        this._O(() => {
          ret.section = this.section();
        });
        this._C("=");
        ret.value = this._R(/^[^\t\n\r \";=\?\\]+/);

        return ret;
      }
    );
  },

  /**
   * Content-Type と Content-Disposition で同じものを使うので共通化
   * また、type/subtype と disposition-type が先頭以外にある例があるので
   * これも同時に解析する
   *
   * parameter-list := *(((type "/" subtype) /
   *                      disposition-type /
   *                      parameter) [";"])
   *
   * @returns {arMIMEParam}
   */
  parameter_list: function() {
    let ret = new arMIMEParams();

    let sections = [];
    let type_found = false;

    this._0N(() => {
      this._O(() => this.CFWS());
      this._ONEOF(
        () => {
          sections.push(this.parameter());
        },
        () => {
          if (type_found) {
            throw arMIMEParserError;
          }

          ret.type = this.token().toLowerCase();
          this._O(() => this.CFWS());
          this._C("/");
          this._O(() => this.CFWS());
          ret.subtype = this.token().toLowerCase();

          type_found = true;
        },
        () => {
          if (type_found) {
            throw arMIMEParserError;
          }

          ret.type = this.token().toLowerCase();

          type_found = true;
        }
      );
      this._O(() => this.CFWS());
      /* セミコロンが無いものをサポート */
      this._O(() => this._C(";"));
    });

    if (!type_found) {
      throw arMIMEParserError;
    }

    let ps = new Map();
    for (let s of sections) {
      let name = s.name;
      let p = null;
      if (ps.has(name)) {
        p = ps.get(name);
      } else {
        p = {
          values: [],
          charset: null,
          language: null
        };
        ps.set(name, p);
      }
      if (s.charset) {
        p.charset = s.charset;
      }
      if (s.language) {
        p.language = s.language;
      }
      p.values.push(s);
    }

    for (let [name, p] of ps) {
      let value = [];
      p.values.sort((a, b) => a.section - b.section);

      for (let s of p.values) {
        if (s.extended) {
          value.push(arUconv.toUnicode(s.value, p.charset));
          this.isUnicode = true;
        } else {
          try {
            /* ここでは encoded-word は使えない事になっているが
             * 使われている事例が沢山ある */
            let context = new arMIMEParser(s.value);
            let t = context.unstructured_ew();
            context.__END();
            value.push(t);
          } catch (e) {
            value.push(s.value);
          }
        }
      }

      ret.addParam(name, value.join(""));
    }

    return ret;
  },

  /**
   * regular-parameter := regular-parameter-name "=" value
   * regular-parameter-name := attribute [section]
   *
   * @return {arMIMEParamSection}
   */
  regular_parameter: function() {
    let ret = new arMIMEParamSection();

    ret.name = this.attribute();
    this._O(() => {
      ret.section = this.section();
    });
    this._C("=");
    ret.value = this.value();

    return ret;
  },

  /**
   * attribute := 1*attribute-char
   *
   * @returns {string}
   */
  attribute: function() {
    return this.attribute_char_1n();
  },

  /**
   * attribute-char := <any (US-ASCII) CHAR except SPACE, CTLs,
   *                    "*", "'", "%", or tspecials>
   *
   * @returns {string}
   */
  RE_attribute_char: /^[^\t\n\r \"%\'-\*,\/:-@\[-\]]/,
  attribute_char: function() {
    return this._R(this.RE_attribute_char);
  },
  RE_attribute_char_1n: /^[^\t\n\r \"%\'-\*,\/:-@\[-\]]+/,
  attribute_char_1n: function() {
    return this._R(this.RE_attribute_char_1n);
  },

  /**
   * section := initial-section / other-sections
   *
   * @returns {number}
   */
  section: function() {
    return this._ONEOF(
      () => this.initial_section(),
      () => this.other_sections()
    );
  },

  /**
   * initial-section := "*0"
   *
   * @returns {number}
   */
  initial_section: function() {
    this._C("*0");

    return 0;
  },

  /**
   * other-sections := "*" ("1" / "2" / "3" / "4" / "5" /
   *                        "6" / "7" / "8" / "9") *DIGIT)
   *
   * @returns {number}
   */
  RE_other_sections: /^\*([1-9][0-9]*)/,
  other_sections: function() {
    let m = this._R_M(this.RE_other_sections);
    let section = Number.parseInt(m[1], 10);

    return section;
  },

  /**
   * extended-parameter := (extended-initial-name "="
   *                        extended-initial-value) /
   *                       (extended-other-names "="
   *                        extended-other-values)
   * extended-initial-name := attribute [initial-section] "*"
   * extended-other-names := attribute other-sections "*"
   * extended-initial-value := [charset] "'" [language] "'"
   *                           extended-other-values
   * charset := <registered character set name>
   * language := <registered language tag [RFC-1766]>
   * Errata ID: 477
   *
   * @return {arMIMEParamSection}
   */
  RE_charset: /^[^\t\n\r \"\(\),\.\/:-@\[-\]\?\']+/,
  RE_language: /^[^\t\n\r \"\(\),\.\/:-@\[-\]\?\']+/,
  extended_parameter: function() {
    return this._ONEOF(
      () => {
        let ret = new arMIMEParamSection();

        ret.name = this.attribute();
        this._O(() => {
          ret.section = this.initial_section();
        });
        this._C("*");
        ret.extended = true;
        this._C("=");
        this._O(() => {
          ret.charset = this._R(this.RE_charset);
        });
        this._C("'");
        this._O(() => {
          ret.language = this._R(this.RE_language);
        });
        this._C("'");
        ret.value = this.extended_other_values();

        return ret;
      },
      () => {
        let ret = new arMIMEParamSection();

        ret.name = this.attribute();
        this._O(() => {
          ret.section = this.other_sections();
        });
        this._C("*");
        ret.extended = true;
        this._C("=");
        ret.value = this.extended_other_values();

        return ret;
      }
    );
  },

  /**
   * extended-other-values := *(ext-octet / attribute-char)
   *
   * @returns {string}
   */
  extended_other_values: function() {
    let ret = [];

    this._0N(() => {
      ret.push(this._ONEOF(
        () => this.ext_octet(),
        () => this.attribute_char()
      ));
    });

    return ret.join("");
  },

  /**
   * ext-octet := "%" 2(DIGIT / "A" / "B" / "C" / "D" / "E" / "F")
   *
   * @returns {string}
   *          デコードした文字列
   */
  RE_ext_octet: /^%([0-9A-Fa-f]{2})/,
  ext_octet: function() {
    let m = this._R_M(this.RE_ext_octet);
    return String.fromCharCode(Number.parseInt(m[1], 16));
  },

  /**
   * encoded-word := "=?" charset ["*" language] "?" encoding "?"
   *                 encoded-text "?="
   * charset = token    ; see section 3
   * encoding = token   ; see section 4
   * encoded-text = 1*<Any printable ASCII character other than "?"
   *                   or SPACE>
   *                ; (but see "Use of encoded-words in message
   *                ; headers", section 5)
   * (From RFC 2047)
   *
   * @returns {Tuple.<boolean,string>}
   *            [true,デコードした文字列]
   *            [false,デコードしてない文字列]
   */
  RE_encoded_text: /^[^ \?]+/,
  encoded_word: function() {
    let orig = [];

    orig.push(this._C("=?"));
    let charset = this.ew_token_noast(); orig.push(charset);
    this._O(() => {
      orig.push(this._C("*"));
      let language = this.ew_token_noast(); orig.push(language);
    });
    orig.push(this._C("?"));
    let encoding = this.ew_token().toUpperCase(); orig.push(encoding);
    orig.push(this._C("?"));
    let text = this._R(this.RE_encoded_text); orig.push(text);
    orig.push(this._C("?="));

    if (encoding == "Q") {
      let context = new arMIMEParser(text);
      let t = context.quoted_printable(true);
      context.__END();
      t = arUconv.toUnicode(t, charset);
      this.isUnicode = true;
      return [true, t];
    } else if (encoding == "B") {
      let t = safe_atob(text);
      t = arUconv.toUnicode(t, charset);
      this.isUnicode = true;
      return [true, t];
    }

    return [false, orig.join("")];
  },

  /* == RFC 2387: The MIME Multipart/Related Content-type == */

  /**
   * related-param   := [ ";" "start" "=" cid ]
   *                    [ ";" "start-info"  "="
   *                       ( cid-list / value ) ]
   *                    [ ";" "type"  "=" type "/" subtype ]
   *                    ; order independent
   * cid-list        := cid cid-list
   * value           := token / quoted-string    ; c.f. [MIME]
   *                       ; value cannot begin with "<"
   *
   * @return {string}
   */
  related_param_start: function() {
    return this.cid();
  },

  /**
   * cid             := msg-id     ; c.f. [822]
   *
   * @return {string}
   */
  cid: function() {
    return this._ONEOF(
      () => this.msg_id(),
      () => {
        this._O(() => this.CFWS());
        this._C("<");
        let s = this.local_part();
        this._C(">");
        this._O(() => this.CFWS());

        return s;
      },
      () => {
        this._O(() => this.CFWS());
        let s = this.local_part();
        this._O(() => this.CFWS());

        return s;
      }
    );
  },

  /* == RFC 2392: Content-ID and Message-ID Uniform Resource Locators == */

  /*
   * content-id    = url-addr-spec
   * message-id    = url-addr-spec
   * url-addr-spec = addr-spec  ; URL encoding of RFC 822 addr-spec
   * cid-url       = "cid" ":" content-id
   * mid-url       = "mid" ":" message-id [ "/" content-id ]
   */

  /* == RFC 2557: MIME Encapsulation of Aggregate Documents, such as HTML == */

  /**
   * content-location =   "Content-Location:" [CFWS] URI [CFWS]
   *
   * RFC 2017 の URL-parameter か URI
   * encoded-word が含まれている場合があるのでこれも対処する
   *
   * @returns {string}
   */
  content_location__value: function() {
    this._O(() => this.CFWS());
    let ret = this._ONEOF(
      () => this.URL_parameter(),
      () => this.unstructured_ew()
    );
    this._O(() => this.CFWS());

    return ret;
  },

  /* == RFC 3629: UTF-8, a transformation format of ISO 10646 == */

  /* 4. Syntax of UTF-8 Byte Sequences */

  /**
   * UTF8-2      = %xC2-DF UTF8-tail
   * UTF8-3      = %xE0 %xA0-BF UTF8-tail / %xE1-EC 2( UTF8-tail ) /
   *               %xED %x80-9F UTF8-tail / %xEE-EF 2( UTF8-tail )
   * UTF8-4      = %xF0 %x90-BF 2( UTF8-tail ) / %xF1-F3 3( UTF8-tail ) /
   *               %xF4 %x80-8F 2( UTF8-tail )
   * UTF8-tail   = %x80-BF
   */

  /* == RFC 3676: The Text/Plain Format and DelSp Parameters == */

  /**
   * flowed-body      = *( paragraph / fixed-line / sig-sep )
   * paragraph        = 1*flowed-line fixed-line
   *                    ; all lines in paragraph MUST be unquoted or
   *                    ; have same quote depth
   * flowed-line      = ( flowed-line-qt / flowed-line-unqt ) flow CRLF
   * flowed-line-qt   = quote ( ( stuffing stuffed-flowed ) /
   *                            unstuffed-flowed )
   * flowed-line-unqt = ( stuffing stuffed-flowed ) / unstuffed-flowed
   * stuffed-flowed   = *text-char
   * unstuffed-flowed = non-sp-quote *text-char
   * fixed-line       = fixed-line-qt / fixed-line-unqt
   * fixed-line-qt    = quote ( ( stuffing stuffed-fixed ) /
   *                            unstuffed-fixed ) CRLF
   * fixed-line-unqt  = ( stuffed-fixed / unstuffed-fixed ) CRLF
   * stuffed-fixed    = *text-char non-sp
   * unstuffed-fixed  = non-sp-quote [ *text-char non-sp ]
   * sig-sep          = [ quote [stuffing] ] "--" SP CRLF
   * quote-mark       = ">"
   * quote            = 1*quote-mark
   * stuffing         = SP ; space-stuffed, added on generation if
   *                       ; needed, deleted on reception
   * flow             = SP ; space before CRLF indicates flowed line,
   *                       ; if DelSp=yes, space was added on generation
   *                       ; and is deleted on reception
   * non-sp-quote     = < any character except NUL, CR, LF, SP, quote-mark >
   * non-sp           = non-sp-quote / quote-mark
   * text-char        = non-sp / SP
   *
   * 末尾の CRLF をオプショナルにする
   *
   * @param   {boolean} delsp
   * @returns {string}
   */
  RE_sig_sep: /^(>*) ?(-- )(?:(\r\n)|$)/,
  RE_flowed_or_fixed_line: /^(>*) ?([^\0\n\r]*?)( ?)(?:(\r\n)|$)/,
  flowed_body: function(delsp) {
    let last_quote = "";
    let last_flowed = false;
    let last_crlf = "";
    let ret = [];

    this._0N(() => {
      this._ONEOF(
        () => {
          let m = this._R_M(this.RE_sig_sep);
          let quote = m[1];
          let line = m[2];
          let crlf = m[3];
          ret.push(quote + line);
          if (crlf) {
            ret.push("\r\n");
          }

          last_flowed = false;
          last_crlf = crlf;
          last_quote = quote;
        },
        () => {
          let m = this._R_M(this.RE_flowed_or_fixed_line);
          let quote = m[1];
          let line = m[2];
          let flow = m[3];
          let crlf = m[4];

          if (last_quote != quote) {
            if (last_flowed) {
              ret.push("\r\n");
              last_flowed = false;
            }
          }

          if (!last_flowed) {
            ret.push(quote);
          }

          ret.push(line);
          if (!delsp) {
            ret.push(flow);
          }
          if (!flow) {
            if (crlf) {
              ret.push("\r\n");
            }
          }

          last_flowed = flow == " ";
          last_crlf = crlf;
          last_quote = quote;
        }
      );
    });

    return ret.join("");
  },

  /* == RFC 4155: The application/mbox Media Type == */

  /* message[RFC 5322] でまとめて解析する */

  /* == RFC 5234: ABNF == */

  /* #B.1.  Core Rules */

  /**
   * CR             =  %x0D
   *                        ; carriage return
   * CRLF           =  CR LF
   *                        ; Internet standard newline
   * LF             =  %x0A
   *                        ; linefeed
   *
   * @returns {string}
   */
  CRLF: function() {
    return this._C("\r\n");
  },

  /**
   * VCHAR          =  %x21-7E
   *                        ; visible (printing) characters
   * VCHAR          =/  UTF8-non-ascii
   * (Updated by RFC 6532)
   *
   * @returns {string}
   */
  /*
    使われてない
    RE_VCHAR_1N: /^[\!-\x7e\x80-\xff]+/,
    VCHAR_1N: function() {
    return this._R(this.RE_VCHAR_1N);
    },
  */

  /**
   *  WSP            =  SP / HTAB
   *                        ; white space
   *
   * @returns {string}
   */
  RE_WSP_0N: /^[\t ]*/,
  WSP_0N: function() {
    return this._R(this.RE_WSP_0N);
  },
  RE_WSP_1N: /^[\t ]+/,
  WSP_1N: function() {
    return this._R(this.RE_WSP_1N);
  },

  /* == RFC 5322: Internet Message Format == */

  /* #3.  Syntax
   * #3.2.  Lexical Tokens
   * #3.2.1.  Quoted characters */

  /**
   * quoted-pair     =   ("\" (VCHAR / WSP)) / obs-qp
   * obs-NO-WS-CTL   =   %d1-8 /            ; US-ASCII control
   *                     %d11 /             ;  characters that do not
   *                     %d12 /             ;  include the carriage
   *                     %d14-31 /          ;  return, line feed, and
   *                     %d127              ;  white space characters
   * obs-qp          =   "\" (%d0 / obs-NO-WS-CTL / LF / CR)
   *
   * @returns {string}
   *          バックスラッシュを除いた文字
   */
  RE_quoted_pair: /^\\([\0-\x7f])/,
  quoted_pair: function() {
    let m = this._R_M(this.RE_quoted_pair);
    return m[1];
  },

  /* #3.2.2.  Folding White Space and Comments */

  /**
   * FWS             =   ([*WSP CRLF] 1*WSP) /  obs-FWS
   *                                        ; Folding white space
   * obs-FWS         =   1*([CRLF] WSP)
   * Errata ID: 1908
   *
   * obs-FWS だけ使う
   *
   * @returns {string}
   */
  FWS: function() {
    let ret = [];

    this._1N(() => {
      this._O(() => this.CRLF());
      ret.push(this.WSP_1N());
    });

    return ret.join("");
  },

  /**
   * ctext           =   %d33-39 /          ; Printable US-ASCII
   *                     %d42-91 /          ;  characters not including
   *                     %d93-126 /         ;  "(", ")", or "\"
   *                     obs-ctext
   * obs-ctext       =   obs-NO-WS-CTL
   * ccontent        =   ctext / quoted-pair / comment / encoded_word
   * comment         =   "(" *([FWS] ccontent) [FWS] ")"
   * (Updated by RFC 2047)
   * ctext           =/  UTF8-non-ascii
   * (Updated by RFC 6532)
   */
  RE_ctext_1n: /^[\x01-\x08\x0b\x0c\x0e-\x1f\!-\'\*-\[\]-\xff]+/,
  comment: function() {
    this._C("(");
    this._0N(() => {
      this._O(() => this.FWS());

      this._ONEOF(
        /* encoded-word をデコードしても使わないので ctext に消費させる */
        () => this._R(this.RE_ctext_1n),
        () => this.quoted_pair(),
        () => this.comment()
      );
    });
    this._O(() => this.FWS());
    this._C(")");
  },

  /**
   * CFWS            =   (1*([FWS] comment) [FWS]) / FWS
   *
   * @returns {string}
   */
  CFWS: function() {
    return this._ONEOF(() => {
      let s = [];

      this._1N(() => {
        let t = [];
        this._O(() => {
          t.push(this.FWS());
        });
        this.comment();
        s.push(t.join(""));
      });
      this._O(() => {
        s.push(this.FWS());
      });

      return s.join("");
    }, () => this.FWS());
  },

  /* #3.2.3.  Atom */

  /**
   * atext           =   ALPHA / DIGIT /    ; Printable US-ASCII
   *                     "!" / "#" /        ;  characters not including
   *                     "$" / "%" /        ;  specials.  Used for atoms.
   *                     "&" / "'" /
   *                     "*" / "+" /
   *                     "-" / "/" /
   *                     "=" / "?" /
   *                     "^" / "_" /
   *                     "`" / "{" /
   *                     "|" / "}" /
   *                     "~"
   * atext           =/  UTF8-non-ascii
   * (Updated by RFC 6532)
   *
   * @returns {string}
   */
  RE_atext_1n: /^[\!#-\'\*\+\-\/-9\x3d\?A-Z\^-\x7e\x80-\xff]+/,
  atext_1n: function() {
    return this._tryFromUTF8(this._R(this.RE_atext_1n));
  },

  /**
   * atom            =   [CFWS] 1*atext [CFWS]
   *
   * @param   {boolean=} noCFWS
   * @returns {string}
   */
  atom: function(noCFWS=false) {
    let ret = "";

    if (!noCFWS) {
      this._O(() => this.CFWS());
    }
    ret = this.atext_1n();
    if (!noCFWS) {
      this._O(() => this.CFWS());
    }

    return ret;
  },

  /**
   * dot-atom-text   =   1*atext *("." 1*atext)
   *
   * @returns {string}
   */
  dot_atom_text: function() {
    let ret = [this.atext_1n()];
    this._0N(() => {
      ret.push(this._C(".") + this.atext_1n());
    });

    return ret.join("");
  },

  /**
   * dot-atom        =   [CFWS] dot-atom-text [CFWS]
   *
   * @returns {string}
   */
  dot_atom: function() {
    let ret = "";

    this._O(() => this.CFWS());
    ret = this.dot_atom_text();
    this._O(() => this.CFWS());

    return ret;
  },

  /* #3.2.4.  Quoted Strings */

  /**
   * qtext           =   %d33 /             ; Printable US-ASCII
   *                     %d35-91 /          ;  characters not including
   *                     %d93-126 /         ;  "\" or the quote character
   *                     obs-qtext
   * obs-qtext       =   obs-NO-WS-CTL
   * qcontent        =   qtext / quoted-pair
   * quoted-string   =   [CFWS]
   *                     DQUOTE ((1*([FWS] qcontent) [FWS]) / FWS) DQUOTE
   *                     [CFWS]
   * Errata ID: 3135
   * qtext           =/  UTF8-non-ascii
   * (Updated by RFC 6532)
   *
   * @param   {boolean=} noCFWS
   * @returns {string}
   *          quote を解除した文字列
   */
  RE_qtext_1n: /^[\x01-\x08\x0b\x0c\x0e-\x1f\!#-\[\]-\xff]+/,
  quoted_string: function(noCFWS=false) {
    let ret = "";

    if (!noCFWS) {
      this._O(() => this.CFWS());
    }
    this._C("\"");
    ret = this._ONEOF(() => {
      let s = [];

      this._1N(() => {
        let t = "";

        this._O(() => {
          t = this.FWS();
        });
        s.push(t + this._ONEOF(
          () => this._tryFromUTF8(this._R(this.RE_qtext_1n)),
          () => this.quoted_pair()
        ));
      });
      this._O(() => {
        s.push(this.FWS());
      });

      return s.join("");
    }, () => this.FWS());
    this._C("\"");
    if (!noCFWS) {
      this._O(() => this.CFWS());
    }

    return ret;
  },

  /* #3.2.5.  Miscellaneous Tokens */

  /**
   * word            =   atom / quoted-string
   *
   * @param   {boolean=} noCFWS
   * @returns {string}
   */
  word: function(noCFWS=false) {
    return this._ONEOF(
      () => this.atom(noCFWS),
      () => this.quoted_string(noCFWS)
    );
  },

  /**
   * phrase          =   1*(encoded-word-seq / word) / obs-phrase
   * obs-phrase      =   (encoded-word-seq / word)
   *                     *(encoded-word-seq / word / "." / CFWS)
   * (Updated by RFC 2047)
   *
   * obs-phrase だけ使う
   *
   * @returns {string}
   *          スペース区切りの文字列
   */
  phrase: function() {
    let ret = [];

    this._O(() => this.CFWS());
    ret.push(this._ONEOF(
      () => this.encoded_word_seq(true),
      () => this.word(true)
    ));
    this._0N(() => {
      let s = [];

      this._O(() => {
        s.push(this.CFWS());
      });
      s.push(this._ONEOF(
        () => this.encoded_word_seq(true),
        () => this.word(true),
        () => this._C(".")
      ));

      ret.push(s.join(""));
    });
    this._O(() => this.CFWS());

    return ret.join("");
  },

  /**
   * unstructured    =   (*([FWS] VCHAR) *WSP) / obs-unstruct
   * obs-utext       =   %d0 / obs-NO-WS-CTL / VCHAR
   * obs-unstruct    =   *( (*CR 1*(obs-utext / FWS)) / 1*LF ) *CR
   * Errata ID: 1905
   *
   * 入力として CRLF のみを想定するので
   * obs-unstruct の独立した CR と LF は無視する
   *
   * [simplified]
   *
   * 前後の FWS は削除する
   *
   * unstructured =   *FWS
   *                     *(*FWS 1*obs-utext) *FWS
   *
   * @returns {string}
   */
  RE_utext_1n: /^[^\t\n\r ]+/,
  unstructured: function() {
    let ret = [];

    this._0N(() => this.FWS());
    this._0N(() => {
      let s = [];
      this._0N(() => {
        s.push(this.FWS());
      });
      s.push(this._R(this.RE_utext_1n));
      ret.push(s.join(""));
    });
    this._0N(() => this.FWS());

    return ret.join("");
  },

  /* #3.3.  Date and Time Specification */

  /**
   * date-time       =   [ day-of-week "," ] date time [CFWS]
   * date            =   day month year
   *
   * @returns {Date}
   */
  date_time: function() {
    let wDay = -1;
    this._O(() => {
      let w = this.day_of_week();
      this._C(",");

      wDay = w;
    });
    let day = this.day();
    let month = this.month();
    let year = this.year();
    let hour = this.hour();
    this._C(":");
    let minute = this.minute();
    let second = 0;
    this._O(() => {
      this._C(":");
      second = this.second();
    });
    let [zone_sign, zone_hour, zone_minute] = this.zone();
    this._O(() => this.CFWS());

    let ret = new Date(Date.UTC(year, month, day,
                                hour - zone_sign * zone_hour,
                                minute - zone_sign * zone_minute,
                                second));

    return ret;
  },

  /**
   * day-of-week     =   ([FWS] day-name) / obs-day-of-week
   * obs-day-of-week =   [CFWS] day-name [CFWS]
   * day-name        =   "Mon" / "Tue" / "Wed" / "Thu" /
   *                     "Fri" / "Sat" / "Sun"
   *
   * obs-day-of-week だけ使う
   *
   * @returns {number}
   *          日曜日からの日数
   */
  RE_day_of_week_Mon: /^Mon/i,
  RE_day_of_week_Tue: /^Tue/i,
  RE_day_of_week_Wed: /^Wed/i,
  RE_day_of_week_Thu: /^Thu/i,
  RE_day_of_week_Fri: /^Fri/i,
  RE_day_of_week_Sat: /^Sat/i,
  RE_day_of_week_Sun: /^Sun/i,
  day_of_week: function() {
    this._O(() => this.CFWS());
    let ret = this._ONEOF(
      () => { this._R(this.RE_day_of_week_Mon); return 1; },
      () => { this._R(this.RE_day_of_week_Tue); return 2; },
      () => { this._R(this.RE_day_of_week_Wed); return 3; },
      () => { this._R(this.RE_day_of_week_Thu); return 4; },
      () => { this._R(this.RE_day_of_week_Fri); return 5; },
      () => { this._R(this.RE_day_of_week_Sat); return 6; },
      () => { this._R(this.RE_day_of_week_Sun); return 0; }
    );
    this._O(() => this.CFWS());

    return ret;
  },

  /**
   * day             =   ([FWS] 1*2DIGIT FWS) / obs-day
   * obs-day         =   [CFWS] 1*2DIGIT [CFWS]
   *
   * obs-day だけ使う
   *
   * @returns {number}
   */
  RE_day: /^[0-9]{1,2}/,
  day: function() {
    this._O(() => this.CFWS());
    let ret = Number.parseInt(this._R(this.RE_day), 10);
    this._O(() => this.CFWS());

    return ret;
  },

  /**
   * month           =   "Jan" / "Feb" / "Mar" / "Apr" /
   *                     "May" / "Jun" / "Jul" / "Aug" /
   *                     "Sep" / "Oct" / "Nov" / "Dec"
   *
   * @returns {number}
   *          1 月 を 0 とする
   */
  RE_month_Jan: /^Jan/i,
  RE_month_Feb: /^Feb/i,
  RE_month_Mar: /^Mar/i,
  RE_month_Apr: /^Apr/i,
  RE_month_May: /^May/i,
  RE_month_Jun: /^Jun/i,
  RE_month_Jul: /^Jul/i,
  RE_month_Aug: /^Aug/i,
  RE_month_Sep: /^Sep/i,
  RE_month_Oct: /^Oct/i,
  RE_month_Nov: /^Nov/i,
  RE_month_Dec: /^Dec/i,
  month: function() {
    let ret = this._ONEOF(
      () => { this._R(this.RE_month_Jan); return 0; },
      () => { this._R(this.RE_month_Feb); return 1; },
      () => { this._R(this.RE_month_Mar); return 2; },
      () => { this._R(this.RE_month_Apr); return 3; },
      () => { this._R(this.RE_month_May); return 4; },
      () => { this._R(this.RE_month_Jun); return 5; },
      () => { this._R(this.RE_month_Jul); return 6; },
      () => { this._R(this.RE_month_Aug); return 7; },
      () => { this._R(this.RE_month_Sep); return 8; },
      () => { this._R(this.RE_month_Oct); return 9; },
      () => { this._R(this.RE_month_Nov); return 10; },
      () => { this._R(this.RE_month_Dec); return 11; }
    );

    return ret;
  },

  /**
   * year            =   (FWS 4*DIGIT FWS) / obs-year
   * obs-year        =   [CFWS] 2*DIGIT [CFWS]
   *
   * obs-year だけ使う
   *
   * @returns {number}
   */
  RE_year: /^[0-9]{2,}/,
  year: function() {
    this._O(() => this.CFWS());
    let ret = Number.parseInt(this._R(this.RE_year), 10);
    this._O(() => this.CFWS());

    if (ret < 50) {
      ret += 2000;
    } else if (ret < 100) {
      ret += 1900;
    }

    return ret;
  },

  /**
   * hour            =   2DIGIT / obs-hour
   * obs-hour        =   [CFWS] 2DIGIT [CFWS]
   *
   * obs-hour だけ使う
   *
   * @returns {number}
   */
  RE_hour: /^[0-9]{2}/,
  hour: function() {
    this._O(() => this.CFWS());
    let ret = Number.parseInt(this._R(this.RE_hour), 10);
    this._O(() => this.CFWS());

    return ret;
  },

  /**
   * minute          =   2DIGIT / obs-minute
   * obs-minute      =   [CFWS] 2DIGIT [CFWS]
   *
   * obs-minute だけ使う
   *
   * @returns {number}
   */
  RE_minute: /^[0-9]{2}/,
  minute: function() {
    this._O(() => this.CFWS());
    let ret = Number.parseInt(this._R(this.RE_minute), 10);
    this._O(() => this.CFWS());

    return ret;
  },

  /**
   * second          =   2DIGIT / obs-second
   * obs-second      =   [CFWS] 2DIGIT [CFWS]
   *
   * obs-second だけ使う
   *
   * @returns {number}
   */
  RE_second: /^[0-9]{2}/,
  second: function() {
    this._O(() => this.CFWS());
    let ret = Number.parseInt(this._R(this.RE_second), 10);
    this._O(() => this.CFWS());

    return ret;
  },

  /**
   * zone            =   (FWS ( "+" / "-" ) 4DIGIT) / obs-zone
   * obs-zone        =   "UT" / "GMT" /     ; Universal Time
   *                                        ; North American UT
   *                                        ; offsets
   *                     "EST" / "EDT" /    ; Eastern:  - 5/ - 4
   *                     "CST" / "CDT" /    ; Central:  - 6/ - 5
   *                     "MST" / "MDT" /    ; Mountain: - 7/ - 6
   *                     "PST" / "PDT" /    ; Pacific:  - 8/ - 7
   *                                        ;
   *                     %d65-73 /          ; Military zones - "A"
   *                     %d75-90 /          ; through "I" and "K"
   *                     %d97-105 /         ; through "Z", both
   *                     %d107-122          ; upper and lower case
   *
   * obs-minute/obs-second が FWS を消費しているのでオプショナルにする
   *
   * @returns {Tuple.<number,number,number>}
   *          符号、時間、分
   */
  RE_zone: /^([\+\-])([0-9]{2})([0-9]{2})/,
  RE_zone_UT: /^UT/i,
  RE_zone_GMT: /^GMT/i,
  RE_zone_EDT: /^EDT/i,
  RE_zone_EST: /^EST/i,
  RE_zone_CDT: /^CDT/i,
  RE_zone_CST: /^CST/i,
  RE_zone_MDT: /^MDT/i,
  RE_zone_MST: /^MST/i,
  RE_zone_PDT: /^PDT/i,
  RE_zone_PST: /^PST/i,
  RE_zone_other: /^[A-Za-z]{3,5}/,
  RE_zone_single: /^[A-IK-Za-ik-z]/,
  zone: function() {
    this._O(() => this.FWS());

    return this._ONEOF(
      () => {
        let m = this._R_M(this.RE_zone);
        let sign = 1;
        if (m[1] == "-") {
          sign = -1;
        }
        return [sign, Number.parseInt(m[2]), Number.parseInt(m[3])];
      },

      () => { this._R(this.RE_zone_UT); return [+1, 0, 0]; },
      () => { this._R(this.RE_zone_GMT); return [+1, 0, 0]; },

      () => { this._R(this.RE_zone_EDT); return [-1, 4, 0]; },
      () => { this._R(this.RE_zone_EST); return [-1, 5, 0]; },
      () => { this._R(this.RE_zone_CDT); return [-1, 5, 0]; },
      () => { this._R(this.RE_zone_CST); return [-1, 6, 0]; },
      () => { this._R(this.RE_zone_MDT); return [-1, 6, 0]; },
      () => { this._R(this.RE_zone_MST); return [-1, 7, 0]; },
      () => { this._R(this.RE_zone_PDT); return [-1, 7, 0]; },
      () => { this._R(this.RE_zone_PST); return [-1, 8, 0]; },

      /* Other multi-character alphabetic time zones */
      () => { this._R(this.RE_zone_other); return [-1, 0, 0]; },

      () => {
        let c = this._R(this.RE_zone_single).toUpperCase();
        if (c == "A") { return [+1, 1, 0]; }
        if (c == "B") { return [+1, 2, 0]; }
        if (c == "C") { return [+1, 3, 0]; }
        if (c == "D") { return [+1, 4, 0]; }
        if (c == "E") { return [+1, 5, 0]; }
        if (c == "F") { return [+1, 6, 0]; }
        if (c == "G") { return [+1, 7, 0]; }
        if (c == "H") { return [+1, 8, 0]; }
        if (c == "I") { return [+1, 9, 0]; }

        if (c == "K") { return [+1, 10, 0]; }
        if (c == "L") { return [+1, 11, 0]; }
        if (c == "M") { return [+1, 12, 0]; }

        if (c == "N") { return [-1, 1, 0]; }
        if (c == "O") { return [-1, 2, 0]; }
        if (c == "P") { return [-1, 3, 0]; }
        if (c == "Q") { return [-1, 4, 0]; }
        if (c == "R") { return [-1, 5, 0]; }
        if (c == "S") { return [-1, 6, 0]; }
        if (c == "T") { return [-1, 7, 0]; }
        if (c == "U") { return [-1, 8, 0]; }
        if (c == "V") { return [-1, 9, 0]; }
        if (c == "W") { return [-1, 10, 0]; }
        if (c == "X") { return [-1, 11, 0]; }
        if (c == "Y") { return [-1, 12, 0]; }

        /* Z */
        return [+1, 0, 0];
      }
    );
  },

  /* #3.4.  Address Specification */

  /**
   * address         =   mailbox / group
   *
   * @returns {arMIMEMailBox}
   */
  address: function() {
    return this._ONEOF(
      () => this.mailbox(),
      () => this.group()
    );
  },

  /**
   * mailbox         =   name-addr / addr-spec
   *
   * @returns {arMIMEMailBox}
   */
  mailbox: function() {
    return this._ONEOF(
      () => this.name_addr(),
      () => {
        let ret = new arMIMEMailBox();

        ret.addr = this.addr_spec();

        return ret;
      }
    );
  },

  /**
   * name-addr       =   [display-name] angle-addr
   * display-name    =   phrase
   *
   * @returns {arMIMEMailBox}
   */
  name_addr: function() {
    let ret = new arMIMEMailBox();

    this._O(() => {
      ret.name = this.phrase();
    });
    ret.addr = this.angle_addr();

    return ret;
  },

  /**
   * angle-addr      =   [CFWS] "<" addr-spec ">" [CFWS] /
   *                     obs-angle-addr
   * obs-angle-addr  =   [CFWS] "<" obs-route addr-spec ">" [CFWS]
   *
   * @return {string}
   */
  angle_addr: function() {
    this._O(() => this.CFWS());
    this._C("<");
    this._O(() => this.obs_route());
    let ret = this.addr_spec();
    this._C(">");
    this._O(() => this.CFWS());

    return ret;
  },

  /**
   * group           =   display-name ":" [group-list] ";" [CFWS]
   *
   * @return {arMIMEMailBox}
   */
  group: function() {
    let ret = new arMIMEMailBox();

    ret.isGroup = true;
    ret.name = this.phrase();
    this._C(":");
    this._O(() => {
      ret.list = this.group_list();
    });
    this._C(";");
    this._O(() => this.CFWS());

    return ret;
  },

  /**
   * mailbox-list    =   (mailbox *("," mailbox)) / obs-mbox-list
   * obs-mbox-list   =   *([CFWS] ",") mailbox *("," [mailbox / CFWS])
   *
   * obs-mbox-list だけ使う
   *
   * @return {Array.<arMIMEMailBox>}
   */
  mailbox_list: function() {
    let ret = [];

    this._0N(() => {
      this._O(() => this.CFWS());
      this._C(",");
    });
    ret.push(this.mailbox());
    this._0N(() => {
      this._C(",");
      this._O(() => this._ONEOF(
        () => {
          ret.push(this.mailbox());
        },
        () => this.CFWS()
      ));
    });

    return ret;
  },

  /**
   * address-list    =   (address *("," address)) / obs-addr-list
   * obs-addr-list   =   *([CFWS] ",") address *("," [address / CFWS])
   *
   * obs-addr-list だけ使う
   *
   * @return {Array.<arMIMEMailBox>}
   */
  address_list: function() {
    let ret = [];

    this._0N(() => {
      this._O(() => this.CFWS());
      this._C(",");
    });
    ret.push(this.address());
    this._0N(() => {
      this._C(",");
      this._O(() => this._ONEOF(
        () => {
          ret.push(this.address());
        },
        () => this.CFWS()
      ));
    });

    return ret;
  },

  /**
   * group-list      =   mailbox-list / CFWS / obs-group-list
   * obs-group-list  =   1*([CFWS] ",") [CFWS]
   *
   * @return {Array.<arMIMEMailBox>}
   */
  group_list: function() {
    let ret;

    this._ONEOF(
      () => {
        ret = this.mailbox_list();
      },
      () => {
        this._1N(() => {
          this._O(() => this.CFWS());
          this._C(",");
        });
        this._O(() => this.CFWS());

        ret = [];
      },
      () => {
        this.CFWS();
        ret = [];
      }
    );

    return ret;
  },

  /* #3.4.1.  Addr-Spec Specification */

  /**
   * addr-spec       =   local-part "@" domain
   *
   * @return {string}
   */
  addr_spec: function() {
    let ret = [];

    ret.push(this.local_part());
    ret.push(this._C("@"));
    ret.push(this.domain());

    return ret.join("");
  },

  /**
   * local-part      =   dot-atom / quoted-string / obs-local-part
   * obs-local-part  =   word *("." word)
   *
   * @return {string}
   */
  local_part: function() {
    return this._ONEOF(
      () => this.quoted_string(),
      () => {
        let s = [this.word()];
        this._0N(() => {
          s.push(this._C(".") + this.word());
        });

        return s.join("");
      }
    );
  },

  /**
   * domain          =   dot-atom / domain-literal / obs-domain
   * obs-domain      =   atom *("." atom)
   *
   * dot-atom の代わりに obs-domain だけ使う
   *
   * @return {string}
   */
  domain: function() {
    return this._ONEOF(
      () => this.domain_literal(),
      () => {
        let s = [this.atom()];
        this._0N(() => {
          s.push(this._C(".") + this.atom());
        });

        return s.join("");
      }
    );
  },

  /**
   * domain-literal  =   [CFWS] "[" *([FWS] dtext) [FWS] "]" [CFWS]
   *
   * @return {string}
   */
  domain_literal: function() {
    let ret = [];

    this._O(() => this.CFWS());
    this._C("[");
    this._0N(() => {
      let s = "";

      this._O(() => {
        s = this.FWS();
      });
      ret.push(s + this._ONEOF(
        () => this.dtext_1n(),
        () => this.quoted_pair()
      ));
    });
    this._O(() => {
      ret.push(this.FWS());
    });
    this._C("]");
    this._O(() => this.CFWS());

    return ret.join("");
  },

  /**
   * dtext           =   %d33-90 /          ; Printable US-ASCII
   *                     %d94-126 /         ;  characters not including
   *                     obs-dtext          ;  "[", "]", or "\"
   * obs-dtext       =   obs-NO-WS-CTL / quoted-pair
   * dtext           =/  UTF8-non-ascii
   * (Updated by RFC 6532)
   *
   * ただし quoted-pair は含まない
   *
   * @returns {string}
   */
  RE_dtext_1n: /^[\x01-\x08\x0b\x0c\x0e-\x1f\!-Z\^-\xff]+/,
  dtext_1n: function() {
    return this._tryFromUTF8(this._R(this.RE_dtext_1n));
  },

  /* #3.5.  Overall Message Syntax */

  /**
   * message         =   (fields / obs-fields)
   *                     [CRLF body]
   * fields          =  *optional-field
   * field-name      =   1*ftext
   * ftext           =   %d33-57 /          ; Printable US-ASCII
   *                     %d59-126           ;  characters not including
   *                                        ;  ":".
   * body            =   (*(*998text CRLF) *998text) / obs-body
   * obs-body        =   *(%d0-127)
   * Errata ID: 1906
   * obs-fields      =  *obs-optional
   * obs-optional    =   field-name *WSP ":" unstructured CRLF
   *
   * mbox-from       = "From" " " *<SPACE or HTAB>
   *                   (<any char except ":", SPACE or HTAB>
   *                    *<any char except CR and LF>) CRLF
   * (From RFC 4155)
   *
   * 解析は後回しなので全て obs-optional として読む
   *
   * @returns {arMIMEPart}
   */
  RE_from_4155: /^From [ \t]*(?:[^: \t\r\n][^\r\n]*)?(?:\r\n|\r|\n)/i,
  RE_ftext_1n: /^[\!-9;-\x7e]+/,
  message: function() {
    let part = new arMIMEPart();

    this._O(() => {
      this._R(this.RE_from_4155);
    });
    this._0N(() => {
      let name = this._R(this.RE_ftext_1n);
      this.WSP_0N();
      this._C(":");
      let value = this.unstructured();
      this.CRLF();

      part.addField(name, value);
    });
    this._O(() => {
      this.CRLF();
      part.body = this._ALL_CHARS();
    });

    return part;
  },

  /* #3.6.  Field Definitions */

  /**
   * msg-id          =   [CFWS] "<" id-left "@" id-right ">" [CFWS]
   * id-left         =   dot-atom-text / obs-id-left
   * obs-id-left     =   local-part
   * id-right        =   dot-atom-text / no-fold-literal / obs-id-right
   * obs-id-right    =   domain
   *
   * @returns {string}
   */
  msg_id: function() {
    let ret = [];

    this._O(() => this.CFWS());
    this._C("<");
    ret.push(this.local_part());
    ret.push(this._C("@"));
    ret.push(this.domain());
    this._C(">");
    this._O(() => this.CFWS());

    return ret.join("");
  },

  /* #3.6.1.  The Origination Date Field */

  /**
   * orig-date       =   "Date:" date-time CRLF
   *
   * @returns {Date}
   */
  orig_date__value: function() {
    return this.date_time();
  },

  /* #3.6.5.  Informational Fields */

  /**
   * subject         =   "Subject:" unstructured CRLF
   *
   * encoded-word が入る
   *
   * subject         =   "Subject:" unstructured-ew CRLF
   *
   * @returns__value {string}
   */
  subject__value: function() {
    return this.unstructured_ew();
  },

  /* #4.  Obsolete Syntax
   * #4.4.  Obsolete Addressing */

  /**
   * obs-route       =   obs-domain-list ":"
   * obs-domain-list =   *(CFWS / ",") "@" domain
   *                     *("," [CFWS] ["@" domain])
   */
  obs_route: function() {
    this._0N(() => {
      this._ONEOF(
        () => this.CFWS(),
        () => this._C(",")
      );
    });
    this._C("@");
    this.domain();
    this._0N(() => {
      this._C(",");
      this._O(() => this.CFWS());
      this._O(() => {
        this._C("@");
        this.domain();
      });
    });

    this._C(":");
  },

  /* == RFC 6532: Internationalized Email Headers == */

  /* #3.  Changes to Message Header Fields
   * #3.1.  UTF-8 Syntax and Normalization */

  /**
   * UTF8-non-ascii  =   UTF8-2 / UTF8-3 / UTF8-4
   * UTF8-2          =   <Defined in Section 4 of RFC3629>
   * UTF8-3          =   <Defined in Section 4 of RFC3629>
   * UTF8-4          =   <Defined in Section 4 of RFC3629>
   */

  /* #3.2.  Syntax Extensions to RFC 5322 */

  /**
   * text    =/  UTF8-non-ascii
   *               ; note that this upgrades the body to UTF-8
   */

  /* == RFC 6838: Media Type Specifications and Registration Procedures == */

  /* FIXME */
});

module.arMIMEParser = arMIMEParser;
module.arMIMEParserError = arMIMEParserError;
module.safe_atob = safe_atob;

})(this);
