"use strict";

this.EXPORTED_SYMBOLS = ["arMIMEPart"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * MIME の各パート情報
 *
 * @constructor
 */
function arMIMEPart() {
  /* ---- Content-Type: (content[RFC2045]) ---- */

  /**
   * 解析結果
   * encode 時に必要
   * @type {arMIMEParams}
   */
  this.contentTypeParams = null;

  /**
   * type[RFC2045]
   * @type {string}
   */
  this.contentType = "";

  /**
   * subtype[RFC2045]
   * @type {string}
   */
  this.contentSubType = "";

  /**
   * MIME-Type
   * @type {string}
   */
  this.mimetype = "";

  /**
   * charset[???]
   * @type {string}
   */
  this.charset = "";

  /**
   * Format Parameter[RFC3676]
   * @type {string}
   */
  this.format = "";

  /**
   * DelSp Parameter[RFC3676]
   * @type {string}
   */
  this.delsp = false;

  /**
   * マルチパートか
   * encode 時に必要
   * @type {boolean}
   */
  this.isMultipart = false;

  /**
   * マルチパートが破損しているか
   * @type {boolean}
   */
  this.isCorrupted = false;

  /**
   * multipart/mixed 相当か
   * 実際には不明な subtype も mixed として扱う
   * @type {boolean}
   */
  this.isMixed = false;

  /**
   * boundary[RFC2046]
   * encode 時に必要
   * @type {string}
   */
  this.boundary = "";

  /**
   * "start"[RFC2387]
   * @type {string}
   */
  this.start = "";

  /**
   * "type"[RFC2387]
   * @type {string}
   */
  this.startMimetype = "";

  /* ---- Content-Location: (content-location[RFC2557]) ---- */

  /**
   * URI[RFC2557]
   * @type {string}
   */
  this.contentLocation = "";

  /* ---- Content-Disposition: (disposition[RFC2138]) ---- */

  /**
   * 解析結果
   * encode 時に必要
   * @type {arMIMEParams}
   */
  this.contentDispositionParams = null;

  /**
   * disposition-type[RFC2138]
   * @type {string}
   */
  this.contentDispositionType = "";

  /**
   * filename-parm[RFC2138]
   * @type {string}
   */
  this.contentDispositionFilename = "";

  /* ---- Date: (orig_date[RFC5322]) ---- */

  /**
   * orig_date[RFC5322]
   * @type {?Date}
   */
  this.date = null;

  /* ---- Subject: (unstructured[RFC5322]) ---- */

  /**
   * Subject
   * @type {string}
   */
  this.subject = "";

  /* ---- Content-ID: (id[RFC2045]) ---- */

  /**
   * msg-id[RFC2045]
   * @type {string}
   */
  this.contentID = "";

  /* ---- Content-Transfer-Encoding: (encoding[RFC2045]) ---- */

  /**
   * mechanism[RFC2045]
   * encode 時に必要
   * @type {string}
   */
  this.contentTransferEncoding = "";

  /* ---- 他 ---- */

  /**
   * body[RFC5322]
   * encode 時に必要
   * @type {string}
   */
  this.body = "";

  /**
   * マルチパートの場合の body-part[RFC2046] の配列
   * encode 時に必要
   * @type {?Array.<arMIMEPart>}
   */
  this.parts = null;

  /**
   * 展開で使用するパラメータ
   * @type {?UnMHTExtractParam}
   */
  this.eParam = null;

  /**
   * 保存で使用するパラメータ
   * @type {?UnMHTSaveParam}
   */
  this.sParam = null;

  /**
   * フィールドのマップ
   *   { 小文字の名前: [[大文字の名前, 値], ...] }
   * @type {Map.<string,Array.<Tuple.<string,string>>>}
   */
  this._fields = new Map();

  Object.seal(this);
}
arMIMEPart.prototype = Object.freeze({
  /**
   * 開始パートを探す
   *
   * @returns {?arMIMEPart}
   *          開始パート
   *          適切なパートが無ければ null
   */
  findStartPart: function() {
    if (!this.isMultipart) {
      return this;
    }

    if (this.isMixed) {
      return this;
    }

    if (this.start) {
      let startContentID = this.start;
      let start
        = this.parts
        .find(p => p.contentID == startContentID);
      if (start) {
        return start.findStartPart();
      }
    }

    if (this.contentSubType == "related") {
      let startMimetype = this.startMimetype;
      let start
        = this.parts
        .find(p => p.mimetype == startMimetype);
      if (start) {
        return start.findStartPart();
      }
    }

    if (this.contentSubType == "alternative") {
      let reversedParts = this.parts.slice().reverse();

      let searchPart = type => {
        let start
          = reversedParts
          .find(p => p.mimetype == type);
        if (start) {
          return start;
        }

        start
          = reversedParts
          .find(p => {
            if (p.isMultipart) {
              let startPart = p.findStartPart();
              if (startPart && startPart.mimetype == type) {
                return true;
              }
            }
            return false;
          });
        if (start) {
          return start.findStartPart();
        }

        return null;
      };

      let p = searchPart("text/html");
      if (p) {
        return p;
      }
      p = searchPart("text/plain");
      if (p) {
        return p;
      }
    }

    if (this.parts.length) {
      return this.parts[0].findStartPart();
    }

    return null;
  },

  /**
   * 複製する
   *
   * @param   {Object} findPart
   *          検索するパート
   *            {
   *              from: {arMIMEPart}
   *                    検索するパート
   *              to:   {arMIMEPart}
   *                    (出力) from に対応する複製先のパート
   *            }
   * @returns {arMIMEPart}
   *          複製
   */
  clone: function(findPart) {
    let newPart = new arMIMEPart();

    if (findPart && this == findPart.from) {
      findPart.to = newPart;
    }

    for (let list of this._fields.values()) {
      for (let [name, value] of list) {
        newPart.addField(name, value);
      }
    }
    if (this.contentTypeParams) {
      newPart.contentTypeParams = this.contentTypeParams.clone();
    }
    newPart.contentType = this.contentType;
    newPart.contentSubType = this.contentSubType;
    newPart.mimetype = this.mimetype;
    newPart.charset = this.charset;
    newPart.format = this.format;
    newPart.delsp = this.delsp;
    newPart.isMultipart = this.isMultipart;
    newPart.isMixed = this.isMixed;
    newPart.boundary = this.boundary;
    newPart.start = this.start;
    newPart.startMimetype = this.startMimetype;
    newPart.contentLocation = this.contentLocation;
    if (this.contentDispositionParams) {
      newPart.contentDispositionParams = this.contentDispositionParams.clone();
    }
    newPart.contentDispositionType = this.contentDispositionType;
    newPart.contentDispositionFilename = this.contentDispositionFilename;
    newPart.contentID = this.contentID;
    newPart.contentTransferEncoding = this.contentTransferEncoding;
    newPart.body = this.body;
    if (this.parts) {
      newPart.parts = this.parts.map(p => p.clone(findPart));
    }

    return newPart;
  },

  /**
   * 破棄
   */
  destruct: function() {
    if (this.parts) {
      for (let p of this.parts) {
        p.destruct();
      }
      this.parts = null;
    }
    this._fields.clear();
    if (this.contentTypeParams) {
      this.contentTypeParams.destruct();
      this.contentTypeParams = null;
    }
    if (this.contentDispositionParams) {
      this.contentDispositionParams.destruct();
      this.contentDispositionParams = null;
    }
    this.body = "";
  },

  /**
   * フィールド名のリスト
   * @type {Array.<string>}
   */
  get fieldNames() {
    let a = [];
    for (let list of this._fields.values()) {
      let [name, value] = list[0];
      a.push(name);
    }
    return a;
  },

  /**
   * フィールドのリスト
   * @type {Array.<Tuple.<string,string>>}
   */
  get fieldList() {
    let a = [];
    for (let list of this._fields.values()) {
      for (let [name, value] of list) {
        a.push([name, value]);
      }
    }
    return a;
  },

  /**
   * フィールドを持っているかを返す
   *
   * @param   {string} name
   *          フィールド名
   * @returns {boolean}
   *          フィールドを持っているか
   */
  hasField: function(name) {
    let nameL = name.toLowerCase();
    return this._fields.has(nameL);
  },

  /**
   * フィールドを返す
   *
   * @param   {string} name
   *          フィールド名
   * @returns {string}
   *          フィールド値
   */
  getField: function(name) {
    let nameL = name.toLowerCase();
    let list = this._fields.get(nameL);
    let [origName, value] = list[0];
    return value;
  },

  /**
   * フィールドのリスト返す
   *
   * @param   {string} name
   *          フィールド名
   * @returns {Array.<string,string>}
   *          実際のフィールド名とフィールド値のリスト
   *          大文字小文字がそれぞれ別の場合あり
   */
  getFieldList: function(name) {
    let nameL = name.toLowerCase();
    return this._fields.get(nameL);
  },

  /**
   * フィールドを追加する
   *
   * @param   {string} name
   *          フィールド名
   * @param   {string} value
   *          フィールド値
   */
  addField: function(name, value) {
    let nameL = name.toLowerCase();
    if (this._fields.has(nameL)) {
      this._fields.get(nameL).push([name, value]);
    } else {
      this._fields.set(nameL, [[name, value]]);
    }
  },

  /**
   * フィールドを上書きする
   *
   * @param   {string} name
   *          フィールド名
   * @param   {string} value
   *          フィールド値
   */
  setField: function(name, value) {
    let nameL = name.toLowerCase();
    this._fields.set(nameL, [[name, value]]);
  },

  /**
   * フィールドを削除する
   *
   * @param   {string} name
   *          フィールド名
   */
  removeField: function(name) {
    let nameL = name.toLowerCase();
    this._fields.delete(nameL);
  },

  /**
   * Content-Type を設定する
   *
   * @param   {arMIMEParams} params
   *          Content-Type
   */
  setContentType: function(params) {
    this.contentTypeParams = params;
  }
});

module.arMIMEPart = arMIMEPart;

})(this);
