"use strict";

this.EXPORTED_SYMBOLS = ["arMIMEUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arPseudoID } = Cu.import("resource://unmht/modules/base/arPseudoID.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyServiceGetter(lazy, "MIMEService",
                                   "@mozilla.org/mime;1",
                                   "nsIMIMEService");

/**
 * 既知の MIME-Type リスト
 * @type {Object.<string,string>}
 */
const KnownExts = Object.freeze({
  "css":   "text/css",
  "mht":   "message/rfc822",
  "mhtml": "message/rfc822",
  "eml":   "message/rfc822",
  "js":    "text/javascript",
  "txt":   "text/plain",
  "jpg":   "image/jpeg",
  "jpeg":  "image/jpeg",
  "gif":   "image/gif",
  "png":   "image/png",
  "bmp":   "image/bmp",
  "ico":   "image/x-icon",
  "svg":   "image/svg+xml",
  "swf":   "application/swf",
  "jar":   "application/java-archive",
  "class": "application/octet-stream"
});

/**
 * MIME ユーティリティ
 *
 * @class
 */
let arMIMEUtils = Object.freeze({
  /**
   * boundary[RFC2046] を作成する
   *
   * @returns {string}
   *          バウンダリ文字列
   */
  createUniqueBoundary: function() {
    return ("----=_Part_"
            + arPseudoID.randHex().toUpperCase()
            + "_"
            + arPseudoID.randHex().toUpperCase()
            + "."
            + Date.now().toString(16).toUpperCase());
  },

  /**
   * 拡張子に対応した MIME-Type を返す
   *
   * @param   {string} ext
   *          拡張子
   * @returns {string}
   *          対応した MIME-Type
   */
  getTypeFromExtension: function(ext) {
    if (/^[sx]?html?$/i.test(ext)) {
      /* XHTML も text/html として扱う */
      return "text/html";
    }

    /* システム設定が特殊場合に供えて既知のものは先にチェック */
    if (ext.toLowerCase() in KnownExts) {
      return KnownExts[ext.toLowerCase()];
    }

    if (ext) {
      /* 未知なのでシステム設定を返す */
      try {
        return lazy.MIMEService.getTypeFromExtension(ext);
      } catch (e) {
      }
    }

    return "application/octet-stream";
  },

  /**
   * ファイルの内容から MIME-Type を推測する
   *
   * @param   {string} body
   *          ファイルの内容
   * @param   {string} fallback
   *          不明な場合の返り値
   * @returns {string}
   *          推測した MIME-Type
   */
  guessTypeFromBody: function(body, fallback) {
    if (/^\s*</.test(body)) {
      return "text/html";
    }

    if (body.startsWith("\xFF\xD8")) {
      return "image/jpeg";
    }

    if (body.startsWith("GIF89a") ||
        body.startsWith("GIF87a")) {
      return "image/gif";
    }

    if (body.startsWith("\x89PNG\x0D\x0A\x1A\x0A")) {
      return "image/png";
    }

    return fallback;
  }
});

module.arMIMEUtils = arMIMEUtils;

})(this);
