"use strict";

this.EXPORTED_SYMBOLS = ["arNetFileLoader",
                         "arNetFileLoaderStatus",
                         "arNetFileLoaderStatusMask"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { LoadContextInfo } = Cu.import("resource://gre/modules/LoadContextInfo.jsm", {});
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arMIMEDecoder } = Cu.import("resource://unmht/modules/base/arMIMEDecoder.jsm", {});

/**
 * 取得状況
 * @enum {number}
 */
const arNetFileLoaderStatus = Object.freeze({
  /** 未取得 */
  NA:     0x0000,

  /** ファイルからの取得 */
  FILE:   0x1001,

  /** キャッシュからの取得 */
  CACHE:       0x2011,
  CACHE_ANON:  0x2012,

  /** ネットワークからの取得 */
  NET:    0x4001,

  /** 取得失敗 */
  NG:     0x8001,
  /** 取得キャンセル */
  CANCEL: 0x8002
});
const arNetFileLoaderStatusMask = Object.freeze({
  FILE:   0x1000,
  CACHE:  0x2000,
  NET:    0x4000,
  ERROR:  0x8000,
});

/**
   * エラーコード
 * @enum {number}
 */
const arNetFileLoaderError = Object.freeze({
  /** HTTP ステータスがエラーの範囲 */
  HTTP_STATUS:      0x1001,
  /** Location フィールドが不正 */
  INVALID_REDIRECT: 0x1002,
  /** ファイルが無い */
  FILE_NOT_FOUND:   0x1003,
  /** URL が不正 */
  INVALID_LOCATION: 0x1004,
  /** ネットワークエラー */
  NETWORK_FAILED:   0x1005
});

/**
 * ファイルロードコンテキスト
 *
 * @constructor
 * @param   {nsIURI} location
 *          対象の URI
 * @param   {?nsIURI} referrer
 *          リファラ
 *          必要なければ null
 * @param   {boolean} useCache
 *          キャッシュを使用するか
 * @param   {arINetFileLoaderListener} listener
 *          イベントリスナ
 * @param   {nsIDOMWindow} window
 *          動作対象のウィンドウ
 * @implements nsIRequestObserver
 * @implements nsIStreamListener
 * @implements nsICacheListener
 */
function arNetFileLoadContext(location, referrer, useCache, listener,
                              window) {
  /**
   * 対象の URI
   * @type {nsIURI}
   */
  this._location = location;

  /**
   * リファラ
   * @type {?nsIURI}
   */
  this._referrer = referrer;

  /**
   * キャッシュを使用するか
   * @type {boolean}
   */
  this._useCache = useCache;

  /**
   * イベントリスナ
   * @type {arINetFileLoaderListener}
   */
  this._listener = listener;

  /**
   * プライベートモードか
   * @type {boolean}
   */
  this._isPrivate = false;
  if (window) {
    try {
      let { PrivateBrowsingUtils } = Cu.import("resource://gre/modules/PrivateBrowsingUtils.jsm", {});
      this._isPrivate = PrivateBrowsingUtils.isWindowPrivate(window);
    } catch (e) {
    }
  }

  /**
   * 取得したデータのサイズ
   * @type {number}
   */
  this._dataLength = 0;

  /**
   * 取得中のチャネル
   * @type {?nsIChannel}
   */
  this._channel = null;

  /**
   * Content-Type
   * @type {string}
   */
  this._contentType = "";

  /**
   * Date
   * @type {string}
   */
  this._lastModified = "";

  /**
   * バイナリ用入力ストリーム
   * @type {nsIBinaryInputStream}
   */
  this._bstream
    = Cc["@mozilla.org/binaryinputstream;1"]
    .createInstance(Ci.nsIBinaryInputStream);

  /**
   * 取得状況
   * @type {arNetFileLoaderStatus}
   */
  this._status = arNetFileLoaderStatus.NA;

  Object.seal(this);
}
arNetFileLoadContext.prototype = Object.freeze({
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIStreamListener,
                                         Ci.nsIRequestObserver,
                                         Ci.nsICacheListener,
                                         Ci.nsICacheEntryOpenCallback]),

  /**
   * 破棄
   * 自分で呼ぶ
   */
  destruct: function() {
    this._location = null;
    this._referrer = null;
    this._listener = null;
    this._channel = null;
    this._bstream = null;
  },

  /**
   * 取得を開始する
   */
  start: function() {
    this._status = arNetFileLoaderStatus.NA;
    this._channel = null;

    this.nextMethod();
  },

  /**
   * 取得をキャンセルする
   */
  cancel: function() {
    this._status = arNetFileLoaderStatus.CANCEL;

    if (this._channel) {
      try {
        this._channel.cancel(Cr.NS_BINDING_ABORTED);
      } catch (e) {
      }
    }

    this.destruct();
  },

  /**
   * 次の取得方法に移行する
   */
  nextMethod: function() {
    let old = this._status;
    switch (this._status) {
      case arNetFileLoaderStatus.NA: {
        if (this._location.scheme == "file") {
          this._status = arNetFileLoaderStatus.FILE;
        } else if (this._useCache) {
          this._status = arNetFileLoaderStatus.CACHE;
        } else {
          this._status = arNetFileLoaderStatus.NET;
        }
        break;
      }
      case arNetFileLoaderStatus.FILE: {
        return;
      }
      case arNetFileLoaderStatus.CACHE: {
        this._status = arNetFileLoaderStatus.CACHE_ANON;
        break;
      }
      case arNetFileLoaderStatus.CACHE_ANON: {
        this._status = arNetFileLoaderStatus.NET;
        break;
      }
      case arNetFileLoaderStatus.NET: {
        return;
      }
      case arNetFileLoaderStatus.NG: {
        return;
      }
      case arNetFileLoaderStatus.CANCEL: {
        return;
      }
    }

    switch (this._status) {
      case arNetFileLoaderStatus.FILE: {
        this._startFile().catch(e => {
          Cu.reportError(e + "\n" + e.stack);
        });
        break;
      }
      case arNetFileLoaderStatus.CACHE: {
          this._startCache(false);
        break;
      }
      case arNetFileLoaderStatus.CACHE_ANON: {
        this._startCache(true);
        break;
      }
      case arNetFileLoaderStatus.NET: {
        this._startNet();
        break;
      }
    }
  },

  /* ==== nsIRequestObserver ==== */

  /**
   * リクエスト開始のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {?nsISupports} context
   *          ユーザ定義
   */
  onStartRequest: function(request, context) {
    this._dataLength = 0;

    let size = 0;
    if (this._channel) {
      let httpChannel = null;
      try {
        httpChannel = this._channel.QueryInterface(Ci.nsIHttpChannel);

        /* アクセス可能かチェックする */
        let tmp = httpChannel.responseStatus;
      } catch (e) {
        httpChannel = null;
      }

      if (httpChannel) {
        if (httpChannel.responseStatus >= 400) {
          this._status = arNetFileLoaderStatus.NG;
          let errorInfo = new arNetFileError(arNetFileLoaderError.HTTP_STATUS,
                                             httpChannel.responseStatus.toString());
          if (this._listener) {
            this._listener.onStop(this, this._status, null,
                                  this._contentType, errorInfo);
          }
          this.destruct();
          return;
        }

        try {
          let contentLength = httpChannel.getResponseHeader("Content-Length");
          size = Number.parseInt(contentLength, 10);
        } catch (e) {
        }
      }
    }

    if (this._listener) {
      this._listener.onStart(this, this._status, this._location, size);
    }
  },

  /**
   * リクエスト終了のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {?nsISupports} context
   *          ユーザ定義
   * @param   {number} statusCode
   *          終了コード
   */
  onStopRequest: function(request, context, statusCode) {
    if (this._status == arNetFileLoaderStatus.CANCEL ||
        this._status == arNetFileLoaderStatus.NG) {
      return;
    }

    if (this._channel) {
      let httpChannel = null;
      try {
        httpChannel = this._channel.QueryInterface(Ci.nsIHttpChannel);
      } catch (e) {
      }

      if (httpChannel) {
        let loc = "";
        try {
          loc = httpChannel.getResponseHeader("Location");
        } catch (e) {
        }
        if (loc) {
          try {
            loc = this._location.resolve(loc);
            this._location = Services.io.newURI(loc, null, null);
          } catch (e) {
            let errorInfo = new arNetFileError(arNetFileLoaderError.INVALID_REDIRECT,
                                               loc);
            if (this._listener) {
              this._listener.onStop(this, arNetFileLoaderStatus.NG, null,
                                    this._contentType, errorInfo);
            }
            this.destruct();
            return;
          }
          this.start();
          return;
        }
        try {
          this._contentType = httpChannel.getResponseHeader("Content-Type");
        } catch (e) {
        }
        try {
          this._lastModified = httpChannel.getResponseHeader("Last-Modified");
        } catch (e) {
        }
      }
    }

    if (this._dataLength === 0 &&
        this._status & arNetFileLoaderStatusMask.CACHE) {
      this.nextMethod();
      return;
    }

    if (this._listener) {
      this._listener.onStop(this, this._status, this._channel,
                            this._contentType, null);
    }
    this.destruct();
  },

  /* ==== nsIStreamListener ==== */

  /**
   * データ到着のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {?nsISupports} context
   *          ユーザ定義
   * @param   {nsIInputStream} inputStream
   *          データを取得するストリーム
   * @param   {number} offset
   *          データの位置
   * @param   {number} count
   *          データの長さ
   */
  onDataAvailable: function(request, context, inputStream, offset, count) {
    if (this._status == arNetFileLoaderStatus.CANCEL ||
        this._status == arNetFileLoaderStatus.NG) {
      return;
    }

    this._bstream.setInputStream(inputStream);
    this._dataLength += count;
    let data = this._bstream.readBytes(count);
    if (this._listener) {
      this._listener.onData(this, this._status, data, count);
    }
  },

  /* ==== nsICacheListener ==== */

  /**
   * キャッシュエントリを使うかどうか
   *
   * @param   {nsICacheEntry} entry
   *          キャッシュの情報
   * @param   {nsIApplicationCache} appCache
   *          アプリケーションキャッシュ
   * @returns {Number}
   *          Ci.nsICacheEntryOpenCallback.*
   */
  onCacheEntryCheck: function (entry, appCache) {
    return Ci.nsICacheEntryOpenCallback.ENTRY_WANTED;
  },

  /**
   * キャッシュエントリが使用可能になったイベント (cache2)
   *
   * @param   {nsICacheEntry} entry
   *          キャッシュの情報
   * @param   {boolean} isNew
   *          新規項目か
   * @param   {nsIApplicationCache} appCache
   *          アプリケーションキャッシュ
   * @param   {nsresult} result
   *          リクエストの結果
   */
  onCacheEntryAvailable: function(entry, isNew, appCache, result) {
    if (!(this._status & arNetFileLoaderStatusMask.CACHE)) {
      return;
    }

    if (!entry || result != Cr.NS_OK) {
      this.nextMethod();
      return;
    }

    if (this._applyResponseHead(entry)) {
      return;
    }

    let istream = entry.openInputStream(0);
    if (istream.isNonBlocking()) {
      let pump = Cc["@mozilla.org/network/input-stream-pump;1"]
        .createInstance(Ci.nsIInputStreamPump);
      pump.init(istream, -1, -1, 0, 0, true);
      pump.asyncRead(this, null);
    } else {
      this.onStartRequest(null, null);

      this.onDataAvailable(null, null, istream, 0, entry.dataSize);
      istream.close();

      this.onStopRequest(null, null, Cr.NS_OK);
    }
  },

  /* ==== private ==== */

  /**
   * ファイルから取得を開始する
   */
  _startFile: Task.async(function*() {
    if (this._listener) {
      this._listener.onStatusChange(this, this._status);
    }

    let path = arFileUtils.URLSpecToPath(this._location.spec);
    let exists = yield OS.File.exists(path);
    if (!exists) {
      let errorInfo = new arNetFileError(arNetFileLoaderError.FILE_NOT_FOUND,
                                         path);
      if (this._listener) {
        this._listener.onStop(this, arNetFileLoaderStatus.NG, null,
                              this._contentType, errorInfo);
      }
      return;
    }

    let data = "";
    try {
      data = yield arFileUtils.readBinary(path);
    } catch (e) {
    }

    if (this._listener) {
      this._listener.onData(this, this._status,
                            data, data.length);
      this._listener.onStop(this, this._status, null,
                            this._contentType, null);
    }
  }),

  /**
   * キャッシュから取得を開始する
   *
   * @param  {boolean} isAnon
   *         anonymous キャッシュか
   */
  _startCache: function(isAnon) {
    if (this._listener) {
      this._listener.onStatusChange(this, this._status);
    }

    let info;
    try {
      /* COMPAT: Firefox 43 or older */
      info = LoadContextInfo.custom(this._isPrivate, isAnon, 0, false);
    } catch (e) {
      try {
        /* COMPAT: Firefox 44 or newer */
        info = LoadContextInfo.custom(this._isPrivate, isAnon, {appId: 0});
      } catch (e2) {
        /* COMPAT: Firefox 51 or newer */
        info = LoadContextInfo.custom(isAnon, {appId: 0});
      }
    }
    let storage = Services.cache2.diskCacheStorage(info, false);

    try {
      storage.asyncOpenURI(this._location, "", Ci.nsICacheStorage.OPEN_READONLY,
                           this);
    } catch (e) {
      this.nextMethod();
    }
  },

  /**
   * HTTP のレスポンスヘッダを必要であれば適用する
   *
   * @param   {nsICacheEntryDescriptor/nsICacheEntry} entry
   *          キャッシュの情報
   * @returns {string}
   *          適用したか
   */
  _applyResponseHead: function(entry) {
    let head = "";
    try {
      head = entry.getMetaDataElement("response-head");
    } catch (e) {
      return false;
    }

    let m = head.match(/^HTTP\/[0-9\.]+ ([0-9]+)[^\r\n]*\r?\n((?:.|[\r\n])*)$/);
    if (!m) {
      return false;
    }

    let [, code, fields] = m;
    if (code >= 400) {
      this.nextMethod();
      return true;
    }

    let part = arMIMEDecoder.decodeMessage(fields);
    if (!part) {
      return false;
    }

    if (part.hasField("Location")) {
      let loc = arMIMEDecoder.decodeLocation(part.getField("Location"));

      try {
        loc = this._location.resolve(loc);
        this._location = Services.io.newURI(loc, null, null);
      } catch (e) {
        this.nextMethod();
        return true;
      }
      this.start();
      return true;
    }

    if (part.hasField("Content-Type")) {
      this._contentType = part.getField("Content-Type");
    }

    if (part.hasField("Last-Modified")) {
      this._lastModified = part.getField("Last-Modified");
    }

    if (part.hasField("Content-Encoding")) {
      let encoding = part.getField("Content-Encoding");

      let converter = null;
      try {
        converter
          = Cc["@mozilla.org/streamconv;1?from=" + encoding + "&to=uncompressed"]
          .createInstance(Ci.nsIStreamConverter);
        converter.asyncConvertData(encoding, "uncompressed", this, null);
      } catch (e) {
        this.nextMethod();
        return true;
      }

      let listener = converter.QueryInterface(Ci.nsIStreamListener);

      let istream = entry.openInputStream(0);

      if (istream.isNonBlocking()) {
        let pump = Cc["@mozilla.org/network/input-stream-pump;1"]
          .createInstance(Ci.nsIInputStreamPump);
        pump.init(istream, -1, -1, 0, 0, true);
        pump.asyncRead(listener, null);
      } else {
        listener.onStartRequest(null, null);

        listener.onDataAvailable(null, null, istream, 0, entry.dataSize);
        istream.close();

        listener.onStopRequest(null, null, Cr.NS_OK);
      }

      return true;
    }

    return false;
  },

  /**
   * ネットワークから取得を開始する
   */
  _startNet: function() {
    if (this._listener) {
      this._listener.onStatusChange(this, this._status);
    }

    let channel;
    try {
      channel = Services.io.newChannelFromURI2(this._location,
                                               null, // aLoadingNode
                                               Services.scriptSecurityManager.getSystemPrincipal(),
                                               null, // aTriggeringPrincipal
                                               Ci.nsILoadInfo.SEC_NORMAL,
                                               Ci.nsIContentPolicy.TYPE_OTHER);
    } catch (e) {
      /* NS_ERROR_UNKNOWN_PROTOCOL 等が投げられる */
      let errorInfo = new arNetFileError(arNetFileLoaderError.INVALID_LOCATION,
                                         this._location.spec);
      if (this._listener) {
        this._listener.onStop(this, arNetFileLoaderStatus.NG, null,
                              this._contentType, errorInfo);
      }
      this.destruct();
      return;
    }

    try {
      channel.QueryInterface(Ci.nsIPrivateBrowsingChannel);
      channel.setPrivate(this._isPrivate);
    } catch (e) {
    }

    let httpChannel = null;
    try {
      httpChannel = channel.QueryInterface(Ci.nsIHttpChannel);
    } catch (e) {
    }
    if (httpChannel) {
      httpChannel.setRequestHeader("Accept", "*/*", false);
      if (this._referrer) {
        httpChannel.referrer = this._referrer;
      }
    }

    this._channel = channel;
    try {
      channel.asyncOpen(this, null);
    } catch (e) {
      let errorInfo = new arNetFileError(arNetFileLoaderError.NETWORK_FAILED,
                                         this._location.spec);
      if (this._listener) {
        this._listener.onStop(this, arNetFileLoaderStatus.NG, null,
                              this._contentType, errorInfo);
      }
      this.destruct();
    }
  }
});

/**
 * エラー情報
 *
 * @constructor
 * @param   {number} code
 *          エラーコード
 * @param   {*} value
 *          エラーごとの付加情報
 */
function arNetFileError(code, value) {
  /**
   * エラーコード
   * @type {arNetFileLoaderError}
   */
  this.code = code;

  /**
   * エラーごとの付加情報
   * @type {*}
   */
  this.value = value;

  Object.seal(this);
}
arNetFileError.prototype = Object.freeze({
  /**
   * 文字列にする
   *
   * @param   {arStringBundle} bundle
   *          StringBundle 管理
   * @returns {string}
   *          文字列
   */
  toStringWithBundle: function(bundle) {
    switch (this.code) {
      case arNetFileLoaderError.HTTP_STATUS: {
        return bundle.get("network.error.http",
                          [this.value]);
      }
      case arNetFileLoaderError.INVALID_REDIRECT: {
        return bundle.get("network.error.redirect",
                          [this.value]);
      }
      case arNetFileLoaderError.FILE_NOT_FOUND: {
        return bundle.get("network.error.not_found");
      }
      case arNetFileLoaderError.INVALID_LOCATION: {
        return bundle.get("network.error.invalid_url");
      }
      case arNetFileLoaderError.NETWORK_FAILED: {
        return bundle.get("network.error.connect");
      }
    }

    return "unknown";
  }
});

/**
 * 取得の結果
 *
 * @constructor
 * @param   {arNetFileLoaderStatus} status
 *          取得状況
 * @param   {string} contentType
 *          Content-Type
 * @param   {string} data
 *          取得したデータ
 * @param   {arNetFileError} errorInfo
 *          エラー情報
 */
function arNetFileResult(status, data, contentType, errorInfo) {
  /**
   * 取得状況
   * @type {arNetFileLoaderStatus}
   */
  this.status = status;

  /**
   * 取得したデータ
   * @type {string}
   */
  this.data = data;

  /**
   * Content-Type
   * @type {string}
   */
  this.contentType = contentType;

  /**
   * エラー情報
   * @type {arNetFileError}
   */
  this.errorInfo = errorInfo;

  Object.seal(this);
}
arNetFileResult.prototype = Object.freeze({
});

/**
 * ネットワークからのファイル取得
 *
 * @class
 */
let arNetFileLoader = Object.freeze({
  /**
   * データを取得してコールバックを呼ぶ
   *
   * @param   {nsIURI} location
   *          対象の URI
   * @param   {?nsIURI} referrer
   *          リファラ
   *          必要なければ null
   * @param   {boolean} useCache
   *          キャッシュを使用するか
   * @param   {nsIDOMWindow} window
   *          動作対象のウィンドウ
   * @returns {Promise.<arNetFileResult>}
   * @resolves 取得の結果
   */
  loadData: function(location, referrer, useCache, window) {
    return new Promise((resolve, reject) => {
      let listener = {
        onStatusChange: function(context, status) {
        },
        onStart: function(context, status, location, size) {
          this._data = "";
        },
        onStop: function(context, status, channel, contentType, errorInfo) {
          resolve(new arNetFileResult(status, this._data, contentType,
                                      errorInfo));
          this._data = "";
        },
        onData: function(context, status, data, count) {
          this._data += data;
        },
        _data: "",
      };
      this.createContext(location, referrer, useCache, listener, window)
        .start();

    });
  },

  /**
   * ファイルロードコンテキストを作成する
   *
   * @param   {nsIURI} location
   *          対象の URI
   * @param   {?nsIURI} referrer
   *          リファラ
   *          必要なければ null
   * @param   {boolean} useCache
   *          キャッシュを使用するか
   * @param   {arINetFileLoaderListener} listener
   *          イベントリスナ
   * @param   {nsIDOMWindow} window
   *          動作対象のウィンドウ
   * @returns {arNetFileLoadContext}
   *          ファイルロードコンテキスト
   */
  createContext: function(location, referrer, useCache, listener, window) {
    return new arNetFileLoadContext(location, referrer, useCache, listener,
                                    window);
  }
});

module.arNetFileLoader = arNetFileLoader;
module.arNetFileLoaderStatus = arNetFileLoaderStatus;
module.arNetFileLoaderStatusMask = arNetFileLoaderStatusMask;

})(this);
