"use strict";

this.EXPORTED_SYMBOLS = ["arPathUtils",
                         "arPathType",
                         "arPathTypeMask"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * パスの種類
 * @enum {number}
 */
const arPathType = Object.freeze({
  /** UNIX のローカル */
  UNIX_LOCAL:               0x001001,
  /** Windows のローカル */
  WINDOWS_LOCAL:            0x001011,
  /** Windows のネットワークドライブ */
  WINDOWS_NETWORK:          0x001012,
  /** Mac OS 9 のローカル */
  MACOS9_LOCAL:             0x001021,
  /** スキーム名の後にスラッシュを含む URI */
  URI:                      0x001101,
  /** スキーム名を含まない URI */
  URI_NO_SCHEME:            0x001102,
  /** スキーム名の後にスラッシュを含まない URI */
  URI_NO_SLASH:             0x011103,
  /** Windows のローカルの URI */
  WINDOWS_LOCAL_FILE_URI:   0x001111,
  /** Windows のネットワークドライブの URI */
  WINDOWS_NETWORK_FILE_URI: 0x001112,
  /** 相対パス */
  RELATIVE:                 0x002001
});

/**
 * パスの種類のマスク
 * @enum {number}
 */
const arPathTypeMask = Object.freeze({
  /** Windows のパス */
  WINDOWS:                  0x000010,
  /** Mac OS 9 のパス */
  MACOX9:                   0x000020,
  /** URI */
  URI:                      0x000100,
  /** 絶対パス */
  ABSOLUTE:                 0x001000,
  /** 相対パス */
  RELATIVE:                 0x002000,
  /** data スキーム */
  DATA:                     0x010000
});

/**
 * パス情報
 *
 * @constructor
 */
function arPathInfo() {
  /**
   * パスの種類
   * @type {arPathType}
  */
  this.type = 0;

  /**
   * パスの区切り
   * @type {string}
   */
  this.sep = "/";

  /**
   * パスの要素
   * @type {Array.<string>}
   */
  this.pathComponents = [];

  /**
   * Windows のドライブ文字
   * @type {string}
   */
  this.driveLetter = undefined;

  /**
   * Windows の file の余分なスラッシュ
   * @type {string}
   */
  this.extraSlash = undefined;

  /**
   * Mac OS 9 のボリュームラベル
   * @type {string}
   */
  this.volumeLabel = undefined;

  /**
   * data: のデータ
   * @type {string}
   */
  this.data = undefined;

  /**
   * URI のスキーム
   * @type {string}
   */
  this.scheme = undefined;

  /**
   * URI のホスト
   * @type {string}
   */
  this.host = undefined;

  /**
   * URI のポート
   * @type {string}
   */
  this.port = undefined;

  /**
   * URI のクエリ
   * @type {string}
   */
  this.query = undefined;

  /**
   * フラグメント
   * @type {string}
   */
  this.fragment = undefined;

  Object.seal(this);
}
arPathInfo.prototype = Object.freeze({
  /**
   * 複製する
   *
   * @returns {arPathInfo}
   */
  clone: function() {
    let info = new arPathInfo();

    info.type = this.type;
    info.sep = this.sep;
    info.pathComponents = this.pathComponents.slice(0);
    info.driveLetter = this.driveLetter;
    info.extraSlash = this.extraSlash;
    info.volumeLabel = this.volumeLabel;
    info.data = this.data;
    info.scheme = this.scheme;
    info.host = this.host;
    info.port = this.port;
    info.query = this.query;
    info.fragment = this.fragment;

    return info;
  }
});

/**
 * 仮想パス操作ユーティリティ
 * ローカルのパスや valid な URI ではない可能性のあるパスを操作する
 *
 * @class
 */
let arPathUtils = Object.freeze({
  /**
   * パスを連結、解決する
   *
   * @param   {string} baseDir
   *          親ディレクトリのパス
   *          末尾にセパレータが無くてもディレクトリ名として扱われる事に注意
   * @param   {string} sub
   *          相対パス、もしくは絶対パス
   * @returns {string}
   *          連結したパス
   */
  resolve: function(baseDir, sub) {
    if (!baseDir) {
      return sub;
    }

    let baseInfo = this.getPathInfo(baseDir);
    let subInfo = this.getPathInfo(sub);

    if (baseInfo.type & arPathTypeMask.DATA ||
        subInfo.type & arPathTypeMask.DATA) {
      return sub;
    }

    if (subInfo.type == arPathType.URI_NO_SCHEME &&
        baseInfo.scheme !== undefined) {
      return baseInfo.scheme + ":" + sub;
    }

    if (subInfo.type & arPathTypeMask.ABSOLUTE &&
        subInfo.type != arPathType.UNIX_LOCAL) {
      /* TYPE_UNIX_LOCAL は URI のパスでもあるのでそのまま返さない */
      return sub;
    }

    if (subInfo.type == arPathType.UNIX_LOCAL) {
      baseInfo.pathComponents = subInfo.pathComponents;
    } else {
      if (baseInfo.pathComponents.length > 0 &&
          baseInfo.pathComponents[baseInfo.pathComponents.length - 1] === "") {
        baseInfo.pathComponents.pop();
      }

      baseInfo.pathComponents
        = baseInfo.pathComponents.concat(subInfo.pathComponents);
    }

    baseInfo.query = subInfo.query;
    baseInfo.fragment = subInfo.fragment;

    this._simplifyPathComponents(baseInfo);

    return this.createPath(baseInfo, false);
  },

  /**
   * パスを比較しやすいように単純化する
   *
   * @param   {string} origpath
   *          対象のパス
   * @returns {string}
   *          単純化したパス
   */
  unique: function(origpath) {
    let info = this.getPathInfo(origpath);
    this._simplifyPathComponents(info);
    let path = "";

    if (info.type == arPathType.URI) {
      if (info.scheme == "http" && info.port == "80") {
        info.port = undefined;
      } else if (info.scheme == "https" && info.port == "443") {
        info.port = undefined;
      }
    }

    switch (info.type) {
      case arPathType.WINDOWS_LOCAL: {
        path = ("file:///" + info.driveLetter.toLowerCase() + ":/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathType.WINDOWS_NETWORK: {
        path = ("file://///" + info.host + "/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathType.WINDOWS_LOCAL_FILE_URI: {
        path = ("file:///" + info.driveLetter.toLowerCase() + ":/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathType.WINDOWS_NETWORK_FILE_URI: {
        path = ("file://///" + info.host + "/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathType.URI: {
        path = (info.scheme + "://" + info.host
                + ((info.port !== undefined) ? (":" + info.port) : "")
                + "/"
                + info.pathComponents.join("/")
                + ((info.query !== undefined) ? ("?" + info.query) : ""));
        break;
      }
      case arPathType.URI_NO_SCHEME: {
        path = ("//" + info.host
                + ((info.port !== undefined) ? (":" + info.port) : "")
                + "/"
                + info.pathComponents.join("/")
                + ((info.query !== undefined) ? ("?" + info.query) : ""));
        break;
      }
      case arPathType.URI_NO_SLASH: {
        path = (info.scheme + ":" + info.data);
        break;
      }
      case arPathType.UNIX_LOCAL: {
        path = ("/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathType.MACOS9_LOCAL: {
        path = (info.volumeLabel + ":"
                + info.pathComponents.join(":"));
        break;
      }
      case arPathType.RELATIVE: {
        path = (info.pathComponents.join("/")
                + ((info.query !== undefined) ? ("?" + info.query) : ""));
        break;
      }
    }

    if (info.fragment !== undefined) {
      path += "#" + info.fragment;
    }

    return path;
  },

  /**
   * フラグメントを削除する
   *
   * @param   {string} origPath
   *          対象のパス
   * @returns {string}
   *          フラグメントを削除したパス
   */
  removeFragment: function(origPath) {
    let [path, fragment] = this.splitFragment(origPath);
    return path;
  },

  /**
   * 親ディレクトリのパスを取得する
   *
   * @param   {string} path
   *          対象のパス
   * @returns {string}
   *          親ディレクトリのパス
   */
  getBaseDir: function(path) {
    let info = this.getPathInfo(path);

    if (info.type & arPathTypeMask.DATA) {
      return path;
    }
    if (info.pathComponents.length) {
      info.pathComponents.pop();
      info.pathComponents.push("");
    }
    info.fragment = undefined;
    info.query = undefined;

    return this.createPath(info);
  },

  /**
   * パス情報を取得する
   *
   * @param   {string} path
   *          対象のパス
   * @returns {arPathInfo}
   *          パス情報
   */
  getPathInfo: function(path) {
    let info = new arPathInfo();
    let m;

    /* -- 先にフラグメントは取っておく -- */

    [path, info.fragment] = this.splitFragment(path);

    /* -- 絶対パス -- */

    m = path.match(/^([A-Za-z]):([\/\\])((?:.|\r|\n)*)$/);
    if (m) {
      /*
       * C:/hoge/fuga.html
       * C:\hoge\fuga.html
       */
      info.type = arPathType.WINDOWS_LOCAL;
      info.driveLetter = m[1];
      info.sep = m[2];
      info.pathComponents = m[3].split(info.sep);
      return info;
    }

    m = path.match(/^\\\\([^\\]+)\\((?:.|\r|\n)*)$/);
    if (m) {
      /*
       * \\host1\hoge\fuga.html
       */
      info.type = arPathType.WINDOWS_NETWORK;
      info.sep = "\\";
      info.host = m[1];
      info.pathComponents = m[2].split(info.sep);
      return info;
    }

    m = path.match(/^file:\/\/(\/?)([A-Za-z]):([\/\\])(((?:.|\r|\n)*)*)$/);
    if (m) {
      /*
       * file://C:/hoge/fuga.html
       * file:///C:\hoge\fuga.html
       */
      info.type = arPathType.WINDOWS_LOCAL_FILE_URI;
      info.extraSlash = m[1];
      info.driveLetter = m[2];
      info.sep = m[3];
      info.pathComponents = m[4].split(info.sep);
      return info;
    }

    m = path.match(/^file:\/\/(\/{1,3})([^\/]+)\/(((?:.|\r|\n)*)*)$/);
    if (m) {
      /*
       * file://///host1/hoge/fuga.html
       */
      info.sep = "/";
      info.type = arPathType.WINDOWS_NETWORK_FILE_URI;
      info.extraSlash = m[1];
      info.host = m[2];
      info.pathComponents = m[3].split(info.sep);
      return info;
    }

    m = path.match(/^([A-Za-z0-9\-]+):\/\/([^\/]*)\/([^\?]*)(\?(((?:.|\r|\n)*)*))?$/);
    if (m) {
      /*
       * http://host1/hoge/fuga.html
       * http://host1/hoge/fuga.cgi?name=value
       */
      info.type = arPathType.URI;
      info.sep = "/";
      info.scheme = m[1];
      info.host = m[2];
      let m2 = info.host.match(/^(.+):([0-9]+)$/);
      if (m2) {
        info.host = m2[1];
        info.port = m2[2];
      }
      info.pathComponents = m[3].split(info.sep);
      if (m[4]) {
        info.query = m[5];
      }
      return info;
    }

    m = path.match(/^(about|data):(((?:.|\r|\n)*)*)$/);
    if (m) {
      /*
       * about:blank
       * data:text/html,hogehoge
       */
      info.type = arPathType.URI_NO_SLASH;
      info.sep = "/";
      info.scheme = m[1];
      info.data = m[2];
      return info;
    }

    m = path.match(/^\/\/([^\/]*)\/([^\?]*)(\?(((?:.|\r|\n)*)*))?$/);
    if (m) {
      /*
       * //host1/hoge/fuga.html
       * //host1/hoge/fuga.cgi?name=value
       */
      info.type = arPathType.URI_NO_SCHEME;
      info.sep = "/";
      info.host = m[1];
      let m2 = info.host.match(/^(.+):([0-9]+)$/);
      if (m2) {
        info.host = m2[1];
        info.port = m2[2];
      }
      info.pathComponents = m[2].split(info.sep);
      if (m[3]) {
        info.query = m[4];
      }
      return info;
    }

    m = path.match(/^\/((?:.|\r|\n)*)$/);
    if (m) {
      /*
       * /hoge/fuga.html
       */
      info.type = arPathType.UNIX_LOCAL;
      info.sep = "/";
      info.pathComponents = m[1].split(info.sep);
      return info;
    }

    m = path.match(/^([^:]+):((?:.|\r|\n)+)$/);
    if (m) {
      /*
       * Macintosh HD:hoge:fuga.html
       */
      info.type = arPathType.MACOS9_LOCAL;
      info.sep = ":";
      info.volumeLabel = m[1];
      info.pathComponents = m[2].split(info.sep);
      return info;
    }

    /* -- 相対パス -- */

    m = path.match(/^(\.\.?([\\\/:])[^\?]*)(\?(((?:.|\r|\n)*)*))?$/);
    if (m) {
      /*
       * ./hoge/fuga.html
       * ./hoge/fuga.html?name=value
       * ..\hoge\fuga.html
       */
      info.type = arPathType.RELATIVE;
      info.sep = m[2];
      if (m[3]) {
        info.query = m[4];
      }
      info.pathComponents = m[1].split(info.sep);
      return info;
    }

    info.type = arPathType.RELATIVE;

    /* **** FIXME: この方法では誤検出もありうる **** */
    if (path.indexOf("/") != -1) {
      /* とりあえず最もありうる / を最優先にする */
      info.sep = "/";
    } else if (path.indexOf("\\") != -1) {
      info.sep = "\\";
    } else {
      info.sep = "/";
    }

    m = path.match(/^([^\?]*)(\?(((?:.|\r|\n)*)*))?$/);
    if (m) {
      /*
       * hoge/fuga.cgi?name=value
       * hoge\\fuga.cgi?name=value
       */
      info.pathComponents = m[1].split(info.sep);
      if (m[2]) {
        info.query = m[3];
      }

      return info;
    }

    /* 基本的に到達しないハズ */
    info.pathComponents = [path];
    return info;
  },

  /**
   * パス情報からパスを生成する
   *
   * @param   {arPathInfo} info
   *          対象のパス情報
   * @param   {boolean} noMHTML
   *          mhtml を無視するか
   * @param   {boolean} noFragment
   *          fragment を無視するか
   * @returns {arPathInfo}
   *          パス
   */
  createPath: function(info, noMHTML, noFragment) {
    let result = "";

    switch (info.type) {
      case arPathType.WINDOWS_LOCAL: {
        result = (info.driveLetter + ":" + info.sep
                  + info.pathComponents.join(info.sep));
        break;
      }
      case arPathType.WINDOWS_NETWORK: {
        result = ("\\\\" + info.host + info.sep
                  + info.pathComponents.join(info.sep));
        break;
      }
      case arPathType.WINDOWS_LOCAL_FILE_URI: {
        result = ("file://" + info.extraSlash
                  + info.driveLetter + ":" + info.sep
                  + info.pathComponents.join(info.sep));
        break;
      }
      case arPathType.WINDOWS_NETWORK_FILE_URI: {
        result = ("file://" + info.extraSlash
                  + info.host + info.sep
                  + info.pathComponents.join(info.sep));
        break;
      }
      case arPathType.URI: {
        result = (info.scheme + "://" + info.host
                  + ((info.port !== undefined) ? (":" + info.port) : "")
                  + info.sep
                  + info.pathComponents.join(info.sep)
                  + ((info.query !== undefined) ? ("?" + info.query) : ""));
        break;
      }
      case arPathType.URI_NO_SCHEME: {
        result = ("//" + info.host
                  + ((info.port !== undefined) ? (":" + info.port) : "")
                  + info.sep
                  + info.pathComponents.join(info.sep)
                  + ((info.query !== undefined) ? ("?" + info.query) : ""));
        break;
      }
      case arPathType.URI_NO_SLASH: {
        result = (info.scheme + ":" + info.data);
        break;
      }
      case arPathType.UNIX_LOCAL: {
        result = (info.sep
                  + info.pathComponents.join(info.sep));
        break;
      }
      case arPathType.MACOS9_LOCAL: {
        result = (info.volumeLabel + info.sep
                  + info.pathComponents.join(info.sep));
        break;
      }
      case arPathType.RELATIVE: {
        result = (info.pathComponents.join(info.sep)
                  + ((info.query !== undefined) ? ("?" + info.query) : ""));
        break;
      }
    }

    if (!noFragment) {
      result += ((info.fragment !== undefined) ? ("#" + info.fragment) : "");
    }

    return result;
  },

  /**
   * 拡張子を取得する
   *
   * @param   {string} path
   *          対象のパス
   * @returns {string}
   *          拡張子
   */
  getExtension: function(path) {
    let info = this.getPathInfo(path);
    if (info.pathComponents.length === 0) {
      return "";
    }

    let leafName = info.pathComponents[info.pathComponents.length - 1];
    if (!leafName) {
      return "";
    }

    let m = leafName.match(/\.([^\.]+)$/);
    if (!m) {
      return "";
    }

    return m[1];
  },

  /**
   * パスとフラグメントを切り離す
   *
   * @param   {string} path
   *          対象のパス
   * @returns {Tuple.<string,string>}
   *          パスとフラグメント
   *          フラグメントは無い場合は undefined
   */
  splitFragment: function(path) {
    let m = path.match(/^((?:&#|[^#])*)(?:#((?:.|\r|\n)*))?$/);
    if (m) {
      return [m[1], m[2]];
    }
    return [path, undefined];
  },

  /* ==== private ==== */

  /**
   * パスの要素を簡略化する
   *
   * @param   {arPathInfo} info
   *          パス情報
   */
  _simplifyPathComponents: function(info) {
    if (info.type == arPathType.DATA) {
      return;
    }
    let tmp = [];
    let lastComponent = "-";
    if (info.type == arPathType.RELATIVE) {
      for (let component of info.pathComponents) {
        if (component == ".") {
          if (tmp.length === 0) {
            tmp.push(component);
          }
        } else if (component == "..") {
          if (tmp.length === 0) {
            tmp.push(component);
          } else if (tmp.length == 1 && tmp[0] == ".") {
            tmp.pop();
            tmp.push(component);
          } else if (tmp.length > 0 && tmp[tmp.length - 1] != "..") {
            tmp.pop();
          } else {
            tmp.push(component);
          }
        } else if (component) {
          tmp.push(component);
        }
        lastComponent = component;
      }
    } else {
      for (let component of info.pathComponents) {
        if (component == ".") {
        } else if (component == "..") {
          tmp.pop();
        } else if (component) {
          tmp.push(component);
        }
        lastComponent = component;
      }
    }
    if (lastComponent === "") {
      tmp.push(lastComponent);
    }
    info.pathComponents = tmp;
  }
});

module.arPathUtils = arPathUtils;
module.arPathType = arPathType;
module.arPathTypeMask = arPathTypeMask;

})(this);
