"use strict";

this.EXPORTED_SYMBOLS = ["arPref", "arPrefClass"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});
let { toUTF8, fromUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});

/**
 * 設定値
 * @typedef {(boolean|string|number)} arPrefValue
 */

/**
 * 設定操作と既定値
 *
 * @constructor
 * @param   {string} prefix
 *          設定のプレフィックス
 * @param   {Object.<string,arPrefValue>} prefs
 *          既定値
 * @param   {function(Map.<string,arPrefValue>):Promise.<void>} onPrefFileLoad
 *          設定ファイルロードイベント
 * @param   {function(Map.<string,arPrefValue>):Promise.<void>} onPrefFileSave
 *          設定ファイルセーブイベント
 */
function arPrefClass(prefix, defPrefs, onPrefFileLoad, onPrefFileSave) {
  /**
   * 設定のプレフィックス
   * @type {string}
   */
  this.prefix = prefix;

  /**
   * バージョン
   *  @type {string}
   */
  this._version = "";

  /**
   * 既定値
   * @type {Map.<string,arPrefValue>}
   */
  this._defPrefs = new Map();

  /**
   * 設定ファイルロードイベント
   *  @type {function(Map.<string,arPrefValue>)}
   */
  this._onPrefFileLoad = onPrefFileLoad;

  /**
   * 設定ファイルセーブイベント
   *  @type {function(Map.<string,arPrefValue>)}
   */
  this._onPrefFileSave = onPrefFileSave;

  /**
   * 設定値読み込み関数マップ
   * @type {Map.<string,function():arPrefValue>}
   */
  this._prefGetter = new Map();

  /**
   * 設定値保存関数マップ
   * @type {Map.<string,function(arPrefValue)>}
   */
  this._prefSetter = new Map();

  if (defPrefs) {
    this._setDefPrefs(defPrefs);
  }

  Object.seal(this);
}
arPrefClass.prototype = Object.freeze({
  /**
   * バージョン
   * @type {string}
   */
  get version() {
    return this._version;
  },

  /**
   * バージョン
   * @type {string}
   */
  set version(version) {
    this._version = version;
    this.setPref("version", version);
  },

  /**
   * 設定名一覧
   * @type {Array.<string>}
   */
  get names() {
    return [...this._defPrefs.keys()];
  },

  /**
   * 設定値の型を返す
   *
   * @param   {string} name
   *          設定名
   * @returns {string}
   *          設定値の型
   * @throws  {Error} Error("UNSUPPORTED_PREFERENCE")
   *          未知の設定名
   */
  getTypeof: function(name) {
    return (typeof(this.getDefPref(name)));
  },

  /* ---- 既定値 ---- */

  /**
   * 既定値を返す
   *
   * @param   {string} name
   *          設定名
   * @returns {string}
   *          設定値の型
   * @throws  {Error} Error("UNSUPPORTED_PREFERENCE")
   *          未知の設定名
   */
  getDefPref: function(name) {
    if (!this._defPrefs.has(name)) {
      throw new Error("UNSUPPORTED_PREFERENCE: " + name);
    }

    return this._defPrefs.get(name);
  },

  /* ---- ローカルのリスナ ---- */

  /**
   * 設定を監視する
   *
   * @param   {string} domain
   *          対象のドメイン
   * @param   {function} func
   *          設定が変更された時に実行する関数
   * @param   {boolean} run
   *          登録と同時に関数を実行するか
   */
  addListener: function(domain, func, run) {
    this.addGlobalListener(this.prefix + domain, func, run);
  },

  /* ---- グローバルのリスナ ---- */

  /**
   * 設定を監視する
   *
   * @param   {string} domain
   *          対象のドメイン
   * @param   {function} func
   *          設定が変更された時に実行する関数
   * @param   {boolean} run
   *          登録と同時に関数を実行するか
   */
  addGlobalListener: function(domain, func, run) {
    let observer = {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

      /* ==== nsIObserver ==== */

      /**
       * 通知を受けたイベント
       *
       * @param   {nsISupports} subject
       *          変更やアクションのあったオブジェクト
       * @param   {string} topic
       *          イベントの種類
       * @param   {string} data
       *          イベントのデータ
       */
      observe: function(subject, topic, data) {
        if (topic == "nsPref:changed") {
          func();
        }
      }
    };

    Services.prefs.addObserver(domain, observer, false);
    arShutdownEvent.addFunc((data, reason) => {
      Services.prefs.removeObserver(domain, observer);
    });

    if (run) {
      func();
    }
  },

  /* ---- 設定操作 ---- */

  /**
   * 設定値を読み込む
   *
   * @param   {string} name
   *          設定名
   * @returns {arPrefValue}
   *          設定値
   *          無ければ既定値を返す
   * @throws  {Error} Error("UNSUPPORTED_PREFERENCE")
   *          未知の設定名
   */
  getPref: function(name) {
    if (!this._prefGetter.has(name)) {
      throw new Error("UNSUPPORTED_PREFERENCE: " + name);
    }
    return this._prefGetter.get(name)();
  },

  /**
   * 設定値を保存する
   *
   * @param   {string} name
   *          設定名
   * @param   {arPrefValue} value
   *          設定値
   * @throws  {Error} Error("UNSUPPORTED_PREFERENCE")
   *          未知の設定名
   */
  setPref: function(name, value) {
    if (!this._prefSetter.has(name)) {
      throw new Error("UNSUPPORTED_PREFERENCE: " + name);
    }
    return this._prefSetter.get(name)(value);
  },

  /**
   * 設定値をクリアする
   *
   * @param   {string} name
   *          設定名
   * @throws  {Error} Error("UNSUPPORTED_PREFERENCE")
   *          未知の設定名
   */
  clearPref: function(name) {
    if (!this._prefSetter.has(name)) {
      throw new Error("UNSUPPORTED_PREFERENCE: " + name);
    }
    return Services.prefs.clearUserPref(this.prefix + name);
  },

  /* ---- ファイルから設定読み込み ---- */

  /**
   * ファイルからの設定の読み込みを開始する
   *
   * @param   {string} filePath
   *          読み込む対象のファイルのパス
   *          null にすると既定値を書いたファイル扱い
   * @returns {Promise.<arPrefFileInputObject>}
   * @resolves 設定入力オブジェクト
   *           実体は {Map.<string,arPrefValue>}
   *             <string 設定名, boolean/string/number 設定値>
   * @rejects {Error} ファイルが不正
   */
  initPrefFromFile: Task.async(function*(filePath) {
    if (!filePath) {
      return new Map();
    }

    let text = yield arFileUtils.readUTF8(filePath);

    let prefs = JSON.parse(text);
    let map = new Map();
    for (let name of Object.keys(prefs)) {
      map.set(name, prefs[name]);
    }

    if (this._onPrefFileLoad) {
      yield this._onPrefFileLoad(map);
    }

    return map;
  }),

  /**
   * ファイルからの設定の読み込みを終了する
   *
   * @param   {arPrefFileInputObject} input
   *          設定入力オブジェクト
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  closePrefFromFile: function(input) {
    /* 特にやる事はない */
    return Promise.resolve();
  },

  /**
   * ファイルから設定値を読み込む
   *
   * @param   {arPrefFileInputObject} input
   *          設定入力オブジェクト
   * @param   {string} name
   *          設定名
   * @returns {arPrefValue}
   *          設定値
   *          無ければ既定値を返す
   * @throws  {Error} Error("UNSUPPORTED_PREFERENCE")
   *          未知の設定名
   */
  getPrefFromFile: function(input, name) {
    if (input.has(name)) {
      return input.get(name);
    }

    return this.getDefPref(name);
  },

  /* ---- ファイルへ設定保存 ---- */

  /**
   * ファイルへの設定の保存を開始する
   *
   * @param   {string} filePath
   *          保存対象のファイル
   * @returns {Promise.<arPrefFileOutputObject>}
   * @resolves 設定出力オブジェクト
   */
  initPrefToFile: function(filePath) {
    return Promise.resolve({
      map: new Map(),
      filePath: filePath
    });
  },

  /**
   * ファイルへの設定の保存を終了する
   *
   * @param   {arPrefFileOutputObject} output
   *          設定出力オブジェクト
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  closePrefToFile: Task.async(function*(output) {
    if (this._onPrefFileSave) {
      yield this._onPrefFileSave(output.map);
    }

    let obj = {};
    for (let [name, value] of output.map) {
      obj[name] = value;
    }

    let data = JSON.stringify(obj, null, "  ");

    yield arFileUtils.createUTF8(output.filePath, data);

    output.map = null;
    output.filePath = null;
  }),

  /**
   * 設定を保存する
   *
   * @param   {arPrefFileOutputObject} output
   *          設定出力オブジェクト
   * @param   {string} name
   *          設定名
   * @param   {arPrefValue} value
   *          既定値
   */
  setPrefToFile: function(output, name, value) {
    output.map.set(name, value);
  },

  /* ==== private ==== */

  /**
   * 既定値を作成する
   *
   * @param   {Object.<string,arPrefValue>} prefs
   *          既定値
   */
  _setDefPrefs: function(prefs) {
    for (let name of Object.keys(prefs)) {
      this._defPrefs.set(name, prefs[name]);
    }
    this._createDefaultBranch();
    this._createPrefGetterSetter();
  },

  /**
   * デフォルト設定を作成する
   */
  _createDefaultBranch: function() {
    let branch = Services.prefs.getDefaultBranch("");
    for (let [name, value] of this._defPrefs) {
      switch (typeof(value)) {
        case "boolean": {
          branch.setBoolPref(this.prefix + name, value);
          break;
        }
        case "string": {
          branch.setCharPref(this.prefix + name, toUTF8(value));
          break;
        }
        case "number": {
          branch.setIntPref(this.prefix + name, value);
          break;
        }
      }
    }
  },

  /**
   * 設定操作関数マップを作成する
   */
  _createPrefGetterSetter: function() {
    for (let [name, value] of this._defPrefs) {
      let pname = this.prefix + name;
      switch (typeof(value)) {
        case "boolean": {
          this._prefGetter.set(name, () => {
            return Services.prefs.getBoolPref(pname);
          });
          this._prefSetter.set(name, value => {
            return Services.prefs.setBoolPref(pname, value);
          });
          break;
        }
        case "string": {
          this._prefGetter.set(name, () => {
            return fromUTF8(Services.prefs.getCharPref(pname));
          });
          this._prefSetter.set(name, value => {
            return Services.prefs.setCharPref(pname, toUTF8(value));
          });
          break;
        }
        case "number": {
          this._prefGetter.set(name, () => {
            return Services.prefs.getIntPref(pname);
          });
          this._prefSetter.set(name, value => {
            return Services.prefs.setIntPref(pname, value);
          });
          break;
        }
      }
    }
  }
});

/**
 * 汎用操作
 * @type {arPrefClass}
 */
let arPref = new arPrefClass();

module.arPref = arPref;
module.arPrefClass = arPrefClass;

})(this);
