"use strict";

this.EXPORTED_SYMBOLS = ["arScoped", "arScopeManager"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});

/**
 * スコープ付きオブジェクト管理
 *
 * @class
 * @template ScopeTargetType
 * @implements arIShutdownEventListener
 */
let arScopeManager = Object.freeze({
  /**
   * オブジェクトを追加する
   *
   * @param   {arScoped} object
   *          追加するオブジェクト
   */
  register: function(object) {
    this._registered.add(object);
    return object;
  },

  /**
   * スコープに関連付いたオブジェクトを破棄する
   *
   * @param   {ScopeTargetType} target
   *          スコープの対象
   */
  leave: function(target) {
    for (let object of this._registered) {
      object.leave(target);
    }
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    for (let object of this._registered) {
      object.destruct(true, data, reason);
    }
    this._registered.clear();
  },

  /* ==== private ==== */

  /**
   * 作成したオブジェクト
   * @type {Set.<arScoped>}
   */
  _registered: new Set(),

  /**
   * 初期化処理
   */
  init: function() {
    arShutdownEvent.addListener(this);
  }
});
arScopeManager.init();

/**
 * スコープ付きオブジェクト
 * スコープを抜けるかシャットダウンで破棄される
 * (スコープを抜けた事は arScoped.prototype.leave によって通知する)
 *
 * @constructor
 * @template ScopeTargetType
 * @param   {arIScopedClass} _class
 *          スコープごとのオブジェクトのクラス
 */
function arScoped(_class) {
  /**
   * スコープごとのオブジェクト
   * @type {Map.<ScopeTargetType,arIScopedClass>}
   */
  this._containers = new Map();

  /**
   * スコープごとのオブジェクトのクラス
   * @type {arIScopedClass}
   */
  this._class = _class;

  Object.seal(this);

  arScopeManager.register(this);
}
arScoped.prototype = Object.freeze({
  /**
   * オブジェクトを作成する
   * 既に存在すればそれを返す
   *
   * @param   {ScopeTargetType} target
   *          スコープ対象
   * @returns {arIScopedClass}
   *          オブジェクト
   */
  create: function(target) {
    if (!this._containers.has(target)) {
      let container = new this._class(target);
      this._containers.set(target, container);

      return container;
    }

    return this._containers.get(target);
  },

  /**
   * オブジェクトを返す
   *
   * @param   {ScopeTargetType} target
   *          スコープ対象
   * @returns {?arIScopedClass}
   *          オブジェクト
   *          無ければ null を返す
   */
  get: function(target) {
    if (!this._containers.has(target)) {
      return null;
    }

    return this._containers.get(target);
  },

  /**
   * 作成された全てのオブジェクトについて関数を適用する
   *
   * @param   {function(ScopeTargetType,arIScopedClass)} f
   *          適用する関数
   *            *param   {ScopeTargetType} target
   *                     スコープ対象
   *            *param   {arIScopedClass} container
   *                     オブジェクト
   */
  forEach: function(f) {
    for (let [target, container] of this._containers) {
      f(target, container);
    }
  },

  /**
   * スコープに関連付いたオブジェクトを破棄する
   *
   * @param   {ScopeTargetType} target
   *          スコープの対象
   */
  leave: function(target) {
    if (!this._containers.has(target)) {
      return;
    }

    let container = this._containers.get(target);
    this._containers.delete(target);
    container.destruct(false, null, null);
  },

  /**
   * 全てのオブジェクトを破棄する
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  destruct: function(isShutdown, data, reason) {
    for (let container of this._containers.values()) {
      container.destruct(isShutdown, data, reason);
    }
    this._containers.clear();
    this._containers = null;
  }
});

module.arScoped = arScoped;
module.arScopeManager = arScopeManager;

})(this);
