"use strict";

this.EXPORTED_SYMBOLS = ["arStylesheetLoader"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyServiceGetter(lazy, "StyleSheetService",
                                   "@mozilla.org/content/style-sheet-service;1",
                                   "nsIStyleSheetService");

/**
 * スタイルシート適用
 *
 * @class
 * @implements arIShutdownEventListener
 */
let arStylesheetLoader = Object.freeze({
  /**
   * スタイルシートを登録する
   *
   * @param   {string} uriSpec
   *          スタイルシートの URI
   * @param   {number} type
   *          スタイルシートの種類
   *            Ci.nsIStyleSheetService.AGENT_SHEET
   *            Ci.nsIStyleSheetService.USER_SHEET
   */
  load: function(uriSpec, type) {
    if (this._sheets.has(uriSpec)) {
      /* 多重登録はしない */
      return;
    }

    let uri = Services.io.newURI(uriSpec, null, null);
    lazy.StyleSheetService.loadAndRegisterSheet(uri, type);
    this._sheets.set(uriSpec, type);
  },

  /**
   * スタイルシートを再登録する
   *
   * @param   {string} uriSpec
   *          スタイルシートの URI
   * @param   {number} type
   *          スタイルシートの種類
   *            Ci.nsIStyleSheetService.AGENT_SHEET
   *            Ci.nsIStyleSheetService.USER_SHEET
   */
  reload: function(uriSpec, type) {
    this.unload(uriSpec, type);
    this.load(uriSpec, type);
  },

  /**
   * スタイルシートの登録を解除する
   * 登録したものはシャットダウン時に自動的に解除される
   *
   * @param   {string} uriSpec
   *          スタイルシートの URI
   * @param   {number} type
   *          スタイルシートの種類
   *            Ci.nsIStyleSheetService.AGENT_SHEET
   *            Ci.nsIStyleSheetService.USER_SHEET
   * @param   {boolean=} noRemove
   *          リストから削除しない (内部用)
   */
  unload: function(uriSpec, type, noRemove=false) {
    let uri = Services.io.newURI(uriSpec, null, null);
    if (lazy.StyleSheetService.sheetRegistered(uri, type)) {
      lazy.StyleSheetService.unregisterSheet(uri, type);
    }

    if (!noRemove) {
      this._sheets.delete(uriSpec);
    }
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    for (let [uriSpec, type] of this._sheets) {
      this.unload(uriSpec, type, true);
    }
  },

  /* ==== private ==== */

  /**
   * 登録したスタイルシートの URI と種類
   * @type {Map.<string,number>}
   */
  _sheets: new Map(),

  /**
   * 初期化処理
   */
  init: function() {
    arShutdownEvent.addListener(this);
  }
});
arStylesheetLoader.init();

module.arStylesheetLoader = arStylesheetLoader;

})(this);
