"use strict";

this.EXPORTED_SYMBOLS = ["arSystemDirectory"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});

/**
 * システムディレクトリ管理
 *
 * @constructor
 * @param   {string} systemName
 *          システムディレクトリ名
 * @implements arIShutdownEventListener
 */
function arSystemDirectory(systemName) {
  /**
   * システムディレクトリ名
   * @type {string}
   */
  this._systemName = systemName;

  /**
   * システムディレクトリ
   * {Profile ディレクトリ}/{_systemName}
   * @type {string}
   */
  this._systemDirectoryPath = null;

  /**
   * キャッシュディレクトリ
   * {Profile ディレクトリ}/{_systemName}/Cache
   * @type {string}
   */
  this._cacheDirectoryPath = null;

  arShutdownEvent.addListener(this);

  Object.seal(this);
}
arSystemDirectory.prototype = Object.freeze({
  /**
   * システムディレクトリを作成する
   *
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  makeSystemDirectory: Task.async(function*() {
    this._systemDirectoryPath = OS.Path.join(OS.Constants.Path.profileDir,
                                             this._systemName);
    let exists = yield OS.File.exists(this._systemDirectoryPath);
    if (!exists) {
      OS.File.makeDir(this._systemDirectoryPath);
    }
  }),

  /**
   * キャッシュディレクトリを作成する
   *
   * @returns {Promise.<void>}
   * @resolves 完了
   */
  makeCacheDirectory: Task.async(function*() {
    yield this.makeSystemDirectory();

    this._cacheDirectoryPath = OS.Path.join(this._systemDirectoryPath,
                                            "Cache");
    let exists = yield OS.File.exists(this._cacheDirectoryPath);
    if (!exists) {
      OS.File.makeDir(this._cacheDirectoryPath);
    }
  }),

  /**
   * キャッシュを削除する
   *
   * @param   {boolean} isShutdown
   *          シャットダウン中か
   */
  clearCacheDirectory: function(isShutdown) {
    let cacheDirectoryPath = this._cacheDirectoryPath;

    if (!cacheDirectoryPath) {
      return;
    }

    let p = Task.spawn(function*() {
      let exists = yield OS.File.exists(cacheDirectoryPath);
      if (!exists) {
        return;
      }

      let resIter = new OS.File.DirectoryIterator(cacheDirectoryPath);
      yield resIter.forEach(entry => {
        return OS.File.remove(entry.path);
      });
    }).catch(e => {
      Cu.reportError(e + "\n" + e.stack);
    });

    if (isShutdown) {
      let { AsyncShutdown } = Cu.import("resource://gre/modules/AsyncShutdown.jsm");
      let name = "arSystemDirectory.clearCacheDirectory";
      AsyncShutdown.profileBeforeChange.addBlocker(name, p);
    }
  },

  /* ---- ディレクトリ取得 ---- */

  /**
   * システムディレクトリ内のファイルを取得する
   *
   * @param   {string} name
   *          ファイル名
   * @returns {Promise.<string>}
   * @resolves ファイル
   */
  getSystemFilePath: Task.async(function*(name) {
    yield this.makeSystemDirectory();

    let path = OS.Path.join(this._systemDirectoryPath, name);

    return path;
  }),

  /**
   * システムディレクトリ内のファイルを取得する
   *
   * @param   {string} name
   *          ファイル名
   * @returns {Promise.<string>}
   * @resolves ファイル
   */
  getCacheFilePath: Task.async(function*(name) {
    yield this.makeCacheDirectory();

    let path = OS.Path.join(this._cacheDirectoryPath, name);

    return path;
  }),

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    let APP_SHUTDOWN = 2;
    this.clearCacheDirectory(reason == APP_SHUTDOWN);
  }
});

module.arSystemDirectory = arSystemDirectory;

})(this);
