"use strict";

this.EXPORTED_SYMBOLS = ["arWinRegUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * Windows レジストリ操作ユーティリティ
 *
 * @class
 */
let arWinRegUtils = Object.freeze({
  /**
   * HKEY_LOCAL_MACHINE 内の値を取得する
   *
   * @param   {string} path
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {?(string|number)}
   *          値
   *          エラーが発生した場合は null
   */
  getHKLMValue: function(path, name) {
    return this._getValue(Ci.nsIWindowsRegKey.ROOT_KEY_LOCAL_MACHINE,
                          path, name);
  },

  /**
   * HKEY_CURRENT_USER 内の値を取得する
   *
   * @param   {string} path
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {?(string|number)}
   *          エラーが発生した場合は null
   *          値
   */
  getHKCUValue: function(path, name) {
    return this._getValue(Ci.nsIWindowsRegKey.ROOT_KEY_CURRENT_USER,
                          path, name);
  },

  /**
   * HKEY_CLASSES_ROOT 内の値を取得する
   *
   * @param   {string} path
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {?(string|number)}
   *          値
   *          エラーが発生した場合は null
   */
  getHKCRValue: function(path, name) {
    return this._getValue(Ci.nsIWindowsRegKey.ROOT_KEY_LOCAL_MACHINE,
                          path, name);
  },

  /* ==== private ==== */

  /**
   * 指定されたルートキー内の値を取得する
   *
   * @param   {number} rootKey
   *          ルートキー
   *            Ci.nsIWindowsRegKey.ROOT_KEY_*
   * @param   {stirng} path
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {?(string|number)}
   *          値
   *          エラーが発生した場合は null
   */
  _getValue: function(rootKey, path, name) {
    let regKey
      = Cc["@mozilla.org/windows-registry-key;1"]
      .createInstance(Ci.nsIWindowsRegKey);

    let value = null;
    try {
      regKey.open(rootKey, path, Ci.nsIWindowsRegKey.ACCESS_READ);

      if (regKey) {
        switch (regKey.getValueType(name)) {
          case Ci.nsIWindowsRegKey.TYPE_STRING: {
            value = regKey.readStringValue(name);
            break;
          }
          case Ci.nsIWindowsRegKey.TYPE_BINARY: {
            value = regKey.readBinaryValue(name);
            break;
          }
          case Ci.nsIWindowsRegKey.TYPE_INT: {
            value = regKey.readIntValue(name);
            break;
          }
          case Ci.nsIWindowsRegKey.TYPE_INT64: {
            value = regKey.readInt64Value(name);
            break;
          }
        }
      }
    } catch (e) {
    }

    regKey.close();

    return value;
  }
});

module.arWinRegUtils = arWinRegUtils;

})(this);
