"use strict";

this.EXPORTED_SYMBOLS = ["arWindowOverlay"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arArrayUtils } = Cu.import("resource://unmht/modules/base/arArrayUtils.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arPrototype } = Cu.import("resource://unmht/modules/base/arPrototype.jsm", {});
let { arScoped } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});

/* global arWindowOverlay */

/**
 * 要素に対するオーバーレイ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          ウィンドウ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 */
function arElementOverlay(window, node) {
  /**
   * ウィンドウ
   * @type {arWindowOverlay}
   */
  this._window = window;

  /**
   * 対象の要素
   * @type {nsIDOMXULElement}
   */
  this._node = node;

  Object.seal(this);
}
arElementOverlay.prototype = Object.freeze({
  /**
   * 対象の要素
   * @type {nsIDOMXULElement}
   */
  get node() {
    return this._node;
  },

  /**
   * イベントハンドラを追加する
   *
   * @param   {string} type
   *          イベントの種類
   * @param   {function(Event)} fn
   *          イベントハンドラ
   */
  addEventListener: function(type, fn) {
    arEvent.add(this._window, this._node, type, fn, false);
    return this;
  },

  /**
   * 要素を追加する
   *
   * @param   {string} nodeName
   *          要素名
   * @param   {Object} attrs
   *          要素の属性
   * @param   {Constructor=} prototype
   *          コンストラクタ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addElement: function(nodeName, attrs, prototype=null) {
    return arWindowOverlay.addElement(this._window, this._node, nodeName, attrs,
                                      prototype);
  },

  /**
   * 要素を追加する
   *
   * @param   {string} ns
   *          ネームスペース
   * @param   {string} nodeName
   *          要素名
   * @param   {Object} attrs
   *          要素の属性
   * @param   {Constructor=} prototype
   *          コンストラクタ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addElementNS: function(ns, nodeName, attrs, prototype=null) {
    return arWindowOverlay.addElement(this._window, this._node, nodeName,
                                      attrs, prototype, ns);
  },

  /**
   * label 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addLabel: function(attrs) {
    return this.addElement("label", attrs);
  },

  /**
   * textbox 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addTextBox: function(attrs) {
    return this.addElement("textbox", attrs);
  },

  /**
   * button 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addButton: function(attrs) {
    return this.addElement("button", attrs);
  },

  /**
   * separator 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addSeparator: function(attrs) {
    return this.addElement("separator", attrs);
  },

  /**
   * menupopup 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addMenuPopup: function(attrs) {
    return this.addElement("menupopup", attrs, arMenuPopupOverlay);
  },

  /**
   * panel 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addPanel: function(attrs) {
    return this.addElement("panel", attrs);
  },

  /**
   * grid 要素を追加する
   *
   * @param   {Object} attrs
   *          grid 要素の属性
   * @param   {Object} params
   *          行や列
   * @returns {arGridOverlay}
   *          要素に対するオーバーレイ
   */
  addGrid: function(attrs, params) {
    return arWindowOverlay.addGrid(this._window, this._node, attrs, params);
  }
});

/**
 * ツールバーに対するオーバーレイ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          ウィンドウ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 * @extends arElementOverlay
 */
function arToolbarOverlay(window, node) {
  arElementOverlay.call(this, window, node);
}
arToolbarOverlay.prototype = Object.freeze({
  /**
   * toolbaritem 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addItem: function(attrs) {
    return this._addToolbarElement("toolbaritem", attrs);
  },

  /**
   * toolbarbutton 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addButton: function(attrs) {
    return this._addToolbarElement("toolbarbutton", attrs);
  },

  /* ==== private ==== */

  /**
   * ツールバーパレットに要素を追加する
   *
   * @param   {string} nodeName
   *          要素名
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  _addToolbarElement: function(nodeName, attrs) {
    let onode = arWindowOverlay.addElement(this._window, null, nodeName, attrs);
    let node = onode.node;
    let window = this._window;

    let toolbox = arArrayUtils.mapfind(
      [
        "navigator-toolbox",
        "mail-toolbox"
      ],
      id => window.document.getElementById(id),
      n => n);
    if (!toolbox) {
      Cu.reportError("toolbox not found.");
      return onode;
    }

    Task.spawn(function*() {
      let i = 0, TRY_MAX = 50;
      for (;;) {
        if (toolbox.palette) {
          break;
        }

        /* onDOMContentLoaded イベント時ではパレットが準備できていない */

        yield arTaskUtils.wait(100);

        i += 1;
        if (i == TRY_MAX) {
          Cu.reportError("toolbox.palette is not defined.");
          return;
        }
      }

      toolbox.palette.appendChild(node);

      let [toolbar, currentset, pos] = arArrayUtils.mapfind(
        window.document.getElementsByTagName("toolbar"),
        toolbar => {
          let currentset = toolbar.getAttribute("currentset").split(",");
          return [toolbar, currentset, currentset.indexOf(node.id)];
        },
        ([toolbar, currentset, pos]) => pos != -1) ||
        [null, null, -1];
      if (!toolbar) {
        return;
      }

      let ref = (pos < currentset.length - 1)
        ? this._findRefNode(toolbar, currentset, pos)
        : null;

      if (!toolbar.toolbox) {
        /* Personal Titlebar で追加したツールバーは toolbox を持たないので
         * パレットに追加したものは insertItem で追加できない */
        node.setAttribute("removable", "true");
        try {
          toolbar.insertBefore(node, ref);
        } catch (e) {
          /* 他のアドオンと操作が競合して失敗するかもしれないので
           * その場合末尾に追加する */
          toolbar.insertBefore(node, null);
        }
      }

      try {
        toolbar.insertItem(node.id, ref);
      } catch (e) {
        /* 他のアドオンと操作が競合して失敗するかもしれないので
         * その場合末尾に追加する */
        toolbar.insertItem(node.id, null);
      }
    }.bind(this)).catch(e => {
      Cu.reportError(e + "\n" + e.stack);
    });

    return onode;
  },

  /**
   * ツールバーの項目の直後の要素を探す
   *
   * @param   {nsIDOMXULElement} toolbar
   *          対象の項目の要素
   * @param   {Array.<string>} currentset
   *          ツールバーの currentset 属性をカンマで分割した配列
   * @param   {number} pos
   *          対象の項目の位置
   * @returns {?nsIDOMXULElement}
   *          ツールバーの項目の直後の要素
   *          見付からなければ null
   */
  _findRefNode: function(toolbar, currentset, pos) {
    return arArrayUtils.mapfind(
      currentset.slice(pos + 1),
      refId => {
        if (refId == "separator" ||
            refId == "spring" ||
            refId == "spacer") {
          /* 複数出現可能な id を持たない要素は出現位置から取得する */
          let refPos = currentset.slice(0, pos)
          .filter(n => n == refId)
          .length;

          let nodes = Array.filter(toolbar.getElementsByTagName("toolbar" + refId),
                                   n => n.parentNode == toolbar);
          if (refPos < nodes.length) {
            return nodes[refPos];
          }

          return null;
        }

        let ref = null;
        try {
          ref = toolbar.querySelector("#" + refId);
        } catch (e) {
          /* Collusion 等で不正な文字列が ID として使用されており
           * そのまま使用すると querySelector が例外を投げる */
          ref = toolbar.ownerDocument.getElementById(refId);
        }
        if (ref) {
          if (ref.parentNode != toolbar) {
            /* forecastfox-toolbar は初期状態で status-bar 内に居るが
             * 設定により後から addon-bar に移動する事があり
             * 状況によっては status-bar 内の forecastfox-toolbar が
             * 取得されてしまう事がある
             * この場合 insertItem の引数としては適切ではない */
            return null;
          }

          return ref;
        }

        return null;
      },
      n => n) || null;
  }
});
arPrototype.extend(arToolbarOverlay, arElementOverlay);

/**
 * menupopup 要素に対するオーバーレイ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          ウィンドウ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 * @extends arElementOverlay
 */
function arMenuPopupOverlay(window, node) {
  arElementOverlay.call(this, window, node);
}
arMenuPopupOverlay.prototype = Object.freeze({
  /**
   * menuitem 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addItem: function(attrs) {
    return this.addElement("menuitem", attrs);
  },

  /**
   * menuseparator 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addSeparator: function(attrs) {
    return this.addElement("menuseparator", attrs);
  },

  /**
   * menu 要素と menupopup 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          menupopup 要素に対するオーバーレイ
   */
  addMenu: function(attrs) {
    let omenu = this.addElement("menu", attrs);
    return arWindowOverlay.addElement(this._window, omenu.node, "menupopup", {}, arMenuPopupOverlay);
  }
});
arPrototype.extend(arMenuPopupOverlay, arElementOverlay);

/**
 * panelmultiview 要素に対するオーバーレイ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          ウィンドウ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 * @extends arElementOverlay
 */
function arPanelMultiViewOverlay(window, node) {
  arElementOverlay.call(this, window, node);
}
arPanelMultiViewOverlay.prototype = Object.freeze({
  /**
   * panelview 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          menupopup 要素に対するオーバーレイ
   */
  addPanelView: function(attrs, label) {
    let node = this.addElement("panelview", attrs, arPanelViewOverlay);
    node.addElement("label", {
      value: label,
      "class": "panel-subview-header"
    });
    node.body = node.addElement("vbox", {
      "class": "panel-subview-body"
    });
    return node;
  }
});
arPrototype.extend(arPanelMultiViewOverlay, arElementOverlay);

/**
 * panelview 要素に対するオーバーレイ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          ウィンドウ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 * @extends arElementOverlay
 */
function arPanelViewOverlay(window, node) {
  this.body = null;

  arElementOverlay.call(this, window, node);
}
arPanelViewOverlay.prototype = Object.freeze({
  /**
   * toolbarbutton 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addItem: function(attrs) {
    return this.body.addElement("toolbarbutton", attrs);
  },

  /**
   * toolbarseparator 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addSeparator: function(attrs) {
    return this.body.addElement("toolbarseparator", attrs);
  },

  /**
   * menu 要素と menupopup 要素を追加する
   * FIXME: 動きがあやしい
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          menupopup 要素に対するオーバーレイ
   */
  addMenu: function(attrs) {
    let omenu = this.body.addElement("menu", attrs);
    return arWindowOverlay.addElement(this._window, omenu.node, "menupopup", {}, arMenuPopupOverlay);
  }
});
arPrototype.extend(arPanelViewOverlay, arElementOverlay);

/**
 * urlbar-icons 要素に対するオーバーレイ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          ウィンドウ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 * @extends arElementOverlay
 */
function arURLBarIconsOverlay(window, node) {
  arElementOverlay.call(this, window, node);
}
arURLBarIconsOverlay.prototype = Object.freeze({
  /**
   * image 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addIcon: function(attrs) {
    return this.addElement("image", attrs);
  }
});
arPrototype.extend(arURLBarIconsOverlay, arElementOverlay);

/**
 * grid 要素に対するオーバーレイ
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          ウィンドウ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 * @extends arElementOverlay
 */
function arGridOverlay(window, node) {
  /**
   * rows 要素のオーバーレイ
   * @type {arElementOverlay}
   */
  this._orows = null;

  /**
   * columns 要素のオーバーレイ
   * @type {arElementOverlay}
   */
  this._ocolumns = null;

  arElementOverlay.call(this, window, node);

  for (let n of Array.from(node.childNodes)) {
    if (n.nodeName.toLowerCase() == "rows") {
      this._orows = new arElementOverlay(window, n);
    } else if (n.nodeName.toLowerCase() == "columns") {
      this._ocolumns = new arElementOverlay(window, n);
    }
  }
}
arGridOverlay.prototype = Object.freeze({
  /**
   * row 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addRow: function(attrs) {
    if (!this._orows) {
      this._orows = this.addElement("rows");
    }
    return this._orows.addElement("row", attrs);
  },

  /**
   * column 要素を追加する
   *
   * @param   {Object} attrs
   *          要素の属性
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addColumn: function(attrs) {
    if (!this._ocolumns) {
      this._ocolumns = this.addElement("columns");
    }
    return this._ocolumns.addElement("column", attrs);
  }
});
arPrototype.extend(arGridOverlay, arElementOverlay);

/**
 * オーバーレイのインスタンス
 *
 * @constructor
 * @param   {nsIXULWindow} window
 *          対象のウィンドウ
 * @implements arIScopedClass
 */
function arWindowOverlayInstance(window) {
  let $ = (id, noWarn) => {
    let node = window.document.getElementById(id);
    if (!node && !noWarn) {
      Cu.reportError("not found: " + id);
    }
    return node;
  };
  let $$ = (selector, noWarn) => {
    let node = window.document.querySelector(selector);
    if (!node && !noWarn) {
      Cu.reportError("not found: " + selector);
    }
    return node;
  };

  let self = this;
  let defineCachedGetter = (prop, func) => {
    let cached = null;
    Object.defineProperty(self, prop, {
      get: function() {
        if (cached) {
          return cached;
        }
        cached = func();
        return cached;
      },
      configurable: false,
      enumerate: true
    });
  };

  /**
   * appcontent
   * @type {arElementOverlay}
   */
  defineCachedGetter("appcontent", () => {
    return new arElementOverlay(window, $("appcontent"));
  });

  /**
   * ファイルメニュー
   * @type {arMenuPopupOverlay}
   */
  defineCachedGetter("fileMenu", () => {
    return new arMenuPopupOverlay(window, $("menu_FilePopup"));
  });

  /**
   * タブメニュー
   * @type {arMenuPopupOverlay}
   */
  defineCachedGetter("tabMenu", () => {
    return new arMenuPopupOverlay(window, $("tabContextMenu"));
  });

  /**
   * ツールメニュー
   * @type {arMenuPopupOverlay}
   */
  defineCachedGetter("toolsMenu", () => {
    let menu = $("menu_ToolsPopup", true);
    if (!menu) {
      /* COMPAT: SeaMonkey, Thunderbird */
      menu = $("taskPopup");
    }
    return new arMenuPopupOverlay(window, menu);
  });

  /**
   * コンテキストメニュー
   * @type {arMenuPopupOverlay}
   */
  defineCachedGetter("contentMenu", () => {
    return new arMenuPopupOverlay(window, $("contentAreaContextMenu"));
  });

  /**
   * フレームのコンテキストメニュー
   * @type {arMenuPopupOverlay}
   */
  defineCachedGetter("frameContentMenu", () => {
    return new arMenuPopupOverlay(window, $$("#frame > menupopup"));
  });

  /**
   * ツールバー
   * @type {arToolbarOverlay}
   */
  defineCachedGetter("toolbar", () => {
    return new arToolbarOverlay(window, null);
  });

  /**
   * ファイルメニュー
   * @type {arMenuPopupOverlay}
   */
  defineCachedGetter("mainPopupSet", () => {
    return new arMenuPopupOverlay(window, $("mainPopupSet"));
  });

  /**
   * パネル
   * @type {arPanelMultiViewOverlay}
   */
  defineCachedGetter("panelUIMultiView", () => {
    let node = $("PanelUI-multiView", true);
    if (!node) {
      /* COMPAT: Thunderbird/SeaMonkey */
      return null;
    }
    return new arPanelMultiViewOverlay(window, node);
  });

  /**
   * ロケーションバーのアイコン
   * @type {arURLBarIconsOverlay}
   */
  defineCachedGetter("URLBarIcons", () => {
    return new arURLBarIconsOverlay(window, $("urlbar-icons"));
  });

  /**
   * ページ情報のグリッド
   * @type {arGridOverlay}
   */
  defineCachedGetter("pageInfoGeneralGrid", () => {
    let node = $("generalGrid", true);
    if (!node) {
      /* COMPAT: SeaMonkey */
      node = arDOMUtils.findParentNode($("urltext", true), "grid");
      if (!node) {
        return null;
      }
    }
    return new arGridOverlay(window, node);
  });

  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 作成した要素
   * @type {Set.<nsIDOMXULElement>}
   */
  this._nodes = new Set();

  Object.seal(this);
}
arWindowOverlayInstance.prototype = Object.freeze({
  /**
   * 要素を作成する
   *
   * @param   {string} nodeName
   *          要素名
   * @param   {string=} ns
   *          ネームスペース
   * @returns {nsIDOMXULElement}
   *          要素
   */
  createElement: function(nodeName, ns) {
    let node = null;
    if (ns) {
      node = this._window.document.createElementNS(ns, nodeName);
    } else {
      node = this._window.document.createElement(nodeName);
    }
    this._nodes.add(node);

    return node;
  },

  /* ==== arIScopedClass ==== */

  /**
   * 破棄する
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  destruct: function(isShutdown, data, reason) {
    for (let node of this._nodes) {
      try {
        node.remove();
      } catch (e) {
        /* 対象が既に閉じている等 */
      }
    }
    this._nodes.clear();
  }
});

/* global -arWindowOverlay */

/**
 * オーバーレイ管理
 *
 * @class
 */
let arWindowOverlay = Object.freeze({
  /**
   * オーバーレイのインスタンスを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {arWindowOverlayInstance}
   *          オーバーレイのインスタンス
   */
  get: function(window) {
    return this._overlays.create(window);
  },

  /**
   * 要素を追加する
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {nsIDOMXULElement} parentNode
   *          親要素
   * @param   {string} nodeName
   *          要素名
   * @param   {Object=} attrs
   *          grid 要素の属性
   * @param   {Constructor=} prototype
   *          コンストラクタ
   * @param   {string=} ns
   *          ネームスペース
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addElement: function(window, parentNode, nodeName,
                       attrs=null, prototype=null, ns=null) {
    let node = this.createElement(window, nodeName, ns);
    if (attrs) {
      if ("id" in attrs) {
        node.id = attrs.id;
      }
      for (let name of Object.keys(attrs)) {
        let value = attrs[name];
        if (value === null) {
          continue;
        }
        if (name == "id" ||
            name == "insertbefore" ||
            name == "insertafter" ||
            name == "POSITION") {
          continue;
        }
        let m = name.match(/^on(.*)$/);
        if (m) {
          arEvent.add(window, node, m[1], value, false);
        } else {
          node.setAttribute(name, value);
        }
      }
    }

    if (parentNode) {
      let ref = null;
      if (attrs) {
        if ("insertbefore" in attrs) {
          ref = arArrayUtils.mapfind(
            attrs["insertbefore"].split(","),
            id => parentNode.querySelector("#" + id),
            n => n && n.parentNode == parentNode) || null;
        } else if ("insertafter" in attrs) {
          let after = arArrayUtils.mapfind(
            attrs["insertafter"].split(","),
            id => parentNode.querySelector("#" + id),
            n => n && n.parentNode == parentNode) || null;
          ref = after ? after.nextSibling : null;
        } else if ("POSITION" in attrs) {
          let pos = attrs["POSITION"];
          if (typeof pos == "object") {
            if ("after" in pos) {
              ref = pos.after.nextSibling;
            } else if ("before" in pos) {
              ref = pos.before;
            }
          } else if (pos == "first") {
            ref = parentNode.firstChild;
          } else if (pos == "last") {
            ref = null;
          }
        }
      }
      parentNode.insertBefore(node, ref);
    }

    if (prototype) {
      return new prototype(window, node);
    }
    return new arElementOverlay(window, node);
  },

  /**
   * グリッドを追加する
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {nsIDOMXULElement} parentNode
   *          親要素
   * @param   {Object} attrs
   *          grid 要素の属性
   * @param   {Object} params
   *          行や列
   * @returns {arGridOverlay}
   *          grid 要素に対するオーバーレイ
   */
  addGrid: function(window, parentNode, attrs, params) {
    let onode = this.addElement(window, parentNode, "grid", attrs, arGridOverlay);

    for (let name of Object.keys(params)) {
      let value = params[name];
      if (name == "columns") {
        for (let attr of value) {
          onode.addColumn(attr);
        }
      } else if (name == "rows") {
        for (let attr of value) {
          onode.addRow(attr);
        }
      }
    }

    return onode;
  },

  /**
   * 要素を作成する
   * シャットダウンやウィンドウを閉じるイベントで自動的に削除される
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {string} nodeName
   *          要素名
   * @param   {string=} ns
   *          ネームスペース
   * @returns {nsIDOMXULElement}
   *          要素
   */
  createElement: function(window, nodeName, ns) {
    return this._overlays.create(window).createElement(nodeName, ns);
  },

  /* ==== private ==== */

  /**
   * 作成した要素リストとオーバーレイ
   * @type {arScopedInstance}
   */
  _overlays: new arScoped(arWindowOverlayInstance)
});

module.arWindowOverlay = arWindowOverlay;

})(this);
