"use strict";

this.EXPORTED_SYMBOLS = ["arWindowUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arScoped } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});

const FirefoxBrowserURI = "chrome://browser/content/browser.xul";
const SeaMonkeyBrowserURI = "chrome://navigator/content/navigator.xul";
const MessengerURI = "chrome://messenger/content/messenger.xul";
const ViewSourceURI = "chrome://global/content/viewSource.xul";
const FirefoxPageInfoURI = "chrome://browser/content/pageinfo/pageInfo.xul";
const SeaMonkeyPageInfoURI = "chrome://navigator/content/pageinfo/pageInfo.xul";

/* global arWindowUtils */

/**
 * ウィンドウごとのプログレスリスナ管理
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @constructor
 * @implements arIScopedClass
 */
function arWebProgressListener(window) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 追加したプログレスリスナ
   * @type {Set.<nsIWebProgressListener>}
   */
  this._listeners = new Set();

  Object.seal(this);
}
arWebProgressListener.prototype = Object.freeze({
  /**
   * ブラウザのプログレスリスナを追加する
   *
   * @param   {nsIWebProgressListener} listener
   *          イベントリスナ
   */
  add: function(listener) {
    let tabbrowser = arWindowUtils.getTabBrowser(this._window);
    if (tabbrowser && "addProgressListener" in tabbrowser) {
      tabbrowser.addProgressListener(listener);
    }

    this._listeners.add(listener);
  },

  /**
   * ブラウザのプログレスリスナを削除する
   *
   * @param   {nsIWebProgressListener} listener
   *          イベントリスナ
   */
  remove: function(listener) {
    let tabbrowser = arWindowUtils.getTabBrowser(this._window);
    if (tabbrowser && "removeProgressListener" in tabbrowser) {
      try {
        tabbrowser.removeProgressListener(listener);
      } catch (e) {
        /* SeaMonkey で失敗する事がある */
      }
    }

    this._listeners.delete(listener);
  },

  /* ==== arIScopedClass ==== */

  /**
   * 破棄する
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  destruct: function(isShutdown, data, reason) {
    for (let listener of this._listeners) {
      this.remove(listener);
    }
    this._listeners.clear();
    this._window = null;
  }
});

/* global -arWindowUtils */

/**
 * Window 操作ユーティリティ
 *
 * @class
 */
let arWindowUtils = Object.freeze({
  /* ---- 種類 ---- */

  /**
   * ブラウザのウィンドウかどうかを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {boolean}
   *          ブラウザのウィンドウかどうか
   */
  isBrowser: function(window) {
    return (window.document.documentURI == FirefoxBrowserURI ||
            window.document.documentURI == SeaMonkeyBrowserURI);
  },

  /**
   * メールのウィンドウかどうかを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {boolean}
   *          メールのウィンドウかどうか
   */
  isMailer: function(window) {
    return (window.document.documentURI == MessengerURI);
  },

  /**
   * ソース表示ウィンドウかどうかを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {boolean}
   *          ソース表示ウィンドウかどうか
   */
  isViewSource: function(window) {
    return (window.document.documentURI == ViewSourceURI);
  },

  /**
   * ページ情報のウィンドウかどうかを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {boolean}
   *          ページ情報のウィンドウかどうか
   */
  isPageInfo: function(window) {
    return (window.document.documentURI == FirefoxPageInfoURI ||
            window.document.documentURI == SeaMonkeyPageInfoURI);
  },

  /* ---- ウィンドウ ---- */

  /**
   * 指定した種類のウィンドウのリストを返す
   *
   * @param   {string}
   *          ウィンドウの種類
   *          null を指定すると全て
   * @returns {Array.<nsIDOMWindow>}
   *          ウィンドウのリスト
   */
  getWindows: function(type) {
    return [...XPCOMUtils.IterSimpleEnumerator(Services.wm.getEnumerator(type),
                                               Ci.nsIDOMWindow)];
  },

  /**
   * ブラウザのウィンドウのリストを返す
   *
   * @returns {Array.<nsIDOMWindow>}
   *          ブラウザのウィンドウのリスト
   */
  getBrowserWindows: function() {
    return this.getWindows("navigator:browser");
  },

  /* ---- ブラウザ ---- */

  /**
   * タブブラウザを取得する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {?(nsIDOMXULElement|BrowserApp)}
   *          ブラウザ (tabbrowser もしくは browser もしくは BrowserApp)
   *          見付からない場合は null
   */
  getTabBrowser: function(window) {
    if (arAppInfo.isFennec) {
      return window.BrowserApp;
    }

    if (this.isMailer(window)) {
      return window.document.getElementById("messagepane");
    }

    return window.document.getElementById("content");
  },

  /**
   * web progress に対応するブラウザを取得する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIWebProgress} webProgress
   *          対象の web progress
   * @returns {?(nsIDOMXULElement|BrowserApp)}
   *          ブラウザ (tabbrowser もしくは browser もしくは BrowserApp)
   *          見付からない場合は null
   */
  getBrowserForWebProgress: function(window, webProgress) {
    for (let browser of arWindowUtils.getBrowsers(window)) {
      if (browser.webProgress.DOMWindowID == webProgress.DOMWindowID) {
        return browser;
      }
    }
    return null;
  },

  /**
   * 現在のブラウザを取得する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {?nsIXULElement}
   *          現在のブラウザ
   *          見付からない場合は null
   */
  getCurrentBrowser: function(window) {
    if (this.isMailer(window)) {
      return window.document.getElementById("messagepane");
    }

    let tabbrowser = this.getTabBrowser(window);
    if (!tabbrowser) {
      return null;
    }

    if (arAppInfo.isFennec) {
      return tabbrowser.selectedTab.browser;
    } else {
      return tabbrowser.selectedBrowser;
    }
  },

  /**
   * ウィンドウのタブの配列を取得する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {Array.<nsIDOMXULElement>}
   *          タブの配列
   */
  getTabs: function(window) {
    let tabbrowser = this.getTabBrowser(window);
    if (!tabbrowser) {
      return [];
    }

    return ("visibleTabs" in tabbrowser)
      ? Array.from(tabbrowser.visibleTabs)
      : Array.from(tabbrowser.tabs);
  },

  /**
   * ウィンドウのブラウザの配列を取得する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {Array.<nsIDOMXULElement>}
   *          ブラウザの配列
   */
  getBrowsers: function(window) {
    if (arAppInfo.isFennec) {
      return this.getTabs(window).map(tab => tab.browser);
    } else {
      return this.getTabs(window).map(tab => tab.linkedBrowser);
    }
  },

  /**
   * ブラウザに対応するタブを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @returns {?nsIDOMXULElement}
   *          対応するタブ
   *          見付からなかったら null
   */
  getTabForBrowser: function(window, browser) {
    let tabbrowser = this.getTabBrowser(window);
    if (!tabbrowser) {
      return null;
    }

    if (arAppInfo.isFennec) {
      return tabbrowser.getTabForBrowser(browser);
    } else {
      return Array.from(tabbrowser.tabs)
        .find(tab => (tab.linkedBrowser == browser)) || null;
    }
  },

  /**
   * ブラウザに対応するタブを閉じる
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   */
  closeTabForBrowser: function(window, browser) {
    let tab = arWindowUtils.getTabForBrowser(window, browser);
    if (!tab) {
      return;
    }

    let tabbrowser = arWindowUtils.getTabBrowser(window);
    if (arAppInfo.isFennec) {
      tabbrowser.closeTab(tab);
    } else {
      tabbrowser.removeTab(tab);
    }
  },

  /* ---- メーラ ---- */

  /**
   * メッセージパネルが閉じているかどうか調べる
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {boolean}
   *          メッセージパネルが閉じているか
   *          メッセージパネルが無い場合は false
   */
  isMessagePaneCollapsed: function(window) {
    if (!this.isMailer(window)) {
      return false;
    }

    let node;
    node = window.document.getElementById("messagepanebox");
    if (node && node.collapsed) {
      return true;
    }
    node = window.document.getElementById("messagepaneboxwrapper");
    if (node && node.collapsed) {
      return true;
    }

    return false;
  },

  /* ---- イベント ---- */

  /**
   * ページのロードイベントのリスナを追加する
   * シャットダウンやウィンドウを閉じるイベントで自動的に削除される
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {function(Event)} func
   *          イベントリスナ
   */
  addPageLoadListener: function(window, func) {
    if (arAppInfo.isFennec) {
      let BrowserApp = window.BrowserApp;
      if (BrowserApp) {
        arEvent.add(window, BrowserApp.deck, "load", func, true);
      }
    } else {
      let appcontent = window.document.getElementById("appcontent");
      if (appcontent) {
        arEvent.add(window, appcontent, "DOMContentLoaded", func, false);
      }
    }
  },

  /**
   * ブラウザのプログレスリスナを追加する
   * シャットダウンやウィンドウを閉じるイベントで自動的に削除される
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIWebProgressListener} listener
   *          イベントリスナ
   */
  addProgressListener: function(window, listener) {
    this._listeners.create(window).add(listener);
  },

  /**
   * ブラウザのプログレスリスナを削除する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIWebProgressListener} listener
   *          イベントリスナ
   */
  removeProgressListener: function(window, listener) {
    this._listeners.create(window).remove(listener);
  },

  /* ---- 探索 ---- */

  /**
   * URL で指定されたウィンドウを探す
   *
   * @param   {string} url
   *          探すウィンドウの URL
   * @returns {?nsIDOMWindow}
   *          URL に一致するウィンドウ
   */
  find: function(url) {
    for (let window of this.getWindows(null)) {
      if (window.document.documentURI == url) {
        return window;
      }
    }

    return null;
  },

  /**
   * URL で指定されたウィンドウにフォーカスする
   *
   * @param   {string} url
   *          フォーカスするウィンドウの URL
   * @returns {boolean}
   *          見付かったか
   */
  focus: function(url) {
    let window = this.find(url);
    if (!window) {
      return false;
    }

    window.focus();
    return true;
  },

  /* ---- ダイアログ ---- */

  /**
   * シャットダウン時にダイアログを閉じる
   *
   * @param   {string} prefix
   *          ダイアログの URI のプレフィックス
   */
  closeOnShutdown: function(prefix) {
    arShutdownEvent.addFunc((data, reason) => {
      for (let window of this.getWindows(null)) {
        if (window.document.documentURI.startsWith(prefix)) {
          /* これはウィンドウに対する非同期的な呼び出しを起こす
           * ウィンドウの unload イベントはモジュールのアンロード後となる */
          window.close();
        }
      }
    });
  },

  /* ==== private ==== */

  /**
   * ウィンドウごとのプログレスリスナ管理
   * @type {arScoped.<arWebProgressListener>}
   */
  _listeners: new arScoped(arWebProgressListener)
});

module.arWindowUtils = arWindowUtils;

})(this);
