"use strict";

this.EXPORTED_SYMBOLS = ["arWindowWatcher"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arScopeManager } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});

/**
 * ウィンドウ監視
 *
 * @class
 * @implements nsIWindowMediatorListener
 * @implements arIShutdownEventListener
 */
let arWindowWatcher = Object.freeze({
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIWindowMediatorListener]),

  /**
   * イベントリスナを登録する
   *
   * @param   {arIWindowEventListener} listener
   *          イベントリスナ
   */
  addListener: function(listener) {
    this._listeners.add(listener);

    for (let window of arWindowUtils.getWindows(null)) {
      listener.onWindowOpen(window);
    }
    for (let window of arWindowUtils.getWindows(null)) {
      listener.onWindowDOMLoad(window);
    }
    for (let window of arWindowUtils.getWindows(null)) {
      listener.onWindowLoad(window);
    }
  },

  /* ==== nsIWindowMediatorListener ==== */

  /**
   * ウィンドウを開いたイベント
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  onOpenWindow: function(window) {
    window = window.QueryInterface(Ci.nsIInterfaceRequestor)
      .getInterface(Ci.nsIDOMWindow);

    for (let listener of this._listeners) {
      listener.onWindowOpen(window);
    }

    this._hookDOMLoad(window);
    this._hookLoad(window);
  },

  /**
   * ウィンドウを閉じたイベント
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  onCloseWindow: function(window) {
    /* onOpenWindow と window が一致しないので使用しない */
  },

  /**
   * ウィンドウを開いたイベント
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {string} title
   *          タイトル
   */
  onWindowTitleChange: function(window, title) {
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    this._listeners.clear();
  },

  /* ==== private ==== */

  /**
   * イベントリスナのリスト
   * @type {Set.<arIWindowEventListener>}
   */
  _listeners: new Set(),

  /**
   * DOM ロードイベントをフックする
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  _hookDOMLoad: function(window) {
    arEvent.oneshot(window, window, "DOMContentLoaded", () => {
      for (let listener of this._listeners) {
        listener.onWindowDOMLoad(window);
      }
    }, false);
  },

  /**
   * ロードイベントをフックする
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  _hookLoad: function(window) {
    arEvent.oneshot(window, window, "load", () => {
      this._hookUnload(window);

      if (arAppInfo.isFennec) {
        this._hookUIReady(window);
      } else {
        for (let listener of this._listeners) {
          listener.onWindowLoad(window);
        }
      }
    }, false);
  },

  /**
   * UI の準備ができたイベントをフックする
   * Fennec のみ
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  _hookUIReady: function(window) {
    if (window.BrowserApp &&
        window.BrowserApp.deck &&
        window.NativeWindow &&
        window.NativeWindow.menu) {
      for (let listener of this._listeners) {
        listener.onWindowLoad(window);
      }
    } else {
      arEvent.oneshot(window, window, "UIReady", () => {
        for (let listener of this._listeners) {
          listener.onWindowLoad(window);
        }
      }, false);
    }
  },

  /**
   * アンロードイベントをフックする
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  _hookUnload: function(window) {
    let f = () => {
      for (let listener of this._listeners) {
        listener.onWindowUnload(window);
      }

      arScopeManager.leave(window);
    };
    arEvent.add(window, window, "unload", f, false);
  },

  /**
   * 初期化処理
   */
  init: function() {
    for (let window of arWindowUtils.getWindows(null)) {
      this._hookUnload(window);
    }

    Services.wm.addListener(this);
    arShutdownEvent.addFunc((data, reason) => {
      Services.wm.removeListener(this);
    });

    arShutdownEvent.addListener(this);
  }
});
arWindowWatcher.init();

module.arWindowWatcher = arWindowWatcher;

})(this);
